#include "CostFunctions.h"
#include "ShapeModel.h"
#include "FindCorrespondence.h"

real_type data_PointToPoint( index_type sourceIndex, const SE3 &tf, ShapeModel &source, ShapeModel &target )
{
	// 1. Apply the transformation to the source shape
	real3_type v = tf * source.vertex[sourceIndex];

	// 2. Find closest point on the target shape
	ClosestPointInfo cpi = target.findClosestPoint(v);

	// 3. Return the distance to the closest point
	return cpi.dist;
}

real_type data_PointToPlane( index_type sourceIndex, const SE3 &tf, ShapeModel &source, ShapeModel &target )
{
	// 1. Apply the transformation to the source shape
	real3_type v = tf * source.vertex[sourceIndex];

	// 2. Find closest point on the target shape
	ClosestPointInfo cpi = target.findClosestPoint(v);

	if (target.markVertex[cpi.index] == VERTEX_NO_ADJACENT_TRIANGLE) {
		return 0.0;
	}

	// 3. Find the plane at the closest point
	real3_type closestPoint, closestN;
	closestPoint = target.vertex[cpi.index];
	closestN = target.normal[cpi.index];

	// 4. Return the point-to-plane distance
	return abs((v - closestPoint) * closestN);
}

real_type data_Hybrid( index_type sourceIndex, const SE3 &tf, ShapeModel &source, ShapeModel &target )
{
	// 1. Apply the transformation to the source shape
	real3_type v = tf * source.vertex[sourceIndex];

	// 2. Find closest point on the target shape
	ClosestPointInfo cpi = target.findClosestPoint(v);

	// 3. Determine the type of point
	bool onBoundary = (target.markVertex[cpi.index] == VERTEX_ON_BOUNDARY);

	if (onBoundary) {
		if (target.markVertex[cpi.index] == VERTEX_NO_ADJACENT_TRIANGLE) {
			return 0.0;
		}

		// 3. Find the plane at the closest point
		real3_type closestPoint, closestN;
		closestPoint = target.vertex[cpi.index];
		closestN = target.normal[cpi.index];

		// 4. Return the point-to-plane distance
		return abs((v - closestPoint) * closestN);
	}
	else {
		// Return the distance to the closest point
		return cpi.dist;
	}
}

real_type data_D6( index_type sourceIndex, const SE3 &tf, ShapeModel &source, ShapeModel &target )
{
	// 1. Apply the transformation to the source shape
	real3_type v = tf * source.vertex[sourceIndex];

	// 2. Find closest point on the target shape
	ClosestPointInfo cpi = target.findClosestPoint(v);

	if ((source.markVertex[sourceIndex] == VERTEX_NO_ADJACENT_TRIANGLE) || 
		(target.markVertex[cpi.index] == VERTEX_NO_ADJACENT_TRIANGLE)) {
		return 0.0;
	}

	// 3. Find the corresponding normals
	real3_type n = tf.rot * source.normal[sourceIndex];
	real3_type closestN = target.normal[cpi.index];
	
	// 4. Return a weighted sum of the point and normal distance
	return sqrt(cpi.dist + 2.0*target.SampleResolution()*(n - closestN).norm());
}

real_type smooth_EdgeLength( index_type sourceIndex1, const SE3 &tf1, index_type sourceIndex2, const SE3 &tf2, ShapeModel &source, ShapeModel &target )
{
	// 1. Apply the transformation to both vertices
	real3_type v1 = source.vertex[sourceIndex1];
	real3_type v2 = source.vertex[sourceIndex2];

	// 2. Measure the change in edge lengths
	real_type lengthBefore = (v1-v2).norm();
	real_type lengthAfter = (tf1*v1 - tf2*v2).norm();

	// 3. Return the absolute difference in lengths
	return abs(lengthBefore - lengthAfter);
}


real_type evaluateDataCost( const std::vector<SE3> &assignedTf, ShapeModel &source, ShapeModel &target, DataFcnType dataFcn )
{
	// Evaluate the data cost for each vertex
	real_type total = 0.0;

	for (index_type i = 0; i < source.numVertices(); i++) {
		total += dataFcn(i, assignedTf[i], source, target);
	}

	return total;
}

real_type evaluateSmoothCost( const std::vector<SE3> &assignedTf, ShapeModel &source, ShapeModel &target, SmoothFcnType smoothFcn )
{
	// Evaluate the data cost for each edge
	real_type total = 0.0;

	if ((source.shapeType != SHAPE_MESH) || (!((Mesh3_real&)source).gd)) {
		return 0.0;
	}

	EdgeList &sourceEdges = ((Mesh3_real&)source).gd->edgeList;

	for (index_type ei = 0; ei < sourceEdges.size(); ei++) {
		Edge &edge = sourceEdges[ei];
		index_type i = edge.vi[0];
		index_type j = edge.vi[1];

		total += smoothFcn(i, assignedTf[i], j, assignedTf[j], source, target);
	}

	return total;
}


real_type evaluateTotalCost( const std::vector<SE3> &assignedTf, ShapeModel &source, ShapeModel &target, DataFcnType dataFcn, SmoothFcnType smoothFcn )
{
	return (evaluateDataCost(assignedTf, source, target, dataFcn) + evaluateSmoothCost(assignedTf, source, target, smoothFcn));
}


data_type evaluateDataCost( const std::vector<SE3> &assignedTf, ShapeModel &source, ShapeModel &target, DataFcnArray dataFcn )
{
	data_type total(dataFcn.size());

	for (index_type i = 0; i < dataFcn.size(); i++) {
		total.val[i] = evaluateDataCost(assignedTf, source, target, dataFcn[i]);
	}

	return total;
}

data_type evaluateSmoothCost( const std::vector<SE3> &assignedTf, ShapeModel &source, ShapeModel &target, SmoothFcnArray smoothFcn )
{
	data_type total(smoothFcn.size());

	for (index_type i = 0; i < smoothFcn.size(); i++) {
		total.val[i] = evaluateSmoothCost(assignedTf, source, target, smoothFcn[i]);
	}

	return total;
}
