#include "FeatureViewer.h"
#include "FindCorrespondence.h"
#include "PointListN.h"
#include "util.h"
#include "scene.h"
#include <iostream>
#include "CorrespViewer.h"

extern "C" {
	#include <SFMT/SFMT.h>
}

using namespace std;

FeatureViewer::FeatureViewer( Scene *sc ) : scene(sc), shiftDim(0), showModes(true), correspViewer(NULL)
{
	// Destroy the object when closing
	setAttribute(Qt::WA_DeleteOnClose);

	// Set the window title.
	setWindowTitle("Feature Space Viewer");
}

void FeatureViewer::init()
{
	// Restore previous viewer state.
	restoreStateFromFile();

	// Context menu
	createContextMenu();

	// Disable the exit key
	setShortcut(EXIT_VIEWER, 0);

	//// Find bounding box of data
	real3_type minPt, maxPt;
	//real_type thresh = 100.0;
	//data_type p(candies.spaceDim());
	//minPt[0] = minPt[1] = minPt[2] = REAL_TYPE_MAX;
	//maxPt[0] = maxPt[1] = maxPt[2] = -REAL_TYPE_MAX;

	//for (index_type i = 0; i < candies.size(); i++) {
	//	p = candies.transforms[i];

	//	for (int d = 0; d < 3; d++) {
	//		if ((abs(p.val[d]) < thresh) && (p.val[d] < minPt[d])) {
	//			minPt[d] = p.val[d];
	//		}
	//		if ((abs(p.val[d]) < thresh) && (p.val[d] > maxPt[d])) {
	//			maxPt[d] = p.val[d];
	//		}
	//	}
	//}
	//cout << "Feature Space Box: " << minPt	<< " -- " << maxPt << endl;

	minPt[0] = minPt[1] = minPt[2] = -M_PI;
	maxPt[0] = maxPt[1] = maxPt[2] = M_PI;

	setSceneBoundingBox(Vec3fToVec(minPt), Vec3fToVec(maxPt));
	camera()->setZClippingCoefficient(10);
	camera()->setZNearCoefficient(1e-2);
	showEntireScene();

	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Reset the window camera
	this->resizeGL(this->width(), this->height());
}

void FeatureViewer::createContextMenu()
{
	contextMenu = new QMenu(tr("&Menu"), this);

	showAllTransformsAct = new QAction(tr("&Plot All Transformations..."), this);
	showAllTransformsAct->setCheckable(true);
	contextMenu->addAction(showAllTransformsAct);

	showFilteredClusterModesAct = new QAction(tr("&Plot Filtered Cluster Modes..."), this);
	showFilteredClusterModesAct->setCheckable(true);
	contextMenu->addAction(showFilteredClusterModesAct);

	showSelectedClusterModesAct = new QAction(tr("&Plot Selected Cluster Modes..."), this);
	showSelectedClusterModesAct->setCheckable(true);
	contextMenu->addAction(showSelectedClusterModesAct);

	showVerifiedTransformsAct = new QAction(tr("&Plot Verified Transformations..."), this);
	showVerifiedTransformsAct->setCheckable(true);
	contextMenu->addAction(showVerifiedTransformsAct);
}

void FeatureViewer::drawVerifiedTransforms()
{
	if (!scene->matchCandidates || !scene->matchCandidates->doneVerify()) {
		return;
	}

	// Draw each verified region
	const std::vector<CorrespondenceList> &clist = scene->matchCandidates->matchList;
	const std::vector<SE3> &tlist = scene->matchCandidates->tfList;
	size_type numRegions = clist.size();

	if (scene->clusterColors.size() < numRegions) {
		// If not enough colors, generate more
		scene->generateClusterColors(numRegions);
	}

	glDisable(GL_LIGHTING);
	glPointSize(3.f);
	glBegin(GL_POINTS);
	int numRegionsShown = 0;
	for (index_type ci = 0; ci < numRegions; ci++) {

		if (clist[ci].size() > scene->minMatchSize_VR) {
			SE3 tf = scene->matchCandidates->rescaleTransform(tlist[ci]);
			data_type twist = tf.log();

			setOpenGLColor(scene->getClusterColor(ci));
			glVertex3f(twist.val[0+shiftDim], twist.val[1+shiftDim], twist.val[2+shiftDim]);

			numRegionsShown++;
		}
	}
	glEnd();

	// Display status string
	setOpenGLColor(foregroundColor().rgba());
	drawText(10, height() - 60, tr("Showing %1/%2 verified regions").arg(numRegionsShown).arg(numRegions));
}

void FeatureViewer::drawFilteredClusterModes()
{
	if (!scene->matchCandidates->doneClustering()) {
		cerr << "Error in FeatureViewer::drawSelectedClusterModes(): Missing pose clustering result." << endl;
		return;
	}

	idxList clusterList;
	int numClusters = scene->matchCandidates->clusterModes.size();
	int minClusterSize = scene->minVerifyClusterSize;
	int maxNumClusters = scene->maxNumVerifyClusters_VR;

	//////////////////////////////////////////////////////////////////////////
	// Find the list of clusters that were chosen
	size_type numVerified = 0;
	for (index_type ci = 0; ci < numClusters; ci++) {
		// Check if cluster is valid
		if (!scene->matchCandidates->validCluster(ci)) {
			continue;
		}

		// Check if cluster is big enough
		size_type clusterSize = scene->matchCandidates->clusterSizes[ci];
		if (clusterSize < minClusterSize) {
			continue;
		}

		// Check if number of verified clusters reached max
		if ((maxNumClusters >= 0) && (numVerified >= maxNumClusters)) {
			break;
		}

		clusterList.push_back(ci);
		numVerified++;
	}

	// Ensure that there's enough colors
	if (scene->clusterColors.size() < scene->matchCandidates->clusterModes.size()) {
		scene->generateClusterColors(scene->matchCandidates->clusterModes.size());
	}

	//////////////////////////////////////////////////////////////////////////
	// Plot the cluster modes
	glDisable(GL_LIGHTING);
	glPointSize(5.f);
	glColor3f(0.3, 0.3, 0.5);
	glBegin(GL_POINTS);
	for (index_type li = 0; li < clusterList.size(); li++) {
		index_type i = clusterList[li];
		data_type p = scene->matchCandidates->clusterModes[i];
		//setOpenGLColor(scene->getClusterColor(i));
		glVertex3f(p.val[0+shiftDim], p.val[1+shiftDim], p.val[2+shiftDim]);
	}
	glEnd();

	drawText(10, height() - 48, tr("Filtering %1/%2 clusters").arg(clusterList.size()).arg(numClusters));
}


void FeatureViewer::drawSelectedClusterModes()
{
	//if (!scene->matchCandidates->doneClustering() || !correspViewer) {
	//	cerr << "Error in FeatureViewer::drawSelectedClusterModes(): Missing pose clustering result or correspViewer." << endl;
	//	return;
	//}
	if (!scene->matchCandidates || !correspViewer || !scene->matchCandidates->size()) {
		cerr << "Error in FeatureViewer::drawSelectedClusterModes(): Missing transformation sampling or correspViewer." << endl;
		return;
	}

	// Select the clusters
	int numClusters = scene->matchCandidates->clusterModes.size();
	idxList pairList = scene->matchCandidates->selectPairsHavingVertices(correspViewer->selected, correspViewer->targetSelected);
	//idxList clusterList = scene->matchCandidates->selectClustersHavingVertices(correspViewer->selected, correspViewer->targetSelected);

	// Ensure that there's enough colors
	if (scene->clusterColors.size() < scene->matchCandidates->clusterModes.size()) {
		scene->generateClusterColors(scene->matchCandidates->clusterModes.size());
	}

	// Plot the cluster modes
	glDisable(GL_LIGHTING);
	glPointSize(3.f);
	glColor3f(0.7, 0.7, 1.0);
	glBegin(GL_POINTS);
	//for (index_type li = 0; li < clusterList.size(); li++) {
	for (index_type li = 0; li < pairList.size(); li++) {
		//index_type i = clusterList[li];
		//data_type p = scene->matchCandidates->clusterModes[i];
		//setOpenGLColor(scene->getClusterColor(i));

		data_type p = scene->matchCandidates->transforms[pairList[li]];
		glVertex3f(p.val[0+shiftDim], p.val[1+shiftDim], p.val[2+shiftDim]);
	}
	glEnd();

	//drawText(10, height() - 36, tr("Selecting %1/%2 clusters").arg(clusterList.size()).arg(numClusters));
	drawText(10, height() - 36, tr("Selecting %1/%2 pairs").arg(pairList.size()).arg(scene->matchCandidates->size()));
}

void FeatureViewer::draw()
{
	if (!scene->matchCandidates) {
		return;
	}
	const MatchCandidates &candies = *scene->matchCandidates;

	if (showVerifiedTransformsAct->isChecked()) {
		drawVerifiedTransforms();
	}
	if (showFilteredClusterModesAct->isChecked()) {
		drawFilteredClusterModes();
	}
	if (showSelectedClusterModesAct->isChecked()) {
		drawSelectedClusterModes();
	}

	if (!showAllTransformsAct->isChecked()) {
		return;
	}
	glDisable(GL_LIGHTING);
	setOpenGLColor(foregroundColor().rgba());
	glPointSize(5);

	if (scene->clusterResultExists()) {
		const idxList &clusterLabels = scene->matchCandidates->clusterLabels;
		const PointListN &clusterModes = scene->matchCandidates->clusterModes;

		if (scene->clusterColors.size() < clusterModes.size()) {
			scene->generateClusterColors(clusterModes.size());
		}

		// Plot the points with the label colors
		glBegin(GL_POINTS);
		for (index_type i = 0; i < candies.size(); i++) {
			data_type p = candies.transforms[i];
			index_type label = clusterLabels[i];

			// Check to see if it passes the minimum cluster size test
			if (!candies.validCluster(label) || (candies.clusterSizes[label] < scene->minVerifyClusterSize)) {
				continue;
			}

			setOpenGLColor(scene->getClusterColor(label));
			glVertex3f(p.val[0+shiftDim], p.val[1+shiftDim], p.val[2+shiftDim]);
		}
		glEnd();

		// Plot the cluster modes
		if (showModes) {
			glPointSize(6.f);
			glBegin(GL_POINTS);
			for (index_type i = 0; i < clusterModes.size(); i++) {
				data_type p = clusterModes[i];

				// Check to see if it passes the minimum cluster size test
				if (!candies.validCluster(i) || (candies.clusterSizes[i] < scene->minVerifyClusterSize)) {
					continue;
				}

				setOpenGLColor(scene->getClusterColor(i));
				glVertex3f(p.val[0+shiftDim], p.val[1+shiftDim], p.val[2+shiftDim]);
			}
			glEnd();
		}

		// Count the number of displayed clusters
		int numClustersShown = 0;
		for (index_type i = 0; i < clusterModes.size(); i++) {
			// Check to see if it passes the minimum cluster size test
			if (candies.validCluster(i) && (candies.clusterSizes[i] >= scene->minVerifyClusterSize)) {
				numClustersShown++;
			}
		}
		setOpenGLColor(foregroundColor().rgba());
		drawText(10, height() - 24, tr("Showing %1/%2 clusters").arg(numClustersShown).arg(clusterModes.size()));
	}
	else {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < candies.size(); i++) {
			data_type p = candies.transforms[i];
			glVertex3f(p.val[0+shiftDim], p.val[1+shiftDim], p.val[2+shiftDim]);
		}
		glEnd();
	}
}


void FeatureViewer::keyPressEvent( QKeyEvent *e )
{
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	if ((e->key()==Qt::Key_S) && (modifiers==Qt::NoButton)) {
		shiftDim++;
		if (shiftDim > 3) {
			shiftDim = 0;
		}
		//shiftDim = 3 - shiftDim;
		displayMessage(tr("Displaying dimensions: [%1,%2,%3]").arg(shiftDim).arg(shiftDim+1).arg(shiftDim+2));
	}
	else if ((e->key()==Qt::Key_M) && (modifiers==Qt::NoButton)) {
		showModes = !showModes;
		displayMessage(tr("Showing Cluster Modes: %1").arg(showModes?"on":"off"));
	}
	else if ((e->key()==Qt::Key_Space) && (modifiers==Qt::NoButton)) {
		// Compute the center location of this window
		QRect win = geometry();
		QPoint center;
		center.setX(win.x() + win.width() / 7);
		center.setY(win.y() + win.height() / 7);

		// Open the menu at this location
		contextMenu->exec(center);
	}
	else {
		QGLViewer::keyPressEvent(e);
	}
}
