#include "FindCorrespondence.h"
#include <SylvainTools/simple_mesh.h>
#include <iostream>
#include <fstream>
#include "PointListN.h"
#include "util.h"
#include <clapack.h>
#include <deque>
#include "geom_util.h"
#include <GraphCutLib/GCoptimization.h>
#include "geodesic.h"

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;

extern "C" {
	void dpotrs_(char *uplo,
				int *n, int *nrhs,
				double *a, int *lda,
				double *b, int *ldb,
				int *info);

	void dpotrf_(char *uplo,
				int *n, double *a, int *lda,
				int *info);

	void dpocon_(char *uplo,
				int *n,
				double *a, int *lda,
				double *anorm, double *rcond,
				double *work, int *iwork,
				int *info);
	
	void dgesv_(int *n, int *nrhs,
				double *a, int *lda, int *ipiv,
				double *b, int *ldb,
				int *info);
}

template<int N, typename Real>
double* ToFortranStyle( Geometry::Square_matrix<N,Real> M )
{
	// Write the matrix in column-major format.

	double *out = new double[N*N];

	for (int i = 0; i < N; i++) {
		for (int j = 0; j < N; j++) {
			out[i+j*N] = M(i,j);
		}
	}

	return out;
}

template<int N, typename Real>
double MatrixNorm1( Geometry::Square_matrix<N,Real> M )
{
	double result = -DBL_MAX;

	// Compute column sums
	for (int j = 0; j < N; j++) {
		double colSum = 0;
		for (int i = 0; i < N; i++) {
			colSum += M(i,j);
		}

		if (colSum > result) {
			result = colSum;
		}
	}

	return result;
}


template<int N,typename Real>
void Cholesky_Solve( const Geometry::Square_matrix<N,Real> &M, Geometry::Vec<N, Real> &v )
{
	char uplo;
	int n, nrhs, lda, ldb, info;
	double *a, *x, *work;
	double rcond, anorm;
	int *iwork;

	uplo = 'U'; /* U/L indicated that the upper/lower triangle of the
				symmetric matrix is stored. */

	nrhs = 1; /* The Fortran routine that I am calling will solve a
			  system of equations with multiple right hand sides.
			  nrhs gives the number of right hand sides, and then b
			  must be a matrix of dimension n by nrhs. */
	n = N;
	lda = N;	// Leading dimension of a and b
	ldb = N;
	a = ToFortranStyle(M); // Convert A to a Fortran style matrix

	/* The Fortran routine replaces the input information in b with the
	results.  Since we are interested in the results, we put the
	initial conditions in the output array and use that in the
	call. */
	x = new double[N];
	for (int i = 0; i < N; i++) {
		x[i] = v[i];
	}

	//// Compute the 1-norm of the input matrix
	//anorm = MatrixNorm1(M);

	//// Estimate the condition number for the matrix; print out warning if it is high
	//work = new double[3*N];
	//iwork = new int[N];

	//// Now the function calls
	dpotrf_(&uplo, &n, a, &lda, &info);
	//dpocon_(&uplo, &n, a, &lda, &anorm, &rcond, work, iwork, &info);

	////cout << "Matrix 1-norm is: " << anorm << endl;
	////cout << "Condition number is: " << rcond << endl;

	//if (rcond < DBL_EPSILON) {
	//	cerr << "Warning in Cholesky_Solve(): Matrix is very close to singular! rcond = " << rcond << "." << endl;
	//}

	// Solve the system
	dpotrs_(&uplo, &n, &nrhs, a, &lda, x, &ldb, &info);

	// Replace the input in v with the output x
	for (int i = 0; i < N; i++) {
		v[i] = x[i];
	}

	// Clean up the memory before returning to the calling program
	delete[] a;
	delete[] x;
	//delete[] work;
	//delete[] iwork;
}


template<int N,typename Real>
void General_Solve( const Geometry::Square_matrix<N,Real> &M, Geometry::Vec<N, Real> &v )
{
	int n, nrhs, lda, ldb, info;
	double *a, *x;
	int *ipiv;
	
	nrhs = 1; /* The Fortran routine that I am calling will solve a
			  system of equations with multiple right hand sides.
			  nrhs gives the number of right hand sides, and then b
			  must be a matrix of dimension n by nrhs. */

	n = N;
	lda = N;	// Leading dimension of a and b
	ldb = N;
	a = ToFortranStyle(M); // Convert A to a Fortran style matrix

	/* The Fortran routine replaces the input information in b with the
	results.  Since we are interested in the results, we put the
	initial conditions in the output array and use that in the
	call. */
	ipiv = new int[N];
	x = new double[N];
	for (int i = 0; i < N; i++) {
		x[i] = v[i];
	}

	// Solve the system
	dgesv_(&n, &nrhs, a, &lda, ipiv, x, &ldb, &info);

	// Replace the input in v with the output x
	for (int i = 0; i < N; i++) {
		v[i] = x[i];
	}

	// Clean up the memory before returning to the calling program
	delete[] a;
	delete[] x;
	delete[] ipiv;
}


real_type FitSphereWeight( const real3_type &query, const real3_type &center, real_type &radSq ) {
	// Compute the squared distance between the center of the evaluation kernel and the query point
	real3_type diff = query - center;
	real_type x2 = diff*diff;

	if (x2 < radSq) {
		// Normalize the domain between 0 and 1
		x2 = 1.0 - (x2 * (1.0 / radSq));
		x2 = x2*x2;
		
		// Evaluate the function (1-x^2)^4
		return x2*x2;
	}

	return 0;
}

bool FitSphere( const PointList &pts, const PointList &normals, real3_type &center, real_type &radSq )
{
	// Fit a 3D sphere to the list of points and normals
	// Use constant weight for the points for now
	const int dim = DIMENSION;
	const real_type beta = 1e6;
	real_type sumWeights = 0;
	real_type nsq, elt;

	// Initialize matrices.  These will be initially zero.
	Geometry::Square_matrix<dim+2, float> DTD;
	Geometry::Vec<dim+2, float> DTb;

	for (index_type i = 0; i < pts.size(); i++) {
		// Construct the design matrix and rhs
		real3_type pt = pts[i];
		real3_type normal = normals[i];
		Geometry::Vec<dim+2,float> ri;
		real_type weight;
		
		weight = FitSphereWeight(pt, center, radSq);
		sumWeights += weight;

		//cout << "Point  #" << i << ": " << pt << endl;
		//cout << "Normal #" << i << ": " << normal << endl;
		//cout << "Weight #" << i << ": " << weight << endl;

		// Add the constraint for the point
		nsq = pt*pt;
		DTD(0,0) += weight;
		DTD(0,dim+1) += weight*nsq;
		DTD(dim+1,dim+1) += weight*nsq*nsq;
		for (int d = 0; d < dim; d++) {
			DTD(0,d+1) += weight*pt[d];
			for (int e = 0; e <= d; e++) {
				DTD(e+1,d+1) += weight*pt[e]*pt[d];
			}
			DTD(d+1,dim+1) += weight*nsq*pt[d];
		}

		weight = weight*beta;
		// Constraint for the normal
		for (int d = 0; d < dim; d++) {
			elt = 2*pt[d];
			DTD(d+1,d+1) += weight;
			DTD(d+1,dim+1) += weight*elt;
			DTD(dim+1,dim+1) += weight*elt*elt;
			DTb[d+1] += weight*normal[d];
			DTb[dim+1] += weight*elt*normal[d];
		}
	}

	if (sumWeights < REAL_TYPE_EPSILON) {
		cerr << "Warning in FitSphere(): Sum of weights " << sumWeights << " is very close to zero!" << endl;
	}

	//cout << "DTD: " << endl << DTD << endl;
	//cout << "DTb: " << DTb << endl << endl;

	// Solve the resulting system
	Cholesky_Solve(DTD, DTb);

	//cout << "Solve result: " << DTb << endl;

	// Store the center
	for (int d = 0; d < dim; d++) {
		center[d] = DTb[d+1];
	}

	// Detect whether if it's a planar feature
	if (abs(DTb[dim+1]) < REAL_TYPE_EPSILON) {
		radSq = DTb[0];

		return false;
	}
	else {
		center = -0.5 / DTb[dim+1] * center;
		radSq = center*center - (DTb[0] / DTb[dim+1]);
		//cout << "Center: " << center << ", radius^2: " << radSq << endl;

		return true;
	}
}




void LocalCoordinateFrame( real3_type n, real3_type &u, real3_type &v )
{
	// Construct a local tangent plane (u,v) perpendicular to n.
	// Helper function for ComputePrincipalCurvatures().

	// 1. Find the minimum absolute component value
	int minComp = 0;
	for (int i = 1; i < 3; i++) {
		if (abs(n[i]) < abs(n[minComp])) {
			minComp = i;
		}
	}

	// 2. Use cross products to find a local frame.
	u[0] = u[1] = u[2] = 0.0;
	u[minComp] = 1.0;
	v = n ^ u;
	v.normalize();
	u = v ^ n;
	u.normalize();
}

bool FitCurvatureTensorLS( const PointList &pts, const PointList &normals, real3_type u, real3_type v, CurvInfo &result )
{
	const size_type npts = pts.size();
	const int dim = DIMENSION;
	real3_type dp, dn;
	real_type dpu, dpv, dnu, dnv;

	if (npts < 3) {
		cout << "Error in FitCurvatureTensorLS(): Not enough constraints for curvature tensor!" << endl;
		return false;
	}

	// Initialize matrices.  These will be initially zero.
	Geometry::Square_matrix<dim, float> ATA;
	Geometry::Vec<dim, float> ATb;

	// Dense sampling method: for all pairs of points
	for (index_type i = 0; i < (npts-1); i++) {
		for (index_type j = i+1; j < npts; j++) {
			//////////////////////////////////////////////////////////////////////////
			// Finite difference of vertex and normal
			dp = pts[j] - pts[i];
			dn = normals[j] - normals[i];
			// Dot product with local coordinate frame
			dpu = dp*u; dpv = dp*v;
			dnu = dn*u; dnv = dn*v;

			//////////////////////////////////////////////////////////////////////////
			// Add to the design matrix: upper-left diagonal
			ATA(0,0) += dpu*dpu;
			ATA(0,1) += dpu*dpv;
			ATA(2,2) += dpv*dpv;

			// Add to the RHS
			ATb[0] += dpu*dnu;
			ATb[1] += dpv*dnu + dpu*dnv;
			ATb[2] += dpv*dnv;
		}
	}

	// Copy over the values to lower-right block of ATA
	ATA(1,1) = ATA(0,0) + ATA(2,2);
	ATA(1,2) = ATA(0,1);

	// Solve the resulting system
	Cholesky_Solve(ATA, ATb);

	//////////////////////////////////////////////////////////////////////////
	// 3. Find eigenvalues / eigenvectors of curvature tensor
	real_type a,b,c,d;
	a = ATb[0]; b = ATb[1]; c = ATb[2];
	d = sqrt((a-c)*(a-c) + 4.0*b*b);
	result.kmax = 0.5*(a + c + d);
	result.kmin = 0.5*(a + c - d);
	real2_type eigmax(-2.0*b, a - c - d);
	real2_type eigmin(-2.0*b, a - c + d);
	eigmax.normalize();
	eigmin.normalize();
	result.vmax = eigmax[0]*u + eigmax[1]*v; result.vmax.normalize();
	result.vmin = eigmin[0]*u + eigmin[1]*v; result.vmin.normalize();

	//// Check result
	//cout << "% Resulting curvature tensor: " << endl;
	//cout << "A = [" << a << ", " << b << "; " << b << ", " << c << "];" << endl;
	//cout << "% Maximum eigenvalue / eigenvector: " << endl;
	//cout << "kmax = " << result.kmax << ";" << endl;
	//cout << "vmax = " << eigmax << ";" << endl;
	//cout << "% Minimum eigenvalue / eigenvector: " << endl;
	//cout << "kmin = " << result.kmin << ";" << endl;
	//cout << "vmin = " << eigmin << ";" << endl;

	return true;
}

bool FitCurvatureTensorRobust( const PointList &pts, const PointList &normals, real3_type u, real3_type v, CurvInfo &result )
{
	const int dim = DIMENSION;
	const size_type numPairs = pts.size() * (pts.size() - 1) / 2;
	const size_type medianIndex = numPairs;
	real3_type dp, dn;
	real_type w1, w2;
	int numItr = 0;
	const int maxItr = 20;
	bool converged = false;
	real_type invSig = 1.0;
	real_type diff;
	const real_type convergeThreshold = 1e-3;
	real_type sigThreshold = REAL_TYPE_MAX;

	// Initialize matrices.  These will be initially zero.
	Geometry::Square_matrix<dim+1, real_type> ATA;
	Geometry::Vec<dim+1, real_type> ATb, lastATb;
	vector< Geometry::Vec<dim+1, real_type> > input;
	vector<real_type> residuals, partialResiduals;
	residuals.resize(2*numPairs);
	partialResiduals.resize(medianIndex+1);
	input.resize(numPairs);

	// Dense sampling method: use all pairs of points for curvature estimate
	index_type pos = 0;
	for (index_type i = 0; i < (pts.size()-1); i++) {
		for (index_type j = i+1; j < pts.size(); j++) {
			//////////////////////////////////////////////////////////////////////////
			// Finite difference of vertex and normal
			dp = pts[j] - pts[i];
			dn = normals[j] - normals[i];

			// Dot product with local coordinate frame & store
			input[pos][0] = dp*u;
			input[pos][1] = dp*v;
			input[pos][2] = dn*u;
			input[pos][3] = dn*v;

			// Initialize residuals
			residuals[2*i] = 1.0;
			residuals[2*i+1] = 1.0;
			
			pos++;
		}
	}

	while (!converged) {
		// Set matrices to zero
		for (index_type i = 0; i < dim+1; i++) {
			for (index_type j = 0; j < dim+1; j++) {
				ATA(i,j) = 0.0;
			}
			ATb[i] = 0.0;
		}

		// Dense sampling method: use all pairs of points for curvature estimate
		for (index_type i = 0; i < input.size(); i++) {
			if (numItr == 0) {
				w1 = w2 = 1.0;
			}
			else {
				if (residuals[2*i] > sigThreshold) {
					w1 = w2 = 0.0;
				}
				else {
					w1 = residuals[2*i]*invSig;   w1 = 1.0 + w1*w1; w1 = 2.0 / (w1*w1); w1 = w1*w1;
					w2 = residuals[2*i+1]*invSig; w2 = 1.0 + w2*w2; w2 = 2.0 / (w2*w2); w2 = w2*w2;
				}
			}

			// Add to the design matrix: upper-left diagonal
			ATA(0,0) += w1*input[i][0]*input[i][0];
			ATA(0,1) += w1*input[i][0]*input[i][1];
			ATA(1,1) += w1*input[i][1]*input[i][1];
			ATA(2,2) += w2*input[i][0]*input[i][0];
			ATA(2,3) += w2*input[i][0]*input[i][1];
			ATA(3,3) += w2*input[i][1]*input[i][1];

			// Add to the RHS
			ATb[0] += w1*input[i][0]*input[i][2];
			ATb[1] += w1*input[i][1]*input[i][2];
			ATb[2] += w2*input[i][0]*input[i][3];
			ATb[3] += w2*input[i][1]*input[i][3];
		}

		//cout << "ATA: " << endl << ATA << endl;
		//cout << "ATb: " << ATb << endl << endl;

		// Solve the resulting system
		Cholesky_Solve(ATA, ATb);

		//cout << "Solve result: " << ATb << endl;

		// Compute residuals, find median
		for (index_type i = 0; i < input.size(); i++) {
			residuals[2*i] = abs(input[i][2] - (input[i][0]*ATb[0] + input[i][1]*ATb[1]));
			residuals[2*i+1] = abs(input[i][3] - (input[i][0]*ATb[2] + input[i][1]*ATb[3]));
			w1 = residuals[2*i]*invSig;   w1 = w1*w1;
			w2 = residuals[2*i+1]*invSig; w2 = w2*w2;
		}
		partial_sort_copy(residuals.begin(), residuals.end(), partialResiduals.begin(), partialResiduals.end());
		w1 = 1.4826 * partialResiduals[medianIndex];
		if ((2.0*w1) < sigThreshold) {
			sigThreshold = 2.0*w1;
			invSig = 1.0/w1;
		}

		// Determine convergence
		if (numItr > 0) {
			diff = (ATb - lastATb).norm();
			if ((numItr >= maxItr) || (diff < convergeThreshold)) {
				converged = true;
			}
		}
		lastATb = ATb;
		numItr++;
	}

	//////////////////////////////////////////////////////////////////////////
	// 3. Find eigenvalues / eigenvectors of curvature tensor
	real_type a,b,c,d,dsq;
	real3_type vmax, vmin;
	a = ATb[0]; b = ATb[1]; c = ATb[2]; d = ATb[3];
	dsq = sqrt((a-d)*(a-d) + 4.0*b*c);
	result.vmax = 0.5*(a - d + dsq)*u + b*v; result.vmax.normalize();
	result.vmin = 0.5*(a - d - dsq)*u + c*v; result.vmin.normalize();
	result.kmax = 0.5*(a + d + dsq);
	result.kmin = 0.5*(a + d - dsq);
	
	return true;
}


QString ComputePrincipalCurvatures( PointList &vertices, const PointList &normals, const distList &scales, CurvInfoList &cl, bool robust )
{
	const size_type	minimumNeighbors = 7;
	const size_type initialNN = 6;
	const real_type nnFactor = 5.0;
	const size_type numScales = scales.size();
	real3_type x, n, u, v;				/**< Vertex & local coordinate frame */
	idxList inRange;
	distList distRange;
	real_type radSq;
	CurvInfo ct;						/**< Curvature tensor */

	cout << "Computing " << (robust?"robust ":"") << "principal curvatures..." << endl;
	cout << "Scales:";
	for (index_type i = 0; i < scales.size(); i++) {
		cout << " " << scales[i];
	}
	cout << endl;

	cl.resize(vertices.size());

	TextProgress(vertices.size(), true);
	for (index_type vi = 0; vi < vertices.size(); vi++) {
		//////////////////////////////////////////////////////////////////////////
		// 1. Compute local frame at the vertex
		x = vertices[vi];
		ct.n = normals[vi];
		LocalCoordinateFrame(ct.n, u, v);
		cl[vi].resize(numScales);

		for (index_type si = 0; si < numScales; si++) {
			//////////////////////////////////////////////////////////////////////////
			if (scales[si] == 0) {
				// Use adaptive neighborhood size: multiple of the distance to the kth nearest neighbor
				vertices.queryKNN(vertices[vi], initialNN, inRange, distRange, true);
				radSq = nnFactor*distRange[initialNN-1];
				radSq = radSq*radSq;
			}
			else {
				// Use pre-specified fixed neighborhood size
				radSq = scales[si];
			}

			//////////////////////////////////////////////////////////////////////////
			// 2. Gather the points within the search radius
			PointList rangePts, rangeNormals;
			vertices.queryRange(x, radSq, inRange, distRange, true);

			if (inRange.size() < minimumNeighbors) {
				vertices.queryKNN(x, minimumNeighbors, inRange, distRange, true);
			}

			rangePts.reserve(inRange.size());
			rangeNormals.reserve(inRange.size());
			for (index_type j = 0; j < inRange.size(); j++) {
				// If the normal differs by an angle more than pi/2
				if (normals[vi] * normals[inRange[j]] <= 0) {
					// Reject
					continue;
				}
				rangePts.add(vertices[inRange[j]]);
				rangeNormals.add(normals[inRange[j]]);
			}

			//////////////////////////////////////////////////////////////////////////
			// 3. Construct and solve least-squares problem
			if (robust) {
				FitCurvatureTensorRobust(rangePts, rangeNormals, u, v, ct);
			}
			else {
				FitCurvatureTensorLS(rangePts, rangeNormals, u, v, ct);
			}

			// Store result
			cl[vi][si] = ct;

			// Check if there's an overflow
			if (!_finite(ct.kmin) || !_finite(ct.kmax)) {
				cerr << "Warning in ComputePrincipalCurvatures(): found NaN value in principal curvature for vertex: " << vi << ", scale: " << scales[si] << "." << endl;

				//// Recompute for debugging
				//cout << "% Local frame (n, u, v):" << endl;
				//cout << "n = " << normals[vi] << "';" << endl;
				//cout << "u = " << u << "';" << endl;
				//cout << "v = " << v << "';" << endl;
				//cout << "% Total " << rangePts.size() << " points:" << endl;
				//for (index_type j = 0; j < rangePts.size(); j++) {
				//	cout << "pts(:," << j+1 << ") = " << rangePts[j] << "';" << endl;
				//	cout << "normals(:," << j+1 << ") = " << rangeNormals[j] << "';" << endl;
				//}
				FitCurvatureTensorLS(rangePts, rangeNormals, u, v, ct);
			}
		}

		TextProgress(vi);
	}

	cout << "...done!" << endl;

	//////////////////////////////////////////////////////////////////////////
	// Return a description of the method used to compute the principal curvatures
	QString result;
	if (robust) {
		result += QString("Robust method");
	}
	else {
		result += QString("Least-squares method");
	}
	result += QString(" using");
	for (index_type i = 0; i < numScales; i++) {
		if (scales[i] == 0) {
			result += QString(" Scale %1: adaptive operating region of %2 * distance to %3 nearest neighbor per vertex")
				.arg(i).arg(nnFactor).arg(initialNN);
		}
		else {
			result += QString(" Scale %1: fixed operating region of radius %2").arg(i).arg(scales[i]);
		}
	}
	return result;
}


data_type ComputeMeanCurvature( const real3_type &query, PointList& allPoints, PointList& allNormals, const distList &scales )
{
	real3_type sample;
	real_type radSq, radSq99;
	data_type features(scales.size());

	for (index_type s = 0; s < scales.size(); s++) {
		// Perform a range query in all points on the original mesh
		idxList inRange;
		distList distRange;
		PointList rangePts, rangeNormals;
		sample = query;
		radSq = scales[s];

		// Compute an "effective radius" where 99% of the weight is within this radius
		// This is to save computation time in the range query
		radSq99 = scales[s] * (FitSphereWeight99Factor * FitSphereWeight99Factor);

		perfTimers["Range query"].AddStart();

		allPoints.queryRange(sample, radSq99, inRange, distRange, true);

		perfTimers["Range query"].Stop();

		// Gather the point locations and the normals
		rangePts.resize(inRange.size());
		rangeNormals.resize(inRange.size());
		for (index_type j = 0; j < inRange.size(); j++) {
			// Throw out the point if the normal is zero
			if (allNormals[inRange[j]].norm() >= REAL_TYPE_EPSILON) {
				rangePts[j] = allPoints[inRange[j]];
				rangeNormals[j] = allNormals[inRange[j]];
			}
		}

		perfTimers["Fit circle"].AddStart();

		// Fit a circle at this position
		if (FitSphere(rangePts, rangeNormals, sample, radSq)) {
			// Store the mean curvature (inverse of the radius) as the feature
			// Also scale the curvatures logarithmically (add one to avoid -Inf singularity for zero curvature)
			features.val[s] = log(1.0 / sqrt(radSq) + 1.0);
		}
		else {
			// If the fit is degenerate, then the curvature is 0.
			features.val[s] = 0;
		}

		perfTimers["Fit circle"].Stop();
	}

	return features;
}




idxList randomSample( int n, int max )
{
	idxList indices;

	if ((n <= 0) || (n >= max)) {
		// Just use all the vertices in the mesh
		indices.resize(max);
		for (index_type i = 0; i < max; i++) {
			indices[i] = i;
		}
	}
	else {
		indices.resize(n);

		// Otherwise, randomly select vertices from the mesh
		for (index_type i = 0; i < n; i++) {
			indices[i] = floor(((double)max) * genrand_real1());
		}
	}

	return indices;
}



MatchCandidates::MatchCandidates( ShapeModel &sm, ShapeModel &tm ) :
sourceModel(sm), targetModel(tm), sourceSample(NULL), targetSample(NULL),
transforms(transformDim), clusterModes(transformDim),
smoothWeight(1.0), dataWeight(1.0), minVerifiedRegionSize(0)
{
}


MatchCandidates::~MatchCandidates()
{
	clearSample();
}

void MatchCandidates::clearSample()
{
	cout << "*** Cleaning all associated ANN data..." << endl;

	// First clear the ANN information of all objects
	transforms.clearANNData();
	clusterModes.clearANNData();
	closestPoints.clearANNData();
	sourceModel.clearANNData();
	targetModel.clearANNData();

	if (sourceSample) {
		sourceSample->clearANNData();
		delete sourceSample;
		sourceSample = NULL;
	}
	if (targetSample) {
		targetSample->clearANNData();
		delete targetSample;
		targetSample = NULL;
	}
}

void MatchCandidates::clearCandidates()
{
	sourceIndex.clear();
	targetIndex.clear();
	transformsSE3.clear();
	transforms.clear();
}

void MatchCandidates::clearClusterResult()
{
	clusterLabels.clear();
	clusterModes.clear();
	clusterSizes.clear();
}

size_type MatchCandidates::size() const
{
	if (sourceIndex.size() != targetIndex.size()) {
		cerr << "Error in MatchCandidates::size(): size of correspondence list is inconsistent!" << endl;
		// TODO: deal with this exception robustly
	}
	return sourceIndex.size();
}

void MatchCandidates::add( index_type si, index_type ti )
{
	SE3 tf;
	if (computeTransform(si, ti, tf)) {
		sourceIndex.push_back(si);
		targetIndex.push_back(ti);
		transformsSE3.push_back(tf);
	}
}

void MatchCandidates::add( index_type si, index_type ti, SE3 tf )
{
	sourceIndex.push_back(si);
	targetIndex.push_back(ti);
	transformsSE3.push_back(tf);
}

void MatchCandidates::subSampleGeometry( size_type ns, size_type nt, real_type pruneMaxRatio )
{
	// Re-sample only if necessary
	if (!sourceSample || (sourceSample->numVertices() != ns)) {
		if (sourceSample) {
			delete sourceSample;
		}
		sourceSample = sourceModel.subSample(ns, pruneMaxRatio);

		if (sourceSample) {
			cout << "*** Subsampling source model: " << sourceSample->numVertices() << " vertices." << endl;
		}
		else {
			cout << "*** Error sampling source model!" << endl;
		}

		clearClusterResult();
	}
	if (!targetSample || (targetSample->numVertices() != nt)) {
		if (targetSample) {
			delete targetSample;
		}
		targetSample = targetModel.subSample(nt, pruneMaxRatio);

		if (targetSample) {
			cout << "***Subsampling target model: " << targetSample->numVertices() << " vertices." << endl;
		}
		else {
			cout << "***Error sampling target model!" << endl;
		}

		clearClusterResult();
	}

	//targetSample = &targetModel;
	//if (targetSample) {
	//	cout << "*** Using full target model: " << targetSample->numVertices() << " vertices." << endl;
	//}
	//clearClusterResult();
}

bool MatchCandidates::computeTransform( index_type si, index_type ti, SE3 &result )
{
	// Compute the transformation based on the local frames
	Geometry::Matrix33f R, Rs, Rt;
	real3_type t, ts, tt;

	// Which scale to use?  Use the last component, usually the smoothest.
	const int sourceComp = sourceSample->frameScales.size()-1;
	const int targetComp = targetSample->frameScales.size()-1;

	for (int d = 0; d < dim; d++) {
		Rs(d,0) = (*sourceSample).frames[si][sourceComp].vmin[d];
		Rs(d,1) = (*sourceSample).frames[si][sourceComp].vmax[d];
		Rs(d,2) = (*sourceSample).frames[si][sourceComp].n[d];
		Rt(d,0) = (*targetSample).frames[ti][targetComp].vmin[d];
		Rt(d,1) = (*targetSample).frames[ti][targetComp].vmax[d];
		Rt(d,2) = (*targetSample).frames[ti][targetComp].n[d];

		// Check for numerical errors
		for (int c = 0; c < dim; c++) {
			if (!_finite(Rs(d,c)) || !_finite(Rt(d,c))) {
				//cerr << "Warning in MatchCandidates::computeTransform(): found NaN value with pair (" << si << "," << ti << ")!" << endl;
				return false;
			}
		}
	}
	ts = (*sourceSample).vertex[si];
	tt = (*targetSample).vertex[ti];

	// Compute the rotation from source to target
	R =	Rt * Rs.transpose();

	// Check negative determinant
	if (det(R) < 0) {
		// Flip a principal direction: find which principal direction has a smaller rotation
		real_type rotMin = (*sourceSample).frames[si][sourceComp].vmin * (*targetSample).frames[ti][targetComp].vmin;
		real_type rotMax = (*sourceSample).frames[si][sourceComp].vmax * (*targetSample).frames[ti][targetComp].vmax;

		// If the minimum principal direction has a larger angle
		if (rotMin < rotMax) {
			for (int d = 0; d < dim; d++) {
				Rt(d,0) = -Rt(d,0);
			}
		}
		// Otherwise, flip the maximum principal direction
		else {
			for (int d = 0; d < dim; d++) {
				Rt(d,1) = -Rt(d,1);
			}
		}

		// Re-compute rotation from source to target
		R =	Rt * Rs.transpose();
	}

	// Write result
	result.rot = R;
	result.tr = tt - R*ts;

	//// Debugging: print info
	//cout << "Source rotation: " << endl << Rs << endl;
	//cout << "Source translation: " << ts << endl;
	//cout << "Target rotation: " << endl << Rt << endl;
	//cout << "Target translation: " << tt << endl;
	//cout << "Result rotation: " << endl << result.rot << endl;
	//cout << "Result translation: " << result.tr << endl;

	return true;
}


void MatchCandidates::normalizeTranslations()
{
	//////////////////////////////////////////////////////////////////////////
	// Step 1: Average the translations about zero.
	const real_type n = (real_type) transformsSE3.size();
	real_type temp, sumNorm, sumNormSq;
	real3_type avgTr(0,0,0);

	// Compute the average
	for (index_type i = 0; i < transformsSE3.size(); i++) {
		avgTr += transformsSE3[i].tr;
	}
	avgTr = avgTr / n;
	
	// Center the points, compute the norm and norm squared
	sumNorm = sumNormSq = 0.0;
	for (index_type i = 0; i < transformsSE3.size(); i++) {
		transformsSE3[i].tr = transformsSE3[i].tr - avgTr;
		temp = transformsSE3[i].tr * transformsSE3[i].tr;
		sumNormSq += temp;
		sumNorm += sqrt(temp);
	}

	//////////////////////////////////////////////////////////////////////////
	// Step 2: Divide by the standard deviation of the norms
	temp = sqrt((sumNormSq - ((sumNorm*sumNorm)/n)) / (n - 1.0));
	for (index_type i = 0; i < transformsSE3.size(); i++) {
		transformsSE3[i].tr = transformsSE3[i].tr / temp;
	}

	// Store the shift and scale
	trShift = avgTr;
	trScale = temp;
}


void MatchCandidates::printInfo() const
{
	cout << "*** Number of Correspondence Candidates: " << size() << endl;
}

void MatchCandidates::writeData( const char *filename ) const
{
	ofstream out;
	out.open(filename, ios_base::binary);

	if (out.fail()) {
		cerr << "Error in MatchCandidates::writeData(): Could not open output file " << filename << " for writing!" << endl;
		return;
	}

	// file format:
	// [INT: dimension of feature space]
	// [UNSIGNED INT: number of correspondences]
	//    [FLOAT: featurePoint[0]] [FLOAT: featurePoint[1]] ... [FLOAT: featurePoint[n]]
	//    [FLOAT: featurePoint[0]] [FLOAT: featurePoint[1]] ... [FLOAT: featurePoint[n]]
	//    ...

	cout << "Saving feature points in " << filename << " ...";
	const unsigned int n = size();
	float value;

	out.write((char*) &transformDim, sizeof(int));
	out.write((char*) &n, sizeof(unsigned int));

	TextProgress(n, true);
	for (unsigned int i = 0; i < n; i++) {
		for (int d = 0; d < transformDim; d++) {
			value = transforms[i].val[d];
			out.write((char*) &value, sizeof(float));
		}
		TextProgress(i);
	}

	out.close();

	cout << "...done!" << endl;
}


void MatchCandidates::populateData()
{
	//////////////////////////////////////////////////////////////////////////
	// Normalize the translations
	cout << "Normalizing translations...";
	normalizeTranslations();
	cout << "done!" << endl;;

	//////////////////////////////////////////////////////////////////////////
	// Using the transforms, populate the transformation space
	cout << "Populating transformation space...";
	transforms.resize(transformsSE3.size());

	TextProgress(transformsSE3.size(), true);
	for (index_type i = 0; i < transformsSE3.size(); i++) {
		data_type temp = transformsSE3[i].log();

		if (!temp.checkFinite()) {
			cout << "Infinite-valued transform pair at index " << i << ": transform " << endl;
			cout << transformsSE3[i];
			cout << " twist vector: " << temp << endl;
		}

		transforms[i] = temp;

		////// Debugging test: test exp transformation
		////SE3 recovered = exp(temp);
		////cout << "------------------------------------------------------------" << endl;
		////cout << "Original Rotation: " << endl << transformsSE3[i].rot << endl;
		////cout << "Original Translation: " << transformsSE3[i].tr << endl;
		////cout << "Recovered Rotation: " << endl << recovered.rot << endl;
		////cout << "Recovered Translation: " << recovered.tr << endl;
		////cout << "------------------------------------------------------------" << endl;
		TextProgress(i);
	}

	cout << "...done!" << endl;
	cout << "*** Found total " << size() << " pairs!" << endl;

	//////////////////////////////////////////////////////////////////////////
	// If a clustering result exists, clear it
	clearClusterResult();
}

int MatchCandidates::spaceDim() const
{
	return transformDim;
}


/**
 * Perform the mean-shift algorithm for every data point.
 * Consolidate the clusters and output cluster labels and means.
 * Based on the C implementation by Piotr Dollar and Sameer Agarwal from Piotr's Matlab Toolbox.
 * A cluster label of 0 means no cluster was assigned, so labels corresponding to actual clusters start with 1.
 *
 * outputs:
 *   clusterLabel    - Cluster labels for each point.
 *   clusterMeans    - Cluster means for each label.
 *   clusterSizes    - Number of members for each cluster.  Sorted in descending order.
 * 
 */
void MatchCandidates::meanShiftBasic( real_type radSq, real_type rate, int maxItr, real_type ANNepsilon )
{
	// Store the current mean for each point
	int iter = 0;
	idxList inRange;
	distList distances;
	bool finished = false;
	const real_type invRadSq = 1.0 / radSq;
	const real_type shiftThreshold = 1e-3;
	PointListN currentMeans(transformDim);
	data_type meanVector(transformDim);
	data_type diff(transformDim);
	std::vector<bool> converged;

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Starting mean-shift clustering..." << endl;
	cout << "*** Squared bandwidth in transformations space: " << radSq << endl;
	cout << "*** Mean-shift rate: " << rate << endl;
	cout << "*** Maximum number of iterations: " << maxItr << endl;
	cout << "*** ANN epsilon for transformation range query: " << ANNepsilon << endl;
	cout << "Progress...";

	//////////////////////////////////////////////////////////////////////////
	// Step 0: Initialize data structures
	currentMeans.setPts(transforms.getData());			/**< Initialize the window for every point */

	// Initialize "is it finished?" variables
	converged.resize(size());
	for (index_type i = 0; i < size(); i++) {
		converged[i] = false;
	}

	transforms.setANNEpsilon(ANNepsilon);

	//////////////////////////////////////////////////////////////////////////
	// Step 1: Iterate the mean shift procedure for each point
	while ((iter < maxItr) && !finished) {
		// Reset "is it finished?" variable
		finished = true;
		QString id = QString("Iteration %1").arg(iter);

		TextProgressWithLabel(size(), true, id);
		for (index_type i = 0; i < size(); i++) {
			if (!converged[i]) {			/**< If the i^th point didn't converge yet */

				converged[i] = true;		/**< Reset "is it finished?" variable */
				meanVector.setZero();		/**< Reset mean vector */


				// Range query on the current window
				transforms.queryRange(currentMeans[i], radSq, inRange, distances, true);


				if (inRange.size()) {
					// Compute the mean using the Epanechnikov (flat) kernel
					for (index_type	j = 0; j < inRange.size(); j++) {
						meanVector += transforms[inRange[j]];
					}
					meanVector *= 1.0 / ((real_type) inRange.size());

					// Blend with the current mean to control the convergence rate
					meanVector = rate*meanVector + (1.0-rate)*currentMeans[i];

					// Determine if the window shifted
					diff = meanVector - currentMeans[i];

					if (diff.normSqrL2() > shiftThreshold) {
						// Assign new mean
						currentMeans[i] = meanVector;

						// Update "is it finished?" variables
						converged[i] = false;
						finished = false;
					}
				}
			}

			TextProgressWithLabel(i, false, id);
		}

		iter++;
	}

	//////////////////////////////////////////////////////////////////////////
	// Step 2: Consolidate the modes
	// In this step, "converged" will mean "assigned a label".
	data_type labelMean(transformDim);
	index_type temp;
	index_type currentLabel = 1;
	const int minimumClusterSize = 2;

	// Initialize space for storing label
	clusterModes.clear();
	clusterSizes.clear();
	clusterLabels.resize(size());
	for (index_type i = 0; i < size(); i++) {
		converged[i] = false;
		clusterLabels[i] = 0;
	}

	// Push back dummy values for "assigned no cluster"
	clusterModes.add(labelMean);
	clusterSizes.push_back(0);

	cout << " consolidating...";
	for (index_type i = 0; i < size(); i++) {
		// For each point, perform a range query on the space of cluster means
		if (!converged[i]) {
			currentMeans.queryRange(currentMeans[i], radSq, inRange, distances, true);

			if (inRange.size() >= minimumClusterSize) {
				index_type numMembers = 0;
				labelMean.setZero();

				// Assign a label to this cluster, and average the modes to get a final cluster mode
				for (index_type	j = 0; j < inRange.size(); j++) {
					temp = inRange[j];

					if (!converged[temp]) {
						clusterLabels[temp] = currentLabel;
						labelMean += currentMeans[temp];
						converged[temp] = true;
						numMembers++;
					}
				}

				// Increment label only if this label was actually assigned
				if (numMembers > 0) {
					currentLabel++;
					clusterModes.add(labelMean / ((real_type)numMembers));
					clusterSizes.push_back(numMembers);
				}
			}
		}
	}

	// Count the number of points with no label
	clusterSizes[0] = 0;
	for (index_type i = 0; i < size(); i++) {
		if (clusterLabels[i] == 0) {
			clusterSizes[0]++;
		}
	}

	// Sort
	sortClusters();

	cout << " done!" << endl;
	cout << "*** Found total " << clusterModes.size() << " clusters!" << endl;
}


void MatchCandidates::sortClusters()
{
	size_type numClusters = clusterSizes.size();

	cout << "sorting...";

	// Sort the modes and re-label
	//cout << "Number of clusters: " << numClusters << endl;
	//cout << "Before sorting: " << endl;

	vector< pair<size_type,int> > sizeSort;
	sizeSort.resize(numClusters);
	for (index_type ci = 0; ci < numClusters; ci++) {
		sizeSort[ci] = pair<size_type,int>(clusterSizes[ci], ci);
		//cout << " --> (" << sizeSort[ci].first << ", " << sizeSort[ci].second << ")" << endl;
	}

	// Sort in descending order, but the label 0 is always at position 0
	struct sizeComp : public binary_function<pair<size_type,int>, pair<size_type,int>, bool> {
		bool operator()(pair<size_type,int> x, pair<size_type,int> y) {
			if (x.second == 0) return true;
			if (y.second == 0) return false;
			return (x.first > y.first);
		}
	};
	sort(sizeSort.begin(), sizeSort.end(), sizeComp());

	// Use the sorted indices to re-label
	idxList sortedLabels,clusterRank;
	PointListN sortedModes(transformDim);
	vector<size_type> sortedSizes;
	sortedLabels.resize(size());
	clusterRank.resize(numClusters);
	sortedModes.resize(numClusters);
	sortedSizes.resize(numClusters);
	for (index_type ci = 0; ci < numClusters; ci++) {
		//cout << " --> (" << sizeSort[ci].first << ", " << sizeSort[ci].second << ")" << endl;
		sortedModes[ci] = clusterModes[sizeSort[ci].second];
		sortedSizes[ci] = clusterSizes[sizeSort[ci].second];
		clusterRank[sizeSort[ci].second] = ci;
	}
	for (index_type pi = 0; pi < size(); pi++) {
		sortedLabels[pi] = clusterRank[clusterLabels[pi]];
	}

	// Copy the result
	clusterLabels = sortedLabels;
	clusterModes = sortedModes;
	clusterSizes = sortedSizes;
}

void MatchCandidates::pruneClusterGeometricConsistency( real_type T_gc )
{
	// Iterate through each cluster and prune correspondences based on geometric consistency

	// 1. Build cluster member list
	std::vector<idxList> memberList;
	memberList.resize(clusterModes.size());
	for (index_type i = 0; i < size(); i++) {
		index_type label = clusterLabels[i];
		memberList[label].push_back(i);
	}

	cout << "Filtering cluster correspondences...";

	// 2. Compute the geometric consistency for all pairs of correspondences
	TextProgress(memberList.size(), true);
	for (index_type ci = 0; ci < memberList.size(); ci++) {
		if (!validCluster(ci)) {
			continue;
		}

		int n = memberList[ci].size();
		idxList removeList;
		removeList.clear();

		// For each correspondence C_1
		for (int i = 0; i < n; i++) {
			index_type si_1 = sourceIndex[memberList[ci][i]];
			index_type ti_1 = targetIndex[memberList[ci][i]];
			int count = 0;

			// Count the number consistent with C_1
			for (int j = 0; j < n; j++) {
				if (i == j) {
					// Excluding the same correspondence
					continue;
				}
				index_type si_2 = sourceIndex[memberList[ci][j]];
				index_type ti_2 = targetIndex[memberList[ci][j]];

				real_type Dgc = GeometricConsistency(sourceSample->vertex[si_1], sourceSample->normal[si_1],
													sourceSample->vertex[si_2], sourceSample->normal[si_2],
													targetSample->vertex[ti_1], targetSample->normal[ti_1],
													targetSample->vertex[ti_2], targetSample->normal[ti_2]);

				// If it's consistent
				if (Dgc < T_gc) {
					count++;
				}
			}

			// Check how many correspondences are consistent with this one
			if (count < (((real_type)n)*0.25)) {
				// Remove this correspondences
				removeList.push_back(i);
			}
		}

		// Reset the label of all the correspondences to remove
		for (int i = 0; i < removeList.size(); i++) {
			clusterLabels[memberList[ci][i]] = 0;
		}

		TextProgress(ci);
	}

	cout << "done!" << endl;
}

bool MatchCandidates::doneClustering() const
{
	if ((!clusterLabels.size()) || (!clusterModes.size())) {
		return false;
	}
	return true;
}

bool MatchCandidates::doneVerify() const
{
	if (!matchList.size() || !tfList.size()) {
		return false;
	}

	return true;
}

SE3 MatchCandidates::getClusterTransform( index_type i ) const
{
	SE3	tf = exp(clusterModes[i]);
	tf.tr = tf.tr * trScale + trShift;			// Un-normalize translation
	return tf;
}

bool MatchCandidates::validCluster( index_type i ) const
{
	return (i > 0);
}

void MatchCandidates::skipVerify( const size_type minClusterSize, const int maxNumClusters )
{
	// Skip the verification step and just copy the clustering result

	if (!doneClustering()) {
		cerr << "Error in MatchCandidates::skipVerify(): Missing pose clustering result." << endl;
		return;
	}

	// Number of pose clusters
	size_type numClusters = clusterModes.size();

	// Initialize correspondence list
	matchList.clear();
	tfList.clear();
	matchList.resize(numClusters);
	tfList.resize(numClusters);

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Skipping match & verify (copying cluster modes)..." << endl;
	cout << "*** Minimum cluster size: " << minClusterSize << endl;
	cout << "*** Maximum number of clusters: " << maxNumClusters << endl;

	size_type numVerified = 0;
	for (index_type ci = 0; ci < numClusters; ci++) {
		// Check if cluster is valid
		if (!validCluster(ci)) {
			continue;
		}

		// Check if cluster is big enough
		size_type clusterSize = clusterSizes[ci];
		if (clusterSize < minClusterSize) {
			continue;
		}

		// Check if number of verified clusters reached max
		if ((maxNumClusters >= 0) && (numVerified >= maxNumClusters)) {
			break;
		}

		// Copy the match list
		for (index_type i = 0; i < size(); i++) {
			if (clusterLabels[i] == ci) {
				real3_type p = sourceSample->vertex[sourceIndex[i]];
				real3_type q = targetSample->vertex[targetIndex[i]];
				ClosestPointInfo modelP = sourceModel.findClosestPoint(p);
				ClosestPointInfo modelQ = targetModel.findClosestPoint(q);

				matchList[ci].add(sourceModel.vertex[modelP.index],
								  targetModel.vertex[modelQ.index],
								  modelP.index,
								  modelQ.index);
			}
		}

		// Store transformation
		tfList[ci] = getClusterTransform(ci);

		// Increment total number verified
		numVerified++;
	}

	cout << "...done!" << endl;
}


void MatchCandidates::verifyRegion( const size_type minClusterSize, const int maxNumClusters, const real_type growRegionDistThreshold, const real_type ANN_Epsilon )
{
	if (!doneClustering()) {
		cerr << "Error in MatchCandidates::verifyRegion(): Missing pose clustering result." << endl;
		return;
	}

	const int expandNeighbors = 4;
	vector<bool> visited;
	deque<index_type> toVisit;
	int numVisited = 0;
	const size_type numSourceVertices = sourceModel.numVertices();

	// Initialize visited list
	visited.resize(numSourceVertices);
	for (index_type i = 0; i < visited.size(); i++) {
		visited[i] = false;
	}

	// Number of pose clusters
	size_type numClusters = clusterModes.size();

	// Initialize correspondence list
	matchList.clear();
	tfList.clear();
	matchList.resize(numClusters);
	tfList.resize(numClusters);

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Starting transformation match & verify..." << endl;
	cout << "*** Minimum mean-shift cluster size: " << minClusterSize << endl;
	cout << "*** Maximum number of mean-shift clusters to verify: " << maxNumClusters << endl;
	cout << "*** Grow region distance threshold squared: " << growRegionDistThreshold << endl;
	cout << "*** ANN epsilon for neighbor & closest point query: " << ANN_Epsilon << endl;

	cout << "Progress...";
	TextProgress(numSourceVertices, true);

	sourceModel.vertex.setANNEpsilon(ANN_Epsilon);
	targetModel.vertex.setANNEpsilon(ANN_Epsilon);

	//////////////////////////////////////////////////////////////////////////
	// For each pose cluster
	size_type numVerified = 0;
	for (index_type ci = 0; ci < numClusters; ci++) {
		// Check if cluster is valid
		if (!validCluster(ci)) {
			continue;
		}

		// Check if cluster is big enough
		size_type clusterSize = clusterSizes[ci];
		if (clusterSize < minClusterSize) {
			continue;
		}

		// Check if number of verified clusters reached max
		if ((maxNumClusters >= 0) && (numVerified >= maxNumClusters)) {
			break;
		}

		//////////////////////////////////////////////////////////////////////////
		// Step 1. Construct list of cluster members (pair candidate index)
		vector<index_type> members;
		members.resize(clusterSize);
		index_type temp = 0;
		for (index_type i = 0; i < size(); i++) {
			if (clusterLabels[i] == ci) {
				members[temp] = i;
				temp++;
			}
		}

		//////////////////////////////////////////////////////////////////////////
		// Step 2. Initialize cluster transformation
		SE3 tf = getClusterTransform(ci);

		// Reset visited list for exploring this cluster.
		for (index_type i = 0; i < visited.size(); i++) {
			visited[i] = false;
		}

		//////////////////////////////////////////////////////////////////////////
		// Step 3. Explore regions using each cluster member
		const int ICP_Delta = clusterSize*10;
		int lastICP = 0;
		QString id = QString("Cluster %1").arg(ci);
		for (index_type mi = 0, mi_end = clusterSize; mi < mi_end; mi++) {
			idxList inRange;
			distList dists;
			real3_type s = sourceSample->vertex[sourceIndex[members[mi]]];

			// Grab nearest neighbor in source mesh and start expanding from this vertex
			ClosestPointInfo source_CPI = sourceModel.findClosestPoint(s);
			toVisit.push_front(source_CPI.index);

			// Iterate until queue is empty
			while (!toVisit.empty()) {
				index_type pi = *toVisit.begin();
				toVisit.pop_front();

				// Check if we visited the vertex
				if (visited[pi]) {
					continue;
				}

				real3_type p = sourceModel.vertex[pi];
				real3_type q = tf*p;

				//////////////////////////////////////////////////////////////////////////
				// Compute distance to target model
				ClosestPointInfo target_CPI = targetModel.findClosestPoint(q);
				real_type distToTarget;
				real3_type targetPoint;
	
				//// Type 1
				//distToTarget = target_CPI.dist;
				//targetPoint = q;

				// Type 2
				targetModel.hybridClosestPoint(q, targetPoint, distToTarget);

				//////////////////////////////////////////////////////////////////////////
				// If distance to target model is within the threshold
				if (distToTarget < growRegionDistThreshold) {
					// Add to the list of correspondences
					matchList[ci].add(p, targetPoint, pi);

					// Refine transformation using ICP
					if (matchList[ci].size() > (lastICP + max(lastICP/4, ICP_Delta))) {
						//cout << "Refining " << matchList[ci].size() << " pairs using ICP..." << endl;
						tf = matchList[ci].registerICP(targetModel, tf);
						lastICP = matchList[ci].size();
					}

					// Expand to neighbors
					sourceModel.queryKNN(p, expandNeighbors, inRange, dists, true);
					for (index_type i = 0; i < inRange.size(); i++) {
						toVisit.push_front(inRange[i]);
					}

					// Mark as vertex as visited
					visited[pi] = true;
					TextProgress(numVisited++);
				}
			}
		}

		//////////////////////////////////////////////////////////////////////////
		// Step 4. Store final transformation
		tfList[ci] = tf;

		// Project target points to closest points on the target shape
		// We may not want to do this with partial shapes
		//matchList[ci].projectToShape(targetModel);

		// Increment total number verified
		numVerified++;

		//////////////////////////////////////////////////////////////////////////
		// Print out some stats
		//if (matchList[ci].size()) {
		//	cout << "Cluster " << ci << " (" << clusterSize << " pairs) --> " << matchList[ci].size() << " matches completed." << endl;
		//}
	}

	cout << "...done!" << endl;
}


void MatchCandidates::sortVerifyResult( size_type minMatchSize, int maxMatchLists )
{
	cout << "*** Sorting match & verify result..." << endl;
	cout << "*** Minimum verified region size: " << minMatchSize << endl;
	cout << "*** Maximum number verified regions: " << maxMatchLists << endl;

	// Sort the list of matched regions
	storeSortedMatchList(matchList, tfList, sortedMatchList, sortedTFList, minMatchSize, maxMatchLists);
	//cout << "Number of matched regions exceeding match size " << minMatchSize << ": " << sortedMatchList.size() << endl;

	// Store the list of inverse transformations
	sortedInverseTFList.resize(sortedTFList.size());
	for (index_type i = 0; i < sortedTFList.size(); i++) {
		sortedInverseTFList[i] = inv(sortedTFList[i]);

		//// Check if we really have the inverse
		//SE3 temp = sortedTFList[i] * sortedInverseTFList[i];
		//cout << "This should be identity: " << temp << endl;
	}

	// Store the minimum match size
	minVerifiedRegionSize = minMatchSize;
	maxNumVerifiedRegions = maxMatchLists;

	cout << "...done!" << endl;
}

void MatchCandidates::refineCluster( index_type ci )
{
	// Bounds checking
	ci = min (ci, matchList.size()-1);

	// Refine transformation using ICP
	if (validCluster(ci)) {
		cout << "Refining cluster " << ci << " (" << matchList[ci].size() << " pairs) using ICP...";

		//////////////////////////////////////////////////////////////////////////
		// Perform one iteration of ICP
		const PointList& sourcePoints = matchList[ci].sourcePoints;
		closestPoints.resize(sourcePoints.size());
		for (index_type i = 0; i < sourcePoints.size(); i++) {
			targetModel.pointSqDistance(tfList[ci]*sourcePoints[i], closestPoints[i]);
		}
		tfList[ci] = registerQuat(sourcePoints, closestPoints);

		//////////////////////////////////////////////////////////////////////////
		// Assign new target points
		for (index_type i = 0; i < sourcePoints.size(); i++) {
			matchList[ci].targetPoints[i] = tfList[ci]*sourcePoints[i];
		}

		cout << "done!" << endl;
	}
}


SE3 MatchCandidates::rescaleTransform( const SE3 &tf ) const
{
	SE3 tfnew;
	tfnew.rot = tf.rot;
	tfnew.tr = (tf.tr - trShift) * trScale;
	return tfnew;
}

SE3 MatchCandidates::getPairTransform( index_type i ) const
{
	SE3	tf = transformsSE3[i];
	tf.tr = tf.tr * trScale + trShift;			// Un-normalize translation
	return tf;
}

idxList MatchCandidates::selectPairsHavingVertices( const idxList &sourceSelected, const idxList &targetSelected ) const
{
	if (!size()) {
		// No transformations!
		return idxList();
	}

	std::vector<bool> sourceVertexMark = closestSampleIndices(sourceModel, *sourceSample, sourceSelected);
	std::vector<bool> targetVertexMark = closestSampleIndices(targetModel, *targetSample, targetSelected);

	// Return pairs that have selected vertices
	idxList foundPairs;

	for (index_type i = 0; i < size(); i++) {
		if (sourceVertexMark[sourceIndex[i]] && targetVertexMark[targetIndex[i]]) {
			foundPairs.push_back(i);
		}
	}

	return foundPairs;
}


std::vector<bool> MatchCandidates::closestSampleIndices( ShapeModel &model, ShapeModel &sample, const idxList &modelIndices ) const
{
	// Initialize a temporary list to just mark the selected vertices
	std::vector<bool> vertexMark;
	vertexMark.resize(sample.numVertices());
	for (index_type i = 0; i < sample.numVertices(); i++) {
		vertexMark[i] = false;
	}
	for (index_type i = 0; i < modelIndices.size(); i++) {
		real3_type v = model.vertex[modelIndices[i]];

		// Find the nearest sample
		idxList	inRange;
		distList dists;
		sample.vertex.queryKNN(v, 1, inRange, dists, true);

		if (dists[0] < REAL_TYPE_EPSILON) {
			vertexMark[inRange[0]] = true;
		}
	}

	return vertexMark;
}

/**
 * From a list of selected vertices on the source model and the target model,
 * Find the clusters containing the corresponding samples.
 */
idxList MatchCandidates::selectClustersHavingVertices( const idxList &sourceSelected, const idxList &targetSelected )
{
	if (!doneClustering()) {
		// No clustering result!
		return idxList();
	}

	idxList foundPairs = selectPairsHavingVertices(sourceSelected, targetSelected);
	std::vector<bool> clusterMark;
	clusterMark.resize(clusterModes.size());
	for (index_type i = 0; i < clusterMark.size(); i++) {
		clusterMark[i] = false;
	}

	// Mark all the cluster labels that have been selected
	for (index_type i = 0; i < foundPairs.size(); i++) {
		clusterMark[clusterLabels[foundPairs[i]]] = true;
	}

	// Put the indices of the clusters in a list
	idxList foundClusters;
	for (index_type i = 0; i < clusterMark.size(); i++) {
		if (clusterMark[i]) {
			foundClusters.push_back(i);
		}
	}

	return foundClusters;
}

idxList MatchCandidates::selectRegionsHavingVertices( const idxList &selected )
{
	if (!doneVerify()) {
		// No clustering result!
		return idxList();
	}

	// Initialize a temporary list to just mark the selected vertices
	std::vector<bool> vertexMark;
	vertexMark.resize(sourceModel.vertex.size());
	for (index_type i = 0; i < sourceModel.vertex.size(); i++) {
		vertexMark[i] = false;
	}
	for (index_type i = 0; i < selected.size(); i++) {
		vertexMark[selected[i]] = true;
	}

	idxList foundClusters;

	// Go through each matchList, select the clusters that contain any of the marked vertices
	for (index_type i = 0; i < sortedMatchList.size(); i++) {
		CorrespondenceList &matches = sortedMatchList[i];

		for (index_type j = 0; j < matches.size(); j++) {
			if (vertexMark[matches.sourceIndex[j]]) {
				foundClusters.push_back(i);
				break;
			}
		}
	}

	return foundClusters;
}



void MatchCandidates::RANSAC_Candidates( size_type max_itr,
										real_type sourceModel_ANNEpsilon, real_type targetModel_ANNEpsilon,
										real_type inlier_queryRadiusSq, real_type inlier_epsilonBandSq,
										real_type minimumInliersFraction )
{
	if (!sourceSample || !targetSample) {
		cerr << "Error in MatchCandidates::RANSAC_Candidates(): Please generate a sample first!" << endl;
		return;
	}

	// Clear the current list of candidates
	clearCandidates();

	// # of vertices in the source / target
	double nSourceSamples = sourceSample->numVertices();
	double nTargetSamples = targetSample->numVertices();

	// Set ANN epsilon for querying closest point on target model
	sourceModel.vertex.setANNEpsilon(sourceModel_ANNEpsilon);
	targetModel.vertex.setANNEpsilon(targetModel_ANNEpsilon);

	// Find closest feature point in source sample
	cout << "Using RANSAC to find candidate correspondences..." << endl;
	cout << "*** # Samples (source/target): (" << nSourceSamples << "/" << nTargetSamples << ")" << endl;
	cout << "*** Number of iterations: " << max_itr << endl;
	cout << "*** ANN Epsilon for closest distance query (source): " << sourceModel_ANNEpsilon << endl;
	cout << "*** ANN Epsilon for closest distance query (target): " << targetModel_ANNEpsilon << endl;
	cout << "*** Squared Radius for gathering inlier candidates: " << inlier_queryRadiusSq << endl;
	cout << "*** Epsilon-band squared distance threshold for inliers: " << inlier_epsilonBandSq << endl;
	cout << "*** Minimum fraction of inliers for good fit: " << minimumInliersFraction << endl;
	cout << "Running...";


	//////////////////////////////////////////////////////////////////////////
	// Repeat for max_itr iterations
	TextProgress(max_itr, true);
	for (index_type itr = 0; itr < max_itr; itr++) {

		//////////////////////////////////////////////////////////////////////////
		// 1. Pick a randomly selected minimal subset.
		// (A random pair of points, one from source and one from target)
		index_type si = floor(nSourceSamples * genrand_real2());
		index_type ti = floor(nTargetSamples * genrand_real2());

		//cout << "Generated pair (source/target): (" << si << "/" << ti << ")" << endl;

		//////////////////////////////////////////////////////////////////////////
		// 2. Fit the model to the subset
		// (a rigid transformation from source to target)
		SE3 tf;
		if (!computeTransform(si, ti, tf)) {
			continue;
		}

		//////////////////////////////////////////////////////////////////////////
		// 3. Apply the transformation and count inliers

		// a. Find all points in the source model within a radius of si
		idxList inRange;
		distList dist;
		sourceModel.queryRange(sourceSample->vertex[si], inlier_queryRadiusSq, inRange, dist, true);

		// b. Count inliers
		int num_inliers = 0;
		for (index_type i = 0; i < inRange.size(); i++) {
			// i. Apply transformation
			real3_type tf_p = tf * sourceModel.vertex[inRange[i]];

			// ii. Compute error measure: find distance to closest point on target model
			real3_type closestPoint;
			real_type closestDist = targetModel.pointSqDistance(tf_p, closestPoint);

			// iii. Is it an inlier?
			if (closestDist < inlier_epsilonBandSq) {
				num_inliers++;
			}
		}

		//////////////////////////////////////////////////////////////////////////
		// 4. If the number of inliers exceeds the threshold, flag the fit as good.

		// Compute percentage of inliers
		float inlier_fraction = ((float)num_inliers) / ((float)inRange.size());

		if (inlier_fraction >= minimumInliersFraction) {
			// Add the point pair & transformation
			add(si, ti, tf);
		}

		TextProgress(itr);
	}

	populateData();			/**< Post-processing */

	cout << "...done! Generated " << size() << " pairs!" << endl;
}


void MatchCandidates::findCandidateCorrespondence( real_type radSq, size_type maxMatches, real_type ANNEpsilon )
{
	if (!sourceSample || !targetSample) {
		cerr << "Error in MatchCandidates::findCandidateCorrespondence(): Please generate a sample first!" << endl;
		return;
	}

	// Find closest feature point in source sample
	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Starting feature-based pairing..." << endl;
	cout << "*** Squared distance for feature space range query: " << radSq << endl;
	cout << "*** Maximum matches per vertex: " << maxMatches << endl;
	cout << "*** ANN Epsilon for feature space range query: " << ANNEpsilon << endl;
	cout << "Progress...";

	// Clear the current list of candidates
	clearCandidates();

	// Set approximate nearest neighbor for feature query
	targetSample->features->setANNEpsilon(ANNEpsilon);

	TextProgress(sourceSample->numVertices(), true);
	for (index_type si = 0; si < sourceSample->numVertices(); si++) {
		real3_type queryPoint = sourceSample->vertex[si];
		data_type queryFeature = (*sourceSample->features)[si];
		idxList inRange;
		distList distances;

		// Query feature space
		targetSample->features->queryRange(queryFeature, radSq, inRange, distances, true);

		// Add the correspondence candidate to the list of candidates
		size_type numMatches = min(inRange.size(), maxMatches);
		for (index_type i = 0; i < numMatches; i++) {
			//real3_type targetPoint = targetSample->vertex[inRange[i]];
			add(si, inRange[i]);
		}

		TextProgress(si);
	}

	cout << "...done!" << endl;

	populateData();			/**< Post-processing */
}

void MatchCandidates::groundTruthCorrespondence()
{
	if (sourceModel.numVertices() != targetModel.numVertices()) {
		cerr << "Error in MatchCandidates::groundTruthCorrespondence(): Source/Target meshes must have per-vertex correspondence!" << endl;
		return;
	}

	cout << "Using ground truth correspondences...";

	// Duplicate the data
	clearSample();
	sourceSample = sourceModel.makeCopy();
	targetSample = targetModel.makeCopy();

	// Generate per-vertex correspondence
	for (index_type si = 0; si < sourceSample->numVertices(); si++) {
		add(si, si);
	}
	populateData();			/**< Post-processing */

	cout << " done!" << endl;
}


void MatchCandidates::spinImageCandidates( size_type maxMatches )
{
	if (!sourceSample) {
		cerr << "Error in MatchCandidates::spinImageCandidates(): Please generate a sample first!" << endl;
		return;
	}

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Starting spin image based pairing..." << endl;
	cout << "*** Expected bin overlap: " << targetSample->spinImageStack.lambda << endl;
	cout << "*** Maximum matches per vertex: " << maxMatches << endl;
	cout << "Progress...";

	// Clear the current list of candidates
	clearCandidates();

	TextProgress(sourceSample->numVertices(), true);
	for (index_type si = 0; si < sourceSample->numVertices(); si++) {
		idxList inRange;
		const SpinImage &P = sourceSample->spinImageStack[si];

		// Perform similarity query
		targetSample->spinImageStack.querySimilar(P, inRange);

		// Add the candidates in range
		size_type numMatches = min(inRange.size(), maxMatches);
		for (index_type i = 0; i < numMatches; i++) {
			add(si, inRange[i]);
		}

		TextProgress(si);
	}

	cout << "...done!" << endl;

	populateData();			/**< Post-processing */
}


void MatchCandidates::storeSortedMatchList( const std::vector<CorrespondenceList> &clistIN,
										    const std::vector<SE3> &tListIN,
											std::vector<CorrespondenceList> &clistOUT,
											std::vector<SE3> &tListOUT,
											int minMatchSize/* = 0*/,
											int maxLists/* = -1*/ )
{
	size_type numClusters = clistIN.size();
	vector< pair<size_type,int> > sizeSort;
	sizeSort.resize(numClusters);
	for (index_type ci = 0; ci < numClusters; ci++) {
		sizeSort[ci] = pair<size_type,int>(clistIN[ci].size(), ci);
	}

	// Sort in descending order
	struct sizeComp : public binary_function<pair<size_type,int>, pair<size_type,int>, bool> {
		bool operator()(pair<size_type,int> x, pair<size_type,int> y) {
			return (x.first > y.first);
		}
	};
	sort(sizeSort.begin(), sizeSort.end(), sizeComp());

	//////////////////////////////////////////////////////////////////////////
	// Store the non-zero correspondence lists
	clistOUT.clear();
	tListOUT.clear();
	for (index_type i = 0; i < numClusters; i++) {
		const CorrespondenceList &matches = clistIN[sizeSort[i].second];

		// Stop if we fall below the minimum match size
		if (matches.size() <= minMatchSize) {
			break;
		}

		// Stop if we reached the maximum number of clusters
		if ((maxLists >= 0) && (i >= maxLists)) {
			break;
		}

		clistOUT.push_back(matches);
		tListOUT.push_back(tListIN[sizeSort[i].second]);
	}
}

int MatchCandidates::GraphCutClassifyVertexFn( int &p )
{
	if (p < sourceModel.numVertices()) {
		return GRAPH_CUT_FROM_SOURCE;
	}
	else {
		// Modify the input index so that we obtain a valid index into the target model
		p = p - sourceModel.numVertices();
		return GRAPH_CUT_FROM_TARGET;
	}
}

bool MatchCandidates::GraphCutIsDummyLabel( int l )
{
	return (l >= ((int)sortedMatchList.size()));
}

void MatchCandidates::precomputeClosestPts()
{
	if (!sortedMatchList.size() || !sortedTFList.size() || (sortedTFList.size() != sortedInverseTFList.size())) {
		// Region match & verify wasn't performed.  Abort.
		cerr << "Error in MatchCandidates::precomputeClosestPts(): Please perform region match & verify first!" << endl;
		return;
	}

	// For each transformation
	const size_type numTransforms = sortedTFList.size();

	// Allocate memory
	closestToSourcePt.resize(sourceModel.numVertices());
	for (index_type i = 0; i < sourceModel.numVertices(); i++) {
		closestToSourcePt[i].resize(numTransforms);
	}
	closestToTargetPt.resize(targetModel.numVertices());
	for (index_type i = 0; i < targetModel.numVertices(); i++) {
		closestToTargetPt[i].resize(numTransforms);
	}

	cout << "Pre-computing closest points according to each transform...";

	TextProgress(numTransforms, true);

	// Compute closest points
	for (index_type ti = 0; ti < numTransforms; ti++) {
		const SE3 &tf = sortedTFList[ti];
		const SE3 &invTf = sortedInverseTFList[ti];
		real3_type transformed;

		// 1. For source vertices: find the closest target vertex after applying the transformation
		for (index_type vi = 0; vi < sourceModel.numVertices(); vi++) {
			transformed = tf * sourceModel.vertex[vi];
			closestToSourcePt[vi][ti] = targetModel.findClosestPoint(transformed);
		}

		// 2. For target vertices: find the closest source vertex after applying the transformation
		for (index_type vi = 0; vi < targetModel.numVertices(); vi++) {
			transformed = invTf * targetModel.vertex[vi];
			closestToTargetPt[vi][ti] = sourceModel.findClosestPoint(transformed);
		}

		TextProgress(ti);
	}

	cout << "...done!" << endl;
}


real_type MatchCandidates::GraphCutSmoothFn( int p1, int p2, int l1, int l2, void *data )
{
	// Define smoothness by measuring how much it preserves the length between p1 and p2.
	MatchCandidates	&mc = *((MatchCandidates*)data);
	int class1 = mc.GraphCutClassifyVertexFn(p1);
	int class2 = mc.GraphCutClassifyVertexFn(p2);
	real_type errorValue;

	//////////////////////////////////////////////////////////////////////////
	if (mc.GraphCutIsDummyLabel(l1) || mc.GraphCutIsDummyLabel(l2)) {
		// Penalize if one is dummy and one is not
		if (!mc.GraphCutIsDummyLabel(l1) || !mc.GraphCutIsDummyLabel(l2)) {
			errorValue = mc.dummyLabelSmoothError;
		}
		else {
			errorValue = 0.0;
		}
	}
	else if (class1 == class2) {
		// If the edge is within one of the shapes
		real3_type d, dp;

		// 1. Look up the corresponding vertices and transformations
		if (class1 == GRAPH_CUT_FROM_SOURCE) {
			d = mc.sourceModel.vertex[p1] - mc.sourceModel.vertex[p2];
			dp = (mc.sortedTFList[l1]*mc.sourceModel.vertex[p1])
				  - (mc.sortedTFList[l2]*mc.sourceModel.vertex[p2]);
		}
		else {
			d = mc.targetModel.vertex[p1] - mc.targetModel.vertex[p2];
			dp = (mc.sortedInverseTFList[l1]*mc.targetModel.vertex[p1])
				  - (mc.sortedInverseTFList[l2]*mc.targetModel.vertex[p2]);
		}

		// 2. Measure length preservation
		real_type dLength = d.norm();
		real_type dpLength = dp.norm();

		// Assign error value as the difference in lengths
		errorValue = min(abs(d.norm() - dp.norm()), mc.smoothTruncateValue);
	}
	//////////////////////////////////////////////////////////////////////////
	// Otherwise, the edge is between the shapes
	else {
		// 1. Check closest point under the current label assignment (both directions)
		int sourcePt, targetPt, sourceLabel, targetLabel;
		real_type sourceMatchError = 0.0, targetMatchError = 0.0;

		if (class1 == GRAPH_CUT_FROM_SOURCE) {
			sourcePt = p1; sourceLabel = l1;
			targetPt = p2; targetLabel = l2;
		}
		else {
			sourcePt = p2; sourceLabel = l2;
			targetPt = p1; targetLabel = l1;
		}

		// Find the closest point using the currently assigned transformation
		ClosestPointInfo &sourcePtMapping = mc.closestToSourcePt[sourcePt][sourceLabel];
		ClosestPointInfo &targetPtMapping = mc.closestToTargetPt[targetPt][targetLabel];

		// Enforce the same label value if this edge is between closest points
		if (sourceLabel != targetLabel) {
			if ((sourcePtMapping.dist < mc.closestPtDistThreshold) && (sourcePtMapping.index == targetPt)) {
					sourceMatchError = mc.mismatchError;
			}
			// Other direction too, if applicable
			if ((targetPtMapping.dist < mc.closestPtDistThreshold) && (targetPtMapping.index == sourcePt)) {
					targetMatchError = mc.mismatchError;
			}
		}

		// 2. Assign error value
		errorValue = sourceMatchError + targetMatchError;
	}

	return mc.smoothWeight * errorValue;
}

real_type MatchCandidates::GraphCutDataFn_PointToPoint( int p, int l, void *data )
{
	// Use closest-point projection error
	MatchCandidates	&mc = *((MatchCandidates*)data);
	real_type errorValue;

	// Check for dummy label
	if (mc.GraphCutIsDummyLabel(l)) {
		errorValue = mc.dummyLabelDataError;
	}
	else {
		// If it's an actual transformation
		int vertexClass = mc.GraphCutClassifyVertexFn(p);
		real_type dist;

		if (vertexClass == GRAPH_CUT_FROM_SOURCE) {
			dist = mc.closestToSourcePt[p][l].dist;
		}
		else {
			dist = mc.closestToTargetPt[p][l].dist;
		}

		// Point-to-point distance
		errorValue = min(dist, mc.dataTruncateValue);
	}

	return mc.dataWeight * errorValue;
}

real_type MatchCandidates::GraphCutDataFn_PointToPlane( int p, int l, void *data )
{
	// Use closest-point projection error
	MatchCandidates	&mc = *((MatchCandidates*)data);
	real_type errorValue;

	// Check for dummy label
	if (mc.GraphCutIsDummyLabel(l)) {
		errorValue = mc.dummyLabelDataError;
	}
	else {
		// If it's an actual transformation
		int vertexClass = mc.GraphCutClassifyVertexFn(p);
		real_type dist;

		if (vertexClass == GRAPH_CUT_FROM_SOURCE) {
			dist = mc.closestToSourcePt[p][l].dist;
		}
		else {
			dist = mc.closestToTargetPt[p][l].dist;
		}

		// Point-to-point distance
		//errorValue = min(dist, mc.dataTruncateValue);

		if (dist < mc.dataTruncateValue) {
			// Point-to-plane distance with point-to-point distance threshold
			// * Use mc.dataTruncateValue as the point-to-point distance threshold

			// 1. Find the points and normals
			real3_type point, closestPoint;
			real3_type n, closestN;
			if (vertexClass == GRAPH_CUT_FROM_SOURCE) {
				const SE3 &tf = mc.sortedTFList[l];
				point = tf * mc.sourceModel.vertex[p];
				n = tf.rot * mc.sourceModel.normal[p];
				closestPoint = mc.targetModel.vertex[mc.closestToSourcePt[p][l].index];
				closestN = mc.targetModel.normal[mc.closestToSourcePt[p][l].index];
			}
			else {
				const SE3 &tf = mc.sortedInverseTFList[l];
				point = tf * mc.targetModel.vertex[p];
				n = tf.rot * mc.targetModel.normal[p];
				closestPoint = mc.sourceModel.vertex[mc.closestToTargetPt[p][l].index];
				closestN = mc.sourceModel.normal[mc.closestToTargetPt[p][l].index];
			}

			// a. Normal-matching error
			//real_type normalError = (1.0 - (n * closestN)) / 2.0;

			// b. Point-to-plane error
			real_type pointToPlaneError = abs((point - closestPoint) * closestN);

			// c. Combine the errors
			//errorValue = pointToPlaneError + mc.dataTruncateValue * normalError;
			errorValue = pointToPlaneError;
		}
		else {
			errorValue = dist;
		}
	}

	return mc.dataWeight * errorValue;
}


real_type MatchCandidates::GraphCutDataFn_Hybrid( int p, int l, void *data )
{
	// Use closest-point projection error
	MatchCandidates	&mc = *((MatchCandidates*)data);
	real_type errorValue;

	// Check for dummy label
	if (mc.GraphCutIsDummyLabel(l)) {
		errorValue = mc.dummyLabelDataError;
	}
	else {
		// If it's an actual transformation
		int vertexClass = mc.GraphCutClassifyVertexFn(p);
		real_type dist;
		bool isClosestPointOnBoundary = false;

		if (vertexClass == GRAPH_CUT_FROM_SOURCE) {
			dist = mc.closestToSourcePt[p][l].dist;
			isClosestPointOnBoundary = (mc.targetModel.markVertex[mc.closestToSourcePt[p][l].index] == VERTEX_ON_BOUNDARY);
		}
		else {
			dist = mc.closestToTargetPt[p][l].dist;
			isClosestPointOnBoundary = (mc.sourceModel.markVertex[mc.closestToTargetPt[p][l].index] == VERTEX_ON_BOUNDARY);
		}

		// Check if the point is NOT on the boundary
		if (!isClosestPointOnBoundary) {
			// Use Point-to-point distance
			errorValue = dist;
		}
		else {
			// Use Point-to-plane distance
			// 1. Find the points and normals
			real3_type point, closestPoint, closestN;

			if (vertexClass == GRAPH_CUT_FROM_SOURCE) {
				const SE3 &tf = mc.sortedTFList[l];
				point = tf * mc.sourceModel.vertex[p];
				closestPoint = mc.targetModel.vertex[mc.closestToSourcePt[p][l].index];
				closestN = mc.targetModel.normal[mc.closestToSourcePt[p][l].index];
			}
			else {
				const SE3 &tf = mc.sortedInverseTFList[l];
				point = tf * mc.targetModel.vertex[p];
				closestPoint = mc.sourceModel.vertex[mc.closestToTargetPt[p][l].index];
				closestN = mc.sourceModel.normal[mc.closestToTargetPt[p][l].index];
			}

			// 2. Assign error
			errorValue = abs((point - closestPoint) * closestN);
		}
	}

	return mc.dataWeight * errorValue;
}

real_type MatchCandidates::GraphCutDataFn_D6( int p, int l, void *data )
{
	// Use closest-point projection error
	MatchCandidates	&mc = *((MatchCandidates*)data);
	real_type errorValue;

	// Check for dummy label
	if (mc.GraphCutIsDummyLabel(l)) {
		errorValue = mc.dummyLabelDataError;
	}
	else {
		// If it's an actual transformation
		int vertexClass = mc.GraphCutClassifyVertexFn(p);
		ShapeModel *eeguh, *juhguh;
		ClosestPointInfo *cpi;
		SE3 tf;

		if (vertexClass == GRAPH_CUT_FROM_SOURCE) {
			eeguh = &mc.sourceModel;
			juhguh = &mc.targetModel;
			cpi = &mc.closestToSourcePt[p][l];
			tf = mc.sortedTFList[l];
		}
		else {
			eeguh = &mc.targetModel;
			juhguh = &mc.sourceModel;
			cpi = &mc.closestToTargetPt[p][l];
			tf = mc.sortedInverseTFList[l];
		}

		index_type i = cpi->index;

		if ((eeguh->markVertex[p] == VERTEX_NO_ADJACENT_TRIANGLE) ||
			(juhguh->markVertex[i] == VERTEX_NO_ADJACENT_TRIANGLE))
		{
			return 0.0;
		}

		real_type dist, res;
		bool isClosestPointOnBoundary;
		real3_type point, closestPoint, n, closestN;

		isClosestPointOnBoundary = (juhguh->markVertex[i] == VERTEX_ON_BOUNDARY);
		res = juhguh->SampleResolution();
		point = tf * eeguh->vertex[p];
		n = tf.rot * eeguh->normal[p];
		closestPoint = juhguh->vertex[i];
		closestN = juhguh->normal[i];

		if (!isClosestPointOnBoundary) {
			dist = cpi->dist;									/**< Point-to-point if not on boundary */
		}
		else {
			dist = abs((point - closestPoint) * closestN);		/**< Point-to-plane if on boundary */
		}

		errorValue  = sqrt(dist + 2.0*res*(n - closestN).norm());		/**< Combined error */
	}

	return mc.dataWeight * errorValue;
}

inline void sortTwo( int &p, int &q )
{
	if (p > q) {
		int temp = q;
		q = p;
		p = temp;
	}
}

void MatchCandidates::makeGraphEdges( GCoptimizationGeneralGraph *gc,
									  GeodesicDistance *source_gd, GeodesicDistance *target_gd,
									  EdgeList sourceExtraEdges, EdgeList targetExtraEdges )
{
	const int siteOffset = sourceModel.numVertices();
	const int totalSites = sourceModel.numVertices() + targetModel.numVertices();
	int p, q;
	vector< vector<int> > neighborIndices;			/**< Adjacent vertex list (for every vertex) */
	neighborIndices.resize(totalSites);

	//////////////////////////////////////////////////////////////////////////
	// Note: We will always store edges as (lower index #, higher index #).
	//////////////////////////////////////////////////////////////////////////
	// 1. Create a neighbor for each edge within the shape
	// a. Source mesh
	for (index_type i = 0; i < source_gd->edgeList.size(); ++i) {
		p = source_gd->edgeList[i].vi[0];
		q = source_gd->edgeList[i].vi[1];

		// Insert edge
		sortTwo(p,q);
		neighborIndices[p].push_back(q);
	}
	// b. Target mesh (offset by number of source mesh vertices)
	for (index_type i = 0; i < target_gd->edgeList.size(); ++i) {
		p = siteOffset + target_gd->edgeList[i].vi[0];
		q = siteOffset + target_gd->edgeList[i].vi[1];

		// Insert edge
		sortTwo(p,q);
		neighborIndices[p].push_back(q);
	}

	//////////////////////////////////////////////////////////////////////////
	// 2. Create extra edges between just boundary vertices of the shape
	// a. Source mesh
	for (index_type i = 0; i < sourceExtraEdges.size(); ++i) {
		p = sourceExtraEdges[i].vi[0];
		q = sourceExtraEdges[i].vi[1];

		// Insert edge
		sortTwo(p,q);
		neighborIndices[p].push_back(q);
	}
	// b. Target mesh (offset by number of source mesh vertices)
	for (index_type i = 0; i < targetExtraEdges.size(); ++i) {
		p = siteOffset + targetExtraEdges[i].vi[0];
		q = siteOffset + targetExtraEdges[i].vi[1];

		// Insert edge
		sortTwo(p,q);
		neighborIndices[p].push_back(q);
	}

	//////////////////////////////////////////////////////////////////////////
	// 3. Create a neighbor between potential correspondences between shapes
	//////////////////////////////////////////////////////////////////////////
	// a. Closest target point for each source/label combination
	for (index_type i = 0; i < sourceModel.numVertices(); ++i) {
		for (index_type j = 0; j < closestToSourcePt[i].size(); ++j) {
			ClosestPointInfo &info = closestToSourcePt[i][j];

			// If the index is valid, and the closest distance is less than the specified threshold
			if (info.dist < closestPtDistThreshold) {
				// Insert edge
				p = i;
				q = siteOffset + info.index;

				sortTwo(p,q);
				neighborIndices[p].push_back(q);
			}
		}
	}

	// b. Closest source point for each target/label combination
	for (index_type i = 0; i < targetModel.numVertices(); ++i) {
		for (index_type j = 0; j < closestToTargetPt[i].size(); ++j) {
			ClosestPointInfo &info = closestToTargetPt[i][j];

			// If the index is valid, and the closest distance is less than the specified threshold
			if (info.dist < closestPtDistThreshold) {
				// Insert edge
				p = info.index;
				q = siteOffset + i;

				sortTwo(p,q);
				neighborIndices[p].push_back(q);
			}
		}
	}


	//////////////////////////////////////////////////////////////////////////
	// Insert a unique list of edges
	for (int i = 0; i < totalSites; ++i) {
		// 1. Sort each neighbor list (in place)
		sort(neighborIndices[i].begin(), neighborIndices[i].end());

		// 2. Insert each edge to the graph
		//cout << "Neighbors of " << i << ":";
		int lastNeighbor = -1;
		for (int j = 0; j < neighborIndices[i].size(); ++j) {
			int neighbor = neighborIndices[i][j];

			if (neighbor != lastNeighbor) {
				//cout << " " << neighbor;
				gc->setNeighbors(i, neighbor);
			}

			lastNeighbor = neighbor;
		}
		//cout << endl;
	}
}


void MatchCandidates::GraphCutRelabel_Forward(int graphCutSolutionType,
											  int graphCutDataTermType,
											  real_type smoothWeight, real_type dataWeight,
											  real_type smoothTruncateValue, real_type dataTruncateValue )
{
	if (!sortedMatchList.size() || !sortedTFList.size() || (sortedTFList.size() != sortedInverseTFList.size())) {
		// Region match & verify wasn't performed.  Abort.
		cerr << "Error in MatchCandidates::GraphCutRelabel(): Please perform region match & verify first!" << endl;
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// 1. Create the graph structure.
	GeodesicDistance *source_gd = NULL;
	if (sourceModel.shapeType == SHAPE_MESH) {
		source_gd = ((Mesh3_real&)sourceModel).gd;

		if (!source_gd) {
			cerr << "Error in MatchCandidates::GraphCutRelabel_Forward(): Please precompute edge data for source mesh!" << endl;
			return;
		}
	}
	else {
		// Not implemented yet.  Abort.
		cerr << "Error in MatchCandidates::GraphCutRelabel_Forward(): Currently only supports meshes." << endl;
		return;
	}

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Starting graph cut optimization (forward)..." << endl;

	// Write descriptions of the graph cut error function
	cout << "*** Smooth term type: edge-length preserve, label consistency" << endl;
	cout << "*** Data term type: ";
	if (graphCutDataTermType == GRAPH_CUT_DATA_POINT_TO_POINT) {
		cout << "point-to-point" << endl;
	}
	else if (graphCutDataTermType == GRAPH_CUT_DATA_POINT_TO_PLANE) {
		cout << "point-to-plane" << endl;
	}
	else if (graphCutDataTermType == GRAPH_CUT_DATA_HYBRID) {
		cout << "hybrid" << endl;
	}
	else if (graphCutDataTermType == GRAPH_CUT_DATA_D6) {
		cout << "point and normal" << endl;
	}

	//////////////////////////////////////////////////////////////////////////
	// 2. Set parameter values.
	cout << "*** Smoothness / data term weights: (" << smoothWeight << "/" << dataWeight << ")" << endl;
	this->smoothWeight = smoothWeight;
	this->dataWeight = dataWeight;

	cout << "*** Maximum smooth term / data term value: " << smoothTruncateValue << "/" << dataTruncateValue << endl;
	this->smoothTruncateValue = smoothTruncateValue;			/**< Set the maximum value for the smooth term. */
	this->dataTruncateValue = dataTruncateValue;				/**< Set the maximum value for the data term. */

	// Set various sizes
	const int numSites = sourceModel.numVertices();
	int numLabels = sortedMatchList.size();

	cout << "*** Number of sites/labels: " << numSites << "/" << numLabels << endl;
	cout << "*** Source model vertices: " << sourceModel.numVertices() << endl;

	// Precompute closest points
	precomputeClosestPts();

	/////////////////////////////////////////////////////////////////////////
	// 3. Perform graph cut optimization.
	GCoptimizationGeneralGraph *gc = new GCoptimizationGeneralGraph(numSites, numLabels);

	try {
		// Use function pointers for the data and smoothness costs
		if (graphCutDataTermType == GRAPH_CUT_DATA_POINT_TO_POINT) {
			gc->setDataCost(&GraphCutDataFn_PointToPoint, this);
		}
		else if (graphCutDataTermType == GRAPH_CUT_DATA_POINT_TO_PLANE) {
			gc->setDataCost(&GraphCutDataFn_PointToPlane, this);
		}
		else if (graphCutDataTermType == GRAPH_CUT_DATA_HYBRID) {
			gc->setDataCost(&GraphCutDataFn_Hybrid, this);
		}
		else if (graphCutDataTermType == GRAPH_CUT_DATA_D6) {
			gc->setDataCost(&GraphCutDataFn_D6, this);
		}
		else {
			cerr << "Unknown data term type " << graphCutDataTermType << "! Aborting graph cut optimization." << endl;
			return;
		}
		gc->setSmoothCost(&GraphCutSmoothFn, this);

		// Set up edges
		for (index_type i = 0; i < source_gd->edgeList.size(); ++i) {
			int p = source_gd->edgeList[i].vi[0];
			int q = source_gd->edgeList[i].vi[1];
			gc->setNeighbors(p, q);
		}

		// Perform optimization
		real_type energy = gc->compute_energy();
		cout << "Before optimization energy is: " << energy << endl;
		if (graphCutSolutionType == GRAPH_CUT_SWAP) {
			cout << "Solving graph cut using swap..." << endl;
			energy = gc->swap();
		}
		else if (graphCutSolutionType == GRAPH_CUT_EXPANSION) {
			cout << "Solving graph cut using expansion..." << endl;
			energy = gc->expansion();
		}
		else {
			cout << "Unknown solution type " << graphCutSolutionType << "!" << endl;
		}
		cout << "After optimization energy is: " << energy << endl;

		// Read back the result
		graphCutLabels_source.resize(sourceModel.numVertices());
		for ( index_type i = 0; i < sourceModel.numVertices(); i++ ) {
			graphCutLabels_source[i] = gc->whatLabel(i);
		}

		std::vector<CorrespondenceList> reLabeledMatchList;
		reLabeledMatchList.clear();
		reLabeledMatchList.resize(sortedMatchList.size());
		for ( index_type i = 0; i < sourceModel.numVertices(); i++ ) {
			index_type label = gc->whatLabel(i);
			real3_type v = sourceModel.vertex[i];
			const SE3 &tf = sortedTFList[label];
			reLabeledMatchList[label].add(v, tf*v);
		}
		storeSortedMatchList(reLabeledMatchList, sortedTFList, sortedMatchListFinal_source, sortedTFListFinal_source);
	}
	catch (GCException e) {
		e.Report();
	}

	// Free memory associated with graph cut data structure
	delete gc;

	cout << "...done!" << endl;
}

void MatchCandidates::GraphCutRelabel( int graphCutSolutionType,
									   int graphCutDataTermType,
									   bool useDummyLabel,
									   int bdySmoothnessKNN,
									   real_type closestPtDistThreshold,
									   real_type mismatchError,
									   real_type dummyLabelSmoothError, real_type dummyLabelDataError,
									   real_type smoothWeight, real_type dataWeight,
									   real_type smoothTruncateValue, real_type dataTruncateValue )
{
	if (!sortedMatchList.size() || !sortedTFList.size() || (sortedTFList.size() != sortedInverseTFList.size())) {
		// Region match & verify wasn't performed.  Abort.
		cerr << "Error in MatchCandidates::GraphCutRelabel(): Please perform region match & verify first!" << endl;
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// 1. Create the graph structure.
	GeodesicDistance *source_gd = NULL, *target_gd = NULL;
	if ((sourceModel.shapeType == SHAPE_MESH) && (targetModel.shapeType == SHAPE_MESH)) {
		source_gd = ((Mesh3_real&)sourceModel).gd;
		target_gd = ((Mesh3_real&)targetModel).gd;

		if (!source_gd || !target_gd) {
			cerr << "Error in MatchCandidates::GraphCutRelabel(): Please precompute edge data for source and target meshes!" << endl;
			return;
		}
	}
	else {
		// Not implemented yet.  Abort.
		cerr << "Error in MatchCandidates::GraphCutRelabel(): Currently only supports meshes." << endl;
		return;
	}

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Starting graph cut optimization (symmetric)..." << endl;

	// Write descriptions of the graph cut error function
	cout << "*** Smooth term type: edge-length preserve, label consistency" << endl;
	cout << "*** Data term type: ";
	if (graphCutDataTermType == GRAPH_CUT_DATA_POINT_TO_POINT) {
		cout << "point-to-point" << endl;
	}
	else if (graphCutDataTermType == GRAPH_CUT_DATA_POINT_TO_PLANE) {
		cout << "point-to-plane" << endl;
	}
	else if (graphCutDataTermType == GRAPH_CUT_DATA_HYBRID) {
		cout << "hybrid" << endl;
	}
	else if (graphCutDataTermType == GRAPH_CUT_DATA_D6) {
		cout << "point and normal" << endl;
	}

	//////////////////////////////////////////////////////////////////////////
	// 2. Set parameter values.
	cout << "*** Using dummy label: " << useDummyLabel << endl;

	cout << "*** Extra smoothness constraints: " << bdySmoothnessKNN << " nearest boundary vertices" << endl;

	cout << "*** Closest point maximum matching distance: " << closestPtDistThreshold << endl;
	this->closestPtDistThreshold = closestPtDistThreshold;

	cout << "*** Error for penalizing inconsistent labels: " << mismatchError << endl;
	this->mismatchError = mismatchError;

	cout << "*** Penalty for assigning the dummy label (no correspondence, smooth/data): (" << dummyLabelSmoothError << "/" << dummyLabelDataError << ")" << endl;
	this->dummyLabelSmoothError = dummyLabelSmoothError;
	this->dummyLabelDataError = dummyLabelDataError;

	cout << "*** Smoothness / data term weights: (" << smoothWeight << "/" << dataWeight << ")" << endl;
	this->smoothWeight = smoothWeight;
	this->dataWeight = dataWeight;

	cout << "*** Maximum smooth term / data term value: " << smoothTruncateValue << "/" << dataTruncateValue << endl;
	this->smoothTruncateValue = smoothTruncateValue;			/**< Set the maximum value for the smooth term. */
	this->dataTruncateValue = dataTruncateValue;				/**< Set the maximum value for the data term. */

	// Set various sizes
	const index_type siteOffset = sourceModel.numVertices();
	const int numSites = sourceModel.numVertices() + targetModel.numVertices();
	int numLabels = sortedMatchList.size();
	if (useDummyLabel) {
		numLabels++;
	}

	cout << "*** Number of sites/labels: " << numSites << "/" << numLabels << endl;
	cout << "*** Source model vertices / Target model vertices: " << sourceModel.numVertices() << "/" << targetModel.numVertices() << endl;

	// Precompute closest points
	precomputeClosestPts();

	/////////////////////////////////////////////////////////////////////////
	// 3. Perform graph cut optimization.
	// We partition the vertices into two groups:
	// 1. The source model vertices (total n vertices) ( indices 0...(n-1) )
	// 2. The target model vertices (total m vertices) ( indices n...(n+m-1) )
	//////////////////////////////////////////////////////////////////////////
	GCoptimizationGeneralGraph *gc = new GCoptimizationGeneralGraph(numSites, numLabels);

	try {
		//////////////////////////////////////////////////////////////////////////
		// Use function pointers for the data and smoothness costs
		if (graphCutDataTermType == GRAPH_CUT_DATA_POINT_TO_POINT) {
			gc->setDataCost(&GraphCutDataFn_PointToPoint, this);
		}
		else if (graphCutDataTermType == GRAPH_CUT_DATA_POINT_TO_PLANE) {
			gc->setDataCost(&GraphCutDataFn_PointToPlane, this);
		}
		else if (graphCutDataTermType == GRAPH_CUT_DATA_HYBRID) {
			gc->setDataCost(&GraphCutDataFn_Hybrid, this);
		}
		else if (graphCutDataTermType == GRAPH_CUT_DATA_D6) {
			gc->setDataCost(&GraphCutDataFn_D6, this);
		}
		else {
			cerr << "Unknown data term type " << graphCutDataTermType << "! Aborting graph cut optimization." << endl;
			return;
		}
		gc->setSmoothCost(&GraphCutSmoothFn, this);

		//////////////////////////////////////////////////////////////////////////
		// Set up edges
		shapeBoundaryKNNEdges(sourceModel, bdySmoothnessKNN, sourceExtraEdges);
		shapeBoundaryKNNEdges(targetModel, bdySmoothnessKNN, targetExtraEdges);
		makeGraphEdges(gc, source_gd, target_gd, sourceExtraEdges, targetExtraEdges);

		//////////////////////////////////////////////////////////////////////////
		// Perform optimization
		real_type energy = gc->compute_energy();
		cout << "Before optimization energy is: " << energy << endl;
		if (graphCutSolutionType == GRAPH_CUT_SWAP) {
			cout << "Solving graph cut using swap..." << endl;
			energy = gc->swap();
		}
		else if (graphCutSolutionType == GRAPH_CUT_EXPANSION) {
			cout << "Solving graph cut using expansion..." << endl;

			if (useDummyLabel) {
				// Try assigning dummy label first
				gc->alpha_expansion(numLabels-1);
			}
			energy = gc->expansion();
		}
		else {
			cout << "Unknown solution type " << graphCutSolutionType << "!" << endl;
		}
		cout << "After optimization energy is: " << gc->compute_energy() << endl;

		//////////////////////////////////////////////////////////////////////////
		// 1. Store the labels
		graphCutLabels_source.resize(sourceModel.numVertices());
		graphCutLabels_target.resize(targetModel.numVertices());
		for ( index_type i = 0; i < sourceModel.numVertices(); i++ ) {
			graphCutLabels_source[i] = gc->whatLabel(i);
		}
		for ( index_type i = 0; i < targetModel.numVertices(); i++ ) {
			graphCutLabels_target[i] = gc->whatLabel(siteOffset + i);
		}

		//////////////////////////////////////////////////////////////////////////
		// 2. Read the result: Source mesh
		std::vector<CorrespondenceList> reLabeledMatchList;

		reLabeledMatchList.clear();
		reLabeledMatchList.resize(sortedMatchList.size());
		for ( index_type i = 0; i < sourceModel.numVertices(); i++ ) {
			index_type label = gc->whatLabel(i);

			// Add the correspondence for non-dummy labels
			real3_type v = sourceModel.vertex[i];
			if (!GraphCutIsDummyLabel(label)) {
				const SE3 &tf = sortedTFList[label];
				reLabeledMatchList[label].add(v, tf*v);
			}
		}
		storeSortedMatchList(reLabeledMatchList, sortedTFList, sortedMatchListFinal_source, sortedTFListFinal_source);

		// 3. Read the result: Target mesh
		reLabeledMatchList.clear();
		reLabeledMatchList.resize(sortedMatchList.size());
		for ( index_type i = 0; i < targetModel.numVertices(); i++ ) {
			index_type label = gc->whatLabel(siteOffset + i);

			// Add the correspondence for non-dummy labels
			real3_type v = targetModel.vertex[i];
			if (!GraphCutIsDummyLabel(label)) {
				const SE3 &tf = sortedInverseTFList[label];
				reLabeledMatchList[label].add(tf*v, v);
			}
		}
		storeSortedMatchList(reLabeledMatchList, sortedInverseTFList, sortedMatchListFinal_target, sortedInverseTFListFinal_target);
	}
	catch (GCException e) {
		e.Report();
	}

	// Free memory associated with graph cut data structure
	delete gc;

	//// Print stats
	//for (index_type i = 0; i < sortedMatchListFinal.size(); i++) {
	//	cout << "Final match list size: list " << i << ", size " << sortedMatchListFinal[i].size() << endl;
	//}

	cout << "...done!" << endl;
}


void MatchCandidates::saveGraphCutResult( const char* filename ) const
{
	ofstream out;
	out.open(filename, ios_base::binary);

	if (out.fail()) {
		cout << "Error in saveGraphCutResult(): could not open file " << filename << " for writing!" << endl;
		return;
	}

	cout << "Saving graph cut result...";

	//////////////////////////////////////////////////////////////////////////
	// Save the list of all the transformations
	unsigned int numTransforms = matchList.size();
	out.write((char*) &numTransforms, sizeof(unsigned int));

	for (index_type i = 0; i < numTransforms; i++) {
		// Store transformation
		tfList[i].saveBinary(out);

		// Store correspondence list
		matchList[i].saveBinary(out);
	}

	//////////////////////////////////////////////////////////////////////////
	// Save the minimum list size parameter
	out.write((char*) &minVerifiedRegionSize, sizeof(unsigned int));
	out.write((char*) &maxNumVerifiedRegions, sizeof(int));

	//////////////////////////////////////////////////////////////////////////
	// Save the graph cut label results
	unsigned int numSourcePoints = graphCutLabels_source.size();
	unsigned int numTargetPoints = graphCutLabels_target.size();
	out.write((char*) &numSourcePoints, sizeof(unsigned int));
	out.write((char*) &numTargetPoints, sizeof(unsigned int));

	// Source shape assignments
	for (index_type i = 0; i < numSourcePoints; i++) {
		unsigned int label = graphCutLabels_source[i];
		out.write((char*) &label, sizeof(unsigned int));
	}

	// Target shape assignments
	for (index_type i = 0; i < numTargetPoints; i++) {
		unsigned int label = graphCutLabels_target[i];
		out.write((char*) &label, sizeof(unsigned int));
	}

	out.close();

	cout << "done!" << endl;
}

void MatchCandidates::loadGraphCutResult( const char* filename )
{
	ifstream in;
	in.open(filename, ios_base::binary);

	if (in.fail()) {
		cout << "Error in loadGraphCutResult(): could not open file " << filename << " for reading!" << endl;
		return;
	}

	cout << "Loading graph cut result...";

	// Clear existing data
	matchList.clear();
	tfList.clear();
	graphCutLabels_source.clear();
	graphCutLabels_target.clear();

	//////////////////////////////////////////////////////////////////////////
	// Read the list of all the transformations
	unsigned int numTransforms;
	in.read((char*) &numTransforms, sizeof(unsigned int));
	matchList.resize(numTransforms);
	tfList.resize(numTransforms);

	for (index_type i = 0; i < numTransforms; i++) {
		// Read transformation
		tfList[i].loadBinary(in);

		// Read correspondence list
		matchList[i].loadBinary(in);
	}

	//////////////////////////////////////////////////////////////////////////
	// Read the minimum list size parameter
	in.read((char*) &minVerifiedRegionSize, sizeof(unsigned int));
	in.read((char*) &maxNumVerifiedRegions, sizeof(int));

	cout << " min verified region size: " << minVerifiedRegionSize;
	cout << " max number of verified regions: " << maxNumVerifiedRegions;

	// Produce the sorted match list
	sortVerifyResult(minVerifiedRegionSize, maxNumVerifiedRegions);

	//////////////////////////////////////////////////////////////////////////
	// Read the graph cut label results
	unsigned int numSourcePoints, numTargetPoints;
	in.read((char*) &numSourcePoints, sizeof(unsigned int));
	in.read((char*) &numTargetPoints, sizeof(unsigned int));
	graphCutLabels_source.resize(numSourcePoints);
	graphCutLabels_target.resize(numTargetPoints);

	// Source shape assignments
	for (index_type i = 0; i < numSourcePoints; i++) {
		unsigned int label;
		in.read((char*) &label, sizeof(unsigned int));
		graphCutLabels_source[i] = label;
	}

	// Target shape assignments
	for (index_type i = 0; i < numTargetPoints; i++) {
		unsigned int label;
		in.read((char*) &label, sizeof(unsigned int));
		graphCutLabels_target[i] = label;
	}

	in.close();

	cout << " ...done!" << endl;
}

void MatchCandidates::shapeBoundaryKNNEdges( ShapeModel &shape, int k, EdgeList &edgeList )
{
	cout << "Creating extra smoothness edges on boundary of shape...";

	// 1. Make a list of all the points on the boundary
	idxList originalIndex;
	PointList boundaryPts;

	for (index_type i = 0; i < shape.numVertices(); i++) {
		if (shape.markVertex[i] == VERTEX_ON_BOUNDARY) {
			originalIndex.push_back(i);
			boundaryPts.add(shape.vertex[i]);
		}
	}

	// 2. For each boundary point, connect to its k-nearest neighbors
	edgeList.clear();
	for (index_type i = 0; i < boundaryPts.size(); i++) {
		real3_type q = boundaryPts[i];

		idxList inRange;
		distList dists;

		boundaryPts.queryKNN(q, k+1, inRange, dists, true);

		for (index_type j = 1; j < inRange.size(); j++) {
			// Create an edge from i to j
			Edge e;
			e.vi[0] = originalIndex[i];
			e.vi[1] = originalIndex[inRange[j]];
			edgeList.push_back(e);
		}
	}

	cout << " done!" << endl;
	cout << "Created " << edgeList.size() << " extra edges!" << endl;
}

real_type MatchCandidates::maxSymmetricHausdorff()
{
	// Check that graph cut result exists
	if (!graphCutLabels_source.size()) {
		return REAL_TYPE_MAX;
	}

	// Check that both shapes are meshes
	if ((sourceModel.shapeType != SHAPE_MESH) || (targetModel.shapeType != SHAPE_MESH)) {
		cerr << "Error in MatchCandidates::maxSymmetricHausdorff(): only implemented for meshes!" << endl;
		return REAL_TYPE_MAX;
	}

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Computing the registration error of all vertices in the shape...";
	real_type maxError1 = -REAL_TYPE_MAX, maxError2 = -REAL_TYPE_MAX;

	//////////////////////////////////////////////////////////////////////////
	// 1. From transformed source to target
	cout << endl << "*** Processing forward...";
	const Mesh3_real &compShapeTarget = (Mesh3_real&)targetModel;
	TextProgress(sourceModel.numVertices(), true);
	for (index_type i = 0; i < sourceModel.numVertices(); i++) {
		SE3 tf = sortedTFList[graphCutLabels_source[i]];
		real3_type ref = tf * sourceModel.vertex[i];
		real3_type v1, v2, v3, closest;
		real_type minDist = REAL_TYPE_MAX;

		// Compute the distance to all the triangles in the target mesh; find the minimum
		for(size_type t = 0,t_end = compShapeTarget.triangle_vertex_index.size();t < t_end;t++){
			const index3_type& tvi = compShapeTarget.triangle_vertex_index[t];
			v1 = compShapeTarget.vertex[tvi[0]];
			v2 = compShapeTarget.vertex[tvi[1]];
			v3 = compShapeTarget.vertex[tvi[2]];

			real_type dist = pointTriangleSqDistance(ref, v1, v2, v3, closest);

			if (dist < minDist) {
				minDist = dist;
			}
		}

		// Keep track of maximum minDist
		if (minDist > maxError1) {
			maxError1 = minDist;
		}
		TextProgress(i);
	}

	//////////////////////////////////////////////////////////////////////////
	// 2. From target to transformed source
	cout << endl << "*** Processing backward...";
	const Mesh3_real &compShapeSource = (Mesh3_real&)sourceModel;
	TextProgress(targetModel.numVertices(), true);
	for (index_type i = 0; i < targetModel.numVertices(); i++) {
		real3_type ref = targetModel.vertex[i];
		real3_type v1, v2, v3, closest;
		real_type minDist = REAL_TYPE_MAX;
		SE3 tf;

		// Compute the distance to all the triangles in the target mesh; find the minimum
		for(size_type t = 0,t_end = compShapeSource.triangle_vertex_index.size();t < t_end;t++){
			const index3_type& tvi = compShapeSource.triangle_vertex_index[t];

			tf = sortedTFList[graphCutLabels_source[tvi[0]]];
			v1 = tf * compShapeSource.vertex[tvi[0]];
			tf = sortedTFList[graphCutLabels_source[tvi[1]]];
			v2 = tf * compShapeSource.vertex[tvi[1]];
			tf = sortedTFList[graphCutLabels_source[tvi[2]]];
			v3 = tf * compShapeSource.vertex[tvi[2]];

			real_type dist = pointTriangleSqDistance(ref, v1, v2, v3, closest);

			if (dist < minDist) {
				minDist = dist;
			}
		}

		// Keep track of maximum minDist
		if (minDist > maxError2) {
			maxError2 = minDist;
		}
		TextProgress(i);
	}

	//////////////////////////////////////////////////////////////////////////
	cout << "done!" << endl;

	// Take the square root to get actual distance
	maxError1 = sqrt(maxError1);
	maxError2 = sqrt(maxError2);

	// Display result
	cout << "*** Hausdorff distance (tf*source --> target): " << maxError1 << endl;
	cout << "*** Hausdorff distance (target --> tf*source): " << maxError2 << endl;

	return max(maxError1, maxError2);
}


void MatchCandidates::packageLabels( std::vector<SE3> &tfList, std::vector<int> &vertexLabels, CorrespondenceList &clist, int minUsedCount/* = 0*/ )
{
	// Print out which labels were used for how many vertices
	// Check if graph cut labels exist
	if (!sortedTFList.size() || !sortedMatchList.size() || !graphCutLabels_source.size()) {
		cerr << "In MatchCandidates::packageLabels(): no graph cut result exists!" << endl;
		return;
	}

	// Total possible labels
	const int maxLabels = sortedTFList.size();
	vector< pair<int,int> > verticesPerLabel;
	verticesPerLabel.resize(maxLabels);
	for (int i = 0; i < verticesPerLabel.size(); i++) {
		verticesPerLabel[i] = pair<int,int>(0,i);
	}

	// Count the number of vertices having the label
	for (index_type i = 0; i < graphCutLabels_source.size(); i++) {
		verticesPerLabel[graphCutLabels_source[i]].first++;
	}

	// Sort in descending order
	struct sizeComp : public binary_function<pair<int,int>, pair<int,int>, bool> {
		bool operator()(pair<int,int> x, pair<int,int> y) {
			return (x.first > y.first);
		}
	};
	sort(verticesPerLabel.begin(), verticesPerLabel.end(), sizeComp());

	// Print out the result
	int totalNonZeroLabels = 0;
	for (int i = 0; i < verticesPerLabel.size(); i++) {
		if (verticesPerLabel[i].first == minUsedCount) {
			break;
		}
		cout << "Label " << verticesPerLabel[i].second << ": " << verticesPerLabel[i].first << " vertices" << endl;
		totalNonZeroLabels++;
	}
	cout << "Total labels used: " << totalNonZeroLabels << endl;


	//////////////////////////////////////////////////////////////////////////
	// Package the result
	if (graphCutLabels_source.size() != sourceModel.numVertices()) {
		cerr << "Error in MatchCandidates::packageLabels(): number of labels doesn't match number of vertices!" << endl;
		return;
	}

	// 1. List of transformations
	tfList.resize(totalNonZeroLabels);
	for (int i = 0; i < tfList.size(); i++) {
		tfList[i] = sortedTFList[verticesPerLabel[i].second];
	}

	// 2. List of assignments: which label for each vertex?
	std::vector<int> translateLabel;
	translateLabel.resize(maxLabels);
	for (int i = 0; i < maxLabels; i++) {
		translateLabel[i] = -1;
	}
	for (int i = 0; i < verticesPerLabel.size(); i++) {
		if (verticesPerLabel[i].first == minUsedCount) {
			break;
		}
		translateLabel[verticesPerLabel[i].second] = i;
	}
	vertexLabels.resize(graphCutLabels_source.size());
	for (int i = 0; i < graphCutLabels_source.size(); i++) {
		vertexLabels[i] = translateLabel[graphCutLabels_source[i]];
	}

	// 3. List of correspondences
	clist.clear();
	clist.reserve(sourceModel.numVertices());
	for (int i = 0; i < sourceModel.numVertices(); i++) {
		// Check if the vertex has a label, is not on the boundary and not isolated
		if ((vertexLabels[i] < 0) ||
			(sourceModel.markVertex[i] != VERTEX_VALID)) {
			continue;
		}

		SE3 &tf = tfList[vertexLabels[i]];
		real3_type v = tf * sourceModel.vertex[i];

		// Find closest point in target model
		ClosestPointInfo cpi = targetModel.findClosestPoint(v);

		// Check if the vertex is not on the boundary and not isolated
		if (targetModel.markVertex[cpi.index] == VERTEX_VALID) {
			clist.add(sourceModel.vertex[i], targetModel.vertex[cpi.index], i, cpi.index);
		}
	}
}



/**
 * Save the local frames at each vertex of a mesh to a binary file.  For reading in MATLAB.
 */
void saveMeshFrames( const char *filename, const ShapeModel* mesh, const CurvInfoList &cl )
{
	float temp;
	ofstream out;
	out.open(filename, ios_base::binary);

	if (out.fail()) {
		cout << "Error in saveFrames(): could not open file " << filename << " for writing!" << endl;
		return;
	}

	if (!mesh || (cl.size() == 0)) {
		cout << "Error in saveFrames(): No mesh / curvature information found!" << endl;
		out.close();
		return;
	}

	if (mesh->vertex.size() != cl.size()) {
		cout << "Number of vertices in mesh does not match curvature info!" << endl;
		out.close();
		return;
	}

	cout << "Saving per-vertex frames...";

	//////////////////////////////////////////////////////////////////////////
	// Write the data
	unsigned int numVertices = cl.size();
	unsigned int numScales = cl[0].size();
	out.write((char*) &numVertices, sizeof(unsigned int));
	out.write((char*) &numScales, sizeof(unsigned int));

	for (index_type i = 0; i < numVertices; i++) {
		// Write the current vertex
		for (int c = 0; c < 3; c++) {
			temp = mesh->vertex[i][c];
			out.write((char*) &temp, sizeof(float));
		}

		for (index_type j = 0; j < numScales; j++) {
			CurvInfo ci = cl[i][j];

			for (int c = 0; c < 3; c++) {
				temp = ci.n[c];
				out.write((char*) &temp, sizeof(float));
			}
			for (int c = 0; c < 3; c++) {
				temp = ci.vmax[c];
				out.write((char*) &temp, sizeof(float));
			}
			for (int c = 0; c < 3; c++) {
				temp = ci.vmin[c];
				out.write((char*) &temp, sizeof(float));
			}
			temp = ci.kmax;
			out.write((char*) &temp, sizeof(float));
			temp = ci.kmin;
			out.write((char*) &temp, sizeof(float));
		}
	}

	out.close();

	cout << "done!" << endl;
}

SE3::SE3() : rot(Geometry::Matrix33f::identity()), tr(0,0,0)
{

}

SE3::SE3( Geometry::Matrix33f R, real3_type t ) : rot(R), tr(t)
{

}

void SE3::setIdentity()
{
	rot = Geometry::Matrix33f::identity();
	tr[0] = tr[1] = tr[2] = 0.0;
}

void SE3::setZero()
{
	for (int i = 0; i < dim; i++) {
		tr[i] = 0.0;
		for (int j = 0; j < dim; j++) {
			rot(i,j) = 0.0;
		}
	}
}

data_type SE3::log() const
{
	// Compute axis-angle representation
	data_type result(twistDim);
	real3_type axis;
	real_type angle;

	// Convert rotation to axis-angle representation
	toAxisAngle(rot, axis, angle);

	if (abs(angle) < REAL_TYPE_EPSILON) {
		// Assign linear velocity = translation & zero angular velocity
		for (int d = 0; d < dim; d++) {
			result.val[d] = tr[d];
			result.val[d+dim] = 0.0;
		}
	}
	else {
		// Debugging: check to see if the axis is unit length
		//real_type axisNorm = axis.norm();
		//if ((axisNorm < (1.0 - REAL_TYPE_EPSILON)) || (axisNorm > (1.0 + REAL_TYPE_EPSILON))) {
		//	cerr << "Warning in SE3::log(): Axis is not unit length! Length = " << axisNorm << "." << endl;
		//}

		// Given the translation, solve for the linear velocity.
		real3_type v = tr;
		Geometry::Matrix33f axisop, ssmAxis, temp;
		temp = -rot;
		for (int d = 0; d < dim; d++) {
			temp(d,d) += 1.0;
		}
		ssmAxis(0,0) = 0.0;      ssmAxis(0,1) = -axis[2]; ssmAxis(0,2) = axis[1];
		ssmAxis(1,0) = axis[2];  ssmAxis(1,1) = 0.0;      ssmAxis(1,2) = -axis[0];
		ssmAxis(2,0) = -axis[1]; ssmAxis(2,1) = axis[0];  ssmAxis(2,2) = 0.0;
		axisop(0,0) = axis[0]*axis[0];
		axisop(1,1) = axis[1]*axis[1];
		axisop(2,2) = axis[2]*axis[2];
		axisop(1,0) = axisop(0,1) = axis[1]*axis[0];
		axisop(2,0) = axisop(0,2) = axis[2]*axis[0];
		axisop(1,2) = axisop(2,1) = axis[1]*axis[2];
		temp = (temp * ssmAxis) + angle*axisop;

		// Solve the system
		General_Solve(temp, v);

		// Assign linear velocity & angular velocity
		for (int d = 0; d < dim; d++) {
			result.val[d] = angle*v[d];
			result.val[d+dim] = angle*axis[d];
		}
	}

	return result;
}

void SE3::applyGL() const
{
	static float m[16];
	OpenGLMatrix(m);
	glMultMatrixf(m);
}

void SE3::OpenGLMatrix( float *o ) const
{
	o[0] = rot(0,0); o[4] = rot(0,1);  o[8] = rot(0,2); o[12] = tr[0];
	o[1] = rot(1,0); o[5] = rot(1,1);  o[9] = rot(1,2); o[13] = tr[1];
	o[2] = rot(2,0); o[6] = rot(2,1); o[10] = rot(2,2); o[14] = tr[2];
	o[3] = 0.0;      o[7] = 0.0;      o[11] = 0.0;      o[15] = 1.0;
}

void SE3::loadBinary( std::ifstream &in )
{
	readBinary(in, rot);
	readBinary(in, tr);
}

void SE3::saveBinary( std::ofstream &out ) const
{
	writeBinary(out, rot);
	writeBinary(out, tr);
}

/**
 * Axis-angle conversion from matrix.  Assumes that the rotation is determinant +1.
 * Grabbed from http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToAngle/index.htm.
 */
void toAxisAngle( const Geometry::Matrix33f R, real3_type &axis, real_type &angle )
{
	const real_type epsilon = 0.1;
	const real_type eepsilon = 0.01;
	const real_type ax = R(2,1)-R(1,2);
	const real_type ay = R(0,2)-R(2,0);
	const real_type az = R(1,0)-R(0,1);
	const real_type trace = R(0,0) + R(1,1) + R(2,2);

	// If the matrix is symmetric
	if ((abs(ax) < eepsilon) && (abs(ay) < eepsilon) && (abs(az) < eepsilon)) {

		// If the matrix is identity
		if ((abs(R(2,1)+R(1,2)) < epsilon) &&
			(abs(R(0,2)+R(2,0)) < epsilon) &&
			(abs(R(1,0)+R(0,1)) < epsilon) &&
			(abs(trace - 3.0) < epsilon)) {
			axis[0] = 1.0;
			axis[1] = 0.0;
			axis[2] = 0.0;
			angle = 0.0;
		}
		else {
			// Otherwise, this matrix is a 180 degree rotation
			angle = M_PI;

			for (int d = 0; d < 3; d++) {
				axis[d] = (R(d,d) + 1.0) / 2.0;
				if (axis[d] > 0)
					axis[d] = sqrt(axis[d]);
				else
					axis[d] = 0.0;
			}

			bool xZero = (abs(axis[0]) < eepsilon);
			bool yZero = (abs(axis[1]) < eepsilon);
			bool zZero = (abs(axis[2]) < eepsilon);
			bool xyPositive = (R(0,1) > 0);
			bool xzPositive = (R(0,2) > 0);
			bool yzPositive = (R(1,2) > 0);
			if (xZero && !yZero && !zZero) {
				if (!yzPositive)
					axis[1] = -axis[1];
			} else if (yZero && !zZero) {
				if (!xzPositive)
					axis[2] = -axis[2];
			} else if (zZero) {
				if (!xyPositive)
					axis[0] = -axis[0];
			}
		}
	}
	else {
		real_type s = sqrt(ax*ax + ay*ay + az*az);

		if (abs(s) < eepsilon)
			s = 1.0;

		s = 1.0 / s;
		axis[0] = ax * s;
		axis[1] = ay * s;
		axis[2] = az * s;
		angle = acos((trace - 1.0) / 2.0);
	}
}

real_type det( Geometry::Matrix33f M )
{
	real_type plus = M(0,0)*M(1,1)*M(2,2) + M(0,1)*M(1,2)*M(2,0) + M(0,2)*M(1,0)*M(2,1);
	real_type minus = M(0,0)*M(1,2)*M(2,1) + M(0,1)*M(1,0)*M(2,2) + M(0,2)*M(1,1)*M(2,0);
	return plus - minus;
}

SE3 exp( data_type twist )
{
	// Extract linear and angular velocity.
	real_type angle;
	real3_type v, w, t;
	for (int d = 0; d < 3; d++) {
		v[d] = twist.val[d];
		w[d] = twist.val[d+3];
	}
	angle = w.norm();
	if (angle < REAL_TYPE_EPSILON) {
		// If the angle is zero, return identity matrix
		return SE3();
	}
	w = w / angle;
	v = v / angle;

	// Compute result
	Geometry::Matrix33f R, ssm, axisop;
	ssm(0,0) = 0.f;   ssm(0,1) = -w[2]; ssm(0,2) = w[1];
	ssm(1,0) = w[2];  ssm(1,1) = 0.f;   ssm(1,2) = -w[0];
	ssm(2,0) = -w[1]; ssm(2,1) = w[0];  ssm(2,2) = 0.f;
	axisop(0,0) = w[0]*w[0];
	axisop(1,1) = w[1]*w[1];
	axisop(2,2) = w[2]*w[2];
	axisop(1,0) = axisop(0,1) = w[1]*w[0];
	axisop(2,0) = axisop(0,2) = w[2]*w[0];
	axisop(1,2) = axisop(2,1) = w[1]*w[2];

	R = ssm*sin(angle) + ssm*ssm*(1.f - cos(angle));
	t = (-R*ssm + angle*axisop) * v;

	// Assign result
	for (int d = 0; d < 3; d++) {
		R(d,d) = R(d,d) + 1;
	}
	return SE3(R, t);
}

std::ostream& operator<<( std::ostream& out, const SE3 &tf )
{
	// Print rotation and translation
	out << std::endl << tf.rot << std::endl << tf.tr << std::endl;
	return out;
}


SE3 registerQuat( const PointList &left, const PointList &right )
{
	real3_type leftCenter, rightCenter;
	Geometry::Matrix33f M;
	size_type n = left.size();

	if (!left.size() || (left.size() != right.size())) {
		cerr << "Error in registerQuat(): No points, or number of points b/t sets do not match!" << endl;
		return SE3();
	}

	//////////////////////////////////////////////////////////////////////////
	// 1. Compute statistics
	for (index_type i = 0; i < n; i++) {
		leftCenter += left[i];
		rightCenter += right[i];
		for (index_type j = 0; j < 3; j++) {
			for (index_type k = 0; k < 3; k++) {
				M(j,k) += left[i][j]*right[i][k];
			}
		}
	}
	leftCenter = leftCenter / ((real_type)n);
	rightCenter = rightCenter / ((real_type)n);
	for (index_type j = 0; j < 3; j++) {
		for (index_type k = 0; k < 3; k++) {
			M(j,k) -= ((real_type)n) * leftCenter[j]*rightCenter[k];
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// 2. Compute eigenvalues
	Geometry::Matrix44f N;
	N(0,0) =   M(0,0) + M(1,1) + M(2,2);
	N(1,1) =   M(0,0) - M(1,1) - M(2,2);
	N(2,2) = - M(0,0) + M(1,1) - M(2,2);
	N(3,3) = - M(0,0) - M(1,1) + M(2,2);
	N(0,1) = N(1,0) = M(1,2) - M(2,1);
	N(0,2) = N(2,0) = M(2,0) - M(0,2);
	N(0,3) = N(3,0) = M(0,1) - M(1,0);
	N(1,2) = N(2,1) = M(0,1) + M(1,0);
	N(1,3) = N(3,1) = M(2,0) + M(0,2);
	N(2,3) = N(3,2) = M(1,2) + M(2,1);

	// Compute the eigenvector
	double *evals = new double[4];
	double **evecs = new double*[4];
	for (int i = 0; i < 4; i++) {
		evecs[i] = new double[4];
	}
	N.eigSymmetric(evals, evecs);

	// Find the index of the maximum eigenvalue
	double maxEval = -DBL_MAX;
	int maxIndex = -1;
	for (int i=0; i<4; i++) {
		if (evals[i] > maxEval) {
			maxIndex = i;
			maxEval = evals[i];
		}
		//cout << "Eigenvalue  " << i << ": " << evals[i] << "\n";
		//cout << "Eigenvector " << i << ": ("
		//	<< evecs[0][i] << ", "
		//	<< evecs[1][i] << ", "
		//	<< evecs[2][i] << ", "
		//	<< evecs[3][i] << ")\n";
	}

	//////////////////////////////////////////////////////////////////////////
	// 3. Construct rotation matrix
	SE3 tf;

	// 1. Compute the rotation according to the maximum eigenvalue's eigenvector
	real_type q0, qx, qy, qz;
	real4_type q;
	for (int i = 0; i < 4; i++) q[i] = evecs[i][maxIndex];
	q.normalize();
	q0 = q[0]; qx = q[1]; qy = q[2]; qz = q[3];

	tf.rot(0,0) = q0*q0 + qx*qx - qy*qy - qz*qz;
	tf.rot(1,1) = q0*q0 - qx*qx + qy*qy - qz*qz;
	tf.rot(2,2) = q0*q0 - qx*qx - qy*qy + qz*qz;
	tf.rot(0,1) = 2.0*(qx*qy - q0*qz);
	tf.rot(1,0) = 2.0*(qy*qx + q0*qz);
	tf.rot(0,2) = 2.0*(qx*qz + q0*qy);
	tf.rot(2,0) = 2.0*(qz*qx - q0*qy);
	tf.rot(1,2) = 2.0*(qy*qz - q0*qx);
	tf.rot(2,1) = 2.0*(qz*qy + q0*qx);

	// 2. Compute the translation as the right centroid minus the rotated left centroid
	tf.tr = rightCenter - (tf.rot * leftCenter);

	//////////////////////////////////////////////////////////////////////////
	//// Debugging: Print matrices
	//cout << "M: " << endl << M << endl;
	//cout << "N: " << endl << N << endl;
	//cout << "(R,t): " << endl << tf << endl;

	//////////////////////////////////////////////////////////////////////////
	// Delete temporary variables
	delete[] evals;
	for (int i = 0; i < 4; i++) {
		delete[] evecs[i];
	}
	delete[] evecs;

	return tf;
}

SE3 inv( const SE3 &tf )
{
	// Take the inverse of a rigid transformation
	SE3 inverse;
	inverse.rot = tf.rot.transpose();
	inverse.tr = - inverse.rot * tf.tr;

	return inverse;
}
