#ifndef __FIND_CORRESPONDENCE_H__
#define __FIND_CORRESPONDENCE_H__

#include "PointListN.h"
#include "Correspondence.h"
#include "ShapeModel.h"
#include "geodesic.h"

class GCoptimizationGeneralGraph;

template<int N,typename Real>
void Cholesky_Solve( const Geometry::Square_matrix<N,Real> &M, Geometry::Vec<N, Real> &v );
template<int N,typename Real>
void General_Solve( const Geometry::Square_matrix<N,Real> &M, Geometry::Vec<N, Real> &v );

real_type det(Geometry::Matrix33f M);

/**
 * Fit a sphere given points and normals.  If the fit is degenerate, then radSq is set to the 
 * distance to the origin and center contains the normal of the plane fitting the points.
 * @param pts      A PointList containing the points to fit.
 * @param normals  A PointList containing the corresponding normals to fit.
 * @param center   This is both input and output; input: center of the weighting kernel, output: center of the fitted sphere
 * @param radSq    This is both input and output; input: radius^2 of the weighting kernel, output: radius^2 of the fitted sphere
 * @return	       This returns false if the fit was degenerate .
 */
bool FitSphere( const PointList &pts, const PointList &normals, real3_type &center, real_type &radSq );


/**
 * Compute per-vertex principal curvatures using the method of [KSN*07].
 */
bool FitCurvatureTensorLS( const PointList &pts, const PointList &normals, real3_type u, real3_type v, CurvInfo &result );
bool FitCurvatureTensorRobust( const PointList &pts, const PointList &normals, real3_type u, real3_type v, CurvInfo &result );
QString ComputePrincipalCurvatures( PointList &vertices, const PointList &normals, const distList &scales, CurvInfoList &cl, bool robust );

/**
 * Save the local frames at each vertex of a mesh to a binary file.  For reading in MATLAB.
 */
void saveMeshFrames( const char *filename, const ShapeModel* mesh, const CurvInfoList &cl );

/**
 * Compute the mean curvature at the query vertex given the points and the normals.
 * @param query       The query point at which to compute the curvature.
 * @param allPoints   List of all points.
 * @param allNormals  List of all normals.
 * @param scales      The list of scales (squared radius for the weight function) at which to compute the curvature.
 */
data_type ComputeMeanCurvature( const real3_type &query, PointList& allPoints, PointList& allNormals, const distList &scales );

/**
 * Returns argument x such that the sum of the weight between [0,x] constitutes 90% of the total weight.
 * This is used to scale visualizations of the weight function as a sphere.
 */
const real_type FitSphereWeight90Factor = 0.5;
const real_type FitSphereWeight99Factor = 0.708;

idxList randomSample( int n, int max );


// Implement Lie groups and algebras

/**
 * Element of the Lie group SE(3).
 */
class SE3 {
public:
	SE3();
	SE3(Geometry::Matrix33f R, real3_type t);

	void setIdentity();
	void setZero();
	data_type log() const;
	inline real3_type operator*( const real3_type &x ) const {
		return rot*x + tr;
	}

	/**
	* Construct a 4x4 OpenGL matrix corresponding to this transformation.
	*/
	void OpenGLMatrix( float *o ) const;
	void applyGL() const;

	Geometry::Matrix33f rot;
	real3_type tr;

	// Reading and writing binary data to file
	void loadBinary( std::ifstream &in );
	void saveBinary( std::ofstream &out ) const;

private:
	static const int dim = 3;
	static const int twistDim = 6;
};

typedef std::vector<SE3> RigidDeform;

SE3 inv( const SE3 &tf );
SE3 exp( data_type twist );
void toAxisAngle( const Geometry::Matrix33f R, real3_type &axis, real_type &angle );
std::ostream& operator<<( std::ostream& out, const SE3 &tf );
SE3 registerQuat( const PointList &left, const PointList &right );

inline SE3 operator*( const SE3 &lhs, const SE3 &rhs )
{
	SE3 result;
	result.rot = lhs.rot * rhs.rot;
	result.tr = lhs.rot*rhs.tr + lhs.tr;
	return result;
}

/**
 * Create an arbitrary orthonormal local coordinate frame with normal n.
 */
void LocalCoordinateFrame( real3_type n, real3_type &u, real3_type &v );


/**
 * Constants for checking if the graph cut vertex is from the source or target.
 */
enum CLASSIFY_GRAPH_CUT_VERTEX {
	GRAPH_CUT_FROM_SOURCE = 1,
	GRAPH_CUT_FROM_TARGET
};

/**
 * Different graph cut solution types: swap or expansion?
 */
enum GRAPH_CUT_SOLUTION_TYPE {
	GRAPH_CUT_SWAP = 1,
	GRAPH_CUT_EXPANSION
};

/**
 * Type of terms in the graph cut error function
 */
enum GRAPH_CUT_ERROR_TYPE {
	GRAPH_CUT_DATA_POINT_TO_POINT = 1,
	GRAPH_CUT_DATA_POINT_TO_PLANE,
	GRAPH_CUT_DATA_HYBRID,
	GRAPH_CUT_DATA_D6
};

/**
 * Type of graph cut instances
 */
enum GRAPH_CUT_INSTANCE_TYPE {
	GRAPH_CUT_INSTANCE_SYMMETRIC = 1,
	GRAPH_CUT_INSTANCE_FORWARD,
	GRAPH_CUT_INSTANCE_BACKWARD
};

/**
 * Keeps track of a list of correspondence candidates and the transformation associated with each candidate.
 * Intended use: First add each candidate using the function add(si, ti), which creates a list of SE(3) transformations.
 * After adding all candidates, run populateData() which will normalize the translations and create twist vectors.
 * Finally, run mean-shift on the twist vectors.
 */
class MatchCandidates {
public:
	MatchCandidates( ShapeModel &sm, ShapeModel &tm );
	~MatchCandidates();

	int spaceDim() const;
	void clearCandidates();
	void clearClusterResult();
	size_type size() const;
	void add( index_type si, index_type ti );
	void add( index_type si, index_type ti, SE3 tf );
	bool computeTransform( index_type si, index_type ti, SE3 &result );
	void normalizeTranslations();
	void printInfo() const;
	void writeData( const char *filename ) const;
	void populateData();

	//////////////////////////////////////////////////////////////////////////
	// Sub-sampling the geometry
	void clearSample();
	void subSampleGeometry( size_type ns, size_type nt, real_type pruneMaxRatio );

	//////////////////////////////////////////////////////////////////////////
	// Generating candidate correspondences
	void groundTruthCorrespondence();
	void findCandidateCorrespondence( real_type radSq, size_type maxMatches, real_type ANNEpsilon );
	void spinImageCandidates( size_type maxMatches );
	void RANSAC_Candidates( size_type max_itr, real_type sourceModel_ANNEpsilon, real_type targetModel_ANNEpsilon, real_type inlier_queryRadiusSq, real_type inlier_epsilonBandSq, real_type minimumInliersFraction );

	SE3 getPairTransform(index_type i) const;
	SE3 rescaleTransform( const SE3 &tf ) const;

	// Correspondence pair candidates
	idxList sourceIndex, targetIndex;				/**< List of the candidate correspondences.  This list contains indices into the SAMPLE OF THE MODEL, not the model itself. */

	// Transformation space
	PointListN transforms;							/**< List of twist vectors for each candidate correspondence */

	//////////////////////////////////////////////////////////////////////////
	// Mean-shift clustering
	bool doneClustering() const;
	void meanShiftBasic(real_type radSq, real_type rate, int maxItr, real_type ANNepsilon);
	void sortClusters();
	void pruneClusterGeometricConsistency( real_type T_gc );
	bool validCluster( index_type i ) const;

	idxList clusterLabels;							/**< Label for each candidate correspondence */
	PointListN clusterModes;						/**< Cluster mean for each cluster mode */
	std::vector<size_type> clusterSizes;			/**< Number of members for each cluster */

	//////////////////////////////////////////////////////////////////////////
	// Dense correspondence using pose clusters
	bool doneVerify() const;
	SE3 getClusterTransform( index_type i ) const;
	void skipVerify( const size_type minClusterSize, const int maxNumClusters );
	void verifyRegion( const size_type minClusterSize, const int maxNumClusters, const real_type growRegionDistThresholdSq, const real_type ANN_Epsilon );

	std::vector<CorrespondenceList> matchList;
	std::vector<SE3> tfList;

	void refineCluster( index_type ci );			/**< Refine the specified cluster using ICP */

	// Select clusters that contain certain vertices
	std::vector<bool> closestSampleIndices( ShapeModel &model, ShapeModel &sample, const idxList &modelIndices ) const;
	idxList selectPairsHavingVertices( const idxList &sourceSelected, const idxList &targetSelected ) const;
	idxList selectClustersHavingVertices( const idxList &sourceSelected, const idxList &targetSelected );
	idxList selectRegionsHavingVertices( const idxList &selected );

	// refineCluster debugging
	PointList closestPoints;

	//////////////////////////////////////////////////////////////////////////
	// Graph-cut based optimal labeling
	void sortVerifyResult( size_type minMatchSize, int maxMatchLists );
	void storeSortedMatchList(const std::vector<CorrespondenceList> &clistIN, const std::vector<SE3> &tListIN,
							  std::vector<CorrespondenceList> &clistOUT, std::vector<SE3> &tListOUT,
							  int minMatchSize = 0, int maxLists = -1 );
	index_type minVerifiedRegionSize;
	int maxNumVerifiedRegions;
	std::vector<CorrespondenceList> sortedMatchList;
	std::vector<SE3> sortedTFList, sortedInverseTFList;

	void precomputeClosestPts();									/**< For each transformation, pre-computes the closest point after applying the transformation (in both directions) */
	bool GraphCutIsDummyLabel( int l );								/**< Is the label the dummy label? */
	int GraphCutClassifyVertexFn( int &p );							/**< Modifies the input so it's a valid index into source/target shape. */
	static real_type GraphCutDataFn_PointToPlane( int p, int l, void *data );
	static real_type GraphCutDataFn_PointToPoint( int p, int l, void *data );
	static real_type GraphCutDataFn_Hybrid( int p, int l, void *data );
	static real_type GraphCutDataFn_D6( int p, int l, void *data );
	static real_type GraphCutSmoothFn( int p1, int p2, int l1, int l2, void *data );
	void shapeBoundaryKNNEdges( ShapeModel &shape, int k, EdgeList &edgeList );
	void makeGraphEdges( GCoptimizationGeneralGraph *gc, GeodesicDistance *source_gd, GeodesicDistance *target_gd, EdgeList sourceExtraEdges, EdgeList targetExtraEdges );
	void GraphCutRelabel( int graphCutSolutionType, int graphCutDataTermType, bool useDummyLabel, int bdySmoothnessKNN, real_type closestPtDistThreshold, real_type mismatchError, real_type dummyLabelSmoothError, real_type dummyLabelDataError, real_type smoothWeight, real_type dataWeight, real_type smoothTruncateValue, real_type dataTruncateValue );
	void GraphCutRelabel_Forward(int graphCutSolutionType, int graphCutDataTermType, real_type smoothWeight, real_type dataWeight, real_type smoothTruncateValue, real_type dataTruncateValue );

	EdgeList sourceExtraEdges, targetExtraEdges;
	std::vector< std::vector<ClosestPointInfo> > closestToSourcePt, closestToTargetPt;
	real_type closestPtDistThreshold;																/**< Graph cut error function parameter (distance threshold for closest point) */
	real_type mismatchError;																		/**< Graph cut error function parameter (error for non-consistent labelling b/t source and target meshes) */
	real_type dummyLabelDataError, dummyLabelSmoothError;											/**< Graph cut error function parameter (error for assigning a dummy label) */
	real_type smoothWeight, dataWeight;																/**< Graph cut error function parameters (weighting for data/smoothness terms) */
	real_type dataTruncateValue, smoothTruncateValue;												/**< Graph cut error function parameters (truncation for data/smoothness terms) */
	std::vector<index_type> graphCutLabels_source, graphCutLabels_target;							/**< Graph cut final labelling */
	std::vector<CorrespondenceList> sortedMatchListFinal_source, sortedMatchListFinal_target;		/**< Graph cut final result (list of correspondences) */
	std::vector<SE3> sortedTFListFinal_source, sortedInverseTFListFinal_target;						/**< Graph cut final result (list of transformations) */

	// Load/Save the correspondence result
	void loadGraphCutResult( const char* filename );
	void saveGraphCutResult( const char* filename ) const;

	// Compute the error of the registration
	real_type maxSymmetricHausdorff();

	// Package graph cut assignment result
	void packageLabels( std::vector<SE3> &tfList, std::vector<int> &vertexLabels, CorrespondenceList &clist, int minUsedCount = 0 );

	//////////////////////////////////////////////////////////////////////////
	// Geometry of the source and target models
	ShapeModel &sourceModel, &targetModel;			/**< Original models */
	ShapeModel *sourceSample, *targetSample;		/**< Random sampling of the original model */

private:
	// Dimension
	static const int dim = 3;
	static const int transformDim = 6;				/**< Dimension of the feature space; initialized in the derived classes */

	// Store transformations
	std::vector<SE3> transformsSE3;					/**< List of SE(3) transformations for each correspondence */
	real3_type trShift;								/**< Shift applied to transformations */
	real_type trScale;								/**< Scale applied to transformations */
};

#endif

