#include "PointList.h"
#include <SylvainTools/simple_mesh.h>
#include "util.h"
#include <ctime>

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;
using namespace Geometry;

PointList::PointList() :
kdTree(NULL), dataPts(NULL), queryEpsilon(0)
{

}

PointList::PointList( const PointList& rhs ) :
queryEpsilon(rhs.queryEpsilon), data(rhs.data), kdTree(NULL), dataPts(NULL)
{
}

PointList::~PointList()
{
	clear();
}

const PointList& PointList::operator=( const PointList & rhs )
{
	if ( this != &rhs ) {
		clear();

		// copy values
		this->queryEpsilon = rhs.queryEpsilon;
		this->data = rhs.data;
	}
	return *this;
}


void PointList::resize( size_type n )
{
	data.resize(n);
}

void PointList::reserve( size_type n )
{
	data.reserve(n);
}

void PointList::clear()
{
	clearANNData();
	data.clear();
}

void PointList::add( real3_type pt )
{
	data.push_back(pt);
}

const ptList* PointList::getData() const
{
	return &data;
}

/**
* Copy the given list of points.
*/
void PointList::setPts( const ptList &pts )
{
	clear();
	resize(pts.size());
	for (size_type i = 0, i_end = pts.size(); i < i_end; i++) {
		data[i] = pts[i];
	}
}

void PointList::removePoints( idxList list )
{
	// Sort the list according to the index number
	std::sort(list.begin(), list.end());

	// Traversing the list in reverse, remove elements from the array
	for (int i = list.size()-1; i >= 0; i--) {
		// Remove each element from the list
		index_type c = list[i];

		// Find the iterator at c
		ptList::iterator itr = data.begin();
		data.erase(itr+c);
	}
}

void PointList::draw( int type, real_type radius /*= 1 */ ) const
{
	// Draw the points in OpenGL.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < data.size(); i++) {
			drawSphere(data[i], radius);
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < data.size(); i++) {
			glVertex3f(data[i][0], data[i][1], data[i][2]);
		}
		glEnd();
	}
}

void PointList::drawWithNames( int type, real_type radius /*= 1 */ ) const
{
	// Draw the points in OpenGL.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < data.size(); i++) {
			glPushName(i);
			drawSphere(data[i], radius);
			glPopName();
		}
	}
	else if (type == POINT_SQUARES) {
		for (index_type i = 0; i < data.size(); i++) {
			glPushName(i);
			glRasterPos3f(data[i][0], data[i][1], data[i][2]);
			glPopName();
		}
	}
}

void PointList::draw( int type, idxList &subset, real_type radius /*= 1 */ ) const
{
	// Draw the highlighted points.
	if (type == POINT_SPHERES) {
		for (index_type i = 0; i < subset.size(); i++) {
			drawSphere(data[subset[i]], radius);
		}
	}
	else if (type == POINT_SQUARES) {
		glBegin(GL_POINTS);
		for (index_type i = 0; i < subset.size(); i++) {
			glVertex3f(data[subset[i]][0], data[subset[i]][1], data[subset[i]][2]);
		}
		glEnd();
	}
}

void PointList::ANNquery( real3_type q, size_type k, float radSq, idxList& indices, distList& distances, bool persistent )
{
	unsigned int nPts = data.size();		// actual number of data points
	ANNpoint queryPt;						// query point
	ANNidxArray nnIdx = NULL;				// near neighbor indices
	ANNdistArray dists = NULL;				// near neighbor distances

	// Reconstruct the k-d tree used for querying the points,
	// If there is no existing persistent data structure available
	if (!persistent || (!dataPts || !kdTree)) {

		//cout << " --> Creating query data structure..." << endl;

		clearANNData();

		//////////////////////////////////////////////////////////////////////////
		// Populate data points
		dataPts = annAllocPts(nPts, dim);

		for (ptList::size_type i = 0; i < nPts; i++) {
			for (int d = 0; d < dim; d++) {
				dataPts[i][d] = data[i][d];

				// Check bad floating point values
				if (!_finite(data[i][d])) {
					cerr << "Error in ANNquery(): Non-finite floating point value at index " << i << "!" << endl;
					// TODO: Deal with this error in a better way
					dataPts[i][d] = 0;
				}
			}
		}

		//////////////////////////////////////////////////////////////////////////
		kdTree = new ANNkd_tree(				// build search structure
								dataPts,		// the data points
								nPts,			// number of points
								dim);			// dimension of space
	}



	//////////////////////////////////////////////////////////////////////////
	// Convert they query point into an ANN-familiar format.
	queryPt = annAllocPt(dim);
	for (int d = 0; d < dim; d++) {			// Copy the query point
		queryPt[d] = q[d];
	}

	//////////////////////////////////////////////////////////////////////////
	// CORE
	//////////////////////////////////////////////////////////////////////////
	if (radSq < 0) {
		//cout << " --> Performing k-nearest query..." << endl;

		// Perform k-nearest-neighbor search
		nnIdx = new ANNidx[k];					// allocate near neighbor indices
		dists = new ANNdist[k];					// allocate near neighbor distances

		kdTree->annkSearch(
							queryPt,			// query point
							k,					// number of near neighbors
							nnIdx,				// nearest neighbors (returned)
							dists,				// distance (returned)
							queryEpsilon);		// error bound
	}
	else {
		//cout << " --> Performing range query..." << endl;

		// Perform range search; initially find the number of points within the range
		k = kdTree->annkFRSearch(
									queryPt,			// query point
									radSq,				// radius of range query
									0,					// initially 0
									NULL,				// return no nearest neighbors
									NULL,				// return no distances
									queryEpsilon);		// error bound

		nnIdx = new ANNidx[k];					// allocate near neighbor indices
		dists = new ANNdist[k];					// allocate near neighbor distances

		// Repeat the search with the *correct* number of neighbors
		kdTree->annkFRSearch(
							queryPt,			// query point
							radSq,				// radius of range query
							k,					// *correct* number of near neighbors
							nnIdx,				// nearest neighbors (returned)
							dists,				// distance (returned)
							queryEpsilon);		// error bound
	}

	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////
	indices.resize(k);
	distances.resize(k);
	for (int i = 0; i < k; i++) {
		indices[i] = nnIdx[i];
		distances[i] = sqrt(dists[i]);
	}

	//////////////////////////////////////////////////////////////////////////
	delete [] nnIdx;		// clean things up
	delete [] dists;
	annDeallocPt(queryPt);
}


void PointList::queryKNN( real3_type q, size_type k, idxList& nearest, distList& dist, bool persistent /*= false*/ )
{
	ANNquery(q, k, -1.0, nearest, dist, persistent);
}


void PointList::queryRange( real3_type q, float radSq, idxList& inRange, distList& dist, bool persistent/* = false*/ )
{
	ANNquery(q, 0, radSq, inRange, dist, persistent);
}

void PointList::clearANNData()
{
	if (dataPts) {
		annDeallocPts(dataPts);
	}
	if (kdTree) {
		delete kdTree;
		kdTree = NULL;
	}
	//annClose();				// done with ANN
}

void PointList::generateRandom()
{
	// Use SFMT to generate random points in the unit interval.
	for (index_type i = 0; i < data.size(); i++) {
		for (int d = 0; d < dim; d++) {
			data[i][d] = genrand_real2();
		}
	}
}

void PointList::setANNEpsilon( double eps )
{
	queryEpsilon = eps;
}

void PointList::extent( real3_type &center, real_type &radius ) const
{
	// Measure the extent of the source points
	center[0] = center[1] = center[2] = 0;
	for (index_type i = 0; i < size(); i++) {
		center += data[i];
	}
	center = center / ((real_type) size());

	radius = -REAL_TYPE_MAX;
	for (index_type i = 0; i < size(); i++) {
		real_type distFromCenter = (data[i] - center).norm();

		if (distFromCenter > radius) {
			radius = distFromCenter;
		}
	}
}
