#include "PointListN.h"
#include <ctime>
#include "util.h"

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;
using namespace Geometry;

PointListN::PointListN( int dim ) :
dim(dim), kdTree(NULL), dataPts(NULL), queryEpsilon(0)
{

}

PointListN::PointListN( const PointListN& rhs ) :
dim(rhs.dim), queryEpsilon(rhs.queryEpsilon), data(rhs.data), kdTree(NULL), dataPts(NULL)
{

}

PointListN::~PointListN()
{
	clear();
}

const PointListN& PointListN::operator=( const PointListN & rhs )
{
	if ( this != &rhs ) {
		clear();

		// copy values
		this->queryEpsilon = rhs.queryEpsilon;
		this->data = rhs.data;
	}
	return *this;
}

unsigned int PointListN::size() const
{
	return data.size();
}

data_type& PointListN::operator[]( index_type i )
{
	return data[i];
}

data_type PointListN::operator[]( index_type i ) const
{
	return data[i];
}

void PointListN::resize( size_type n )
{
	// Create dummy variable
	data_type dummy(dim);
	data.resize(n, dummy);
}

void PointListN::clear()
{
	clearANNData();
	data.clear();
}

void PointListN::add( data_type pt )
{
	data.push_back(pt);
}

const std::vector< data_type >* PointListN::getData() const
{
	return &data;
}

/**
* Copy the given list of points.
*/
void PointListN::setPts( const std::vector<data_type> *pts )
{
	clear();
	resize(pts->size());
	for (size_type i = 0; i < pts->size(); i++) {
		data[i] = (*pts)[i];
	}
}

void PointListN::removePoints( idxList list )
{
	// Sort the list according to the index number
	std::sort(list.begin(), list.end());

	// Traversing the list in reverse, remove elements from the array
	for (int i = list.size()-1; i >= 0; i--) {
		// Remove each element from the list
		index_type c = list[i];

		// Find the iterator at c
		std::vector<data_type>::iterator itr = data.begin();
		data.erase(itr+c);
	}
}

void PointListN::ANNquery( data_type q, size_type k, float radSq, idxList& indices, distList& distances, bool persistent )
{
	unsigned int nPts = data.size();		// actual number of data points
	ANNpoint queryPt;						// query point
	ANNidxArray nnIdx = NULL;				// near neighbor indices
	ANNdistArray dists = NULL;				// near neighbor distances

	// Reconstruct the k-d tree used for querying the points,
	// If there is no existing persistent data structure available
	if (!persistent || (!dataPts || !kdTree)) {

		//cout << " --> Creating query data structure..." << endl;

		clearANNData();

		//////////////////////////////////////////////////////////////////////////
		// Populate data points
		dataPts = annAllocPts(nPts, dim);

		for (int i = 0; i < nPts; i++) {
			for (int d = 0; d < dim; d++) {
				dataPts[i][d] = data[i].val[d];

				// Check bad floating point values
				if (!_finite(data[i].val[d])) {
					cerr << "Error in ANNquery(): Non-finite floating point value at index " << i << "!" << endl;
					// TODO: Deal with this error in a better way
					dataPts[i][d] = 0;
				}
			}
		}

		//////////////////////////////////////////////////////////////////////////
		kdTree = new ANNkd_tree(				// build search structure
								dataPts,		// the data points
								nPts,			// number of points
								dim);			// dimension of space
	}


	//////////////////////////////////////////////////////////////////////////
	// Convert the query point into an ANN-familiar format.
	queryPt = annAllocPt(dim);
	for (int d = 0; d < dim; d++) {			// Copy the query point
		queryPt[d] = q.val[d];
	}

	//////////////////////////////////////////////////////////////////////////
	// CORE
	//////////////////////////////////////////////////////////////////////////
	if (radSq < 0) {
		//cout << " --> Performing k-nearest query..." << endl;

		// Perform k-nearest-neighbor search
		nnIdx = new ANNidx[k];					// allocate near neighbor indices
		dists = new ANNdist[k];					// allocate near neighbor distances

		kdTree->annkSearch(
							queryPt,			// query point
							k,					// number of near neighbors
							nnIdx,				// nearest neighbors (returned)
							dists,				// distance (returned)
							queryEpsilon);		// error bound
	}
	else {
		//cout << " --> Performing range query..." << endl;

		// Perform range search; initially find the number of points within the range
		k = kdTree->annkFRSearch(
									queryPt,			// query point
									radSq,				// radius of range query
									0,					// initially 0
									NULL,				// return no nearest neighbors
									NULL,				// return no distances
									queryEpsilon);		// error bound

		nnIdx = new ANNidx[k];					// allocate near neighbor indices
		dists = new ANNdist[k];					// allocate near neighbor distances

		// Repeat the search with the *correct* number of neighbors
		kdTree->annkFRSearch(
							queryPt,			// query point
							radSq,				// radius of range query
							k,					// *correct* number of near neighbors
							nnIdx,				// nearest neighbors (returned)
							dists,				// distance (returned)
							queryEpsilon);		// error bound
	}

	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////
	indices.resize(k);
	distances.resize(k);
	for (int i = 0; i < k; i++) {
		indices[i] = nnIdx[i];
		distances[i] = sqrt(dists[i]);
	}

	//////////////////////////////////////////////////////////////////////////
	delete [] nnIdx;		// clean things up
	delete [] dists;
	annDeallocPt(queryPt);
}


void PointListN::queryKNN( data_type q, size_type k, idxList& nearest, distList& dist, bool persistent /*= false*/ )
{
	ANNquery(q, k, -1.0, nearest, dist, persistent);
}


void PointListN::queryRange( data_type q, float radSq, idxList& inRange, distList& dist, bool persistent/* = false*/ )
{
	ANNquery(q, 0, radSq, inRange, dist, persistent);
}

void PointListN::clearANNData()
{
	if (dataPts) {
		annDeallocPts(dataPts);
	}
	if (kdTree) {
		delete kdTree;
		kdTree = NULL;
	}
	//annClose();				// done with ANN
}


void PointListN::generateRandom() {
	// Use SFMT to generate random points in the unit interval.
	for (index_type i = 0; i < data.size(); i++) {
		for (int d = 0; d < dim; d++) {
			data[i].val[d] = genrand_real2();
		}
	}
}

void PointListN::setANNEpsilon( double eps )
{
	queryEpsilon = eps;
}

int PointListN::getDim() const
{
	return dim;
}
