#ifndef __POINT_LIST_N_H__
#define __POINT_LIST_N_H__

#include <ANN/ANN.h>
#include "types.h"

/**
* A list of points in N-dimensions, with k-nearest neighbor and range search capabilities.
* Uses the ANN library to find nearest neighbors.
*/
class PointListN {
public:
	PointListN( int dim );									/**< Constructor */
	~PointListN();											/**< Destructor */
	PointListN(const PointListN& rhs);						/**< Copy Constructor */
	const PointListN& operator=(const PointListN & rhs);	/**< Assignment */

	/**
	 * Returns the dimensionality of the points.
	 */
	int getDim() const;

	/**
	 * Returns the number of points contained in this list.
	 */
	size_type size() const;

	/**
	 * Resize the array to contain n points.
	 * @param n An integer argument.
	 */
	void resize(size_type n);

	/**
	 * Return a reference to the i^th point in the list.
	 * @param i An integer argument.
	 * @return The reference to the i^th point, which is modifiable.
	 */
	data_type& operator[](index_type i);

	/**
	 * Return the i^th point in the list.
	 * @param i An integer argument.
	 * @return The i^th point. This is not modifiable.
	 */
	data_type operator[](index_type i) const;

	/**
	 * Add a point to the list.
	 * @param pt A data point with three components.
	 */
	void add(data_type pt);

	/**
	 * Returns a pointer to the vector containing the list of all points.
	 * @return A pointer to a ptList.
	 */
	const std::vector<data_type>* getData() const;

	/**
	 * Copy an existing ptList to the current list of points.
	 * This function clears the current list and replaces it with the input list of points.
	 * @param pts A pointer to an existing ptList.
	 */
	void setPts(const std::vector<data_type> *pts);

	/**
	 * Perform a k-nearest-neighbor query.  Refer to the function ANNquery for details.
	 */
	void queryKNN(data_type q, size_type k, idxList& nearest, distList& dist, bool persistent = false);

	/**
	* Perform a range query.  Refer to the function ANNquery for details.
	*/
	void queryRange(data_type q, float radSq, idxList& inRange, distList& dist, bool persistent = false);

	/**
	 * Clear the persistent ANN data structures, if they exist.
	 */
	void clearANNData();

	/**
	 * Remove points from the list.
	 */
	void removePoints(idxList list);

	/**
	 * Clear the list of points.
	 */
	void clear();

	/**
	 * Generate random points in the unit interval.
	 */
	void generateRandom();

	void setANNEpsilon( double eps );

private:
	const int dim;								/**< Dimension */
	std::vector<data_type> data;				/**< List of points */

	/**
	 * ANN-related functionality.
	 */
	ANNpoint* dataPts;							/**< Persistent data array for multiple queries. */
	ANNkd_tree* kdTree;							/**< Persistent search structure for multiple queries. */
	double queryEpsilon;						/**< Error bound for approximate nearest neighbor search. */

	/**
	* Perform either a k-nearest-neighbor query or a range query depending on the input parameters.
	* To perform a k-nearest neighbor query, set "radSq" to a negative number
    * Setting "radSq" to a positive number retrieves all points within distance sqrt(radSq) of the query point.
	* @param q The query point.
	* @param k The number of nearest neighbors to retrieve.
	* @param radSq The squared radius of the range query to perform.
	* @param indices The indices of the retrieved points.
	* @param distances The distance to the retrieved points.
	* @param persistent Tell the function whether to keep a persistent data structure for the query.
	*/
	void ANNquery( data_type q, size_type k, float radSq, idxList& indices, distList& distances, bool persistent );
};

#endif
