#include "ShapeModel.h"
#include "FileSequence.h"
#include <ctime>
#include "util.h"
#include "geom_util.h"
#include "scene.h"
#include <fstream>
#include "PointListN.h"
#include <QDir>
#include "geodesic.h"
#include <vtk-5.0/vtkQuadricDecimation.h>
#include <vtk-5.0/vtkOBJReader.h>
#include <vtk-5.0/vtkTriangleFilter.h>
#include <glm/glm.h>

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;

ShapeModel::ShapeModel( int type ) : shapeType(type), features(NULL), dataDir("preprocessing/"), useRootDir(true), sampleResolution(0.0)
{

}

ShapeModel::~ShapeModel()
{
	if (features) {
		delete features;
	}
}

ShapeModel* loadShape( QString baseDir, QString filename, const Scene &scene, bool processFrames/* = true*/ )
{
	// TODO: Expand this implementation to detect the shape type

	// *Assumption:* For now, assume this is a mesh
	Mesh3_real *shape = new Mesh3_real();

	shape->setFile(baseDir, filename);
	shape->loadGeometry();
	shape->printInfo();
	shape->setSampleResolution(scene.pointSize());

	// Compute / load additional data
	shape->computeNormals(true);
	shape->markBoundary_ComputeResolution();

	if (processFrames) {
		distList frameScales = scene.makeSceneSqRadii(scene.scaleFactors_PC);
		shape->computeFrames(frameScales);

		distList featureScales = scene.makeSceneSqRadii(scene.scaleFactors_MC);
		shape->computeFeatures(scene.featureType, featureScales, scene.ANNEpsilon_FM);

		if (scene.featureType == FEATURE_SPIN_IMAGES) {
			shape->computeSpinImages(scene.binSizeMultiple_SI, scene.imageWidth_SI, scene.maxSupportAngle_SI);

			//// Allocate Spin Image Stack
			//shape->spinImageStack.resize(shape->numVertices());
			//shape->spinImageStack.binSize = shape->SampleResolution()*scene.binSizeMultiple_SI;
			//shape->spinImageStack.imageWidth = scene.imageWidth_SI;
			//shape->spinImageStack.supportAngle = scene.maxSupportAngle_SI;
		}
	}

	return shape;
}


void ShapeModel::setFile( QString baseDir, QString filename )
{
	// Convert all the \ to /
	baseDir = baseDir.replace('\\', QString("/"));
	filename = filename.replace('\\', QString("/"));

	rootDir = baseDir;

	// Append ending / if it's missing
	if (!rootDir.endsWith(QString("/"))) {
		rootDir = rootDir + QString("/");
	}

	// Parse the filename into a path and file
	splitPath(filename, modelDir, this->filename);

	// Remove any overlap between baseDir and rootDir
	if (modelDir.startsWith(rootDir)) {
		modelDir = modelDir.right(modelDir.length() - rootDir.length());
	}
	else {
		useRootDir = false;
	}

	//cout << "Extracted directories:" << endl;
	//cout << "root: " << qPrintable(rootDir) << endl;
	//cout << "modelDir: " << qPrintable(modelDir) << endl;
	//cout << "file: " << qPrintable(this->filename) << endl;
}

void ShapeModel::attachSuffix( QString suffix /*= QString("copy") */ )
{
	QString file = removeExt(filename);
	QString ext = filename.right(filename.length() - file.length());
	int num = 0;

	//cout << "file: " << qPrintable(file) << ", ext: " << qPrintable(ext) << endl;

	// Generate a filename so that its geometry, frame, and feature files are unique on disk
	do {
		QString suff;
		if (num > 0) {
			suff = QString("%1%2").arg(suffix).arg(num);
		}
		else {
			suff = suffix;
		}
		filename = QString("%1_%2%4").arg(file).arg(suff).arg(ext);
		num++;
	} while (QFile::exists(geometryFile()) || QFile::exists(frameFile()) || QFile::exists(featureFile()));

	cout << "new filename: " << qPrintable(filename) << endl;
}

QString ShapeModel::getFilename() const
{
	return filename;
}

QString ShapeModel::uid() const
{
	return modelDir + filename;
}

QString ShapeModel::geometryFile() const
{
	if (useRootDir) {
		return rootDir + modelDir + filename;
	}
	else {
		return modelDir + filename;
	}
}

QString ShapeModel::frameFile() const
{
	// Construct the frame file location
	QDir qdir;
	QString path = rootDir + dataDir + modelDir;
	QString file = removeExt(filename) + QString("-frames.data");

	//cout << "Location is: " << qPrintable(path+file) << endl;

	// Guarantee that the directory structure exists (create if needed)
	if (!qdir.mkpath(path)) {
		cerr << "Could not make path: " << qPrintable(path) << endl;
	}

	return path+file;
}

QString ShapeModel::featureFile() const
{
	// Construct the frame file location
	QDir qdir;
	QString path = rootDir + dataDir + modelDir;
	QString file = removeExt(filename) + QString("-features.data");

	//cout << "Location is: " << qPrintable(path+file) << endl;

	// Guarantee that the directory structure exists (create if needed)
	if (!qdir.mkpath(path)) {
		cerr << "Could not make path: " << qPrintable(path) << endl;
	}

	return path+file;
}


QString ShapeModel::spinImageFile() const
{
	// Construct the frame file location
	QDir qdir;
	QString path = rootDir + dataDir + modelDir;
	QString file = removeExt(filename) + QString("-spinImages.data");

	//cout << "Location is: " << qPrintable(path+file) << endl;

	// Guarantee that the directory structure exists (create if needed)
	if (!qdir.mkpath(path)) {
		cerr << "Could not make path: " << qPrintable(path) << endl;
	}

	return path+file;
}


size_type ShapeModel::numVertices() const
{
	return vertex.size();
}

void ShapeModel::queryKNN( real3_type q, size_type k, idxList& nearest, distList& dist, bool persistent /*= false*/ )
{
	vertex.queryKNN(q, k, nearest, dist, persistent);
}

void ShapeModel::queryRange( real3_type q, float radSq, idxList& inRange, distList& dist, bool persistent /*= false*/ )
{
	vertex.queryRange(q, radSq, inRange, dist, persistent);
}


void ShapeModel::computeFrames( const distList &scales, bool force /*= false */ )
{
	computeNormals();

	// Try loading the frames from disk
	if (!force && loadFrames()) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Set ANN to maximum precision
	vertex.setANNEpsilon(0);

	// Compute principal curvatures
	frameScales = scales;
	frames.clear();
	frames.resize(numVertices());

	const size_type	minimumNeighbors = 7;
	const size_type initialNN = 6;
	const real_type nnFactor = 5.0;
	const size_type numScales = frameScales.size();

	cout << "*** Computing principal curvatures..." << endl;
	cout << "****** Scales:";
	for (index_type i = 0; i < scales.size(); i++) {
		cout << " " << scales[i];
	}
	cout << endl;


	TextProgress(numVertices(), true);
	for (index_type vi = 0; vi < numVertices(); vi++) {

		// Allocate space for all the frames
		frames[vi].resize(numScales);

		// Skip over invalid vertices
		if (markVertex[vi] != VERTEX_VALID) {
			continue;
		}

		// Compute local frame at the vertex
		real3_type x, u, v;							/**< Vertex & local coordinate frame */
		x = vertex[vi];
		LocalCoordinateFrame(normal[vi], u, v);

		for (index_type si = 0; si < numScales; si++) {
			// 1. Gather the points within the search radius
			PointList rangePts, rangeNormals;
			idxList inRange;
			distList distRange;

			vertex.queryRange(x, scales[si], inRange, distRange, true);

			if (inRange.size() < minimumNeighbors) {
				vertex.queryKNN(x, minimumNeighbors, inRange, distRange, true);
			}

			rangePts.reserve(inRange.size());
			rangeNormals.reserve(inRange.size());
			for (index_type j = 0; j < inRange.size(); j++) {
				// Reject if there is no adjacent triangle (no normal vector!!) or
				// Reject if the normal differs by an angle more than pi/2
				if ((markVertex[inRange[j]] == VERTEX_NO_ADJACENT_TRIANGLE) ||
					(normal[vi] * normal[inRange[j]] <= 0)) {
					continue;
				}

				rangePts.add(vertex[inRange[j]]);
				rangeNormals.add(normal[inRange[j]]);
			}

			// 2. Construct and solve least-squares problem
			CurvInfo ct;
			FitCurvatureTensorLS(rangePts, rangeNormals, u, v, ct);

			// Store result
			ct.n = normal[vi];
			if (!ct.checkOrthonormal()) {
				//cerr << "Warning in computeFrames(): frame " << vi << " (scale " << scales[si] << ") is not orthonormal.  Correcting frame." << endl;
				ct.makeOrthonormal();
			}
			frames[vi][si] = ct;

			// Check if there's an overflow
			if (!_finite(ct.kmin) || !_finite(ct.kmax)) {
				cerr << "Warning in computeFrames(): found NaN value in principal curvature for vertex: " << vi << ", scale: " << scales[si] << "." << endl;

				//// Recompute for debugging
				//cout << "% Local frame (n, u, v):" << endl;
				//cout << "n = " << normals[vi] << "';" << endl;
				//cout << "u = " << u << "';" << endl;
				//cout << "v = " << v << "';" << endl;
				//cout << "% Total " << rangePts.size() << " points:" << endl;
				//for (index_type j = 0; j < rangePts.size(); j++) {
				//	cout << "pts(:," << j+1 << ") = " << rangePts[j] << "';" << endl;
				//	cout << "normals(:," << j+1 << ") = " << rangeNormals[j] << "';" << endl;
				//}
				//FitCurvatureTensorLS(rangePts, rangeNormals, u, v, ct);
			}
		}

		TextProgress(vi);
	}

	cout << "...done!" << endl;

	//////////////////////////////////////////////////////////////////////////
	// Return a description of the method used to compute the principal curvatures
	QString curvDesc("Least-squares method");
	curvDesc += QString(" using");
	for (index_type i = 0; i < numScales; i++) {
		if (frameScales[i] == 0) {
			curvDesc += QString(" Scale %1: adaptive operating region of %2 * distance to %3 nearest neighbor per vertex")
								.arg(i).arg(nnFactor).arg(initialNN);
		}
		else {
			curvDesc += QString(" Scale %1: fixed operating region of radius %2").arg(i).arg(scales[i]);
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// Save to disk: Get description from user of the save file
	QString text = QString("Principal curvatures for %1 computed on %2. %3.")
		.arg(filename)
		.arg(QDateTime::currentDateTime().toString(QString("MM/dd/yyyy, hh:mm:ss ap")))
		.arg(curvDesc);

	saveFrames(text);

	cout << "Computed principal curvatures: " << qPrintable(curvDesc) << endl;
	cout << "Saved principal curvature data from: " << qPrintable(frameFile()) << endl;
	cout << "Save description: " << qPrintable(text) << endl;
}


/**
* Save the principal curvatures.
* File format:
* [UNSIGNED INT] Text description length
*
* for each character in description:
* [CHAR] Character in description
*
* [UNSIGNED INT] Number of vertices
* [UNSIGNED INT] Number of scales
*
* for each scale:
*   [FLOAT] Scale value (0 is adaptive scale)
*
* for each vertex:
*   for each scale:
*   [FLOAT] [FLOAT] [FLOAT] normal vector
*   [FLOAT] [FLOAT] [FLOAT] maximum principal curvature vector
*   [FLOAT] [FLOAT] [FLOAT] minimum principal curvature vector
*   [FLOAT] maximum principal curvature
*   [FLOAT] minimum principal curvature
*/
void ShapeModel::saveFrames( const QString &description )
{
	float temp;
	ofstream out;
	out.open(qPrintable(frameFile()), ios_base::binary);

	if (out.fail()) {
		cout << "Error in saveFrames(): could not open file " << qPrintable(frameFile()) << " for writing!" << endl;
		return;
	}

	if ((frames.size() == 0) || (frameScales.size() == 0)) {
		cout << "Error in saveFrames(): No curvature information or scale information found!" << endl;
		out.close();
		return;
	}

	cout << "Saving principal curvatures...";

	//////////////////////////////////////////////////////////////////////////
	// Write the description
	unsigned int descriptionLength = description.size();
	out.write((char*) &descriptionLength, sizeof(unsigned int));
	for (index_type i = 0; i < descriptionLength; i++) {
		char ch = description.at(i).toLatin1();
		out.write((char*) &ch, sizeof(char));
	}

	//////////////////////////////////////////////////////////////////////////
	// Write the data
	unsigned int numVertices = frames.size();
	unsigned int numScales = frameScales.size();
	out.write((char*) &numVertices, sizeof(unsigned int));
	out.write((char*) &numScales, sizeof(unsigned int));

	for (index_type i = 0; i < numScales; i++) {
		temp = frameScales[i];
		out.write((char*) &temp, sizeof(float));
	}

	for (index_type i = 0; i < numVertices; i++) {
		for (index_type j = 0; j < numScales; j++) {
			CurvInfo ci = frames[i][j];

			for (int c = 0; c < 3; c++) {
				temp = ci.n[c];
				out.write((char*) &temp, sizeof(float));
			}
			for (int c = 0; c < 3; c++) {
				temp = ci.vmax[c];
				out.write((char*) &temp, sizeof(float));
			}
			for (int c = 0; c < 3; c++) {
				temp = ci.vmin[c];
				out.write((char*) &temp, sizeof(float));
			}
			temp = ci.kmax;
			out.write((char*) &temp, sizeof(float));
			temp = ci.kmin;
			out.write((char*) &temp, sizeof(float));
		}
	}

	out.close();

	cout << "done!" << endl;
}


bool ShapeModel::loadFrames()
{
	float temp;
	ifstream in;
	in.open(qPrintable(frameFile()), ios_base::binary);
	if (in.fail()) {
		cout << "Error in loadFrames(): could not open file " << qPrintable(frameFile()) << " for reading!" << endl;
		return false;
	}

	cout << "*** Loading frames...";

	//////////////////////////////////////////////////////////////////////////
	// Read the description
	QString text;
	unsigned int descriptionLength;
	in.read((char*) &descriptionLength, sizeof(unsigned int));
	text.resize(descriptionLength);
	//cout << "Description length is: " << descriptionLength << endl;
	for (index_type i = 0; i < descriptionLength; i++) {
		char ch;
		in.read((char*) &ch, sizeof(char));
		text[i] = QChar(ch);
	}
	//cout << "Description is: " << qPrintable(text) << endl;

	// Extract source file name from description
	QRegExp rx("Principal curvatures for ([-.\\w]+)");
	QString sourceFile;
	if (rx.indexIn(text, 0) != -1) {
		sourceFile = rx.cap(1);

		if (!geometryFile().endsWith(sourceFile)) {
			cerr << "Error in loadFrames(): file in description does not match current file!" << endl;
			return false;
		}
	}
	else {
		// We cannot use the data because we cannot find the corresponding mesh.
		cout << "No source file name in description!  Clearing data." << endl;
		frames.clear();
		frameScales.clear();
		return false;
	}

	//////////////////////////////////////////////////////////////////////////
	// Read the data
	unsigned int numVertices;
	unsigned int numScales;
	in.read((char*) &numVertices, sizeof(unsigned int));
	in.read((char*) &numScales, sizeof(unsigned int));

	// Check the data
	if (numVertices != this->numVertices()) {
		cerr << "Error in loadFrames(): number of vertices do not match!" << endl;
		return false;
	}

	//cout << "Number of vertices: " << numVertices << endl;
	//cout << "Number of scales: " << numScales << endl;
	frames.resize(numVertices);
	frameScales.resize(numScales);

	for (index_type i = 0; i < numScales; i++) {
		in.read((char*) &temp, sizeof(float));
		frameScales[i] = temp;
	}

	for (index_type i = 0; i < numVertices; i++) {
		frames[i].resize(numScales);

		for (index_type j = 0; j < numScales; j++) {
			CurvInfo ci;

			for (int c = 0; c < 3; c++) {
				in.read((char*) &temp, sizeof(float));
				ci.n[c] = temp;
			}
			for (int c = 0; c < 3; c++) {
				in.read((char*) &temp, sizeof(float));
				ci.vmax[c] = temp;
			}
			for (int c = 0; c < 3; c++) {
				in.read((char*) &temp, sizeof(float));
				ci.vmin[c] = temp;
			}
			in.read((char*) &temp, sizeof(float));
			ci.kmax = temp;
			in.read((char*) &temp, sizeof(float));
			ci.kmin = temp;

			frames[i][j] = ci;
		}
	}

	in.close();

	cout << " done!" << endl;

	return true;
}


void ShapeModel::computeFeatures( int featureType, const distList &scales, real_type eps )
{
	if (featureType == FEATURE_PRINCIPAL_CURVATURES) {
		makePrincipalCurvatureFeatures();
	}
	else if (featureType == FEATURE_PRINCIPAL_CURVATURE_RATIO) {
		makePrincipalCurvatureRatioFeatures();
	}
	else if (featureType == FEATURE_MEAN_CURVATURE) {
		makeMeanCurvatureFeatures(scales, eps);
	}
	else if (featureType == FEATURE_PRINCIPAL_CURVATURE_NORM) {
		makePrincipalCurvatureNormFeatures();
	}
}


void ShapeModel::makePrincipalCurvatureFeatures()
{
	//////////////////////////////////////////////////////////////////////////
	// Derive features from principal curvatures
	// Just copy the principal curvature values
	// If we have estimated the principal curvatures on "n" different scales,
	// Then the features will have 2*n components

	//////////////////////////////////////////////////////////////////////////
	// 1. Allocate space for features
	featureScales.resize(frameScales.size()*2);
	for (index_type vi = 0; vi < frameScales.size(); vi++) {
		featureScales[2*vi] = frameScales[vi];
		featureScales[2*vi+1] = frameScales[vi];
	}

	if (features) delete features;
	features = new PointListN(featureScales.size());
	features->resize(numVertices());

	// Display status
	cout << "Populating feature values (copying principal curvatures)..." << endl;
	cout << "Scales:";
	for (index_type vi = 0; vi < featureScales.size(); vi++) {
		cout << " " << featureScales[vi];
	}
	cout << endl;

	// 2. Populate features
	for (index_type vi = 0; vi < numVertices(); vi++) {
		if (markVertex[vi] == VERTEX_VALID) {
			for (index_type si = 0; si < frameScales.size(); si++) {
				(*features)[vi].val[2*si] = frames[vi][si].kmin;
				(*features)[vi].val[2*si+1] = frames[vi][si].kmax;
			}
		}
	}

	cout << "...done!" << endl;
}

void ShapeModel::makePrincipalCurvatureRatioFeatures()
{
	//////////////////////////////////////////////////////////////////////////
	// Derive features from principal curvatures
	// Compute the ratio of the principal curvatures (min / max) over multiple scales

	//////////////////////////////////////////////////////////////////////////
	// 1. Allocate space for features
	featureScales.resize(frameScales.size());
	for (index_type i = 0; i < frameScales.size(); i++) {
		featureScales[i] = frameScales[i];
	}

	if (features) delete features;
	features = new PointListN(featureScales.size());
	features->resize(numVertices());

	// Display status
	cout << "*** Computing features (ratio of principal curvatures)..." << endl;
	cout << "****** Scales:";
	for (index_type i = 0; i < featureScales.size(); i++) {
		cout << " " << featureScales[i];
	}
	cout << endl;

	// 2. Populate features
	for (index_type vi = 0; vi < numVertices(); vi++) {
		if (markVertex[vi] == VERTEX_VALID) {
			for (index_type si = 0; si < frameScales.size(); si++) {
				(*features)[vi].val[si] = frames[vi][si].principalCurvatureRatio();
			}
		}
	}

	cout << "...done!" << endl;
}


void ShapeModel::makePrincipalCurvatureNormFeatures()
{
	//////////////////////////////////////////////////////////////////////////
	// Derive features from principal curvatures
	// Compute the norm of the principal curvatures (min, max) over multiple scales

	//////////////////////////////////////////////////////////////////////////
	// 1. Allocate space for features
	featureScales.resize(1);
	featureScales[0] = frameScales[frameScales.size()-1];

	if (features) delete features;
	features = new PointListN(featureScales.size());
	features->resize(numVertices());

	// Display status
	cout << "*** Computing features (norm of multi-scale principal curvatures)..." << endl;
	cout << "****** Scales:";
	for (index_type i = 0; i < featureScales.size(); i++) {
		cout << " " << featureScales[i];
	}
	cout << endl;

	// 2. Populate features
	for (index_type vi = 0; vi < numVertices(); vi++) {
		if (markVertex[vi] == VERTEX_VALID) {
			data_type allCurvatures(frameScales.size());

			for (index_type si = 0; si < frameScales.size(); si++) {
				//allCurvatures.val[2*si] = frames[vi][si].kmin;
				//allCurvatures.val[2*si+1] = frames[vi][si].kmax;
				allCurvatures.val[si] = frames[vi][si].kmin;
			}

			(*features)[vi].val[0] = allCurvatures.normL2();
		}
	}

	cout << "...done!" << endl;
}

void ShapeModel::makeMeanCurvatureFeatures( const distList &scales, real_type eps )
{
	//////////////////////////////////////////////////////////////////////////
	// Try loading from disk
	if (loadFeatures()) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Compute mean curvature feature vector
	// Allocate space for features
	if (features) delete features;
	features = new PointListN(scales.size());
	features->resize(numVertices());
	featureScales = scales;

	// Set ANN Epsilon for MC estimation
	vertex.setANNEpsilon(eps);

	// Compute features
	cout << "Computing mean curvature features..." << endl;
	cout << "Scales:";
	for (index_type i = 0; i < featureScales.size(); i++) {
		cout << " " << featureScales[i];
	}
	cout << endl;

	TextProgress(numVertices(), true);
	for (index_type i = 0; i < numVertices(); i++) {
		(*features)[i] = ComputeMeanCurvature(vertex[i], vertex, normal, scales);
		TextProgress(i);
	}

	cout << " done!" << endl;

	//////////////////////////////////////////////////////////////////////////
	// Save to disk
	saveFeatures();
}

/**
* Save the mean curvatures.
* File format:

* [UNSIGNED INT] Number of vertices
* [UNSIGNED INT] Number of scales
*
* for each scale:
*   [FLOAT] Scale value (0 is adaptive scale)
*
* for each vertex:
*   for each scale:
*   [FLOAT] mean curvature value
*/
void ShapeModel::saveFeatures()
{
	float temp;
	ofstream out;
	out.open(qPrintable(featureFile()), ios_base::binary);

	if (out.fail()) {
		cout << "Error in saveFeatures(): could not open file " << qPrintable(frameFile()) << " for writing!" << endl;
		return;
	}

	if ((features->size() == 0) || (featureScales.size() == 0)) {
		cout << "Error in saveFeatures(): No curvature information or scale information found!" << endl;
		out.close();
		return;
	}

	cout << "Saving mean curvatures to " << qPrintable(featureFile()) << "...";

	//////////////////////////////////////////////////////////////////////////
	// Write the data
	unsigned int numVertices = features->size();
	unsigned int numScales = featureScales.size();
	out.write((char*) &numVertices, sizeof(unsigned int));
	out.write((char*) &numScales, sizeof(unsigned int));

	for (index_type i = 0; i < numScales; i++) {
		temp = featureScales[i];
		out.write((char*) &temp, sizeof(float));
	}

	for (index_type i = 0; i < numVertices; i++) {
		for (index_type j = 0; j < numScales; j++) {
			temp = (*features)[i].val[j];
			out.write((char*) &temp, sizeof(float));
		}
	}

	out.close();

	cout << "done!" << endl;
}

bool ShapeModel::loadFeatures()
{
	float temp;
	ifstream in;
	in.open(qPrintable(featureFile()), ios_base::binary);
	if (in.fail()) {
		cout << "Error in loadFeatures(): could not open file " << qPrintable(featureFile()) << " for reading!" << endl;
		return false;
	}

	cout << " features";

	//////////////////////////////////////////////////////////////////////////
	// Write the data
	unsigned int numVertices, numScales;
	in.read((char*) &numVertices, sizeof(unsigned int));
	in.read((char*) &numScales, sizeof(unsigned int));

	// Check the data
	if (numVertices != this->numVertices()) {
		cerr << "Error in loadFeatures(): number of vertices do not match!" << endl;
		return false;
	}

	if (features) {
		delete features;
	}
	features = new PointListN(numScales);
	features->resize(numVertices);
	featureScales.resize(numScales);

	for (index_type i = 0; i < numScales; i++) {
		in.read((char*) &temp, sizeof(float));
		featureScales[i] = temp;
	}

	for (index_type i = 0; i < numVertices; i++) {
		for (index_type j = 0; j < numScales; j++) {
			in.read((char*) &temp, sizeof(float));
			(*features)[i].val[j] = temp;
		}
	}

	in.close();

	return true;
}

void ShapeModel::drawFeatureValues( const Scene &scene, int comp ) const
{
	if (!features || !featureScales.size()) {
		return;
	}

	// Clamp comp to be in the range
	int compMax = featureScales.size();
	comp = min(compMax-1, comp);

	// Splat the computed curvatures as points
	glDisable(GL_LIGHTING);
	glPointSize(5.f);
	glBegin(GL_POINTS);
	for (index_type i = 0; i < features->size(); i++) {
		if ((*features)[i].normL2() > scene.threshold_FM) {
			real3_type v = vertex[i];

			setOpenGLColor(scene.getColormapColor((*features)[i].val[comp]));
			glVertex3f(v[0], v[1], v[2]);
		}
	}
	glEnd();

	// Draw the feature estimation scales
	drawSqRadii(featureScales, comp);
}

void ShapeModel::drawFrames( const Scene &scene, int comp, int curvDisplayMode ) const
{
	real3_type temp;

	// Clamp comp to be in the range
	int compMax = frameScales.size();
	comp = min(compMax-1, comp);

	// Draw the principal curvature frame
	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	for (index_type i = 0; i < frames.size(); i++) {
		if (curvDisplayMode == CURVINFO_MIN_AND_MAX_CURVATURES) {
			real3_type u, v;

			// Draw the minimum curvature principal direction
			temp = frames[i][comp].vmin;
			u = vertex[i];
			v = vertex[i] + scene.pointSize()*temp;
			glVertex3f(u[0], u[1], u[2]);
			glVertex3f(v[0], v[1], v[2]);

			// Draw the maximum curvature principal direction
			temp = frames[i][comp].vmax;
			u = vertex[i];
			v = vertex[i] + scene.pointSize()*temp;
			glVertex3f(u[0], u[1], u[2]);
			glVertex3f(v[0], v[1], v[2]);
		}
		else {
			if (curvDisplayMode == CURVINFO_NORMALS) {
				temp = frames[i][comp].n;
			}
			else if (curvDisplayMode == CURVINFO_MIN_CURVATURE) {
				temp = frames[i][comp].vmin;
			}
			else if (curvDisplayMode == CURVINFO_MAX_CURVATURE) {
				temp = frames[i][comp].vmax;
			}

			real3_type u = vertex[i];
			real3_type v = vertex[i] + scene.pointSize()*temp;

			glVertex3f(u[0], u[1], u[2]);
			glVertex3f(v[0], v[1], v[2]);
		}
	}
	glEnd();
}

void ShapeModel::drawCurvatureValues( const Scene &scene, int comp, int curvDisplayMode ) const
{
	real_type temp;

	// Clamp comp to be in the range
	int compMax = frameScales.size();
	comp = min(compMax-1, comp);

	// Splat the computed curvatures as points
	glDisable(GL_LIGHTING);
	glPointSize(5.f);
	glBegin(GL_POINTS);
	for (index_type i = 0; i < frames.size(); i++) {
		if (curvDisplayMode == CURVINFO_MIN_CURVATURE) {
			temp = frames[i][comp].kmin;
		}
		else if (curvDisplayMode == CURVINFO_MAX_CURVATURE) {
			temp = frames[i][comp].kmax;
		}

		if (temp > scene.threshold_FM) {
			real3_type v = vertex[i];

			setOpenGLColor(scene.getColormapColor(temp));
			glVertex3f(v[0], v[1], v[2]);
		}
	}
	glEnd();

	// Draw the feature estimation scales
	drawSqRadii(frameScales, comp);
}


ShapeModel* ShapeModel::makeCopy() const
{
	ShapeModel *shape = new ShapeModel(*this);
	shape->features = new PointListN(*features);		/**< Deep-copy the feature vectors */
	shape->attachSuffix();

	return shape;
}

void ShapeModel::loadGeometry()
{
	// TODO!!!
	// Implement loading of point geometry.
}

void ShapeModel::computeNormals( bool force /*= false */ )
{
	// TODO!!!
	// Implement computing normals with point geometry.
}

void ShapeModel::draw() const
{
	// Splat the points
	glDisable(GL_LIGHTING);
	glPointSize(5.f);
	glBegin(GL_POINTS);
	for (index_type i = 0; i < vertex.size(); i++) {
		real3_type v = vertex[i];
		glVertex3f(v[0], v[1], v[2]);
	}
	glEnd();
}

void ShapeModel::drawWithStatus() const
{
	// Splat the points
	glDisable(GL_LIGHTING);
	glPointSize(5.f);
	glBegin(GL_POINTS);
	for (index_type i = 0; i < vertex.size(); i++) {
		real3_type v = vertex[i];

		switch(markVertex[i]) {
			case VERTEX_NO_ADJACENT_TRIANGLE:
				glColor3f(1.0,0.0,0.0);
				break;
			case VERTEX_ON_BOUNDARY:
				glColor3f(0.0,0.0,1.0);
				break;
			default:
				glColor3f(1.0,1.0,1.0);
				break;
		}

		glVertex3f(v[0], v[1], v[2]);
	}
	glEnd();
}

void ShapeModel::drawSubset( idxList set ) const
{
	// Splat the points
	glDisable(GL_LIGHTING);
	glPointSize(5.f);
	glBegin(GL_POINTS);
	for (index_type i = 0; i < set.size(); i++) {
		real3_type v = vertex[set[i]];
		glVertex3f(v[0], v[1], v[2]);
	}
	glEnd();
}

idxList ShapeModel::pruneInvalidVertices( const idxList &list ) const
{
	// Prune sample points that are isolated or on the boundary
	idxList newList;
	newList.reserve(list.size());

	for (index_type i = 0; i < list.size(); i++) {
		index_type vi = list[i];

		if (markVertex[vi] == VERTEX_VALID) {
			// Add to the list
			newList.push_back(vi);
		}
	}

	return newList;
}

idxList ShapeModel::pruneSamplePrincipalCurvatureRatio( const idxList &list, real_type maxRatio ) const
{
	// Prune sample points based on the ratio of principal curvatures
	idxList newList;
	newList.reserve(list.size());

	for (index_type i = 0; i < list.size(); i++) {
		index_type vi = list[i];

		// Compute the ratio of the principal curvatures in all scales
		bool ok = true;

		for (index_type si = 0; si < frameScales.size(); si++) {
			real_type ratio = frames[vi][si].principalCurvatureRatio();
			
			// If the ratio of principal curvatures exceeds the maximum ratio threshold
			if ((ratio == 0.0) || (ratio >= maxRatio)) {
				// Don't use the point
				ok = false;
				break;
			}
		}

		if (ok) {
			// Add to the list
			newList.push_back(vi);
		}
	}

	return newList;
}

ShapeModel* ShapeModel::pickSample( idxList subset, ShapeModel *shape/* = NULL*/ ) const
{
	if (!shape) {
		shape = new ShapeModel(SHAPE_POINTS);
	}

	// Generate a unique filename for this shape
	shape->rootDir = this->rootDir;
	shape->modelDir = this->modelDir;
	shape->dataDir = this->dataDir;
	shape->filename = this->filename;
	shape->sampleResolution = this->sampleResolution;
	shape->attachSuffix(QString("sample"));

	// Allocate the space
	size_type n = subset.size();
	shape->box_min = box_min;
	shape->box_max = box_max;
	shape->vertex.resize(n);
	shape->normal.resize(n);
	shape->markVertex.resize(n);
	if (frames.size()) {
		shape->frames.resize(n);
		shape->frameScales = frameScales;
	}
	if (features && features->size()) {
		shape->features = new PointListN(featureScales.size());
		shape->features->resize(n);
		shape->featureScales = featureScales;
	}
	if (spinImageStack.size()) {
		shape->spinImageStack.resize(n);
		shape->spinImageStack.copyParams(spinImageStack);
	}

	// Store the geometry
	for (index_type i = 0; i < n; i++) {
		index_type si = subset[i];
		shape->vertex[i] = vertex[si];
		shape->normal[i] = normal[si];
		shape->markVertex[i] = markVertex[si];
		if (frames.size()) {
			shape->frames[i] = frames[si];
		}
		if (features && features->size()) {
			(*shape->features)[i] = (*features)[si];
		}
		if (spinImageStack.size()) {
			shape->spinImageStack[i] = spinImageStack[si];
		}
	}

	return shape;
}

ShapeModel* ShapeModel::subSample( size_type n, real_type pruneMaxRatio ) const
{
	// Generate a random sample
	idxList idx = randomSample(n, numVertices());
	if (!idx.size()) {
		cerr << "*** Warning in ShapeModel::subSample(): No random samples selected!" << endl;
		return NULL;
	}

	// Prune isolated vertices and boundary vertices
	idx = pruneInvalidVertices(idx);
	if (!idx.size()) {
		cerr << "*** Warning in ShapeModel::subSample(): No valid vertices selected!" << endl;
		return NULL;
	}

	// Prune samples based on ratio of principal curvatures
	idx = pruneSamplePrincipalCurvatureRatio(idx, pruneMaxRatio);
	if (!idx.size()) {
		cerr << "*** Warning in ShapeModel::subSample(): No vertices with stable principal directions!" << endl;
		return NULL;
	}

	return pickSample(idx);
}

void ShapeModel::sampleVertices( size_type n, PointList &result ) const
{
	// Just subsample the vertices
	result.clear();

	// Generate a random sample
	idxList idx = randomSample(n, numVertices());
	if (!idx.size()) {
		return;
	}

	result.resize(idx.size());

	// Store the geometry
	for (index_type i = 0; i < n; i++) {
		index_type si = idx[i];
		result[i] = vertex[si];
	}
}

ClosestPointInfo ShapeModel::findClosestPoint( const real3_type &p )
{
	ClosestPointInfo result;
	idxList inRange;
	distList dists;

	// Use ANN library to find the closest point
	queryKNN(p, 1, inRange, dists, true);

	/**< TODO: We should really have one estimation of the frames, not multiple scales. */
	// Which scale to use?  Use the last component, usually the smoothest.
	//const int useComp = frameScales.size()-1;

	if (inRange.size()) {
		result.index = inRange[0];
		result.dist = dists[0];
		//result.position = vertex[inRange[0]];
		//result.frameInfo = &frames[inRange[0]][useComp];
		//result.valid = true;
	}

	return result;
}

void ShapeModel::hybridClosestPoint( const real3_type &p, real3_type &closest, real_type &dist )
{
	ClosestPointInfo cpi = findClosestPoint(p);

	if (markVertex[cpi.index] == VERTEX_ON_BOUNDARY) {
		// Point-to-plane distance if on boundary
		real3_type v = vertex[cpi.index];
		real3_type n = normal[cpi.index];

		// Signed distance
		dist = (p - v) * n;

		// Project to plane
		closest = p - dist*n;

		// Absolute distance
		dist = abs(dist);
	}
	else {
		// Point-to-point distance (otherwise)
		dist = cpi.dist;
		closest = vertex[cpi.index];
	}
}


real_type ShapeModel::pointSqDistance( const real3_type &p, real3_type &closestPoint )
{
	idxList inRange;
	distList dists;

	// Use ANN library to find the closest point
	queryKNN(p, 1, inRange, dists, true);

	if (inRange.size()) {
		closestPoint = vertex[inRange[0]];
		return dists[0] * dists[0];
	}
	else {
		return REAL_TYPE_MAX;
	}
}

real_type ShapeModel::pointSqDistance( const real3_type &p, real3_type &closestPoint, CurvInfo &frameInfo )
{
	idxList inRange;
	distList dists;

	// Use ANN library to find the closest point
	queryKNN(p, 1, inRange, dists, true);

	// Which scale to use?  Use the last component, usually the smoothest.
	const int useComp = frameScales.size()-1;

	if (inRange.size()) {
		closestPoint = vertex[inRange[0]];
		/**< TODO: We should really have one estimation of the frames, not multiple scales. */
		frameInfo = frames[inRange[0]][useComp];
		return dists[0] * dists[0];
	}
	else {
		return REAL_TYPE_MAX;
	}
}

real_type ShapeModel::pointSqDistance( const real3_type &p, index_type &closestIndex )
{
	idxList inRange;
	distList dists;

	// Use ANN library to find the closest point
	queryKNN(p, 1, inRange, dists, true);

	if (inRange.size()) {
		closestIndex = inRange[0];
		return dists[0] * dists[0];
	}
	else {
		return REAL_TYPE_MAX;
	}
}

void ShapeModel::drawSqRadii( const distList &sqRadii, int comp ) const
{
	glEnable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glCullFace(GL_BACK);
	glEnable(GL_BLEND);
	glDepthMask(GL_FALSE);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	//////////////////////////////////////////////////////////////////////////
	// Draw the spheres representing the different scales so that we can see them
	int longestAxis = 0;
	real_type longestAxisLength;
	real3_type length = box_max - box_min;
	for (int d = 1; d < 3; d++) {
		if (length[d] > length[longestAxis]) {
			longestAxis = d;
		}
	}
	real3_type pos = box_min;
	pos[longestAxis] += (box_max[longestAxis] - box_min[longestAxis]) / 4.0;
	//real_type radFac = FitSphereWeight99Factor;

	// Clamp comp to be in the range
	int compMax = sqRadii.size();
	comp = min(compMax-1, comp);

	for (int i = 0; i < sqRadii.size(); i++) {
		real_type rad = sqrt(sqRadii[i]);

		pos[longestAxis] += rad;
		if (i == comp) {
			setOpenGLMaterialColor(RealColor(1.0,1.0,0), 0.5);
		}
		else {
			setOpenGLMaterialColor(RealColor(0.8,0.8,1.0), 0.5);
		}
		//drawSphere(pos, rad*radFac, 30, 30);
		drawSphere(pos, rad, 30, 30);
		pos[longestAxis] += rad;
	}
	//////////////////////////////////////////////////////////////////////////

	glDepthMask (GL_TRUE);
	glDisable(GL_BLEND);
	glDisable(GL_CULL_FACE);
}

void ShapeModel::transformVertex( const SE3 &tf, index_type i )
{
	transformVertex(*this, tf, i);
}

void ShapeModel::transformVertex( const ShapeModel &shape, const SE3 &tf, index_type i )
{
	// Apply the rigid transformation to the vertex and frames
	vertex[i] = tf*shape.vertex[i];
	normal[i] = tf.rot*shape.normal[i];
	normal[i].normalize();

	for (index_type j = 0; j < frameScales.size(); j++) {
		frames[i][j].n = tf.rot*shape.frames[i][j].n;
		frames[i][j].vmax = tf.rot*shape.frames[i][j].vmax;
		frames[i][j].vmin = tf.rot*shape.frames[i][j].vmin;
		frames[i][j].makeOrthonormal();
	}
}

void ShapeModel::clearANNData()
{
	vertex.clearANNData();
	normal.clearANNData();
	if (features) {
		features->clearANNData();
	}
}

void ShapeModel::printInfo() const
{
	cout << "*** Filename: " << qPrintable(uid()) << endl;
	cout << "*** Number of vertices: " << vertex.size() << endl;
	if (shapeType == SHAPE_MESH) {
		cout << "*** Number of faces: " << ((Mesh3_real*)this)->triangle_vertex_index.size() << endl;
	}
	//cout << "*** Bounding box range: " << box_min << " " << box_max << endl;
	cout << "*** Bounding box diagonal length: " << (box_max - box_min).norm() << endl;
}

real_type ShapeModel::SampleResolution() const
{
	return sampleResolution;
}

void ShapeModel::setSampleResolution(real_type res)
{
	sampleResolution = res;
}


void ShapeModel::computeSpinImages( real_type binSize_MR, int imageWidth, real_type supportAngle )
{
	//////////////////////////////////////////////////////////////////////////
	// Try loading from disk
	if (!imageWidth || loadSpinImages()) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Compute Spin Images
	// 1. Allocate space if necessary
	if (spinImageStack.size() != numVertices()) {
		spinImageStack.resize(numVertices());
	}
	spinImageStack.binSize = sampleResolution*binSize_MR;
	spinImageStack.imageWidth = imageWidth;
	spinImageStack.supportAngle = supportAngle;
	distList numBins;										/**< Keep track of # bins filled in spin image */
	numBins.reserve(numVertices());

	// Compute features
	cout << "Computing spin image features..." << endl;
	cout << "*** Bin Size: " << spinImageStack.binSize << "(" << binSize_MR << " times the mesh resolution)" << endl;
	cout << "*** Image Width: " << imageWidth << endl;
	cout << "*** Support Angle: " << supportAngle << " degrees" << endl;

	TextProgress(numVertices(), true);
	for (index_type i = 0; i < numVertices(); i++) {
		computeSpinImage(i);

		const SpinImage &si = spinImageStack[i];
		if (!si.invalid()) {
			numBins.push_back(si.countBins());
		}

		TextProgress(i);
	}

	//////////////////////////////////////////////////////////////////////////
	// Compute lambda: half of median # of bins
	int medianIndex = ceil((((double)numBins.size()) - 1.0) / 2.0);
	std::partial_sort(numBins.begin(), numBins.begin() + medianIndex + 1, numBins.end());
	real_type median;
	if (numBins.size() % 2 == 0) {
		// If even, then average the two medians
		median = (numBins[medianIndex] + numBins[medianIndex-1]) / 2.0;
	}
	else {
		median = numBins[medianIndex];
	}
	spinImageStack.lambda = median / 2.0;

	cout << "*** Expected overlap: " << spinImageStack.lambda << " bins" << endl;

	//////////////////////////////////////////////////////////////////////////
	cout << " done!" << endl;

	// Save spin images
	saveSpinImages();
}

void ShapeModel::computeSpinImage( index_type i )
{
	//cout << "Computing spin image feature..." << endl;
	//cout << "*** Bin Size: " << spinImageStack.binSize << endl;
	//cout << "*** Image Width: " << spinImageStack.imageWidth << endl;
	//cout << "*** Support Angle: " << spinImageStack.supportAngle << " degrees" << endl;

	// Check that the vertex is valid
	if (markVertex[i] == VERTEX_NO_ADJACENT_TRIANGLE) {
		// Just use a placeholder spin image that has no data
		spinImageStack[i].resize(spinImageStack.imageWidth);
		spinImageStack[i].reset();
		return;
	}

	// Store the current vertex and normal
	real3_type p = vertex[i];
	real3_type n = normal[i];
	real_type minCosAngle = cos(spinImageStack.supportAngle / 180.0 * M_PI);

	// Allocate space for the spin image
	spinImageStack[i].resize(spinImageStack.imageWidth);
	spinImageStack[i].reset();

	// Accumulate spin image for the rest of the vertices
	for (index_type j = 0; j < numVertices(); j++) {
		// Don't accumulate the same point
		if (i == j) {
			continue;
		}

		// Support angle test
		if ( (n * normal[j]) > minCosAngle ) {
			// Compute spin map coordinates and accumulate
			spinImageStack[i].accumulate(SpinMap(p, n, vertex[j]), spinImageStack.binSize);
		}
	}
}

void ShapeModel::saveSpinImages()
{
	float temp;
	ofstream out;
	out.open(qPrintable(spinImageFile()), ios_base::binary);

	if (out.fail()) {
		cout << "Error in saveSpinImages(): could not open file " << qPrintable(spinImageFile()) << " for writing!" << endl;
		return;
	}

	cout << "Saving spin images to " << qPrintable(spinImageFile()) << "...";

	//////////////////////////////////////////////////////////////////////////
	// Write the data
	unsigned int numVertices = spinImageStack.size();
	out.write((char*) &numVertices, sizeof(unsigned int));
	out.write((char*) &spinImageStack.binSize, sizeof(real_type));
	out.write((char*) &spinImageStack.imageWidth, sizeof(int));
	out.write((char*) &spinImageStack.supportAngle, sizeof(real_type));
	out.write((char*) &spinImageStack.lambda, sizeof(real_type));

	for (index_type vi = 0; vi < numVertices; vi++) {
		for (index_type i = 0; i < spinImageStack.imageWidth; i++) {
			for (index_type j = 0; j < spinImageStack.imageWidth; j++) {
				out.write((char*) &spinImageStack[vi](i,j), sizeof(real_type));
			}
		}
	}

	out.close();

	cout << "done!" << endl;
}

bool ShapeModel::loadSpinImages()
{
	float temp;
	ifstream in;
	in.open(qPrintable(spinImageFile()), ios_base::binary);
	if (in.fail()) {
		cout << "Error in loadSpinImages(): could not open file " << qPrintable(spinImageFile()) << " for reading!" << endl;
		return false;
	}

	cout << "*** Loading spin images...";

	//////////////////////////////////////////////////////////////////////////
	// Read the data
	unsigned int numVertices;
	in.read((char*) &numVertices, sizeof(unsigned int));

	// Check the data
	if (numVertices != this->numVertices()) {
		cerr << "Error in loadSpinImages(): number of vertices do not match!" << endl;
		return false;
	}

	in.read((char*) &spinImageStack.binSize, sizeof(real_type));
	in.read((char*) &spinImageStack.imageWidth, sizeof(int));
	in.read((char*) &spinImageStack.supportAngle, sizeof(real_type));
	in.read((char*) &spinImageStack.lambda, sizeof(real_type));
	spinImageStack.resize(numVertices);

	for (index_type vi = 0; vi < numVertices; vi++) {
		// Allocate space for the spin image
		spinImageStack[vi].resize(spinImageStack.imageWidth);
		for (index_type i = 0; i < spinImageStack.imageWidth; i++) {
			for (index_type j = 0; j < spinImageStack.imageWidth; j++) {
				in.read((char*) &spinImageStack[vi](i,j), sizeof(real_type));
			}
		}
	}

	in.close();

	cout << "done!" << endl;

	cout << "Loaded spin image features..." << endl;
	cout << "*** Bin Size: " << spinImageStack.binSize << endl;
	cout << "*** Image Width: " << spinImageStack.imageWidth << endl;
	cout << "*** Support Angle: " << spinImageStack.supportAngle << " degrees" << endl;
	cout << "*** Expected overlap: " << spinImageStack.lambda << " bins" << endl;

	return true;
}

void ShapeModel::saveSingleSpinImage( const char *filename, index_type i )
{
	if (!spinImageStack.size() || spinImageStack[i].invalid()) {
		cerr << "Error in saveSingleSpinImage(): No Spin Image Information Found!" << endl;
		return;
	}

	ofstream out;
	out.open(filename, ios_base::binary);

	if (out.fail()) {
		cerr << "Error in saveSingleSpinImage(): could not open file " << filename << " for writing!" << endl;
		return;
	}

	cout << "Saving spin image at vertex " << i << "...";

	//////////////////////////////////////////////////////////////////////////
	// Write the data
	SpinImage &image = spinImageStack[i];
	float temp;

	for (index_type i = 0; i < image.width(); i++) {
		for (index_type j = 0; j < image.width(); j++) {
			temp = image(i, j);
			out.write((char*) &temp, sizeof(float));
		}
	}

	out.close();

	cout << "done!" << endl;
}



Mesh3_real::Mesh3_real() : ShapeModel(SHAPE_MESH), gd(NULL)
{
	gd = new GeodesicDistance(*this);
}


Mesh3_real::~Mesh3_real()
{
	delete gd;
}

void Mesh3_real::loadGeometry()
{
	GLMmodel *model;
	GLfloat *modelNormals;
	int numNormals;
	bool useFacet = false;
	model = glmReadOBJ(qPrintable(geometryFile()));

	// Optional unitize argument
	//if (unitize) {
	//	glmUnitize(model);
	//}

	// If there are no normals
	if (model->numnormals == 0) {
		// generate facet normals at each vertex
		glmFacetNormals(model);
		numNormals = model->numfacetnorms;
		modelNormals = model->facetnorms;
		useFacet = true;
		//glmVertexNormals(model, 90);
		//numNormals = model->numnormals;
		//modelNormals = model->normals;
	}
	else {
		numNormals = model->numnormals;
		modelNormals = model->normals;
	}

	// Allocate memory
	vertex.resize(model->numvertices);
	normal.resize(numNormals);
	markVertex.resize(model->numvertices);

	// Initialize the bounding box for the mesh
	for (int d = 0; d < 3; d++) {
		box_min[d] = REAL_TYPE_MAX;
		box_max[d] = -REAL_TYPE_MAX;
	}		

	// Assign vertices
	// GLM reads into a 1-based array
	for(size_type v = 0,v_end = vertex.size();v < v_end;v++){
		for(size_type n = 0;n < 3;n++){
			vertex[v][n] = model->vertices[3 * (v+1) + n];

			// Calculate the bounding box of the mesh
			if (vertex[v][n] < box_min[n]) {
				box_min[n] = vertex[v][n];
			}
			else if (vertex[v][n] > box_max[n]) {
				box_max[n] = vertex[v][n];
			}
		}

		// Reset status flag
		markVertex[v] = VERTEX_VALID;
	}

	// Assign normals
	// GLM reads into a 1-based array
	for(size_type v = 0,v_end = normal.size();v < v_end;v++){
		for(size_type n = 0;n < 3;n++){
			normal[v][n] = modelNormals[3 * (v+1) + n];
		}
	}

	const size_type max_v = vertex.size()-1;
	size_type index = 0;

	triangle_vertex_index.resize(model->numtriangles);
	triangle_normal_index.resize(model->numtriangles);

	// GLM reads into a 1-based array
	for(size_type t = 0,t_end = triangle_vertex_index.size();
		t < t_end;++t){

			GLMtriangle& triangle = model->triangles[t];

			index3_type& tvi = triangle_vertex_index[index];
			index3_type& tni = triangle_normal_index[index];

			bool valid = true;

			for(size_type i = 0;i < 3;i++){
				const index_type tvii = triangle.vindices[i]-1;
				tvi[i] = tvii;
				tni[i] = (useFacet ? t : triangle.nindices[i]-1);

				valid = valid && (tvii <= max_v);
			}

			if (valid) index++;
	} // END OF for t

	triangle_vertex_index.resize(index);
	triangle_normal_index.resize(index);

	glmDelete(model);
}

void Mesh3_real::draw() const
{
	glBegin(GL_TRIANGLES);

	for(size_type t = 0,t_end = triangle_vertex_index.size();t < t_end;t++){

		const index3_type& tvi = triangle_vertex_index[t];
		//const index3_type& tni = triangle_normal_index[t];

		for(size_type i = 0;i < 3;i++){

			const real3_type& v = vertex[tvi[i]];
			//const real3_type& n = normal[tni[i]];
			const real3_type& n = normal[tvi[i]];

			glNormal3f(n[0],n[1],n[2]);
			glVertex3f(v[0],v[1],v[2]);
		}
	}

	glEnd();
}

void Mesh3_real::drawWithStatus() const
{
	glDisable(GL_LIGHTING);
	glBegin(GL_TRIANGLES);

	for(size_type t = 0,t_end = triangle_vertex_index.size();t < t_end;t++){

		const index3_type& tvi = triangle_vertex_index[t];
		//const index3_type& tni = triangle_normal_index[t];

		for(size_type i = 0;i < 3;i++){

			const real3_type& v = vertex[tvi[i]];
			//const real3_type& n = normal[tni[i]];
			const real3_type& n = normal[tvi[i]];

			switch(markVertex[tvi[i]]) {
			case VERTEX_NO_ADJACENT_TRIANGLE:
				glColor3f(1.0,0.0,0.0);
				break;
			case VERTEX_ON_BOUNDARY:
				glColor3f(0.0,0.0,1.0);
				break;
			default:
				glColor3f(1.0,1.0,1.0);
				break;
			}

			glNormal3f(n[0],n[1],n[2]);
			glVertex3f(v[0],v[1],v[2]);
		}
	}

	glEnd();

	// Splat the points without any adjacent triangles
	glDisable(GL_LIGHTING);
	glPointSize(5.f);
	glColor3f(1.0,0.0,0.0);
	glBegin(GL_POINTS);
	for (index_type i = 0; i < vertex.size(); i++) {
		real3_type v = vertex[i];

		if(markVertex[i] == VERTEX_NO_ADJACENT_TRIANGLE) {
			glVertex3f(v[0], v[1], v[2]);
		}
	}
	glEnd();
}

void Mesh3_real::drawSubset( idxList set ) const
{
	glBegin(GL_TRIANGLES);

	// Initialize a temporary list to just mark the selected vertices
	std::vector<bool> vertexMark;
	vertexMark.resize(vertex.size());
	for (index_type i = 0; i < vertex.size(); i++) {
		vertexMark[i] = false;
	}
	for (index_type i = 0; i < set.size(); i++) {
		vertexMark[set[i]] = true;
	}
	
	// Draw the triangles where all three vertices are marked
	for(size_type t = 0,t_end = triangle_vertex_index.size();t < t_end;t++)
	{
		const index3_type& tvi = triangle_vertex_index[t];
		//const index3_type& tni = triangle_normal_index[t];

		//////////////////////////////////////////////////////////////////////////
		// Check if all three vertices are selected
		bool draw = true;
		for(size_type i = 0;i < 3;i++){
			if (!vertexMark[tvi[i]]) {
				draw = false;
			}
		}

		//////////////////////////////////////////////////////////////////////////
		// If yes, then draw them
		if (draw) {
			for(size_type i = 0;i < 3;i++){

				const real3_type& v = vertex[tvi[i]];
				//const real3_type& n = normal[tni[i]];
				const real3_type& n = normal[tvi[i]];

				glNormal3f(n[0],n[1],n[2]);
				glVertex3f(v[0],v[1],v[2]);
			}
		}
	}

	glEnd();
}

ShapeModel* Mesh3_real::makeCopy() const
{
	ShapeModel *shape = new Mesh3_real(*this);
	if (this->features) {
		shape->features = new PointListN(*features);		/**< Deep-copy the feature vectors */
	}
	if (this->gd) {
		*((Mesh3_real*)shape)->gd = *this->gd;				/**< Deep-copy the edge information */
	}
	shape->attachSuffix();

	return shape;
}


void Mesh3_real::sampleVertices( size_type n, PointList &result ) const
{
	int numDense = 20*n;
	PointList dense;
	dense.resize(numDense);

	// Outline:
	// 1. Pick a triangle of the mesh according to its surface area
	size_type numTriangles = triangle_vertex_index.size();
	float *area = new float[numTriangles];
	float *cArea = new float[numTriangles+1];
	float avgArea, totalArea, pr, sample;

	for (index_type ti = 0; ti < numTriangles; ti++) {
		index3_type tri = triangle_vertex_index[ti];
		real3_type v[3], d[2], n;
		for (int i = 0; i < 3; i++) {
			v[i] = vertex[tri[i]];
		}

		d[0] = v[1] - v[0];
		d[1] = v[2] - v[0];
		n = d[0]^d[1];

		area[ti] = 0.5 * n.norm();

		//cout << "sa " << ti << ": " << area[ti] << endl;
	}

	// Construct the CDF for random sampling
	ComputeStep1dCDF(area, numTriangles, &avgArea, cArea);
	totalArea = avgArea*((real_type)numTriangles);

	//cout << "total area: " << totalArea << endl;

	// Pick a triangle based on surface area
	// Generate a random number in [0,1] using SFMT
	// Initialize random number generator
	init_gen_rand((unsigned)time( NULL ));

	//vector<int> countPicked;
	//countPicked.resize(numTriangles);

	//for (Simple_mesh::index_type i = 0; i < numTriangles; i++) {
	//	countPicked[i] = 0;
	//}

	for (int si = 0; si < numDense; si++) {
		float pf;
		index_type pick;

		sample = genrand_real2();
		pf = SampleStep1d(area, cArea, avgArea, numTriangles, sample, &pr);
		pick = floorf(pf*numTriangles);
		//cout << "Picked: " << pick << endl;
		//countPicked[pick]++;

		// Pick a random point within this triangle
		float sum = 2;
		float u1, u2;
		while (sum > 1) {
			u1 = genrand_real2();
			u2 = genrand_real2();
			sum = u1 + u2;
		}

		index3_type tri = triangle_vertex_index[pick];
		real3_type v[3];
		for (int i = 0; i < 3; i++) {
			v[i] = vertex[tri[i]];
		}

		real3_type pt = u1*(v[1] - v[0]) + u2*(v[2] - v[0]) + v[0];
		dense[si] = pt;

		//cout << "Sampled point: " << pt << endl;
	}

	//for (index_type i = 0; i < numTriangles; i++) {
	//	cout << "% picked for triangle " << i << ": " << ((float)countPicked[i])/((float)numSamples)*100.f;
	//	cout << " / % surface area: " << area[i] / totalArea * 100.f << endl;
	//}

	//////////////////////////////////////////////////////////////////////////
	// Now, pick a random point and remove all points within a certain 
	// radius around it; iterate until we reach numSamples.
	//
	// Pick the radius so that there will be numSamples points distributed on
	// the surface if we iterate indefinitely.  This can be accomplished by
	// taking the total surface area and determining how many samples can
	// fit in this area given the minimum distance between any pair is the radius.
	// Approximate this by finding, how many circles fit in this area?

	idxList inRadius;
	distList dist;
	float rsq = totalArea / (((real_type)n)*M_PI);

	result.resize(n);

	for (index_type i = 0; i < n; i++) {
		// Pick a random vertex
		index_type pick = floorf((real_type)dense.size() * genrand_real1());

		// Find all the vertices within a radius of that vertex
		inRadius.clear();
		dist.clear();
		dense.queryRange(dense[pick], rsq, inRadius, dist);

		// Assign the vertex to the list
		result[i] = dense[pick];

		// Remove all the vertices within that radius
		dense.removePoints(inRadius);
	}

	delete[] area;
	delete[] cArea;
}

void Mesh3_real::markBoundary_ComputeResolution()
{
	gd->createEdgeList();
	int numBoundaryEdges = 0;
	distList edgeLengths;
	edgeLengths.resize(gd->edgeList.size());

	for (index_type i = 0; i < gd->edgeList.size(); i++) {
		if (gd->edgeList[i].isBoundary()) {
			//cout << "Boundary edge detected: (" << gd->edgeList[i].vi[0] << "," << gd->edgeList[i].vi[1] << ")" << endl;

			// Mark the two endpoints of this edge
			markVertex[gd->edgeList[i].vi[0]] = VERTEX_ON_BOUNDARY;
			markVertex[gd->edgeList[i].vi[1]] = VERTEX_ON_BOUNDARY;

			numBoundaryEdges++;
		}

		// Compute length of the edge
		real3_type v0 = vertex[gd->edgeList[i].vi[0]];
		real3_type v1 = vertex[gd->edgeList[i].vi[1]];
		edgeLengths[i] = (v0 - v1).norm();
	}

	//int totalEdges = edgeLengths.size();
	//int medianIndex = ceil((((double)totalEdges) - 1.0) / 2.0);
	//std::partial_sort(edgeLengths.begin(), edgeLengths.begin() + medianIndex + 1, edgeLengths.end());

	//if (totalEdges % 2 == 0) {
	//	// If even, then average the two medians
	//	sampleResolution = (edgeLengths[medianIndex] + edgeLengths[medianIndex-1]) / 2.0;
	//}
	//else {
	//	sampleResolution = edgeLengths[medianIndex];
	//}

	cout << "****** Number of boundary edges detected: " << numBoundaryEdges << endl;
	//cout << "****** Mesh resolution: " << sampleResolution << endl;
}

/**
* Implementation of Algorithm by Max (1999), "Weights for computing vertex normals from facet normals."
* Outline at: http://graphics.ucsd.edu/twiki/bin/view.pl/Main/PerFacePerVertexCurvatures
*/
void Mesh3_real::computeNormals( bool force/* = false*/ )
{
	if (!force && (normal.size() == vertex.size())) {
		return;
	}

	cout << "*** Computing normals on triangle mesh...";

	// Allocate memory for storing normals at each vertex
	// real3_type already initializes values to zero.
	normal.clear();
	normal.resize(numVertices());

	// For each face
	for (index_type fi = 0; fi < triangle_vertex_index.size(); fi++) {
		const index3_type& tri = triangle_vertex_index[fi];
		real3_type u = vertex[tri[1]] - vertex[tri[0]];
		real3_type v = vertex[tri[2]] - vertex[tri[0]];
		real3_type w = vertex[tri[2]] - vertex[tri[1]];
		real3_type n = u^v;
		real_type a = u*u;
		real_type b = v*v;
		real_type c = w*w;

		normal[tri[0]] += n / (a*b);
		normal[tri[1]] += n / (a*c);
		normal[tri[2]] += n / (b*c);
	}

	// Normalize the normal vector at each vertex
	int numIsolatedVertices = 0;

	for (index_type vi = 0; vi < numVertices(); vi++) {
		normal[vi].normalize();

		// Check for numerical errors
		if (!_finite(normal[vi][0]) || !_finite(normal[vi][1]) || !_finite(normal[vi][2])) {
			//cerr << "Warning in Mesh3_real::computeNormals(): Non-finite valued normal! Possibly no adjacent triangles." << endl;

			// Mark the current vertex as having no adjacent triangles
			markVertex[vi] = VERTEX_NO_ADJACENT_TRIANGLE;
			numIsolatedVertices++;
		}
	}

	cout << " done!" << endl;

	cout << "****** Number of isolated vertices (no adjacent triangles): " << numIsolatedVertices << endl;
}


real_type Mesh3_real::pointSqDistance( const real3_type &p, real3_type &closestPoint )
{
	return ShapeModel::pointSqDistance(p, closestPoint);

	//real3_type temp;
	//real_type tempDistSq;
	//real_type distSq = REAL_TYPE_MAX;

	////////////////////////////////////////////////////////////////////////////
	//// Method 1: Naive method with quick bound
	//// Iterate over all triangles, but compute a quick lower bound
	//// to eliminate far-away triangles
	//for(size_type t = 0,t_end = triangle_vertex_index.size();t < t_end;t++){
	//	const index3_type& tvi = triangle_vertex_index[t];
	//	const real3_type& t0 = vertex[tvi[0]];
	//	const real3_type& t1 = vertex[tvi[1]];
	//	const real3_type& t2 = vertex[tvi[2]];

	//	if (distSq > lowerBoundTriangleSqDistance(p, t0, t1, t2)) {
	//		// Update with real distance
	//		tempDistSq = pointTriangleSqDistance(p, t0, t1, t2, temp);

	//		if (distSq > tempDistSq) {
	//			distSq = tempDistSq;
	//			closestPoint = temp;
	//		}
	//	}
	//}

	//return distSq;
}

void Mesh3_real::drawFeatureValues( const Scene &scene, int comp ) const
{
	// Plot the mesh along with the curvature value
	if (!features) {
		return;
	}

	// Clamp comp to be in the range
	int compMax = featureScales.size();
	comp = min(compMax-1, comp);

	glDisable(GL_LIGHTING);
	glBegin(GL_TRIANGLES);
	for(size_type t = 0,t_end = triangle_vertex_index.size();t < t_end;t++){
		const index3_type& tvi = triangle_vertex_index[t];
		//const index3_type& tni = triangle_normal_index[t];

		for(size_type i = 0;i < 3;i++){
			const real3_type& v = vertex[tvi[i]];
			//const real3_type& n = vertex_normal[tni[i]];

			setOpenGLColor(scene.getColormapColor((*features)[tvi[i]].val[comp]));
			//setOpenGLMaterialColor(color);
			//glNormal3f(n[0],n[1],n[2]);
			glVertex3f(v[0],v[1],v[2]);
		}
	}
	glEnd();

	// Draw the feature estimation scales
	drawSqRadii(featureScales, comp);
}

void Mesh3_real::drawCurvatureValues( const Scene &scene, int comp, int curvDisplayMode ) const
{
	if (!frames.size()) {
		return;
	}

	// Clamp comp to be in the range
	int compMax = frameScales.size();
	comp = min(compMax-1, comp);

	// Draw the mesh with the computed normals
	if (curvDisplayMode == 2 || curvDisplayMode == 3) {
		glDisable(GL_LIGHTING);
	}
	glBegin(GL_TRIANGLES);
	for(size_type t = 0,t_end = triangle_vertex_index.size();t < t_end;t++){
		const index3_type& tvi = triangle_vertex_index[t];
		//const index3_type& tni = triangle_normal_index[t];

		for(size_type i = 0;i < 3;i++){
			const real3_type& v = vertex[tvi[i]];
			const real3_type& n = frames[tvi[i]][comp].n;

			if (curvDisplayMode == 2) {
				setOpenGLColor(scene.getColormapColor(frames[tvi[i]][comp].kmax));
			}
			else if (curvDisplayMode == 3) {
				setOpenGLColor(scene.getColormapColor(frames[tvi[i]][comp].kmin));
			}

			glNormal3f(n[0],n[1],n[2]);
			glVertex3f(v[0],v[1],v[2]);
		}
	}
	glEnd();

	// Draw the feature estimation scales
	drawSqRadii(frameScales, comp);
}

void Mesh3_real::loadGeometryVTK( double decimateFactor /*= 0.0 */ )
{
	// Use VTK classes to load the geometry
	vtkOBJReader *poly = vtkOBJReader::New();
	poly->SetFileName(qPrintable(geometryFile()));

	// convert polygonal mesh into triangle mesh
	vtkTriangleFilter *tri = vtkTriangleFilter::New();
	tri->SetInput(poly->GetOutput());
	vtkPolyData *triData = tri->GetOutput();

	// decimate
	vtkQuadricDecimation *decimate = vtkQuadricDecimation::New();
	decimate->SetTargetReduction(decimateFactor);
	decimate->SetInput(triData);
	vtkPolyData *decimateData = decimate->GetOutput();

	// Do everything
	decimate->Update();

	cout << "Original triangle mesh: " << triData->GetNumberOfPoints() << " vertices, " << triData->GetNumberOfCells() << " triangles." << endl;
	cout << "Decimated triangle mesh: " << decimateData->GetNumberOfPoints() << " vertices, " << decimateData->GetNumberOfCells() << " triangles." << endl;

	//////////////////////////////////////////////////////////////////////////
	// Read the geometry data
	// 1. Allocate memory
	index_type nVerts = decimateData->GetNumberOfPoints();
	vertex.resize(nVerts);
	normal.resize(nVerts);
	markVertex.resize(nVerts);

	index_type nTris = decimateData->GetNumberOfCells();
	triangle_vertex_index.resize(nTris);
	triangle_normal_index.clear();					/**< We are not going to use triangle-based indexing of normals */

	// 2. Read list of points
	for (index_type i = 0; i < nVerts; i++) {
		double data[3];
		decimateData->GetPoint(i, data);

		for (int c = 0; c < 3; c++) {
			vertex[i][c] = data[c];

			// Calculate the bounding box of the mesh
			if (vertex[i][c] < box_min[c]) {
				box_min[c] = vertex[i][c];
			}
			else if (vertex[i][c] > box_max[c]) {
				box_max[c] = vertex[i][c];
			}
		}
	}

	// 3. Read list of triangles
	int *data = new int[10];
	int numData = 0;
	for (int i = 0; i < decimateData->GetNumberOfCells(); i++) {
		decimateData->GetCellPoints(i, numData, data);

		if (numData == 3) {
			for (int c = 0; c < 3; c++) {
				triangle_vertex_index[i][c] = data[c];
			}
		}
		else {
			cerr << "Non-triangle detected at index " << i << endl;
		}
	}

	// Clean-up
	delete data;
	decimate->Delete();
	tri->Delete();
	poly->Delete();
}

Mesh3_real* Mesh3_real::simplifyGeometry( double factor, double scale/* = 1.0*/ ) const
{
	Mesh3_real *shape = new Mesh3_real();
	shape->rootDir = this->rootDir;
	shape->modelDir = this->modelDir;
	shape->dataDir = this->dataDir;
	shape->filename = this->filename;
	shape->sampleResolution = this->sampleResolution;

	cout << "***Simplifying to target factor: " << factor << "..." << endl;
	cout << "***Scaling geometry by: " << scale << "..." << endl;

	shape->loadGeometryVTK(factor);
	shape->attachSuffix(QString("simplified"));

	// Scale the model
	for (index_type i = 0; i < shape->numVertices(); i++) {
		shape->vertex[i] = scale * shape->vertex[i];
	}
	shape->box_min = scale * shape->box_min;
	shape->box_max = scale * shape->box_max;

	// Save to disk
	shape->writeGeometryOBJ();

	// Perform only basic pre-processing
	shape->computeNormals(true);
	shape->markBoundary_ComputeResolution();

	return shape;
}

void Mesh3_real::writeGeometryOBJ() const
{
	// Based on http://www.nada.kth.se/~asa/Ray/vertface2obj.m

	// Check if the file already exists
	if (QFile::exists(qPrintable(geometryFile()))) {
		cerr << "Error in Mesh3_real::writeGeometryOBJ(): File " << qPrintable(geometryFile()) << " already exists." << endl;
		return;
	}

	ofstream out;
	out.open(qPrintable(geometryFile()));

	if (out.fail()) {
		cerr << "Error in Mesh3_real::writeGeometryOBJ(): Could not open output file " << qPrintable(geometryFile()) << " for writing!" << endl;
		return;
	}

	cout << "Saving vertex and triangle data in " << qPrintable(geometryFile()) << " ...";

	// Write vertex data
	for (index_type i = 0; i < numVertices(); i++) {
		out << "v";
		for (int c = 0; c < 3; c++) {
			out << " " << vertex[i][c];
		}
		out << "\n";
	}

	// Write triangle data
	out << "g foo\n";
	for (index_type i = 0; i < triangle_vertex_index.size(); i++) {
		out << "f";
		for (int c = 0; c < 3; c++) {
			out << " " << triangle_vertex_index[i][c]+1;
		}
		out << "\n";
	}
	out << "g\n";

	out.close();

	cout << "...done!" << endl;
}

Mesh3_real* Mesh3_real::pickSubset( idxList subset ) const
{
	Mesh3_real* shape = new Mesh3_real();
	shape = (Mesh3_real*)pickSample(subset, shape);

	// Copy over the connectivity information
	shape->triangle_vertex_index.clear();
	shape->triangle_normal_index.clear();

	// Initialize a temporary list to just mark the selected vertices
	std::vector<bool> vertexMark;
	std::vector<index_type> subsetIndex;
	vertexMark.resize(vertex.size());
	subsetIndex.resize(vertex.size());
	for (index_type i = 0; i < vertex.size(); i++) {
		vertexMark[i] = false;
		subsetIndex[i] = 0;
	}
	for (index_type i = 0; i < subset.size(); i++) {
		vertexMark[subset[i]] = true;
		subsetIndex[subset[i]] = i;
	}

	// Draw the triangles where all three vertices are marked
	for(size_type t = 0,t_end = triangle_vertex_index.size();t < t_end;t++)
	{
		const index3_type& tvi = triangle_vertex_index[t];
		//const index3_type& tni = triangle_normal_index[t];

		//////////////////////////////////////////////////////////////////////////
		// Check if all three vertices are selected
		bool draw = true;
		for(size_type i = 0;i < 3;i++){
			if (!vertexMark[tvi[i]]) {
				draw = false;
			}
		}

		//////////////////////////////////////////////////////////////////////////
		// If yes, then draw them
		if (draw) {
			index3_type triIndex;
			for(size_type i = 0;i < 3;i++){
				triIndex[i] = subsetIndex[tvi[i]];
			}
			shape->triangle_vertex_index.push_back(triIndex);
			//shape->triangle_normal_index.push_back(tni);
		}
	}

	return shape;
}
