#ifndef	__SHAPE_MODEL_H__
#define __SHAPE_MODEL_H__

#include <QString>
#include "PointList.h"
#include "SpinImage.h"
#include <limits>

class Scene;
class ShapeModel;
class PointListN;
class SE3;
class GeodesicDistance;

enum ShapeType {
	SHAPE_MESH = 0,
	SHAPE_POINTS
};

enum VertexStatus {
	VERTEX_VALID = 0,
	VERTEX_ON_BOUNDARY,
	VERTEX_NO_ADJACENT_TRIANGLE
};

enum CURVINFO_DISPLAY {
	CURVINFO_NORMALS = 1,
	CURVINFO_MIN_CURVATURE,
	CURVINFO_MAX_CURVATURE,
	CURVINFO_MIN_AND_MAX_CURVATURES
};

enum FEATURE_TYPE {
	FEATURE_PRINCIPAL_CURVATURES = 1,
	FEATURE_PRINCIPAL_CURVATURE_RATIO,
	FEATURE_PRINCIPAL_CURVATURE_NORM,
	FEATURE_MEAN_CURVATURE,
	FEATURE_SPIN_IMAGES
};

/**
 * Generic shape loading function.
 */
ShapeModel* loadShape( QString baseDir, QString filename, const Scene &scene, bool processFrames = true );

/**
 * Class for holding the principal curvature frame.
 */
class CurvInfo {
public:
	CurvInfo() : kmin(0.0), kmax(0.0) {};

	inline void makeOrthonormal() {
		// Perform Gram-Schmidt orthogonalization
		n.normalize();
		vmax = vmax - (vmax*n)*n;
		vmax.normalize();
		vmin = vmin - (vmin*n)*n - (vmin*vmax)*vmax;
		vmin.normalize();
	}

	inline bool checkOrthonormal() {
		static const real_type eps = 1e-5;

		// 1. Check orthogonality
		if ((abs(n*vmax) > eps) || (abs(n*vmin) > eps) || (abs(vmin*vmax) > eps)) {
			return false;
		}

		// 2. Check length
		if ((n.norm() > (1 + eps)) || (n.norm() < (1 - eps)) ||
			(vmax.norm() > (1 + eps)) || (vmax.norm() < (1 - eps)) ||
			(vmin.norm() > (1 + eps)) || (vmin.norm() < (1 - eps)))
		{
			return false;
		}

		return true;
	}

	inline real_type principalCurvatureRatio() const {
		if (abs(kmax) <= REAL_TYPE_EPSILON) {
			return 0.0;
		}
		else {
			return abs(kmin / kmax);
		}
	}

	real3_type n, vmin, vmax;
	real_type kmin, kmax;
};

typedef std::vector< std::vector<CurvInfo> > CurvInfoList;


/**
* Class for holding information about the closest point on a shape.
*/
class ClosestPointInfo {
public:
	ClosestPointInfo() : index(0), dist(std::numeric_limits<real_type>::infinity()) { };
	//ClosestPointInfo() : index(0), frameInfo(NULL), dist(REAL_TYPE_MAX), valid(false), indexValid(false) { };

	index_type index;					/**< Index of the closest point on the shape */
	real_type dist;						/**< Distance (NOT squared!) to the closest point */

	//real3_type position;				/**< Position of the closest point */
	//const CurvInfo *frameInfo;			/**< Pointer to vertex's local frame information */
	//bool valid;							/**< Flag information for storing validity */
};

/**
 * A general description of a shape.  This class describes a shape with point samples.
 */
class ShapeModel {
public:
	//////////////////////////////////////////////////////////////////////////
	// Constructor / Destructor
	ShapeModel( int type );
	~ShapeModel();

	const int shapeType;											/**< Type of the shape. */
	virtual ShapeModel* makeCopy() const;							/**< Make and return a copy of the derived object. */

	bool useRootDir;												/**< Boolean indicating whether to use the root dir */
	void setFile( QString baseDir, QString filename );				/**< Set path/filename, which we will treat as a unique identifier. */
	QString uid() const;											/**< Returns the model dir + filename: the unique identifier. */
	QString getFilename() const;									/**< Returns just the filename. */
	QString geometryFile() const;									/**< Get the complete location for the geometry file. */
	QString frameFile() const;										/**< The location for saving frames. */
	QString featureFile() const;									/**< The location for saving features. */
	QString spinImageFile() const;									/**< Location for saving spin images. */
	void attachSuffix( QString suffix = QString("copy") );			/**< Attach a suffix to the filename. */
	void printInfo() const;											/**< Print information about the shape */

	real_type SampleResolution() const;								/**< The vertex sample resolution (mesh resolution) */
	void setSampleResolution(real_type res);						/**< Set the mesh resolution */

	//////////////////////////////////////////////////////////////////////////
	// Load/Save/Access functions: automatically load from disk on demand
	// If data doesn't exist, then automatically create and save it
	inline size_type numVertices() const;
	virtual void loadGeometry();									/**< Parse the filename and load it */
	virtual void computeNormals( bool force = false );				/**< Compute normals */

	// Compute / load / save local frames
	void computeFrames( const distList &scales, bool force = false );
	void saveFrames( const QString &description );
	bool loadFrames();

	// Compute / load / save feature vectors
	void computeFeatures( int featureType, const distList &scales, real_type eps );
	void makePrincipalCurvatureFeatures();
	void makePrincipalCurvatureRatioFeatures();
	void makePrincipalCurvatureNormFeatures();
	void makeMeanCurvatureFeatures( const distList &scales, real_type eps );
	void saveFeatures();
	bool loadFeatures();

	// Compute / Load / Save Spin Images
	void allocateSpinImageStack();
	void computeSpinImages( real_type binSize_MR, int imageWidth, real_type supportAngle );
	void computeSpinImage( index_type i );
	void saveSingleSpinImage( const char *filename, index_type i );
	void saveSpinImages();
	bool loadSpinImages();

	void transformVertex( const SE3 &tf, index_type i );								/**< Transform a vertex & associated frames */
	void transformVertex( const ShapeModel &shape, const SE3 &tf, index_type i );		/**< Apply transform to another shape and assign it to this one*/

	//////////////////////////////////////////////////////////////////////////
	// Data
	PointList vertex, normal;							/**< Vertex & Normal Data */
	real3_type box_min, box_max;						/**< Bounding box */
	CurvInfoList frames;								/**< Local frame (per vertex) */
	PointListN *features;								/**< Multi-scale feature vector (per vertex) */
	distList frameScales, featureScales;				/**< Scale values for frames and features */
	std::vector<int> markVertex;						/**< Status flag for each vertex */
	SpinImageStack spinImageStack;						/**< Spin images for each vertex */

	// K-d tree

	//////////////////////////////////////////////////////////////////////////
	// Query functions

	// Neighbor query (e.g. adjacent vertices on a mesh, knn query if a point rep'n)

	// K-nearest neighbor query & Range query
	void queryKNN(real3_type q, size_type k, idxList& nearest, distList& dist, bool persistent = false);
	void queryRange(real3_type q, float radSq, idxList& inRange, distList& dist, bool persistent = false);

	// Closest point query (e.g. what is the closest point on the mesh, 1-nn query if a point rep'n)
	virtual ClosestPointInfo findClosestPoint( const real3_type &p );
	virtual void hybridClosestPoint( const real3_type &p, real3_type &closest, real_type &dist );
	virtual real_type pointSqDistance( const real3_type &p, real3_type &closestPoint );
	virtual real_type pointSqDistance( const real3_type &p, real3_type &closestPoint, CurvInfo &frameInfo );
	virtual real_type pointSqDistance( const real3_type &p, index_type &closestIndex );

	void clearANNData();

	//////////////////////////////////////////////////////////////////////////
	// Display functions (OpenGL)

	// Displaying vertex data, normals, local frames, feature vectors, k-d tree, etc. etc.
	virtual void draw() const;
	virtual void drawWithStatus() const;
	virtual void drawSubset( idxList set ) const;
	virtual void drawFeatureValues( const Scene &scene, int comp ) const;
	virtual void drawFrames( const Scene &scene, int comp, int curvDisplayMode ) const;
	virtual void drawCurvatureValues( const Scene &scene, int comp, int curvDisplayMode ) const;
	void drawSqRadii( const distList &sqRadii, int comp ) const;

	//////////////////////////////////////////////////////////////////////////
	// Sampling functions
	idxList pruneInvalidVertices( const idxList &list ) const;
	idxList pruneSamplePrincipalCurvatureRatio( const idxList &list, real_type maxRatio ) const;
	ShapeModel* subSample( size_type n, real_type pruneMaxRatio ) const;
	virtual void sampleVertices( size_type n, PointList &result ) const;
	ShapeModel* pickSample( idxList subset, ShapeModel *shape = NULL ) const;

protected:

	QString rootDir;						/**< Base directory path. */
	QString modelDir, dataDir;				/**< Directories containing the models & the data. */
	QString filename;						/**< The model's filename. */

	real_type sampleResolution;				/**< The model's sample resolution. */

};


/**
 * Derived Mesh class.
 */
class Mesh3_real : public ShapeModel {
public:
	Mesh3_real();
	~Mesh3_real();

	virtual ShapeModel* makeCopy() const;
	virtual void loadGeometry();

	Mesh3_real* simplifyGeometry( double factor, double scale = 1.0 ) const;		/**< Return a new simplified mesh */
	void loadGeometryVTK( double decimateFactor = 0.0 );							/**< Load mesh using VTK and use Quadric Decimation to simplify the mesh */
	void writeGeometryOBJ() const;

	void markBoundary_ComputeResolution();											/**< Mark boundary edges and find the median edge length */
	virtual void computeNormals( bool force = false );				/**< Compute normals per vertex by simply averaging the normals of the incident triangles. */

	//////////////////////////////////////////////////////////////////////////
	// Querying
	virtual real_type pointSqDistance( const real3_type &p, real3_type &closestPoint );

	//////////////////////////////////////////////////////////////////////////
	// Sampling
	virtual void sampleVertices( size_type n, PointList &result ) const;			/**< Sample points uniformly (based on surface area) on a mesh. */
	Mesh3_real* pickSubset( idxList subset ) const;									/**< Return a subset of the mesh */

	//////////////////////////////////////////////////////////////////////////
	// Drawing
	virtual void draw() const;
	virtual void drawWithStatus() const;
	virtual void drawSubset( idxList set ) const;
	virtual void drawFeatureValues( const Scene &scene, int comp ) const;
	virtual void drawCurvatureValues( const Scene &scene, int comp, int curvDisplayMode ) const;

	//////////////////////////////////////////////////////////////////////////
	// Additional Data
	idx3List triangle_vertex_index;
	idx3List triangle_normal_index;
	GeodesicDistance *gd;
};

#endif
