#include "ShapeSlicer.h"
#include "scene.h"
#include "viewer.h"
#include <algorithm>
using namespace std;

ShapeSlicer::ShapeSlicer( Scene *sc, ShapeModel* s ) :
scene(sc), shape(s), pointSelect(3), selectionMode_(NONE),
contextMenu(NULL), loadSliceResultAct(NULL), saveSliceResultAct(NULL)
{
	// Destroy the object when closing
	setAttribute(Qt::WA_DeleteOnClose);

	if (!shape) {
		// Immediately kill
		close();
	}

	shapeColor[0] = 1.0; shapeColor[1] = 0.7; shapeColor[2] = 0.7;

	selectAll();
}

ShapeSlicer::~ShapeSlicer()
{
	if (contextMenu) delete contextMenu;
	if (saveSliceResultAct) delete saveSliceResultAct;
	if (loadSliceResultAct) delete loadSliceResultAct;
}

void ShapeSlicer::init()
{
	createContextMenu();

	//////////////////////////////////////////////////////////////////////////
	// Disable the exit key
	setShortcut(EXIT_VIEWER, 0);
	setWindowTitle("Shape Slicer");

	// Center the view on the source and target
	setSceneBoundingBox(qglviewer::Vec(shape->box_min[0], shape->box_min[1], shape->box_min[2]),
		qglviewer::Vec(shape->box_max[0], shape->box_max[1], shape->box_max[2]));
	//camera()->setZClippingCoefficient(100);
	//camera()->setZNearCoefficient(1e-5);
	showEntireScene();

	//////////////////////////////////////////////////////////////////////////
	// Set up the material
	glEnable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);

	// Used to display semi-transparent selection rectangle
	glBlendFunc(GL_ONE, GL_ONE);

	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Set snapshot parameters
	setSnapshotQuality(0);
	setSnapshotFormat(tr("PNG"));

	// Set the max select buffer size
	setSelectBufferSize(100000);

	//////////////////////////////////////////////////////////////////////////
	cout << "...done!" << endl;
}

void ShapeSlicer::draw()
{
	// Draw the target shape
	glEnable(GL_LIGHTING);
	setOpenGLMaterialColor(shapeColor);
	shape->drawSubset(selected);

	// Draws rectangular selection area. Could be done in postDraw() instead.
	if (selectionMode_ == TOGGLE)
		drawSelectionRectangle();
}

void ShapeSlicer::drawWithNames()
{
	glPointSize(5.0);
	shape->vertex.drawWithNames(POINT_SQUARES);
}

void ShapeSlicer::drawSelectionRectangle() const
{
	startScreenCoordinatesSystem();
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);

	glColor4f(0.0, 0.0, 0.3, 0.3);
	glBegin(GL_QUADS);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glLineWidth(2.0);
	glColor4f(0.4, 0.4, 0.5, 0.5);
	glBegin(GL_LINE_LOOP);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);
	stopScreenCoordinatesSystem();
}

void ShapeSlicer::mousePressEvent( QMouseEvent *e )
{
	// Start selection.
	rectangle_ = QRect(e->pos(), e->pos());

	if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ControlModifier)) {
		selectionMode_ = TOGGLE;
	}
	else {
		QGLViewer::mousePressEvent(e);
	}
}

void ShapeSlicer::mouseMoveEvent(QMouseEvent* e)
{
	if (selectionMode_ == TOGGLE) {
		// Updates rectangle_ coordinates and redraws rectangle
		rectangle_.setBottomRight(e->pos());
		updateGL();
	}
	else {
		QGLViewer::mouseMoveEvent(e);
	}
}

void ShapeSlicer::mouseReleaseEvent(QMouseEvent* e)
{
	if (selectionMode_ == TOGGLE) {
		// Actual selection on the rectangular area.
		// Possibly swap left/right and top/bottom to make rectangle_ valid.
		rectangle_ = rectangle_.normalized();

		int rw, rh;
		rw = rectangle_.width();
		rh = rectangle_.height();

		if (rw <= 0) rw = 1;
		if (rh <= 0) rh = 1;

		// Define selection window dimensions
		setSelectRegionWidth(rw);
		setSelectRegionHeight(rh);

		// Compute rectangle center and perform selection
		select(rectangle_.center());

		// reset selection mode
		selectionMode_ = NONE;

		// Update display to show new selected objects
		updateGL();
	}
	else {
		QGLViewer::mouseReleaseEvent(e);
	}
}

void ShapeSlicer::endSelection( const QPoint& )
{
	// Flush GL buffers
	glFlush();

	// Get the number of objects that were seen through the pick matrix frustum. Reset GL_RENDER mode.
	GLint nbHits = glRenderMode(GL_RENDER);

	//cout << "Number of objects selected: " << nbHits << endl;

	// Interpret results : each object created 4 values in the selectBuffer().
	// (selectBuffer())[4*i+3] is the id pushed on the stack.
	if (nbHits > 0) {
		// If the rectangle is small, then select a single object of minimum depth.
		if ((rectangle_.width() < pointSelect) || (rectangle_.height() < pointSelect)) {
			GLuint zMin = (selectBuffer())[1];
			int minId = (selectBuffer())[3];

			for (int i=1; i<nbHits; ++i) {
				if ((selectBuffer())[4*i+1] < zMin) {
					zMin = (selectBuffer())[4*i+1];
					minId = (selectBuffer())[4*i+3];
				}
			}

			addIdSelection(minId);
		}
		else {
			for (int i=0; i<nbHits; ++i) {
				addIdSelection((selectBuffer())[4*i+3]);
			}
		}
	}
}

void ShapeSlicer::keyPressEvent( QKeyEvent *e )
{
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	if ((e->key()==Qt::Key_Space) && (modifiers==Qt::NoButton)) {
		// Compute the center location of this window
		QRect win = geometry();
		QPoint center;
		center.setX(win.x() + win.width() / 7);
		center.setY(win.y() + win.height() / 7);

		// Open the menu at this location
		contextMenu->exec(center);
		updateGL();
	}
	else if ((e->key()==Qt::Key_A) && (modifiers==Qt::CTRL)) {
		// Select all vertices
		selectAll();
		updateGL();
	}
	else {
		QGLViewer::keyPressEvent(e);
	}
}

void ShapeSlicer::createContextMenu()
{
	contextMenu = new QMenu(tr("&Menu"), this);

	saveSliceResultAct = new QAction(tr("&Save Slice Result..."), this);
	contextMenu->addAction(saveSliceResultAct);
	connect(saveSliceResultAct, SIGNAL(triggered()), this, SLOT(saveSliceResult()));

	loadSliceResultAct = new QAction(tr("&Load Slice Result..."), this);
	contextMenu->addAction(loadSliceResultAct);
	connect(loadSliceResultAct, SIGNAL(triggered()), this, SLOT(loadSliceResult()));

	addToSceneAct = new QAction(tr("&Add Slice to Scene..."), this);
	contextMenu->addAction(addToSceneAct);
	connect(addToSceneAct, SIGNAL(triggered()), this, SLOT(addToScene()));
}

void ShapeSlicer::addIdSelection( index_type id )
{
	idxList::iterator found = std::find(selected.begin(), selected.end(), id);
	if (found == selected.end()) {
		// Uncomment to enable toggling
		//selected.push_back(id);
	}
	else {
		selected.erase(found);
	}
}


void ShapeSlicer::saveSliceResult()
{
	// Add a new frame to the animation.
	QString file = QFileDialog::getSaveFileName(this, tr("Save Mesh"), tr(""), tr("Meshes (*.obj)"));

	if (!file.size()) {
		return;
	}

	if (shape->shapeType == SHAPE_MESH) {
		Mesh3_real* newShape = ((Mesh3_real*)shape)->pickSubset(selected);
		newShape->setFile(tr(scene->rootDir.c_str()), file);
		newShape->writeGeometryOBJ();
	}
	else {
		QMessageBox::information(this, tr("Error!"), tr("Saving out a point set is not implemented yet!"), QMessageBox::Ok);
	}
}

void ShapeSlicer::loadSliceResult()
{
	// Add a new frame to the animation.
	QString file = QFileDialog::getOpenFileName(this, tr("Open Mesh"), tr(""), tr("Meshes (*.obj)"));

	if (file.size() > 0) {
		shape = loadShape(tr(scene->rootDir.c_str()), file, *scene, false);
		selectAll();
	}
}

void ShapeSlicer::selectAll()
{
	selected.resize(shape->numVertices());
	for (index_type i = 0; i < selected.size(); i++) {
		selected[i] = i;
	}
}

void ShapeSlicer::addToScene()
{
	if (shape->shapeType == SHAPE_MESH) {
		Mesh3_real* newShape = ((Mesh3_real*)shape)->pickSubset(selected);
		scene->getAnimation()->add(newShape);
	}
	else {
		QMessageBox::information(this, tr("Error!"), tr("Adding a point set to the scene is not implemented yet!"), QMessageBox::Ok);
	}
}