// Adapted from the multiSelect example
#include "TransformViewer.h"
#include <QMouseEvent>
#include "scene.h"
#include "geom_util.h"

using namespace qglviewer;
using namespace std;

real_type globalPointSize;

SE3 getFrameTf(const qglviewer::Frame &fr)
{
	// Convert the frame to a rigid transformation
	SE3 tf;
	float rotM[3][3];

	// Translation
	tf.tr = VecToVec3f(fr.translation());

	// Rotation
	fr.rotation().getRotationMatrix(rotM);
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			tf.rot(i,j) = rotM[i][j];
		}
	}

	return tf;
}

void TransformObject::draw() const
{
	glPushMatrix();
	getFrameTf(frame).applyGL();
		// Draw the point
		glPointSize(3.f);
		glBegin(GL_POINTS);
		glVertex3f(position[0],position[1], position[2]);
		glEnd();

		// Draw the normal
		real3_type v = position + globalPointSize*normal;
		glBegin(GL_LINES);
		glVertex3f(position[0], position[1], position[2]);
		glVertex3f(v[0], v[1], v[2]);
		glEnd();
	glPopMatrix();


	//////////////////////////////////////////////////////////////////////////
	// Manually apply the transformation to the point and the normal
	//SE3 tf = getFrameTf(frame);
	//real3_type pos = tf * position;
	//glPointSize(3.f);
	//glBegin(GL_POINTS);
	//glVertex3f(pos[0],pos[1], pos[2]);
	//glEnd();

	//// Draw the normal
	//real3_type v = pos + globalPointSize*(tf.rot * normal);
	//glBegin(GL_LINES);
	//glVertex3f(pos[0], pos[1], pos[2]);
	//glVertex3f(v[0], v[1], v[2]);
	//glEnd();
}

void ManipulatedFrameSetConstraint::clearSet()
{
	objects_.clear();
}

void ManipulatedFrameSetConstraint::addObjectToSet(TransformObject* o)
{
	objects_.append(o);
}

void ManipulatedFrameSetConstraint::constrainTranslation(qglviewer::Vec &translation, Frame *const)
{
	for (QList<TransformObject*>::iterator it=objects_.begin(), end=objects_.end(); it != end; ++it)
		(*it)->frame.translate(translation);
}

void ManipulatedFrameSetConstraint::constrainRotation(qglviewer::Quaternion &rotation, Frame *const frame)
{
	// A little bit of math. Easy to understand, hard to guess (tm).
	// rotation is expressed in the frame local coordinates system. Convert it back to world coordinates.
	const Vec worldAxis = frame->inverseTransformOf(rotation.axis());
	const Vec pos = frame->position();
	const float angle = rotation.angle();

	for (QList<TransformObject*>::iterator it=objects_.begin(), end=objects_.end(); it != end; ++it)
	{
		// Rotation has to be expressed in the object local coordinates system.
		Quaternion qObject((*it)->frame.transformOf(worldAxis), angle);
		(*it)->frame.rotate(qObject);

		// Comment these lines only rotate the objects
		Quaternion qWorld(worldAxis, angle);
		// Rotation around frame world position (pos)
		(*it)->frame.setPosition(pos + qWorld.rotate((*it)->frame.position() - pos));
	}
}


TransformViewer::TransformViewer( Scene *sc, ShapeModel *source, ShapeModel *target ) :
selectionMode_(NONE), scene(sc), sourceShape(source), targetShape(target), showSource(true), showTarget(true),
dataCost(NULL), smoothCost(NULL), assignedTf(NULL), useSameColor(false), showRegistrationError(false)
{
	// Destroy the object when closing
	setAttribute(Qt::WA_DeleteOnClose);

	// Set the window title.
	setWindowTitle("Transformation Viewer / Editor");
}

TransformViewer::~TransformViewer()
{
	// Delete all the objects
	for (int i=0; i<int(objects_.size()); i++) {
		delete objects_.at(i);
	}
	
	// Delete data and smoothness costs
	if (dataCost) {
		delete dataCost;
	}
	if (smoothCost) {
		delete smoothCost;
	}

	// Delete assigned transformations
	if (assignedTf) {
		delete assignedTf;
	}
}

void TransformViewer::init()
{
	restoreStateFromFile();

	if (!scene || !sourceShape || !targetShape) {
		// Close immediately.
		this->close();
		return;
	}

	globalPointSize = scene->pointSize();

	// Create an object for each point on the source shape.
	for (index_type i = 0; i < sourceShape->numVertices(); i++) {
		TransformObject* o = new TransformObject();
		//o->frame.setPosition(Vec(i/float(nb), j/float(nb), 0.0f));
		o->position = sourceShape->vertex[i];
		o->normal = sourceShape->normal[i];
		objects_.append(o);
	}

	// Initialize assigned transformations array
	assignedTf = new std::vector<SE3>();
	assignedTf->resize(objects_.size());

	// Specify which data and smooth functions to evaluate
	dataFunctions.push_back(&data_PointToPoint);
	dataFunctions.push_back(&data_PointToPlane);
	dataFunctions.push_back(&data_Hybrid);
	dataFunctions.push_back(&data_D6);
	smoothFunctions.push_back(&smooth_EdgeLength);

	// Specifiy names of the data and smooth functions
	dataFunctionNames.push_back(tr("Point-to-point"));
	dataFunctionNames.push_back(tr("Point-to-plane"));
	dataFunctionNames.push_back(tr("Hybrid"));
	dataFunctionNames.push_back(tr("Point and Normal Distance"));
	smoothFunctionNames.push_back(tr("Edge length difference"));

	// Store the current costs
	dataCost = new data_type(dataFunctions.size());
	smoothCost = new data_type(smoothFunctions.size());


	//////////////////////////////////////////////////////////////////////////
	// A ManipulatedFrameSetConstraint will apply displacements to the selection
	setManipulatedFrame(new ManipulatedFrame());
	manipulatedFrame()->setConstraint(new ManipulatedFrameSetConstraint());

	// Set signal to update cost whenever the assignment changes
	connect(manipulatedFrame(), SIGNAL(manipulated()), this, SLOT(evaluateCost()));

	// Disable using current color as color of material
	glDisable(GL_COLOR_MATERIAL);

	// Used to display semi-transparent selection rectangle
	glBlendFunc(GL_ONE, GL_ONE);

	// Disable the exit key
	setShortcut(EXIT_VIEWER, 0);

	//////////////////////////////////////////////////////////////////////////
	// Center the view on the source and target
	real3_type box_min, box_max;
	for (int d = 0; d < 3; d++) {
		box_min[d] = min(sourceShape->box_min[d], targetShape->box_min[d]);
		box_max[d] = max(sourceShape->box_max[d], targetShape->box_max[d]);
	}
	setSceneBoundingBox(qglviewer::Vec(box_min[0], box_min[1], box_min[2]),
		qglviewer::Vec(box_max[0], box_max[1], box_max[2]));
	//camera()->setZClippingCoefficient(100);
	//camera()->setZNearCoefficient(1e-5);
	showEntireScene();

	//////////////////////////////////////////////////////////////////////////
	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Set snapshot parameters
	setSnapshotQuality(0);
	setSnapshotFormat(tr("PNG"));

	// Set the max select buffer size
	setSelectBufferSize(100000);
}

QString TransformViewer::helpString() const
{
	QString text("<h2>m u l t i S e l e c t </h2>");
	text += "This example illustrates an application of the <code>select()</code> function that ";
	text += "enables the selection of several objects.<br><br>";
	text += "Object selection is preformed using the left mouse button. Press <b>Shift</b> to add objects ";
	text += "to the selection, and <b>Alt</b> to remove objects from the selection.<br><br>";
	text += "Individual objects (click on them) as well as rectangular regions (click and drag mouse) can be selected. ";
	text += "To do this, the selection region size is modified and the <code>endSelection()</code> function ";
	text += "has been overloaded so that <i>all</i> the objects of the region are taken into account ";
	text += "(the default implementation only selects the closest object).<br><br>";
	text += "The selected objects can then be manipulated by pressing the <b>Control</b> key. ";
	text += "Other set operations (parameter edition, deletion...) can also easily be applied to the selected objects.";
	return text;
}


//  D r a w i n g   f u n c t i o n

void TransformViewer::draw()
{
	glDisable(GL_LIGHTING);

	// Draws selected objects only.
	glColor3f(0.9, 0.3, 0.3);
	for (QList<int>::const_iterator it=selection_.begin(), end=selection_.end(); it != end; ++it)
		objects_.at(*it)->draw();

	glEnable(GL_LIGHTING);


	if (showRegistrationError) {
		drawRegistrationError();
		drawColormap();
	}
	else {
		if (showSource) {
			if (useSameColor) {
				setOpenGLMaterialColor(scene->targetColor);
			}
			else {
				setOpenGLMaterialColor(scene->sourceColor);
			}

			if (sourceShape->shapeType == SHAPE_MESH) {
				const Mesh3_real &temp = *((Mesh3_real*)sourceShape);

				// Draw deformed source shape
				glBegin(GL_TRIANGLES);
				for(size_type t = 0,t_end = temp.triangle_vertex_index.size();t < t_end;t++){
					const index3_type& tvi = temp.triangle_vertex_index[t];

					for(size_type i = 0;i < 3;i++){
						SE3 tf = getFrameTf(objects_[tvi[i]]->frame);
						real3_type v = tf * objects_[tvi[i]]->position;
						real3_type n = tf.rot * objects_[tvi[i]]->normal;

						glNormal3f(n[0],n[1],n[2]);
						glVertex3f(v[0],v[1],v[2]);
					}
				}
				glEnd();
			}
			else {
				sourceShape->draw();
			}
		}

		if (showTarget) {
			setOpenGLMaterialColor(scene->targetColor);
			targetShape->draw();
		}
	}

	// Draws manipulatedFrame (the set's rotation center)
	if (manipulatedFrame()->isManipulated())
	{
		glPushMatrix();
		glMultMatrixd(manipulatedFrame()->matrix());
		drawAxis(0.5);
		glPopMatrix();
	}

	// Draws rectangular selection area. Could be done in postDraw() instead.
	if (selectionMode_ != NONE)
		drawSelectionRectangle();


	// Show current error function costs
	glDisable(GL_LIGHTING);
	setOpenGLColor(foregroundColor().rgba());
	const int yOffset = 15;
	if (dataCost && smoothCost) {
		for (index_type i = 0; i < dataFunctions.size(); i++) {
			int j = dataFunctions.size() - i - 1;
			int x = 10;
			int y = height() - yOffset*(2 + j + smoothFunctions.size() + dataFunctions.size());
			drawText(x, y, tr("Total cost (%1): %2").arg(dataFunctionNames[i]).arg(dataCost->val[i] + smoothCost->val[0]));
		}
	}
	if (dataCost) {
		for (index_type i = 0; i < dataFunctions.size(); i++) {
			int j = dataFunctions.size() - i - 1;
			int x = 10;
			int y = height() - yOffset*(2 + j + smoothFunctions.size());
			drawText(x, y, tr("Data cost (%1): %2").arg(dataFunctionNames[i]).arg(dataCost->val[i]));
		}
	}
	if (smoothCost) {
		for (index_type i = 0; i < smoothFunctions.size(); i++) {
			int j = smoothFunctions.size() - i - 1;
			int x = 10;
			int y = height() - yOffset*(2 + j);
			drawText(x, y, tr("Smooth cost (%1): %2").arg(smoothFunctionNames[i]).arg(smoothCost->val[i]));
		}
	}
}


//   C u s t o m i z e d   m o u s e   e v e n t s

void TransformViewer::mousePressEvent(QMouseEvent* e)
{
	// Start selection. Mode is ADD with Shift key and TOGGLE with Alt key.
	rectangle_ = QRect(e->pos(), e->pos());

	if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ShiftModifier))
		selectionMode_ = ADD;
	else
		if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier))
			selectionMode_ = REMOVE;
		else
		{
			if (e->modifiers() == Qt::ControlModifier)      
				startManipulation();
			QGLViewer::mousePressEvent(e);
		}
}

void TransformViewer::mouseMoveEvent(QMouseEvent* e)
{
	if (selectionMode_ != NONE)
	{
		// Updates rectangle_ coordinates and redraws rectangle
		rectangle_.setBottomRight(e->pos());

		updateGL();
	}
	else
		QGLViewer::mouseMoveEvent(e);
}

void TransformViewer::mouseReleaseEvent(QMouseEvent* e)
{
	if (selectionMode_ != NONE)
	{
		// Actual selection on the rectangular area.
		// Possibly swap left/right and top/bottom to make rectangle_ valid.
		rectangle_ = rectangle_.normalized();

		// Define selection window dimensions
		setSelectRegionWidth(rectangle_.width());
		setSelectRegionHeight(rectangle_.height());
		// Compute rectangle center and perform selection
		select(rectangle_.center());
		// Update display to show new selected objects
		updateGL();
	}
	else {
		QGLViewer::mouseReleaseEvent(e);
	}
}


//   C u s t o m i z e d   s e l e c t i o n   p r o c e s s

void TransformViewer::drawWithNames()
{
	for (int i=0; i<int(objects_.size()); i++)
	{
		glPushName(i);
		objects_.at(i)->draw();
		glPopName();
	}
}

void TransformViewer::endSelection(const QPoint&)
{
	// Flush GL buffers
	glFlush();

	// Get the number of objects that were seen through the pick matrix frustum. Reset GL_RENDER mode.
	GLint nbHits = glRenderMode(GL_RENDER);

	if (nbHits > 0)
	{
		// Interpret results : each object created 4 values in the selectBuffer().
		// (selectBuffer())[4*i+3] is the id pushed on the stack.
		for (int i=0; i<nbHits; ++i)
			switch (selectionMode_)
		{
			case ADD    : addIdToSelection((selectBuffer())[4*i+3]); break;
			case REMOVE : removeIdFromSelection((selectBuffer())[4*i+3]);  break;
			default : break;
		}
	}
	selectionMode_ = NONE;
}

void TransformViewer::startManipulation()
{
	ManipulatedFrameSetConstraint* mfsc = (ManipulatedFrameSetConstraint*)(manipulatedFrame()->constraint());
	mfsc->clearSet();

	for (QList<int>::const_iterator it=selection_.begin(), end=selection_.end(); it != end; ++it)
	{
		mfsc->addObjectToSet(objects_[*it]);
	}

	updateManipulatedFrameCenter();
}


void TransformViewer::updateManipulatedFrameCenter()
{
	real3_type averagePosition;
	for (QList<int>::const_iterator it=selection_.begin(), end=selection_.end(); it != end; ++it)
	{
		SE3 tf = getFrameTf(objects_[*it]->frame);
		averagePosition += tf * objects_[*it]->position;
	}

	if (selection_.size() > 0)
		manipulatedFrame()->setPosition(Vec3fToVec(averagePosition) / selection_.size());
}


//   S e l e c t i o n   t o o l s

void TransformViewer::addIdToSelection(int id)
{
	if (!selection_.contains(id))
		selection_.push_back(id);
}

void TransformViewer::removeIdFromSelection(int id)
{
	selection_.removeAll(id);
}

void TransformViewer::drawSelectionRectangle() const
{
	startScreenCoordinatesSystem();
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);

	glColor4f(0.0, 0.0, 0.3, 0.6);
	glBegin(GL_QUADS);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glDisable(GL_BLEND);

	glLineWidth(2.0);
	setOpenGLColor(foregroundColor().rgba());
	glBegin(GL_LINE_LOOP);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glEnable(GL_LIGHTING);
	stopScreenCoordinatesSystem();
}

void TransformViewer::keyPressEvent( QKeyEvent *e )
{
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	if ((e->key()==Qt::Key_F) && (modifiers==Qt::NoButton)) {
		if (showSource && showTarget) {
			showSource = true;
			showTarget = false;
		}
		else if (showSource && !showTarget) {
			showSource = false;
			showTarget = true;
		}
		else if (!showSource && showTarget) {
			showSource = true;
			showTarget = true;
		}
		updateGL();
	}
	else if ((e->key()==Qt::Key_C) && (modifiers==Qt::NoButton)) {
		evaluateCost();
		updateGL();
	}
	else if ((e->key()==Qt::Key_S) && (modifiers==Qt::NoButton)) {
		useSameColor = !useSameColor;
		updateGL();
	}
	else if ((e->key()==Qt::Key_M) && (modifiers==Qt::CTRL)) {
		// Save the transformed mesh
		QString file = QFileDialog::getSaveFileName(this, tr("Save Mesh"), scene->lastDir, tr("Meshes (*.obj)"));

		if (!file.size()) {
			return;
		}

		if (sourceShape->shapeType == SHAPE_MESH) {
			Mesh3_real* newShape = (Mesh3_real*) ((Mesh3_real*)sourceShape)->makeCopy();

			// Deform shape
			for(size_type vi = 0; vi < newShape->numVertices() ; vi++){
				SE3 tf = getFrameTf(objects_[vi]->frame);
				real3_type v = tf * objects_[vi]->position;
				real3_type n = tf.rot * objects_[vi]->normal;

				newShape->vertex[vi] = v;
				newShape->normal[vi] = n;
			}

			newShape->setFile(tr(scene->rootDir.c_str()), file);
			newShape->writeGeometryOBJ();
		}
		else {
			QMessageBox::information(this, tr("Error!"), tr("Saving out a point set is not implemented yet!"), QMessageBox::Ok);
		}
	}
	else if ((e->key()==Qt::Key_E) && (modifiers==Qt::CTRL)) {
		computeRegistrationError();
		showRegistrationError = true;
		updateGL();
	}
	else if ((e->key()==Qt::Key_R) && (modifiers==Qt::CTRL)) {
		showRegistrationError = !showRegistrationError;
		updateGL();
	}
	else {
		QGLViewer::keyPressEvent(e);
	}
}

std::vector<SE3>& TransformViewer::getAssignedTf()
{
	if (assignedTf->size() != objects_.size()) {
		assignedTf->resize(objects_.size());
	}

	for (index_type i = 0; i < assignedTf->size(); i++) {
		(*assignedTf)[i] = getFrameTf(objects_[i]->frame);
	}

	return *assignedTf;
}

void TransformViewer::setAssignedTf( index_type i, const SE3& tf )
{
	//if (assignedTf) {
	//	(*assignedTf)[i] = tf;
	//}

	// Update the frame of each object
	qglviewer::Frame &frame = objects_[i]->frame;
	qglviewer::Quaternion q;
	double rotM[3][3];

	// Rotation
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			rotM[i][j] = tf.rot(i,j);
		}
	}
	q.setFromRotationMatrix(rotM);
	frame.setRotation(q);

	// Translation
	frame.setTranslation(Vec3fToVec(tf.tr));
}


void TransformViewer::evaluateCost()
{
	// Get assigned transforms
	std::vector<SE3> &assignedTf = getAssignedTf();

	// Evaluate the cost
	*dataCost = evaluateDataCost(assignedTf, *sourceShape, *targetShape, dataFunctions);
	*smoothCost = evaluateSmoothCost(assignedTf, *sourceShape, *targetShape, smoothFunctions);

	//// Print the cost
	//cout << "-------------------------------------------------------------------------------" << endl;
	//cout << "Evaluating cost functions:" << endl;
	//for (index_type i = 0; i < dataFunctions.size(); i++) {
	//	cout << "*** Data cost (" << i << ", " << qPrintable(dataFunctionNames[i]) << "): " << dataCost->val[i] << endl;
	//}

	//for (index_type i = 0; i < smoothFunctions.size(); i++) {
	//	cout << "*** Smooth cost (" << i << ", " << qPrintable(smoothFunctionNames[i]) << "): " << smoothCost->val[i] << endl;
	//}
}

void TransformViewer::computeRegistrationError()
{
	if (sourceShape->shapeType != SHAPE_MESH) {
		return;
	}

	cout << "Computing the registration error of all vertices in the shape...";
	
	const Mesh3_real &compShape = *((Mesh3_real*)sourceShape);

	registrationError.resize(targetShape->numVertices());

	TextProgress(targetShape->numVertices(), true);
	for (index_type i = 0; i < targetShape->numVertices(); i++) {
		real3_type ref = targetShape->vertex[i];
		real3_type v1, v2, v3, closest;
		real_type minDist = REAL_TYPE_MAX;
		SE3 tf;

		// Compute the distance to all the triangles in the target mesh; find the minimum
		for(size_type t = 0,t_end = compShape.triangle_vertex_index.size();t < t_end;t++){
			const index3_type& tvi = compShape.triangle_vertex_index[t];

			tf = getFrameTf(objects_[tvi[0]]->frame);
			v1 = tf * objects_[tvi[0]]->position;
			tf = getFrameTf(objects_[tvi[1]]->frame);
			v2 = tf * objects_[tvi[1]]->position;
			tf = getFrameTf(objects_[tvi[2]]->frame);
			v3 = tf * objects_[tvi[2]]->position;

			real_type dist = pointTriangleSqDistance(ref, v1, v2, v3, closest);

			if (dist < minDist) {
				minDist = dist;
			}
		}

		// Divide the minimum error by the scene size
		registrationError[i] = sqrt(minDist) / scene->totalScene_boundingBoxDiagonal;

		//cout << "Error in vertex " << i << ": " << registrationError[i] << endl;
		TextProgress(i);
	}

	cout << "done!" << endl;
}

void TransformViewer::drawRegistrationError()
{
	// Draw the error on the target mesh
	if (registrationError.size() != targetShape->numVertices()) {
		return;
	}

	if (targetShape->shapeType == SHAPE_MESH) {
		const Mesh3_real &drawShape = *((Mesh3_real*)targetShape);

		glDisable(GL_LIGHTING);
		glBegin(GL_TRIANGLES);
		for(size_type t = 0,t_end = drawShape.triangle_vertex_index.size();t < t_end;t++){
			const index3_type& tvi = drawShape.triangle_vertex_index[t];

			for(size_type i = 0;i < 3;i++){
				const real3_type& v = drawShape.vertex[tvi[i]];

				setOpenGLColor(scene->getColormapColor(registrationError[tvi[i]]));
				glVertex3f(v[0],v[1],v[2]);
			}
		}
		glEnd();
	}
}

void TransformViewer::drawColormap()
{
	const int subdivisions = 100;
	float w = width();
	float h = height();
	float mapWidth = 30;
	float mapOffset = 20;
	float top = 0.25*h;
	float bottom = 0.75*h;
	float left = w - mapWidth - mapOffset;
	float right = w - mapOffset;
	float current = top;
	float offset = abs(bottom - top) / ((float)subdivisions);

	startScreenCoordinatesSystem();
	glDisable(GL_LIGHTING);

	glBegin(GL_QUAD_STRIP);
	for (int i = 0; i < subdivisions; i++) {
		RealColor currentColor = scene->colormap.getColor(((float)subdivisions - i - 1) / ((float)subdivisions - 1));
		setOpenGLColor(currentColor);
		glVertex2i(left,  current);
		glVertex2i(right, current);
		current += offset;
	}
	glEnd();

	glEnable(GL_LIGHTING);
	stopScreenCoordinatesSystem();

	// Draw the min and max labels
	drawText(left - mapOffset, top, tr("%1").arg(scene->colormapMax));
	drawText(left - mapOffset, bottom, tr("%1").arg(scene->colormapMin));
}