#include "geodesic.h"
#include <limits.h>
#include <QString>

using namespace std;

GeodesicDistance::GeodesicDistance() : input(NULL) {

}

/**
 * Initialize class with an input mesh.
 */
GeodesicDistance::GeodesicDistance( const Mesh3_real &inputMesh ) : input(&inputMesh) {

}

/**
 * Print the edge list.
 */
void GeodesicDistance::printEdgeList(ostream& out/* = std::cout*/, int numDigits/* = 0*/) {
	QString format("e%1: (v%2, v%3) (t%4, t%5)");
	QString str;

	for (index_type ei = 0, eEnd = edgeList.size(); ei < eEnd; ei++) {
		Edge &current = edgeList[ei];
		str = format.arg(ei, numDigits)
					.arg(current.vi.x(), numDigits)
					.arg(current.vi.y(), numDigits)
					.arg(current.ti.x(), numDigits)
					.arg(current.ti.y(), numDigits);

		out << qPrintable(str) << endl;
	}
	out << endl;
}

/**
 * Print the adjacency list.
 */
void GeodesicDistance::printAdjList(ostream& out/* = std::cout*/, int numDigits/* = 0*/) {
	QString basicFormat("v%1:%2");
	QString pairFormat(" (v%1, e%2)");
	QString str;

	for (index_type vi = 0, vEnd = adjList.size(); vi < vEnd; vi++) {
		QString listStr("");

		// print the list of adjacent vertices
		for (index_type ai = 0, aEnd = adjList[vi].size(); ai < aEnd; ai++) {
			listStr = listStr + pairFormat.arg(adjList[vi][ai].x(),numDigits).arg(adjList[vi][ai].y(),numDigits);
		}

		str = basicFormat.arg(vi, numDigits).arg(listStr);

		out << qPrintable(str) << endl;
	}
	out << endl;
}


/**
 * Create a list of edges (along with their adjacent triangles) for the input mesh.
 *
 * We actually maintain two lists: the adjacency list for vertices, and the list
 * of edges.  The adjacency list stores, for each vertex v0, a list of adjacent
 * vertices v_adj along with the edge index # for the edge (v0 -> v_adj).
 * We use this adjacency list to check if the edge is already present in the
 * edge list, avoiding duplicates.
 *
 * To generate the list of edges, for each triangle, we add the three edges into
 * the edge list while checking for duplicates in the adjacency list.
 *
 */
void GeodesicDistance::createEdgeList() {
	edgeList.clear();
	adjList.clear();

	index_type numVertices = input->numVertices();
	index_type numTriangles = input->triangle_vertex_index.size();

	adjList.resize(numVertices);

	// Construct a list of edges
	for (index_type ti = 0, tNum = numTriangles;
		 ti < tNum; ti++)
	{
		// Grab the vertices
		index3_type tri = input->triangle_vertex_index[ti];

		// For each edge
		for (int i = 1; i <= 3; i++) {
			int i0 = i - min(i,2); int i1 = i - i0;		// iterates in (0, 1), (0, 2), (1, 2)
			int v0, v1;									// grabs the vertices of the edges
			bool found;
			index_type j;

			// Check to see if the edge is already in the adjList
			// Do both orders of the edge (v0->v1 and v1->v0)
			for (int temp = 0; temp < 2; temp++) {
				// The endpoints of the edge
				if (temp == 0) {
					v0 = tri[i0]; v1 = tri[i1];
				}
				else {
					v0 = tri[i1]; v1 = tri[i0];
				}

				found = false;
				for (j = 0; j < adjList[v0].size(); j++) {
					if (adjList[v0][j].x() == v1) {
						found = true;
						break;
					}
				}

				if (found) {
					// Update the edge with the current triangle index
					edgeList[adjList[v0][j].y()].ti[1] = ti;
				}
				else {
					index_type ei;

					if (temp == 0) {
						// Create a new edge
						Edge ne;
						ne.vi[0] = v0; ne.vi[1] = v1;
						ne.ti[0] = ti;

						// push back onto the edgeList
						edgeList.push_back(ne);

						ei = edgeList.size()-1;
					}
					else {
						// The edge must be stored already in the adjList
						found = false;
						for (j = 0; j < adjList[v1].size(); j++) {
							if (adjList[v1][j].x() == v0) {
								found = true;
								break;
							}
						}
						if (found) {
							ei = adjList[v1][j].y();
						}
						else {
							// This case should not happen
							throw "createEdgeList(): Reverse edge not found in adjList!";
						}
					}

					// Add to the adjacency list
					adjList[v0].push_back(index2_type(v1, ei));
				}
			}
		}
	}
}

/**
 * Find the min and max degree, and store a list of boundary vertices.
 */
void GeodesicDistance::computeStats() {
	// Compute the min and max degree of a vertex
	minDegree = input->vertex.size();
	maxDegree = 0;
	for (index_type vi = 0, vEnd = adjList.size(); vi < vEnd; vi++) {
		unsigned int deg = adjList[vi].size();
		if (deg < minDegree) {
			minDegree = deg;
		}
		if (deg > maxDegree) {
			maxDegree = deg;
		}
	}

	// Find all boundary edges
	bdyEdges.clear();
	for (index_type ei = 0, eEnd = edgeList.size(); ei < eEnd; ei++) {
		if (edgeList[ei].isBoundary()) {
			bdyEdges.push_back(ei);
		}
	}

	cout << "Minimum Degree: " << minDegree << endl;
	cout << "Maximum Degree: " << maxDegree << endl;
	cout << "Number of Boundary Edges: " << bdyEdges.size() << endl;
}


/**
 * Compute geodesic distance function from the source vertex.
 *
 * Outline of algorithm:
 *   -# Initialize priority queue (add initial windows)
 *   -# While priority queue is not empty
 *     - Remove a window from the top of the queue (minimum distance)
 *     - Propagate the window
 *
 * Notes: use STL priority_queue (in <queue>)
 * This method will just manage the priority queue (also generating the initial windows).
 * There will be another method dedicated to propagating the windows.
 *
 */
void GeodesicDistance::computeGeodesic( index_type vStart )
{
	index_type numVertices = input->vertex.size();
	index_type numTriangles = input->triangle_vertex_index.size();
	index_type numEdges = edgeList.size();

	// Initialize data structure
	allWindows.clear();
	allWindows.resize(numEdges);

	//////////////////////////////////////////////////////////////////////////
	// TODO: Implementation; see paper.
	//////////////////////////////////////////////////////////////////////////
}

void GeodesicDistance::setInput( const Mesh3_real &input_mesh )
{
	input = &input_mesh;
}

EdgeWindow::EdgeWindow() :
end(0,0), dist(0,0), leftSide(true), distSource(0) {

}

EdgeWindow::EdgeWindow( real_type b0, real_type b1,
										  real_type d0, real_type d1,
										  bool side, real_type sigma ) :
end(b0, b1), dist(d0, d1), leftSide(side), distSource(sigma) {

}

Edge::Edge() : vi(0, 0), ti(0, 0) {

}

/**
 * Determine if the edge is on the boundary.
 *
 * This method assumes that the edge list has been constructed in full.
 * After construction of the edge list, triangle index 0 can only be on the first element,
 * since triangle 0 is the first triangle to be processed and its three edges are added in the first iteration.
 * Therefore, if the second triangle index is the default value 0, then it is a boundary edge.
 */
bool Edge::isBoundary()
{
	return (ti.y() == 0);
}