#ifndef __GEODESIC_H__
#define __GEODESIC_H__

#include "ShapeModel.h"
#include "util.h"

/**
* Edge type.
* Contains the endpoints of the edge (two vertices) and the two adjacent triangles.
* In particular we know that the edge is on the boundary if there is only one
* adjacent triangle.
*/
class Edge {
public:
	Edge();
	bool isBoundary();

	index2_type vi;		// adjacent vertices
	index2_type ti;		// adjacent triangles
};

typedef std::vector<Edge> EdgeList;

/**
* Window type.
* This is used to store geodesic distance on the edges of the mesh.
* There can be multiple windows for each edge.
*/
class EdgeWindow {
public:
	EdgeWindow();
	EdgeWindow(real_type b0, real_type b1, real_type d0, real_type d1, bool side, real_type sigma);

private:
	real2_type end;
	real2_type dist;
	bool leftSide;
	real_type distSource;
};

typedef std::vector<EdgeWindow> WindowList;			/**< List of windows, to be associated with each edge. */

/**
 * Class for computing geodesic distance on a mesh.
 * This class contains all the algorithms for calculating geodesic distance.
 * The algorithm is from Surazhsky et al. (Siggraph 2005).
 */
class GeodesicDistance {
public:
	GeodesicDistance();
	GeodesicDistance( const Mesh3_real &inputMesh );
	
	void setInput( const Mesh3_real &input_mesh );
	void createEdgeList();							/**< Create the edge list */
	void computeStats();							/**< Compute degree stats and mark boundary vertices */
	void printEdgeList(std::ostream& out = std::cout, int numDigits = 0);	/**< Debugging: print the edge list */
	void printAdjList(std::ostream& out = std::cout, int numDigits = 0);	/**< Debugging: print the adjacency list */

	//////////////////////////////////////////////////////////////////////////
	const Mesh3_real *input;							/**< Input mesh */
	std::vector<WindowList> allWindows;					/**< List of windows on each edge */
	EdgeList edgeList;									/**< Edge list*/
	std::vector<std::vector<index2_type> > adjList;		/**< Vertex adjacency list */
	unsigned int minDegree, maxDegree;					/**< Degree statistics */
	std::vector<index_type> bdyEdges;					/**< Mark boundary edges */

	//////////////////////////////////////////////////////////////////////////
	void computeGeodesic(index_type vStart);		/**< Compute the geodesic distance from vStart */

};

#endif
