// Adapted from:
// Wild Magic Source Code
// David Eberly
// http://www.geometrictools.com
// Copyright (c) 1998-2007
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or (at
// your option) any later version.  The license is available for reading at
// either of the locations:
//     http://www.gnu.org/copyleft/lgpl.html
//     http://www.geometrictools.com/License/WildMagicLicense.pdf
//
// Version: 4.0.1 (2007/05/06)

#include "geom_util.h"

real_type pointTriangleSqDistance( const real3_type &p, const real3_type &t0, const real3_type &t1, const real3_type &t2, real3_type &closestPoint )
{
	real3_type kDiff = t0 - p;
	real3_type kEdge0 = t1 - t0;
	real3_type kEdge1 = t2 - t0;
	real_type fA00 = kEdge0 * kEdge0;
	real_type fA01 = kEdge0 * kEdge1;
	real_type fA11 = kEdge1 * kEdge1;
	real_type fB0 = kDiff * kEdge0;
	real_type fB1 = kDiff * kEdge1;
	real_type fC = kDiff * kDiff;
	real_type fDet = abs(fA00*fA11-fA01*fA01);
	real_type fS = fA01*fB1-fA11*fB0;
	real_type fT = fA01*fB0-fA00*fB1;
	real_type fSqrDistance;

	if (fS + fT <= fDet)
	{
		if (fS < (real_type)0.0)
		{
			if (fT < (real_type)0.0)  // region 4
			{
				if (fB0 < (real_type)0.0)
				{
					fT = (real_type)0.0;
					if (-fB0 >= fA00)
					{
						fS = (real_type)1.0;
						fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
					}
					else
					{
						fS = -fB0/fA00;
						fSqrDistance = fB0*fS+fC;
					}
				}
				else
				{
					fS = (real_type)0.0;
					if (fB1 >= (real_type)0.0)
					{
						fT = (real_type)0.0;
						fSqrDistance = fC;
					}
					else if (-fB1 >= fA11)
					{
						fT = (real_type)1.0;
						fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
					}
					else
					{
						fT = -fB1/fA11;
						fSqrDistance = fB1*fT+fC;
					}
				}
			}
			else  // region 3
			{
				fS = (real_type)0.0;
				if (fB1 >= (real_type)0.0)
				{
					fT = (real_type)0.0;
					fSqrDistance = fC;
				}
				else if (-fB1 >= fA11)
				{
					fT = (real_type)1.0;
					fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
				}
				else
				{
					fT = -fB1/fA11;
					fSqrDistance = fB1*fT+fC;
				}
			}
		}
		else if (fT < (real_type)0.0)  // region 5
		{
			fT = (real_type)0.0;
			if (fB0 >= (real_type)0.0)
			{
				fS = (real_type)0.0;
				fSqrDistance = fC;
			}
			else if (-fB0 >= fA00)
			{
				fS = (real_type)1.0;
				fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
			}
			else
			{
				fS = -fB0/fA00;
				fSqrDistance = fB0*fS+fC;
			}
		}
		else  // region 0
		{
			// minimum at interior point
			real_type fInvDet = ((real_type)1.0)/fDet;
			fS *= fInvDet;
			fT *= fInvDet;
			fSqrDistance = fS*(fA00*fS+fA01*fT+((real_type)2.0)*fB0) +
				fT*(fA01*fS+fA11*fT+((real_type)2.0)*fB1)+fC;
		}
	}
	else
	{
		real_type fTmp0, fTmp1, fNumer, fDenom;

		if (fS < (real_type)0.0)  // region 2
		{
			fTmp0 = fA01 + fB0;
			fTmp1 = fA11 + fB1;
			if (fTmp1 > fTmp0)
			{
				fNumer = fTmp1 - fTmp0;
				fDenom = fA00-2.0f*fA01+fA11;
				if (fNumer >= fDenom)
				{
					fS = (real_type)1.0;
					fT = (real_type)0.0;
					fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
				}
				else
				{
					fS = fNumer/fDenom;
					fT = (real_type)1.0 - fS;
					fSqrDistance = fS*(fA00*fS+fA01*fT+2.0f*fB0) +
						fT*(fA01*fS+fA11*fT+((real_type)2.0)*fB1)+fC;
				}
			}
			else
			{
				fS = (real_type)0.0;
				if (fTmp1 <= (real_type)0.0)
				{
					fT = (real_type)1.0;
					fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
				}
				else if (fB1 >= (real_type)0.0)
				{
					fT = (real_type)0.0;
					fSqrDistance = fC;
				}
				else
				{
					fT = -fB1/fA11;
					fSqrDistance = fB1*fT+fC;
				}
			}
		}
		else if (fT < (real_type)0.0)  // region 6
		{
			fTmp0 = fA01 + fB1;
			fTmp1 = fA00 + fB0;
			if (fTmp1 > fTmp0)
			{
				fNumer = fTmp1 - fTmp0;
				fDenom = fA00-((real_type)2.0)*fA01+fA11;
				if (fNumer >= fDenom)
				{
					fT = (real_type)1.0;
					fS = (real_type)0.0;
					fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
				}
				else
				{
					fT = fNumer/fDenom;
					fS = (real_type)1.0 - fT;
					fSqrDistance = fS*(fA00*fS+fA01*fT+((real_type)2.0)*fB0) +
						fT*(fA01*fS+fA11*fT+((real_type)2.0)*fB1)+fC;
				}
			}
			else
			{
				fT = (real_type)0.0;
				if (fTmp1 <= (real_type)0.0)
				{
					fS = (real_type)1.0;
					fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
				}
				else if (fB0 >= (real_type)0.0)
				{
					fS = (real_type)0.0;
					fSqrDistance = fC;
				}
				else
				{
					fS = -fB0/fA00;
					fSqrDistance = fB0*fS+fC;
				}
			}
		}
		else  // region 1
		{
			fNumer = fA11 + fB1 - fA01 - fB0;
			if (fNumer <= (real_type)0.0)
			{
				fS = (real_type)0.0;
				fT = (real_type)1.0;
				fSqrDistance = fA11+((real_type)2.0)*fB1+fC;
			}
			else
			{
				fDenom = fA00-2.0f*fA01+fA11;
				if (fNumer >= fDenom)
				{
					fS = (real_type)1.0;
					fT = (real_type)0.0;
					fSqrDistance = fA00+((real_type)2.0)*fB0+fC;
				}
				else
				{
					fS = fNumer/fDenom;
					fT = (real_type)1.0 - fS;
					fSqrDistance = fS*(fA00*fS+fA01*fT+((real_type)2.0)*fB0) +
						fT*(fA01*fS+fA11*fT+((real_type)2.0)*fB1)+fC;
				}
			}
		}
	}

	// account for numerical round-off error
	if (fSqrDistance < (real_type)0.0)
	{
		fSqrDistance = (real_type)0.0;
	}

	closestPoint = t0 + fS*kEdge0 + fT*kEdge1;
	//m_afTriangleBary[1] = fS;
	//m_afTriangleBary[2] = fT;
	//m_afTriangleBary[0] = (real_type)1.0 - fS - fT;
	return fSqrDistance;
}


real_type pointBoxSqDistance ( const real3_type &p, const real3_type &boxMin, const real3_type &boxMax, real3_type &closestPoint )
{
	// work in the box's coordinate system
	real3_type boxCenter = (boxMin + boxMax) / (real_type)2.0;
	real3_type halfLength = (boxMax - boxMin) / (real_type)2.0;
	real3_type kDiff = p - boxCenter;

	// compute squared distance and closest point on box
	real_type fSqrDistance = (real_type)0.0, fDelta;
	real3_type kClosest;

	for (int i = 0; i < 3; i++)
	{
		kClosest[i] = kDiff[i];
		if (kClosest[i] < -halfLength[i])
		{
			fDelta = kClosest[i] + halfLength[i];
			fSqrDistance += fDelta*fDelta;
			kClosest[i] = -halfLength[i];
		}
		else if (kClosest[i] > halfLength[i])
		{
			fDelta = kClosest[i] - halfLength[i];
			fSqrDistance += fDelta*fDelta;
			kClosest[i] = halfLength[i];
		}
	}

	closestPoint = boxCenter + kClosest;

	return fSqrDistance;
}


real_type lowerBoundTriangleSqDistance( const real3_type &p, const real3_type &t0, const real3_type &t1, const real3_type &t2 )
{
	// Find bounding box of triangle
	static real3_type boxMin, boxMax;
	boxMin[0] = boxMin[1] = boxMin[2] = REAL_TYPE_MAX;
	boxMax[0] = boxMax[1] = boxMax[2] = -REAL_TYPE_MAX;

	for (int d = 0; d < 3; d++) {
		if (t0[d] < boxMin[d])      boxMin[d] = t0[d];
		else if (t0[d] > boxMax[d]) boxMax[d] = t0[d];
		if (t1[d] < boxMin[d])      boxMin[d] = t1[d];
		else if (t1[d] > boxMax[d]) boxMax[d] = t1[d];
		if (t2[d] < boxMin[d])      boxMin[d] = t2[d];
		else if (t2[d] > boxMax[d]) boxMax[d] = t2[d];
	}

	//////////////////////////////////////////////////////////////////////////
	// Method 1: Lower-bound via distance to bounding box of triangle
	static real3_type temp;
	return pointBoxSqDistance(p, boxMin, boxMax, temp);

	//////////////////////////////////////////////////////////////////////////
	// Method 2: Approximate bounding sphere as center of box with length half of diagonal
	//real3_type center = (boxMax + boxMin) / (real_type)2.0;
	//real_type r = (boxMax - boxMin).norm() / (real_type)2.0;
	//real3_type diff = p - center;
	//return max(0.0, diff*diff - r*r);
}
