#include "scene.h"
#include <ctime>
#include <utility>
#include <iostream>

using namespace std;

extern "C" {
	#include <SFMT/SFMT.h>
}

const char *globalForegroundColor = "black";
const char *globalBackgroundColor = "white";

Scene::Scene() : animation(NULL), target(NULL), source(NULL),
cl(NULL), matchCandidates(NULL),
defaultLoadFrames(0), defaultSkipFrames(1),
sourceFrame(0), targetFrame(0),
autoProcessMode(AUTO_PROCESS_NONE),
currentBone(0), colormapMin(0), colormapMax(0),
robust_PC(false), ANNEpsilon_MC(0),
featureSamplesSource(1000), featureSamplesTarget(1000), featureType(FEATURE_PRINCIPAL_CURVATURES),
ANNEpsilon_FM(0), radius_FM(0), maxCandidates_FM(100), threshold_FM(0),
maxDist_FM(100), pruneMaxRatio_FM(1.0),
ANNEpsilon_MS(0), radius_MS(0), maxIter_MS(100), rate_MS(1),
ANNEpsilon_VR(0), minVerifyClusterSize(0), minMatchSize_VR(0), maxNumVerifyClusters_VR(-1), maxNumLabels_GC(-1),
growRegionDistThreshold_VR(1), skipVerification_VR(false), graphCutSolutionType_GC(2), useDummyLabel_GC(false),
mismatchError_GC(10000), dummyLabelSmoothError_GC(100), dummyLabelDataError_GC(100), bdySmoothnessKNN_GC(0), closestPtDistThreshold_GC(1),
smoothWeight_GC(1.0), dataWeight_GC(1.0), smoothTruncateValue_GC(100.0), dataTruncateValue_GC(100.0),
dataTermType_GC(GRAPH_CUT_DATA_POINT_TO_POINT), graphCutInstance_GC(GRAPH_CUT_INSTANCE_SYMMETRIC),
RANSAC_max_itr(0), RANSAC_sourceModel_ANNEpsilon(0), RANSAC_targetModel_ANNEpsilon(0),
RANSAC_inlier_queryRadius(0), RANSAC_inlier_epsilonBand(0), RANSAC_minimumInliersPercentage(0),
showClusterMode(0)
{
	//////////////////////////////////////////////////////////////////////////
	// Pick random numbers
	//////////////////////////////////////////////////////////////////////////
	// Initialize random number generator
	init_gen_rand((unsigned)time( NULL ));
	for (index_type i = 0; i < 10000; i++) {
		// generate a random number
		cpi.push_back(genrand_real1());
	}

	//////////////////////////////////////////////////////////////////////////
	// Initialize color defaults
	sourceColor.resize(3);
	targetColor.resize(3);
	pointColor.resize(3);
	selectedColor.resize(3);
	sourceColor[0] = 0.7; sourceColor[1] = 0.7; sourceColor[2] = 1.0;
	targetColor[0] = 1.0; targetColor[1] = 0.7; targetColor[2] = 0.7;
	pointColor[0] = 1.0; pointColor[1] = 1.0; pointColor[2] = 0.0;
	selectedColor[0] = 0.0; selectedColor[1] = 1.0; selectedColor[2] = 1.0;
}

Scene::~Scene()
{
	clear();
}


void Scene::solveCorrespondences()
{
	initMatchCandidates(*source, *target);

	if (!clusterResultExists()) {
		perfTimers["Sub-sampling geometry"].Start();
		matchCandidates->subSampleGeometry(featureSamplesSource, featureSamplesTarget, pruneMaxRatio_FM);
		perfTimers["Sub-sampling geometry"].Stop();

		perfTimers["Populating correspondence candidates"].Start();
		if (featureType == FEATURE_SPIN_IMAGES) {
			matchCandidates->spinImageCandidates(maxCandidates_FM);
		}
		else {
			matchCandidates->findCandidateCorrespondence(radius_FM*radius_FM, maxCandidates_FM, ANNEpsilon_FM);
		}
		perfTimers["Populating correspondence candidates"].Stop();

		perfTimers["Mean-shift clustering"].Start();
		matchCandidates->meanShiftBasic(radius_MS*radius_MS,rate_MS,maxIter_MS,ANNEpsilon_MS);
		perfTimers["Mean-shift clustering"].Stop();

		perfTimers["Extracting dense correspondences (ICP)"].Start();
		if (skipVerification_VR) {
			matchCandidates->skipVerify(minVerifyClusterSize, maxNumVerifyClusters_VR);
		}
		else {
			matchCandidates->verifyRegion(minVerifyClusterSize, maxNumVerifyClusters_VR, makeSceneRadius(growRegionDistThreshold_VR), ANNEpsilon_VR);
		}
		matchCandidates->sortVerifyResult(minMatchSize_VR, maxNumLabels_GC);
		perfTimers["Extracting dense correspondences (ICP)"].Stop();

		perfTimers["Graph cut optimization"].Start();
		if (graphCutInstance_GC == GRAPH_CUT_INSTANCE_SYMMETRIC) {
			matchCandidates->GraphCutRelabel(
											graphCutSolutionType_GC,
											dataTermType_GC,
											useDummyLabel_GC,
											bdySmoothnessKNN_GC,
											makeSceneRadius(closestPtDistThreshold_GC),
											makeSceneRadius(mismatchError_GC),
											makeSceneRadius(dummyLabelSmoothError_GC),
											makeSceneRadius(dummyLabelDataError_GC),
											smoothWeight_GC,
											dataWeight_GC,
											makeSceneRadius(smoothTruncateValue_GC),
											makeSceneRadius(dataTruncateValue_GC));
		}
		else if (graphCutInstance_GC == GRAPH_CUT_INSTANCE_FORWARD) {
			matchCandidates->GraphCutRelabel_Forward(
													graphCutSolutionType_GC,
													dataTermType_GC,
													smoothWeight_GC,
													dataWeight_GC,
													makeSceneRadius(smoothTruncateValue_GC),
													makeSceneRadius(dataTruncateValue_GC));
		}
		else {
			cerr << "Unknown graph cut instance type " << graphCutInstance_GC << "!" << endl;
		}
		perfTimers["Graph cut optimization"].Stop();

		cout << "-------------------------------------------------------------------------------" << endl;
		cout << "Running time summary" << endl;
		printTime("Sub-sampling geometry");
		printTime("Populating correspondence candidates");
		printTime("Mean-shift clustering");
		printTime("Extracting dense correspondences (ICP)");
		printTime("Graph cut optimization");
		cout << "-------------------------------------------------------------------------------" << endl;
	}
}

void Scene::clear()
{
	clearAnimation();
}

void Scene::clearAnimation()
{
	if (animation) {
		delete animation;
		animation = NULL;
		target = source = NULL;
	}
}

void Scene::createNewAnimation() {
	if (animation) {
		clearAnimation();
	}
	animation = new ShapeAnimation();
}

void Scene::loadNewAnimation( const FileSequence &fs )
{
	createNewAnimation();
	animation->load(QString(rootDir.c_str()), fs, *this);
}

void Scene::addAnimationFrame( QString filename, bool processFrames/* = true*/ )
{
	animation->add(QString(rootDir.c_str()), filename, *this, processFrames);
}


ShapeAnimation* Scene::getAnimation()
{
	return animation;
}

void Scene::Update( Subject* theChangedSubject )
{
	// Notify the observers of the scene
	Notify();
}

void Scene::setXMLParamList( xmlParam &p )
{
	p.addParam("sourceColor", "floatArray", (void*)&sourceColor);
	p.addParam("targetColor", "floatArray", (void*)&targetColor);
	p.addParam("pointColor", "floatArray", (void*)&pointColor);
	p.addParam("selectedColor", "floatArray", (void*)&selectedColor);
	p.addParam("rootDir", "string", (void*)&rootDir);
	p.addParam("defaultFile", "string", (void*)&defaultFile);
	p.addParam("defaultLoadFrames", "int", (void*)&defaultLoadFrames);
	p.addParam("defaultSkipFrames", "int", (void*)&defaultSkipFrames);
	p.addParam("sourceFrame", "int", (void*)&sourceFrame);
	p.addParam("targetFrame", "int", (void*)&targetFrame);
	p.addParam("correspondenceFile", "string", (void*)&correspondenceFile);
	p.addParam("defaultColormapFile", "string", (void*)&defaultColormapFile);
	p.addParam("defaultSourceFile", "string", (void*)&defaultSourceFile);
	p.addParam("defaultTargetFile", "string", (void*)&defaultTargetFile);
	p.addParam("autoProcessMode", "int", (void*)&autoProcessMode);
	p.addParam("autoSaveCorrespondenceResult", "string", (void*)&autoSaveCorrespondenceResult);

	// Correspondence parameters
	p.addParam("scaleFactors_MC", "floatArray", (void*)&scaleFactors_MC);
	p.addParam("ANNEpsilon_MC", "float", (void*)&ANNEpsilon_MC);
	p.addParam("colormapMin", "float", (void*)&colormapMin);
	p.addParam("colormapMax", "float", (void*)&colormapMax);
	p.addParam("featureSamplesSource", "int", (void*)&featureSamplesSource);
	p.addParam("featureSamplesTarget", "int", (void*)&featureSamplesTarget);
	p.addParam("featureType", "int", (void*)&featureType);
	p.addParam("ANNEpsilon_FM", "float", (void*)&ANNEpsilon_FM);
	p.addParam("radius_FM", "float", (void*)&radius_FM);
	p.addParam("maxCandidates_FM", "int", (void*)&maxCandidates_FM);
	p.addParam("threshold_FM", "float", (void*)&threshold_FM);
	p.addParam("ANNEpsilon_MS", "float", (void*)&ANNEpsilon_MS);
	p.addParam("radius_MS", "float", (void*)&radius_MS);
	p.addParam("maxIter_MS", "float", (void*)&maxIter_MS);
	p.addParam("rate_MS", "float", (void*)&rate_MS);
	p.addParam("scaleFactors_PC", "floatArray", (void*)&scaleFactors_PC);
	p.addParam("robust_PC", "bool", (void*)&robust_PC);
	p.addParam("minVerifyClusterSize", "int", (void*)&minVerifyClusterSize);
	p.addParam("maxDist_FM", "float", (void*)&maxDist_FM);
	p.addParam("ANNEpsilon_VR", "float", (void*)&ANNEpsilon_VR);
	
	p.addParam("totalScene_boundingBoxDiagonal", "float", (void*)&totalScene_boundingBoxDiagonal);

	p.addParam("RANSAC_max_itr", "int", (void*)&RANSAC_max_itr);
	p.addParam("RANSAC_sourceModel_ANNEpsilon", "float", (void*)&RANSAC_sourceModel_ANNEpsilon);
	p.addParam("RANSAC_targetModel_ANNEpsilon", "float", (void*)&RANSAC_targetModel_ANNEpsilon);
	p.addParam("RANSAC_inlier_queryRadius", "float", (void*)&RANSAC_inlier_queryRadius);
	p.addParam("RANSAC_inlier_epsilonBand", "float", (void*)&RANSAC_inlier_epsilonBand);
	p.addParam("RANSAC_minimumInliersPercentage", "float", (void*)&RANSAC_minimumInliersPercentage);

	p.addParam("pruneMaxRatio_FM", "float", (void*)&pruneMaxRatio_FM);
	p.addParam("minMatchSize_VR", "int", (void*)&minMatchSize_VR);
	p.addParam("maxNumVerifyClusters_VR", "int", (void*)&maxNumVerifyClusters_VR);
	p.addParam("skipVerification_VR", "bool", (void*)&skipVerification_VR);
	p.addParam("maxNumLabels_GC", "int", (void*)&maxNumLabels_GC);
	p.addParam("growRegionDistThreshold_VR", "float", (void*)&growRegionDistThreshold_VR);
	p.addParam("graphCutSolutionType_GC", "int", (void*)&graphCutSolutionType_GC);
	p.addParam("useDummyLabel_GC", "bool", (void*)&useDummyLabel_GC);
	p.addParam("bdySmoothnessKNN_GC", "int", (void*)&bdySmoothnessKNN_GC);
	p.addParam("closestPtDistThreshold_GC", "float", (void*)&closestPtDistThreshold_GC);
	p.addParam("mismatchError_GC", "float", (void*)&mismatchError_GC);
	p.addParam("dummyLabelSmoothError_GC", "float", (void*)&dummyLabelSmoothError_GC);
	p.addParam("dummyLabelDataError_GC", "float", (void*)&dummyLabelDataError_GC);
	p.addParam("smoothWeight_GC", "float", (void*)&smoothWeight_GC);
	p.addParam("dataWeight_GC", "float", (void*)&dataWeight_GC);
	p.addParam("smoothTruncateValue_GC", "float", (void*)&smoothTruncateValue_GC);
	p.addParam("dataTruncateValue_GC", "float", (void*)&dataTruncateValue_GC);
	p.addParam("dataTermType_GC", "int", (void*)&dataTermType_GC);
	p.addParam("graphCutInstance_GC", "int", (void*)&graphCutInstance_GC);

	// Spin Image Parameters
	p.addParam("binSizeMultiple_SI", "float", (void*)&binSizeMultiple_SI);
	p.addParam("imageWidth_SI", "int", (void*)&imageWidth_SI);
	p.addParam("maxSupportAngle_SI", "float", (void*)&maxSupportAngle_SI);

	// Compute alignment distance
	p.addParam("alignDistanceFile", "string", (void*)&alignDistanceFile);
}

CorrespondenceList* Scene::getCorrespondences()
{
	return cl;
}

RealColor Scene::getColormapColor( real_type val ) const
{
	return colormap.getColor((val - colormapMin) / (colormapMax - colormapMin));
}

void Scene::regenerateColors()
{
	for (index_type l = 0; l < clusterColors.size(); l++) {
		// Get a random color from the colormap
		clusterColors[l] = colormap.getColor(genrand_real2());
	}
}

void Scene::generateClusterColors( size_type n )
{
	// Pick colors from the colormap for each label
	clusterColors.resize(n);

	for (index_type l = 0; l < n; l++) {
		// Get a random color from the colormap
		clusterColors[l] = colormap.getColor(genrand_real2());
	}
}

RealColor Scene::getClusterColor( index_type i )
{
	return clusterColors[i];
}

size_type Scene::getClusterSize( index_type i )
{
	return matchCandidates->clusterSizes[i];
}

void Scene::initMatchCandidates( ShapeModel &sm, ShapeModel &tm, bool forceReset/* = false*/ )
{
	// Re-init only if necessary
	if (forceReset || !matchCandidates ||
		(&sm != &matchCandidates->sourceModel) ||
		(&tm != &matchCandidates->targetModel)) {

		// Initialize structure
		if (matchCandidates) {
			delete matchCandidates;
		}
		matchCandidates = new MatchCandidates(sm, tm);
		cout << "Creating new MatchCandidates object..." << endl;
	}
}

bool Scene::clusterResultExists() const
{
	if (!matchCandidates || (!matchCandidates->doneClustering() && !matchCandidates->doneVerify())) {
		return false;
	}

	return true;
}

real_type Scene::makeSceneRadius( real_type percent ) const
{
	return totalScene_boundingBoxDiagonal * percent / 100.0;
}

real_type Scene::makeSceneRadiusSq( real_type percent ) const
{
	real_type radius = makeSceneRadius(percent);
	return (radius * radius);
}

distList Scene::makeSceneSqRadii( const FloatArray &percentList ) const
{
	// Generate scales as a percentage of the bounding box diagonal
	distList result;
	result.resize(percentList.size());

	for (int i = 0; i < percentList.size(); i++) {
		result[i] = makeSceneRadiusSq(percentList[i]);
	}

	return result;
}

void Scene::clearCorrespondenceList()
{
	if (cl) {
		delete cl;
		cl = NULL;
	}
}

real_type Scene::pointSize( real_type frac /*= 0.01 */ ) const
{
	return frac*totalScene_boundingBoxDiagonal;
}

void Scene::init()
{
	lastDir = QString(rootDir.c_str());

	if (defaultColormapFile.length() > 0) {
		cout << "Opening default colormap: " << defaultColormapFile << endl;
		colormap.load(defaultColormapFile.c_str());
	}
}

