#ifndef __SCENE_H__
#define __SCENE_H__

#include "subject.h"
#include "ShapeAnimation.h"
#include "geodesic.h"
#include <vector>
#include <QString>
#include <libXMLParam/xmlparam.h>
#include "Colormap.h"
#include "FindCorrespondence.h"


// Global foreground and background colors for OpenGL viewer windows
extern const char *globalForegroundColor;
extern const char *globalBackgroundColor;

enum AutoProcessModes {
	AUTO_PROCESS_NONE = 0,
	AUTO_PROCESS_CORRESPONDENCE,
	AUTO_READ_CORRESPONDENCE_RESULT,
	AUTO_COMPUTE_ALIGNMENT_DISTANCE
};

class Scene : public Subject, public Observer {
public:
	Scene();
	virtual ~Scene();

	// Initializing the scene
	void init();

	// Clearing the data
	void clear();
	void clearAnimation();

	// Loading and Accessing the Mesh Animation
	void createNewAnimation();
	void loadNewAnimation(const FileSequence &fs);
	void addAnimationFrame( QString filename, bool processFrames = true );
	ShapeAnimation* getAnimation();

	// Updating the display
	virtual void Update(Subject* theChangedSubject);

	// Deformation parameters
	ShapeModel *source, *target;		/**< Source and target meshes for the deformation */

	// Get correspondences
	CorrespondenceList* getCorrespondences();

	// Useful global variables
	QString lastDir;

	// Get a colormap color
	RealColor getColormapColor(real_type val) const;

	// Get the scene point size; automatically determined based on length of the bounding box diagonal
	real_type pointSize( real_type frac = 0.01 ) const;

	// Reading deformation parameters from the file
	void setXMLParamList(xmlParam &p);

	FloatArray sourceColor, targetColor, pointColor, selectedColor;		/**< Color settings */

	std::string rootDir;							/**< Root data directory */
	std::string defaultFile;						/**< File to open at startup */
	int defaultLoadFrames;							/**< How many frames (at most) to load at startup */
	int defaultSkipFrames;							/**< Skip over frames when loading */
	int sourceFrame, targetFrame;					/**< Source and target mesh frame (animation frame #) */

	std::string defaultSourceFile;					/**< Default source mesh to open at startup */
	std::string defaultTargetFile;					/**< Default target mesh to open at startup */
	int autoProcessMode;							/**< Automatic processing mode */
	std::string autoSaveCorrespondenceResult;		/**< Automatic save location for correspondence result */

	std::string correspondenceFile;					/**< File to open for correspondences */
	std::string defaultColormapFile;				/**< Colormap to use */

	// Display settings
	int currentBone;						/**< Current bone we are displaying */
	Colormap colormap;						/**< Colormap */
	float colormapMin, colormapMax;			/**< Min and max values for assigning colormap values */

	float totalScene_boundingBoxDiagonal;	/**< A measure of scene size */

	// Correspondence parameters
	bool robust_PC;							/**< Whether to generate robust principal curvatures? */
	FloatArray scaleFactors_PC;				/**< Scales for generating the principal curvature frames (percentage of bbox diagonal) */
	FloatArray scaleFactors_MC;				/**< Sizes for the curvature descriptor (percentage of bbox diagonal) */
	float ANNEpsilon_MC;					/**< Range query epsilon for mean curvature estimation */
	int featureSamplesSource;				/**< Number of features to compute in the source mesh */
	int featureSamplesTarget;				/**< Number of features to compute in the target mesh */
	int featureType;						/**< Type of features to compute in each mesh */
	float ANNEpsilon_FM;					/**< Range query epsilon for feature matching */
	float radius_FM;						/**< Range query radius for feature matching */
	int maxCandidates_FM;					/**< Maximum number of candidates for feature matching */
	float threshold_FM;						/**< Minimum norm of a feature used for matching */
	float maxDist_FM;						/**< Maximum Euclidean distance for correspondence pairing (% of bbox diagonal) */
	float pruneMaxRatio_FM;					/**< Maximum ratio of principal curvatures for point pruning */

	float ANNEpsilon_MS;					/**< Range query epsilon for range query in feature space within mean-shift clustering */
	float radius_MS;						/**< Radius of kernel for mean-shift clustering */
	float maxIter_MS;						/**< Maximum # of iterations for mean-shift clustering */
	float rate_MS;							/**< Rate of convergence for mean-shift clustering */

	float ANNEpsilon_VR;					/**< Range query epsilon for region verification */
	int minVerifyClusterSize;					/**< Minimum size of a cluster for it to be matched & verified */
	int minMatchSize_VR;						/**< Minimum size of verified region */
	int maxNumVerifyClusters_VR;				/**< Maximum number of clusters for verification */
	int maxNumLabels_GC;						/**< Maximum number of labels for the graph cut */
	real_type growRegionDistThreshold_VR;		/**< Distance threshold for growing the verified region */
	int graphCutSolutionType_GC;				/**< What solution method to use in graph cut optimization */
	bool useDummyLabel_GC;						/**< Whether to assign a dummy label in the graph cut optimization? */
	int bdySmoothnessKNN_GC;					/**< Connecting k-nearest boundary vertices for smoothness constraints */
	real_type mismatchError_GC;					/**< Penalty for inconsistent labeling */
	real_type dummyLabelSmoothError_GC;			/**< Smoothness penalty for assigning the dummy label */
	real_type dummyLabelDataError_GC;			/**< Data penalty for assigning the dummy label */
	real_type closestPtDistThreshold_GC;		/**< Closest point distance threshold for graph cut */
	real_type smoothWeight_GC;					/**< Weight of smoothness term for graph cut error function */
	real_type dataWeight_GC;					/**< Weight of data term for graph cut error function */
	real_type smoothTruncateValue_GC;			/**< Maximum value of the smooth term for graph cut error function */
	real_type dataTruncateValue_GC;				/**< Maximum value of the data term for graph cut error function */
	int dataTermType_GC;						/**< Type of data term in error function */
	int graphCutInstance_GC;					/**< Graph cut instance setup: symmetric, forward, or backward */

	int RANSAC_max_itr;							/**< RANSAC correspondence: Number of iterations */
	float RANSAC_sourceModel_ANNEpsilon;		/**< RANSAC correspondence: ANN Epsilon for closest distance query */
	float RANSAC_targetModel_ANNEpsilon;		/**< RANSAC correspondence: ANN Epsilon for closest distance query */
	float RANSAC_inlier_queryRadius;			/**< RANSAC correspondence: Radius for gathering inlier candidates */
	float RANSAC_inlier_epsilonBand;			/**< RANSAC correspondence: Epsilon-band distance threshold for inliers */
	float RANSAC_minimumInliersPercentage;		/**< RANSAC correspondence: Minimum percentage of inliers for good fit */

	// Spin image parameters
	real_type binSizeMultiple_SI;				/**< Spin Images: Bin Size parameter as a multiple of the mesh resolution */
	int imageWidth_SI;							/**< Spin Images: Image Width */
	real_type maxSupportAngle_SI;				/**< Spin Images: Maximum Support Angle (in degrees) */

	// Skip verification step?
	bool skipVerification_VR;					/**< Skip verification step? */

	// Compute and save alignment distance
	std::string alignDistanceFile;				/**< Text filename to store the resulting distance */

	// Create radius values from the total scene size
	real_type makeSceneRadius( real_type percent ) const;
	real_type makeSceneRadiusSq( real_type percent ) const;
	distList makeSceneSqRadii( const FloatArray &percentList ) const;

	// Store colors for each cluster
	void regenerateColors();
	void generateClusterColors( size_type n );
	RealColor getClusterColor( index_type i );
	size_type getClusterSize( index_type i );
	std::vector<RealColor> clusterColors;

	// Store clustering data
	void initMatchCandidates( ShapeModel &sm, ShapeModel &tm, bool forceReset = false );
	bool clusterResultExists() const;
	MatchCandidates *matchCandidates;

	// Debugging: Visualizing the cluster modes.
	int showClusterMode;

	// Store initial bone locations and influence radii.
	distList initialRadii;
	RigidDeform initialTransforms;

	// Solve and package correspondence result
	void solveCorrespondences();

private:
	ShapeAnimation *animation;				/**< A mesh animation sequence */

	// Sampling correspondences
	std::vector<double> cpi;				/**< Random numbers to pick correspondence indices */
	CorrespondenceList *cl;					/**< New correspondence class (with only positions, not any mesh vertex indices) */

	// Sampling the mesh
	PointList sourceSample;
	void clearCorrespondenceList();

	// Debugging: save the previous solution
	RigidDeform lastSolution;
};

#endif
