#ifndef __TYPES_H__
#define __TYPES_H__

#include <SylvainTools/geom.h>
#include <QTextStream>
#include <cfloat>

#undef min
#undef max

// Type definitions
typedef float real_type;
typedef Geometry::Vec2<real_type> real2_type;
typedef Geometry::Vec3<real_type> real3_type;
typedef Geometry::Vec<4,real_type> real4_type;
typedef Geometry::Square_matrix<3,real_type> matrix3_type;

typedef unsigned int index_type;
typedef Geometry::Vec2<index_type> index2_type;
typedef Geometry::Vec3<index_type> index3_type;

typedef unsigned int size_type;

typedef std::vector<index_type> idxList;
typedef std::vector<real_type> distList;
typedef std::vector<index3_type> idx3List;
typedef std::vector<real3_type> ptList;

#define REAL_TYPE_EPSILON FLT_EPSILON
#define REAL_TYPE_MAX FLT_MAX

typedef real3_type RealColor;
typedef Geometry::Vec3i IntColor;

// Data type definition.
class data_type {
public:
	inline data_type(int d) {
		val.resize(d);
		setZero();
	}

	inline void setZero() {
		for (int i = 0; i < val.size(); i++) {
			val[i] = 0;
		}
	}

	inline bool checkFinite() {
		for (int i = 0; i < val.size(); i++) {
			if (!_finite(val[i])) {
				return false;
			}
		}
		return true;
	}

	inline data_type& operator+=(const data_type& v) {
		for (unsigned int i = 0; i < val.size(); i++) {
			val[i] += v.val[i];
		}
		return *this;
	}

	inline data_type& operator*=(const real_type r) {
		for (unsigned int i = 0; i < val.size(); i++) {
			val[i] *= r;
		}
		return *this;
	}

	inline void normalizeL1() {
		real_type total = normL1();

		if (total > 0) {
			for (unsigned int i = 0; i < val.size(); i++) {
				val[i] /= total;
			}
		}
	}

	inline real_type normL1() {
		real_type total = 0;
		for (unsigned int i = 0; i < val.size(); i++) {
			total += abs(val[i]);
		}

		return total;
	}

	inline void normalizeL2() {
		real_type total = normL2();

		if (total > 0) {
			for (unsigned int i = 0; i < val.size(); i++) {
				val[i] /= total;
			}
		}
	}

	inline real_type normSqrL2() {
		real_type total = 0;
		for (unsigned int i = 0; i < val.size(); i++) {
			total += val[i]*val[i];
		}

		return total;
	}

	inline real_type normL2() {
		return sqrt(normSqrL2());
	}

	inline real_type findMin() const {
		real_type elt = REAL_TYPE_MAX;

		for (unsigned int i = 0; i < val.size(); i++) {
			elt = std::min(elt, val[i]);
		}

		return elt;
	}

	inline real_type findMax() const {
		real_type elt = -REAL_TYPE_MAX;

		for (unsigned int i = 0; i < val.size(); i++) {
			elt = std::max(elt, val[i]);
		}

		return elt;
	}

	std::vector<real_type> val;
};

/**
 * Add two data vectors.
 */
inline data_type operator+(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res.val[i] = v1.val[i] + v2.val[i];
	}
	return res;
}

/**
* Subtract two data vectors.
*/
inline data_type operator-(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res.val[i] = v1.val[i] - v2.val[i];
	}
	return res;
}

/**
 * Component-wise multiply.
 */
inline data_type operator*(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res.val[i] = v1.val[i] * v2.val[i];
	}
	return res;
}

/**
 * Component-wise divide.
 */
inline data_type operator/(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res.val[i] = v1.val[i] / v2.val[i];
	}
	return res;
}


/**
 * Dot product.
 */
inline real_type dot(const data_type& v1, const data_type& v2) {
	// No size checking here!
	real_type res = 0;
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res += v1.val[i] * v2.val[i];
	}
	return res;
}

inline data_type operator*(const real_type r, const data_type& v) {
	data_type res(v.val.size());
	for (unsigned int i = 0; i < v.val.size(); i++) {
		res.val[i] = r * v.val[i];
	}
	return res;
}

inline data_type operator*(const data_type& v, const real_type r) {
	data_type res(v.val.size());
	for (unsigned int i = 0; i < v.val.size(); i++) {
		res.val[i] = v.val[i] * r;
	}
	return res;
}

inline data_type operator/(const data_type& v, const real_type r) {
	data_type res(v.val.size());
	for (unsigned int i = 0; i < v.val.size(); i++) {
		res.val[i] = v.val[i] / r;
	}
	return res;
}


inline std::ostream& operator<<(std::ostream& out, const data_type &v) {
	out << "[";
	for (unsigned int i = 0; i < v.val.size(); i++) {
		if (i > 0) {
			out << ",";
		}
		out << v.val[i];
	}
	out << "]";
	return out;
}

inline QTextStream& operator<<(QTextStream& out, const data_type &v) {
	out << "[";
	for (int i = 0; i < v.val.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v.val[i];
	}
	out << "]";
	return out;
}



#endif
