#include "util.h"
#include <iomanip>
#include <QRegExp>
#include <QFile>
#include <QScrollBar>
#include <QLineEdit>
#include <glut.h>
#include "ShapeModel.h"

stdext::hash_map<std::string, TimerPair> perfTimers;

/**
* Write the contents of a vector to the stream.
*/
std::ostream& operator<<(std::ostream& out, const Geometry::Vec3f &v) {
	out << "[";
	for (int i = 0; i < 3; i++) {
		//out << std::setprecision(2) << setiosflags(std::ios_base::fixed)<< v[i] << ",";
		//out << std::setprecision(5) << setiosflags(std::ios_base::scientific)<< v[i] << ",";
		out << v[i] << ",";
	}
	out << "\b]";
	return out;
}

QTextStream& operator<<(QTextStream& out, const Geometry::Vec3f &v) {
	out << "[" << v[0] << "," << v[1] << "," << v[2] << "]";
	return out;
}

std::ostream& operator<<(std::ostream& out, const FloatArray &v) {
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v[i];
	}
	out << "]";
	return out;
}

QTextStream& operator<<(QTextStream& out, const FloatArray &v) {
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v[i];
	}
	out << "]";
	return out;
}

// Grabbed from pbrt 1.03, core/mc.cpp
void ComputeStep1dCDF(float *f, int nSteps, float *c, float *cdf) {
	// Compute integral of step function at $x_i$
	int i;
	cdf[0] = 0.;
	for (i = 1; i < nSteps+1; ++i)
		cdf[i] = cdf[i-1] + f[i-1] / nSteps;
	// Transform step function integral into cdf
	*c = cdf[nSteps];
	for (i = 1; i < nSteps+1; ++i)
		cdf[i] /= *c;
}

// Grabbed from pbrt 1.03, core/mc.cpp
float SampleStep1d(float *f, float *cdf, float c, int nSteps, float u, float *pdf) {
	// Find surrounding cdf segments
	float *ptr = std::lower_bound(cdf, cdf+nSteps+1, u);
	int offset = (int) (ptr-cdf-1);
	// Return offset along current cdf segment
	u = (u - cdf[offset]) / (cdf[offset+1] - cdf[offset]);
	*pdf = f[offset] / c;
	return (offset + u) / nSteps;
}


void setOpenGLColor( QRgb color )
{
	unsigned char r = (unsigned char) qRed(color);
	unsigned char g = (unsigned char) qGreen(color);
	unsigned char b = (unsigned char) qBlue(color);
	unsigned char a = (unsigned char) qAlpha(color);
	glColor4ub(r,g,b,a);
}

void setOpenGLColor( const RealColor &color, real_type alpha/* = 1.0*/ )
{
	glColor4f(color[0], color[1], color[2], alpha);
}

void setOpenGLColor( const FloatArray &color, real_type alpha/* = 1.0*/ )
{
	if (color.size() < 3) {
		std::cerr << "Error in setOpenGLMaterialColor(): not enough colors!" << std::endl;
		return;
	}

	glColor4f(color[0], color[1], color[2], alpha);
}

float interpSlider(QScrollBar *slider) {
	return ((float)slider->value()) / ((float)(slider->maximum() - slider->minimum()));
}

void setInterpSlider(QScrollBar *slider, float interp) {
	slider->setValue((slider->maximum() - slider->minimum())*interp + slider->minimum());
}

float QLineEditFloatValue(QLineEdit *edit) {
	bool ok;
	float value = edit->text().toFloat(&ok);

	if (!ok) {
		// Just set the value to zero
		value = 0.f;
		edit->setText(QString("%1").arg(value));
	}

	return value;
}

void outputToTextStream(QTextStream &out, int ind, Geometry::Matrix33f rot, Geometry::Vec3f tr) {
	// Print
	out << "R{" << ind+1 << "} = ";
	out << "[" << rot(0,0) << ", " << rot(0,1) << ", " << rot(0,2) << "\n        ";
	out <<        rot(1,0) << ", " << rot(1,1) << ", " << rot(1,2) << "\n        ";
	out <<        rot(2,0) << ", " << rot(2,1) << ", " << rot(2,2) << "];\n";

	out << "T{" << ind+1 << "} = ";
	out << "[" << tr[0] << ", " << tr[1] << ", " << tr[2] << "];\n";
	out << "\n";
}

void drawBoxOpenGL( const real3_type &minPos, const real3_type &maxPos )
{
	real3_type centerPos = (minPos + maxPos) / 2;

	glBegin(GL_LINES);
	// diagonal
	//glVertex3f(minPos[0], minPos[1], minPos[2]);
	//glVertex3f(maxPos[0], maxPos[1], maxPos[2]);

	// 3 lines from min pos
	//glColor3f(1.f,0.f,0.f);
	glVertex3f(minPos[0], minPos[1], minPos[2]);
	glVertex3f(maxPos[0], minPos[1], minPos[2]);

	//glColor3f(0.f,1.f,0.f);
	glVertex3f(minPos[0], minPos[1], minPos[2]);
	glVertex3f(minPos[0], maxPos[1], minPos[2]);

	//glColor3f(0.f,0.f,1.f);
	glVertex3f(minPos[0], minPos[1], minPos[2]);
	glVertex3f(minPos[0], minPos[1], maxPos[2]);

	// 2 lines from min pos with x set to max pos
	//glColor3f(1.f,1.f,1.f);
	glVertex3f(maxPos[0], minPos[1], minPos[2]);
	glVertex3f(maxPos[0], maxPos[1], minPos[2]);

	glVertex3f(maxPos[0], minPos[1], minPos[2]);
	glVertex3f(maxPos[0], minPos[1], maxPos[2]);

	// 2 lines from min pos with y set to max pos
	glVertex3f(minPos[0], maxPos[1], minPos[2]);
	glVertex3f(maxPos[0], maxPos[1], minPos[2]);

	glVertex3f(minPos[0], maxPos[1], minPos[2]);
	glVertex3f(minPos[0], maxPos[1], maxPos[2]);

	// 2 lines from min pos with z set to max pos
	glVertex3f(minPos[0], minPos[1], maxPos[2]);
	glVertex3f(maxPos[0], minPos[1], maxPos[2]);

	glVertex3f(minPos[0], minPos[1], maxPos[2]);
	glVertex3f(minPos[0], maxPos[1], maxPos[2]);

	// 3 lines from max pos
	glVertex3f(maxPos[0], maxPos[1], maxPos[2]);
	glVertex3f(minPos[0], maxPos[1], maxPos[2]);

	glVertex3f(maxPos[0], maxPos[1], maxPos[2]);
	glVertex3f(maxPos[0], minPos[1], maxPos[2]);

	glVertex3f(maxPos[0], maxPos[1], maxPos[2]);
	glVertex3f(maxPos[0], maxPos[1], minPos[2]);
	glEnd();	// End of: bounding box drawing

	//glBegin(GL_LINES);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0]-1, centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1]-1, centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]-1);
	//glEnd();
}

void drawSphere( const real3_type pos, const float radius, GLint slices/* = 10*/, GLint stacks/* = 10*/ )
{
	glPushMatrix();
		glTranslatef(pos[0], pos[1], pos[2]);
		glutSolidSphere(radius, slices, stacks);
	glPopMatrix();
}


Geometry::Vec3f VecToVec3f( qglviewer::Vec v )
{
	return Geometry::Vec3f(v[0], v[1], v[2]);
}

qglviewer::Vec Vec3fToVec( Geometry::Vec3f v )
{
	return qglviewer::Vec(v[0], v[1], v[2]);
}

real_type meshNorm( const ShapeModel &mesh1, const ShapeModel &mesh2 )
{
	size_type numVertices = mesh1.numVertices();
	real_type totalError = 0;

	if (numVertices != mesh2.numVertices()) {
		std::cerr << "Error in meshNorm(): shapes have different number of vertices!";
		return -1;
	}

	for (index_type i = 0; i < numVertices; i++) {
		real3_type diff = mesh1.vertex[i] - mesh2.vertex[i];
		totalError += diff.square_norm();
	}

	return totalError;
}

void setOpenGLMaterialColor( const FloatArray floatColor, real_type alpha /*= 1.f*/ )
{
	if (floatColor.size() < 3) {
		std::cerr << "Error in setOpenGLMaterialColor(): not enough colors!" << std::endl;
		return;
	}

	static GLfloat color[4];
	for (int d = 0; d < 3; d++) {
		color[d] = floatColor[d];
	}
	color[3] = alpha;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
}

void setOpenGLMaterialColor( const RealColor floatColor, real_type alpha /*= 1.f */ )
{
	static GLfloat color[4];
	for (int d = 0; d < 3; d++) {
		color[d] = floatColor[d];
	}
	color[3] = alpha;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
}

void TextProgress( double x, bool init/* = false*/ )
{
	static const int bins = 100;
	static int currentBin, binIncrement;
	static double binSize;

	// Initialize progress display
	if (init) {
		binSize = x / ((double)bins);
		currentBin = 0;
		binIncrement = std::max(1, (int)(1.f/binSize));
		std::cout << " - " << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
	}
	// Update progress display
	else {
		if (x >= ((double)currentBin)*binSize) {
			currentBin += binIncrement;
			std::cout << "\b\b\b\b";
			if (currentBin >= 1000.1) {
				std::cout << "\b";
			}
			if (currentBin >= 10000.1) {
				std::cout << "\b";
			}
			std::cout << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
		}
	}
}

void TextProgressWithLabel( double x, bool init /*= false*/, QString id /*= QString() */ )
{
	static const int bins = 100;
	static int currentBin, binIncrement;
	static double binSize;
	int idLength = id.length();

	// Initialize progress display
	if (init) {
		binSize = x / ((double)bins);
		currentBin = 0;
		binIncrement = std::max(1, (int)(1.f/binSize));
		std::cout << "id " << qPrintable(id) << " - " << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
	}
	// Update progress display
	else {
		if (x >= ((double)currentBin)*binSize) {
			currentBin += binIncrement;

			// Backspace
			for (int i = 0; i < idLength+10; i++) {
				std::cout << "\b";
			}

			// Print out next progress
			std::cout << "id " << qPrintable(id) << " - " << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
		}
	}
}


Geometry::Matrix33f rotateAxis(const Geometry::Vec3f &axis, const float cosAngle) {
	float sinAngle = sqrtf(1.f - cosAngle*cosAngle);

	// use rodrigues' formula
	Geometry::Matrix33f ssm;
	ssm(0,0) = 0.f;   ssm(0,1) = -axis[2]; ssm(0,2) = axis[1];
	ssm(1,0) = axis[2];  ssm(1,1) = 0.f;   ssm(1,2) = -axis[0];
	ssm(2,0) = -axis[1]; ssm(2,1) = axis[0];  ssm(2,2) = 0.f;

	Geometry::Matrix33f rot, id;
	id = Geometry::Matrix33f::identity();
	rot = id + ssm*sinAngle + ssm*ssm*(1.f - cosAngle);

	return rot;
}

void writeBinary( std::ofstream &out, const Geometry::Matrix33f m )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			temp = m(i,j);
			out.write((char*) &temp, sizeof(float));
		}
	}
}

void writeBinary( std::ofstream &out, const Geometry::Vec3f v )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		temp = v[i];
		out.write((char*) &temp, sizeof(float));
	}
}

void readBinary( std::ifstream &in, Geometry::Matrix33f &m )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			in.read((char*) &temp, sizeof(float));
			m(i,j) = temp;
		}
	}
}

void readBinary( std::ifstream &in, Geometry::Vec3f &v )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		in.read((char*) &temp, sizeof(float));
		v[i] = temp;
	}
}

