#ifndef __UTIL_H__
#define __UTIL_H__

#include <algorithm>
#include <SylvainTools/geom.h>
#include <iostream>
#include <QString>
#include <QtOpenGL>
#include <QTextStream>
#include <Windows.h>
#include <QColor>
#include <libXMLParam/xmlParam.h>
#include "types.h"
#include <QGLViewer/qglviewer.h>
#include <hash_map>
#include <fstream>

#undef min
#undef max

// Useful functions
std::ostream& operator<<(std::ostream& out, const Geometry::Vec3f &v);
QTextStream& operator<<(QTextStream& out, const Geometry::Vec3f &v);
std::ostream& operator<<(std::ostream& out, const FloatArray &v);
QTextStream& operator<<(QTextStream& out, const FloatArray &v);

void ComputeStep1dCDF(float *f, int nSteps, float *c, float *cdf);
float SampleStep1d(float *f, float *cdf, float c, int nSteps, float u, float *pdf);



// High-performance timer collection. adapted from
// http://blog.kowalczyk.info/kb/high-resolution-timer-for-timing-code-fragments.html
// very simple, high-precision (at least in theory) timer for timing API calls
struct TimerPair {
	TimerPair() : totalDuration(0.0) {
		QueryPerformanceFrequency(&freq);
	}

	void Start() {
		totalDuration = 0.0;
		QueryPerformanceCounter(&mTimeStart);
	}

	void AddStart() {
		QueryPerformanceCounter(&mTimeStart);
	}

	void Stop() {
		QueryPerformanceCounter(&mTimeStop);
		totalDuration += (double) (mTimeStop.QuadPart - mTimeStart.QuadPart);
	}

	double GetDurationInSecs() {
		return (totalDuration / ((double)freq.QuadPart));
	}

	LARGE_INTEGER freq;
	LARGE_INTEGER mTimeStart;
	LARGE_INTEGER mTimeStop;
	double totalDuration;
};

// A collection of performance timers
extern stdext::hash_map<std::string, TimerPair> perfTimers;

// Print timer
inline void printTime(const std::string str) {
	std::cout << "***** " << str << " took " << perfTimers[str].GetDurationInSecs() << " seconds." << std::endl;
}


void setOpenGLColor(QRgb color);
void setOpenGLColor( const RealColor &color, real_type alpha = 1.0 );
void setOpenGLColor( const FloatArray &color, real_type alpha = 1.0 );
void setOpenGLMaterialColor( const RealColor floatColor, real_type alpha = 1.f );
void setOpenGLMaterialColor( const FloatArray floatColor, real_type alpha = 1.f );

// UI controller utility functions
class QScrollBar;
class QLineEdit;
float interpSlider(QScrollBar *slider);
void setInterpSlider(QScrollBar *slider, float interp);
float QLineEditFloatValue(QLineEdit *edit);
void outputToTextStream(QTextStream &out, int ind, Geometry::Matrix33f rot, Geometry::Vec3f tr);

// Box drawing
void drawBoxOpenGL( const real3_type &minPos, const real3_type &maxPos );

// Sphere drawing
void drawSphere( const real3_type pos, const float radius, GLint slices = 10, GLint stacks = 10 );
qglviewer::Vec Vec3fToVec( Geometry::Vec3f v );
Geometry::Vec3f VecToVec3f( qglviewer::Vec v );
QPoint clickedCoord( QMouseEvent *e );

// Enumerate correspondence display modes.
enum DISPLAYMODE {
	DISPLAY_SOURCE = 1,
	DISPLAY_TARGET,
	DISPLAY_SOURCE_AND_TARGET
};

// Enumerate deformation types.
enum DeformTypes {
	DEFORM_EMBEDDED = 1,
	DEFORM_WEIGHTED,
	DEFORM_EM
};

/**
 * Compute total error between two meshes.
 * Meshes must be in per-vertex correspondence.
 */
class ShapeModel;
real_type meshNorm(const ShapeModel &mesh1, const ShapeModel &mesh2);

/**
 * Display a text progress bar.
 * @param x     The current index (when update), or the maximum index (when init).
 * @param init  As true, this initializes the progress counter.  As false, it updates the counter.
 */
void TextProgress( double x, bool init = false );
void TextProgressWithLabel( double x, bool init = false, QString id = QString() );


// Rotation matrix about an axis.
Geometry::Matrix33f rotateAxis(const Geometry::Vec3f &axis, const float cosAngle);

void writeBinary( std::ofstream &out, const Geometry::Matrix33f m );
void writeBinary( std::ofstream &out, const Geometry::Vec3f v );
void readBinary( std::ifstream &in, Geometry::Matrix33f &m );
void readBinary( std::ifstream &in, Geometry::Vec3f &v );

#endif
