#include "viewer.h"
#include <fstream>
#include "scene.h"
#include "CorrespViewer.h"
#include "FindCorrespondence.h"
#include "FeatureViewer.h"
#include "CorrespEdit.h"
#include "geom_util.h"
#include "FileSequence.h"
#include "ShapeSlicer.h"
#include "TransformViewer.h"

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;

Viewer::Viewer(Scene *s) :
scene(s), displayFrame(0), sceneChanged(false),
deformEdit(NULL), maxGraphNodes(1000), maxGraphNeighbors(50), maxOverlapFactor(0.1f),
correspViewer(NULL), tempList(NULL), interactionMode_(NONE), sourceFeatures(NULL),
targetFeatures(NULL), displayScale(0),
autoLaunchDone(false), maxFeatureSpaceRange(0.03), correspEdit(NULL), featureViewer(NULL),
curvDisplayMode(1), shapeSlicer(NULL), transformViewer(NULL), sourceQueryIndex(-1)
{
	// Attach to subject
	scene->Attach(this);
}

Viewer::~Viewer()
{
	// Detach from subject
	scene->Detach(this);

	// Close all sub-windows
	if (correspEdit) {
		correspEdit->close();
	}
	if (featureViewer) {
		featureViewer->close();
	}
	if (correspViewer) {
		correspViewer->close();
	}
}



void Viewer::drawFrame( const ShapeModel* frame )
{
	// Show samples, if they exist
	if (showSamplesAct->isChecked()) {
		if (scene->matchCandidates) {
			if (scene->matchCandidates->sourceSample && (&scene->matchCandidates->sourceModel == frame)) {
				frame = scene->matchCandidates->sourceSample;
			}
			if (scene->matchCandidates->targetSample && (&scene->matchCandidates->targetModel == frame)) {
				frame = scene->matchCandidates->targetSample;
			}
		}
	}

	// Display the shape
	if (showFeaturesAct->isChecked()) {
		frame->drawFeatureValues(*scene, displayScale);
	}
	else if (showCurvaturesAct->isChecked()) {
		frame->drawCurvatureValues(*scene, displayScale, curvDisplayMode);
	}
	else if (showFramesAct->isChecked()) {
		frame->drawFrames(*scene, displayScale, curvDisplayMode);
	}
	else if (showStatusAct->isChecked()) {
		frame->drawWithStatus();
	}
	else if (showGeometryAct->isChecked()) {
		frame->draw();
	}

	// Draw vertices
	if (showVerticesAct->isChecked()) {
		setOpenGLMaterialColor(scene->pointColor);
		for (index_type i = 0; i < frame->numVertices(); i++) {
			drawSphere(frame->vertex[i], scene->pointSize() / 2.0);
		}
	}

	// Draw frames
	if (showVertexFramesAct->isChecked()) {
		int compMax = frame->frameScales.size();
		int comp = min(compMax-1, displayScale);

		for (index_type i = 0; i < frame->numVertices(); i++) {
			const CurvInfo &coordFrame = frame->frames[i][comp];
			qglviewer::Vec from = Vec3fToVec(frame->vertex[i]);
			qglviewer::Vec vmin = Vec3fToVec(coordFrame.vmin);
			qglviewer::Vec vmax = Vec3fToVec(coordFrame.vmax);
			qglviewer::Vec vn = Vec3fToVec(coordFrame.n);

			setOpenGLMaterialColor(RealColor(0.6, 0.6, 1.0));
			drawArrow(from, from + scene->pointSize()*vmin, scene->pointSize() / 25.0);

			setOpenGLMaterialColor(RealColor(1.0, 0.6, 0.6));
			drawArrow(from, from + scene->pointSize()*vmax, scene->pointSize() / 25.0);

			setOpenGLMaterialColor(scene->pointColor);
			drawArrow(from, from + scene->pointSize()*vn, scene->pointSize() / 25.0);
		}
	}



	// Display text status
	if (showCurvaturesAct->isChecked() || showFramesAct->isChecked()) {
		setOpenGLColor(foregroundColor().rgba());

		switch(curvDisplayMode) {
			case CURVINFO_NORMALS:
				drawText(10, height() - 24, tr("Showing frame normals"));
				break;
			case CURVINFO_MIN_CURVATURE:
				drawText(10, height() - 24, tr("Showing frame minimum principal curvatures"));
				break;
			case CURVINFO_MAX_CURVATURE:
				drawText(10, height() - 24, tr("Showing frame maximum principal curvatures"));
				break;
			default:
				break;
		}
	}
}


/**
 * Draw the scene.
 */
void Viewer::draw() {
	if (!autoLaunchDone) {
		autoLaunch();
	}

	//////////////////////////////////////////////////////////////////////////
	// Store current state
	//////////////////////////////////////////////////////////////////////////
	GLboolean colorMaterial, lighting;
	GLint polygonMode;
	glGetBooleanv(GL_LIGHTING, &lighting);
	glGetBooleanv(GL_COLOR_MATERIAL, &colorMaterial);
	glGetIntegerv(GL_POLYGON_MODE, &polygonMode);

	//////////////////////////////////////////////////////////////////////////
	// Set state
	//////////////////////////////////////////////////////////////////////////
	// Toggle wireframe display
	if (showWireframe()) {
		glDisable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else {
		//glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		//glCullFace(GL_BACK);
	}

	// Set up the material
	glEnable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);
	
	//////////////////////////////////////////////////////////////////////////
	// Draw Geometry
	GLfloat color[4];
	// Set alpha
	color[3] = 1.0f;

	if (scene->getAnimation()) {
		if (renderTransparent()) {
			glEnable(GL_BLEND);
			glDepthMask (GL_FALSE);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			color[3] = 0.7f;
		}

		if (showAllFramesAct->isChecked()) {
			color[0] = 0.7f;  color[1] = 0.7f;  color[2] = 1.0f;
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

			for (int i = 0; i < scene->getAnimation()->numFrames(); i += 10) {
				drawFrame((*scene->getAnimation())[i]);
			}
		}
		else {
			if (showTargetModel()) {
				color[0] = 1.0f;  color[1] = 0.7f;  color[2] = 0.7f;
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
				drawFrame(scene->target);
			}
			if (showSourceModel()) {
				color[0] = 0.7f;  color[1] = 0.7f;  color[2] = 1.0f;
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

				drawFrame(scene->source);
			}
		}

		// Draw the bounding box, if necessary
		if (showBbox()) {
			scene->getAnimation()->drawBox(*scene->target);
		}

		if (renderTransparent()) {
			glDepthMask (GL_TRUE);
			glDisable(GL_BLEND);
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// Feature proximity testing
	drawSphereTest();

	// Draw the selected model point
	if (scene->matchCandidates && sourceQueryIndex >= 0) {
		glDisable(GL_LIGHTING);
		setOpenGLColor(foregroundColor().rgba());
		glPointSize(5.f);
		glBegin(GL_POINTS);
		real3_type v = scene->matchCandidates->sourceModel.vertex[sourceQueryIndex];
		glVertex3f(v[0], v[1], v[2]);
		glEnd();
		glEnable(GL_LIGHTING);
	}

	
	// Draw correspondence candidates
	setOpenGLMaterialColor(scene->pointColor);
	sourceQuery.draw(POINT_SPHERES, scene->pointSize());
	setOpenGLMaterialColor(scene->selectedColor);
	targetMatch.draw(POINT_SPHERES, scene->pointSize());

	//////////////////////////////////////////////////////////////////////////
	// Visualize pose clustering result
	if (visualizeClusterAct->isChecked()) {
		visualizeClusterMode();
	}

	//////////////////////////////////////////////////////////////////////////
	// Draw axis at scene center
	if (axisIsDrawn()) {
		qglviewer::Vec center = sceneCenter();
		glPushMatrix();
		glTranslatef(center[0], center[1], center[2]);
		QGLViewer::drawAxis(sceneRadius()); // Or any scale
		glPopMatrix();
	}

	//////////////////////////////////////////////////////////////////////////
	// Restore OpenGL state
	//////////////////////////////////////////////////////////////////////////
	glPolygonMode(GL_FRONT_AND_BACK, polygonMode);

	//if (colorMaterial)glEnable(GL_COLOR_MATERIAL);
	//else glDisable(GL_COLOR_MATERIAL);

	//if (lighting) glEnable(GL_LIGHTING);
	//else glDisable(GL_LIGHTING);
	glDisable(GL_LIGHTING);
}

void Viewer::pointListTest() {
	GLfloat color[4];

	//////////////////////////////////////////////////////////////////////////
	// Unit test for testing capabilities of PointList
	if (!tempList) {
		// Create a random point list
		tempList = new PointList();
		tempList->resize(1000);
		tempList->generateRandom();
	}
	color[0] = 1.f;  color[1] = 1.f;  color[2] = 0.0f;  color[3] = 1.0f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	tempList->draw(POINT_SPHERES, 0.005);

	idxList list;
	distList dist;
	float rad = 1;
	real3_type q(0, 0, 0);
	//tempList->queryRange(q, rad*rad, list, dist, true);
	tempList->queryKNN(q, 100, list, dist, true);
	color[0] = 0.f;  color[1] = 0.f;  color[2] = 1.f;  color[3] = 1.f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	tempList->draw(POINT_SPHERES, list, 0.006);

	glEnable(GL_BLEND);
	glDepthMask (GL_FALSE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	color[0] = 0.f;  color[1] = 1.f;  color[2] = 1.f;  color[3] = 0.5f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	drawSphere(q, rad, 15, 15);
	glDepthMask (GL_TRUE);
	glDisable(GL_BLEND);
}


void Viewer::colormapTest() {
	// Colormap rendering test using spiral
	const float nbSteps = 200.0;
	glBegin(GL_QUAD_STRIP);
	for (int i=0; i<nbSteps; ++i)
	{
		const RealColor color = scene->colormap.getColor(i/(nbSteps-1));
		const float ratio = i/nbSteps;
		const float angle = 21.0f*ratio;
		const float c = cos(angle);
		const float s = sin(angle);
		const float r1 = 1.0f - 0.8f*ratio;
		const float r2 = 0.8f - 0.8f*ratio;
		const float alt = ratio - 0.5f;
		const float nor = 0.5f;
		const float up = sqrt(1.0f-nor*nor);

		setOpenGLColor(color);
		setOpenGLMaterialColor(color);
		//cout << "Color: " << i/(nbSteps-1) << ", " << color << endl;
		glNormal3f(nor*c, up, nor*s);
		glVertex3f(r1*c, alt, r1*s);
		setOpenGLMaterialColor(color);
		glNormal3f(nor*c, up, nor*s);
		glVertex3f(r2*c, alt+0.05f, r2*s);
	}
	glEnd();
}

void Viewer::drawSphereTest()
{
	if (scene->source) {
		glDisable(GL_LIGHTING);
		glColor3f(1,1,1);

		// Plot the selected points
		glPointSize(5.f);
		glBegin(GL_POINTS);
		for (index_type i = 0; i < inRange.size(); i++) {
			real3_type p = scene->source->vertex[inRange[i]];
			glVertex3f(p[0],p[1],p[2]);
		}
		glEnd();

		// Plot the selected normals
		glBegin(GL_LINES);
		for (index_type i = 0; i < inRange.size(); i++) {
			real3_type p1 = scene->source->vertex[inRange[i]];
			real3_type p2 = p1 + 5*scene->pointSize()*scene->source->normal[inRange[i]];
			glVertex3f(p1[0],p1[1],p1[2]);
			glVertex3f(p2[0],p2[1],p2[2]);
		}
		glEnd();

		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glCullFace(GL_BACK);
		glEnable(GL_BLEND);
		glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		// Draw fitted sphere
		setOpenGLMaterialColor(RealColor(0.8,0.8,1.0), 0.5);
		drawSphere(fitCenter, fitRadius, 30, 30);

		glDepthMask (GL_TRUE);
		glDisable(GL_BLEND);
		glDisable(GL_CULL_FACE);
	}

	// Test Sphere Fitting
	if (interactionMode_ == DEFINE_SPHERE) {
		glDisable(GL_LIGHTING);
		glColor3f(1,1,1);
		glBegin(GL_LINES);
		glVertex3f(clickStartPos[0],clickStartPos[1],clickStartPos[2]);
		glVertex3f(clickEndPos[0],clickEndPos[1],clickEndPos[2]);
		glEnd();

		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glCullFace(GL_BACK);
		glEnable(GL_BLEND);
		glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		setOpenGLMaterialColor(RealColor(0.8,0.8,1.0), 0.5);
		drawSphere(clickStartPos, (clickEndPos-clickStartPos).norm(), 30, 30);
		glDepthMask (GL_TRUE);
		glDisable(GL_BLEND);
		glDisable(GL_CULL_FACE);
	}	
}


void Viewer::animate()
{
	if (sceneChanged) {
		//cerr << "Scene contents updated!" << endl;
		updateGL();

		sceneMutex.lock();
		sceneChanged = false;
		sceneMutex.unlock();
	}
}


/**
 * Initialize the viewer.
 */
void Viewer::init()
{
	// Create context menu.
	createContextMenu();

	// Restore previous viewer state.
	restoreStateFromFile();

	/////////////////////////////////////////////////////
	//       Keyboard shortcut customization           //
	//      Changes standard action key bindings       //
	/////////////////////////////////////////////////////

	// Define 'Control+Q' as the new exit shortcut (default was 'Escape')
	setShortcut(EXIT_VIEWER, Qt::CTRL+Qt::Key_Q);

	// Set 'Control+F' as the FPS toggle state key.
	setShortcut(DISPLAY_FPS, Qt::CTRL+Qt::Key_F);

	// Set 'Control+Y' as the camera fly mode toggle key.
	setShortcut(CAMERA_MODE, Qt::CTRL+Qt::Key_Y);

	// Disable some shortcuts
	setShortcut(MOVE_CAMERA_LEFT, Qt::NoButton);
	setShortcut(MOVE_CAMERA_RIGHT, Qt::NoButton);
	setShortcut(MOVE_CAMERA_UP, Qt::NoButton);
	setShortcut(MOVE_CAMERA_DOWN, Qt::NoButton);

	// Disable draw grid toggle shortcut (default was 'G')
	//setShortcut(DRAW_GRID, 0);

	// Add custom key description (see keyPressEvent).
	setKeyDescription(Qt::Key_Space, "Opens Viewer Menu");
	setKeyDescription(Qt::CTRL+Qt::Key_Left, "Shows previous frame in the animation");
	setKeyDescription(Qt::CTRL+Qt::Key_Right, "Shows next frame in the animation");
	setKeyDescription(Qt::KeypadModifier+Qt::Key_Enter, "Generate deformation graph");
	setKeyDescription(Qt::CTRL+Qt::Key_Return, "Generates and solves for the deformation");
	setKeyDescription(Qt::SHIFT+Qt::Key_Return, "Step one iteration in the deformation solver");
	setKeyDescription(Qt::Key_F, "Flip between displaying source and target model");

	//////////////////////////////////////////////////////////////////////////
	// Initialize rendering
	//////////////////////////////////////////////////////////////////////////
	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Set snapshot parameters
	setSnapshotQuality(0);
	setSnapshotFormat(tr("PNG"));
	//QDir::setCurrent(tr(scene->rootDir.c_str()));

	//// Opens help window
	//help();

	// Start the animation
	startAnimation();

	// Initialize scene variables
	scene->init();
}

void Viewer::autoLaunch()
{
	// Check the current directory
	cout << "Current Directory is: " << qPrintable(QDir::currentPath()) << endl;

	// If there is a specific source and target file, load it
	if (scene->defaultSourceFile.length() && scene->defaultTargetFile.length()) {
		scene->createNewAnimation();
		scene->addAnimationFrame(tr(scene->defaultSourceFile.c_str()));
		scene->addAnimationFrame(tr(scene->defaultTargetFile.c_str()));

		scene->sourceFrame = 0;
		scene->targetFrame = 1;
		scene->source = (*scene->getAnimation())[scene->sourceFrame];
		scene->target = (*scene->getAnimation())[scene->targetFrame];

		// Select target mesh for vieweing
		displayFrame = 1;
		fitViewToShape(scene->target);
		setTargetModelShowAct->setChecked(true);
	}
	else if (scene->defaultFile.length()) {
		// Open the default file, if specified
		cout << "Opening file: " << scene->defaultFile << endl;
		open(scene->defaultFile.c_str(), scene->defaultLoadFrames, scene->defaultSkipFrames);

		// Bounds checking on the frame number
		int maxFrame = scene->getAnimation()->numFrames() - 1;
		if (scene->sourceFrame > maxFrame) {
			scene->sourceFrame = maxFrame;
			cout << "Source Frame # too large; clamping to " << scene->sourceFrame << "." << endl;
		}
		if (scene->targetFrame > maxFrame) {
			scene->targetFrame = maxFrame;
			cout << "Target Frame # too large; clamping to " << scene->targetFrame << "." << endl;
		}

		scene->source = (*scene->getAnimation())[scene->sourceFrame];
		scene->target = (*scene->getAnimation())[scene->targetFrame];

		//testCurvatures();
		//testFeatureSelect();
		//testFeatureMatch();

		//CreateCorrespViewer();
	}

	autoProcess();

	//////////////////////////////////////////////////////////////////////////
	// Just sample the source and target meshes
	scene->initMatchCandidates(*scene->source, *scene->target);
	scene->matchCandidates->subSampleGeometry(scene->featureSamplesSource, scene->featureSamplesTarget, scene->pruneMaxRatio_FM);

	autoLaunchDone = true;
}

void Viewer::autoProcess()
{
	if (scene->autoProcessMode == AUTO_PROCESS_NONE) {
		return;
	}
	else if (scene->autoProcessMode == AUTO_PROCESS_CORRESPONDENCE) {
		// Compute the correspondence
		scene->solveCorrespondences();

		// Save the correspondence result
		scene->matchCandidates->saveGraphCutResult(scene->autoSaveCorrespondenceResult.c_str());

		// Exit the program
		exit(0);
	}
	else if (scene->autoProcessMode == AUTO_READ_CORRESPONDENCE_RESULT) {
		// Load the correspondence result
		scene->initMatchCandidates(*scene->source, *scene->target);
		scene->matchCandidates->loadGraphCutResult(scene->autoSaveCorrespondenceResult.c_str());

		// Display correspondence result
		CreateCorrespViewer();
		correspViewer->DisplayCorrespondenceResult(true);
		correspViewer->updateDisplay();

		// Display transformed result
		CreateTransformViewer();
	}
	else if (scene->autoProcessMode == AUTO_COMPUTE_ALIGNMENT_DISTANCE) {
		// Load the correspondence result
		scene->initMatchCandidates(*scene->source, *scene->target);
		scene->matchCandidates->loadGraphCutResult(scene->autoSaveCorrespondenceResult.c_str());

		// Add the number to a text file
		ofstream out;
		out.open(scene->alignDistanceFile.c_str(), ios_base::out | ios_base::app);

		if (out.fail()) {
			cerr << "Error in saving alignment distance: could not open output file " << scene->alignDistanceFile.c_str() << " for writing!" << endl;
		}
		else {
			real_type error = ComputeAlignDistance();

			out << scene->autoSaveCorrespondenceResult.c_str() << ": Maximum Symmetric Hausdorff Error " << error << " ("
				<< error / scene->totalScene_boundingBoxDiagonal * 100.0 << "% of bbox diagonal)\n";

			out.close();
		}

		// Exit the program
		exit(0);
	}
}

QString Viewer::helpString() const
{
	QString text("<h2>S i m p l e V i e w e r</h2>");
	text += "Use the mouse to move the camera around the object. ";
	text += "You can respectively revolve around, zoom and translate with the three mouse buttons. ";
	text += "Left and middle buttons pressed together rotate around the camera view direction axis<br><br>";
	text += "Pressing <b>Alt</b> and one of the function keys (<b>F1</b>..<b>F12</b>) defines a camera keyFrame. ";
	text += "Simply press the function key again to restore it. Several keyFrames define a ";
	text += "camera path. Paths are saved when you quit the application and restored at next start.<br><br>";
	text += "Press <b>F</b> to display the frame rate, <b>A</b> for the world axis, ";
	text += "<b>Alt+Return</b> for full screen mode and <b>Control+S</b> to save a snapshot. ";
	text += "See the <b>Keyboard</b> tab in this window for a complete shortcut list.<br><br>";
	text += "Double clicks automates single click actions: A left button double click aligns the closer axis with the camera (if close enough). ";
	text += "A middle button double click fits the zoom of the camera and the right button re-centers the scene.<br><br>";
	text += "A left button double click while holding right button pressed defines the camera <i>Revolve Around Point</i>. ";
	text += "See the <b>Mouse</b> tab and the documentation web pages for details.<br><br>";
	text += "Press <b>Escape</b> to exit the viewer.";
	return text;
}

void Viewer::add()
{
	// Add a new frame to the animation.
	QStringList files = QFileDialog::getOpenFileNames(this, tr("Open Mesh Animation"), scene->lastDir, tr("Meshes (*.obj)"));

	if (files.size() > 0) {
		// Input from the user: process the local frame information?
		QMessageBox::StandardButton reply = QMessageBox::question(this, tr("Process local frames?"),
												tr("Load / compute local curvature frames when adding?"),
												QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
		bool processFrames = true;

		if (reply == QMessageBox::Cancel) {
			return;
		}

		if (reply == QMessageBox::No) {
			processFrames = false;
		}

		QStringList list = files;
		QStringList::Iterator it = list.begin();
		while(it != list.end()) {
			// Add the frame to the animation
			scene->addAnimationFrame(*it, processFrames);
			++it;
		}
	}
}


void Viewer::open(const char *file/* = NULL*/, int maxFrames/* = 0*/, int skipFrames/* = 1*/) {
	QString fileName;
	if (file) {
		fileName = QString(file);
	}
	else {
		fileName = QFileDialog::getOpenFileName(this, tr("Open Mesh Animation"), scene->lastDir, tr("Meshes (*.obj)"));
	}
	if (fileName.isEmpty()) {
		return;
	}

	FileSequence fs(fileName);

	if (!fs.valid()) {
		return;
	}

	scene->lastDir = grabPath(fileName);
	int numGoodFrames = 0;
	fs.skipFrame = skipFrames;

	if (fs.hasNumber()) {
		QString newFile = fs[fs.startFrame];
		int frame = fs.startFrame;

		while(QFile::exists(newFile)) {
			//cout << "Good frame " << numGoodFrames << " in file! " << qPrintable(newFile) << endl;
			numGoodFrames++;
			frame += fs.skipFrame;
			newFile = fs[frame];
		}

		fs.endFrame = frame - fs.skipFrame;
	}

	if ((maxFrames > 0) && (numGoodFrames > maxFrames)) {
		// Clamp the number of frames we are loading, if necessary
		fs.endFrame = fs.startFrame + (maxFrames - 1)*fs.skipFrame;
	}

	//////////////////////////////////////////////////////////////////////////
	// Load all the frames

	// Clear all the current data
	scene->clear();

	// Load the new data
	scene->loadNewAnimation(fs);
	displayFrame = 0;

	// Center the view on the new data
	ShapeModel &temp = *(*scene->getAnimation())[displayFrame];
	fitViewToShape(&temp);

	// Set the target mesh
	scene->source = scene->target = &temp;
	setTargetModelShowAct->setChecked(true);
}

bool Viewer::showWireframe() {
	return setWireframeModeAct->isChecked();
}


bool Viewer::showBbox() {
	return setBboxModeAct->isChecked();
}

bool Viewer::showSourceModel()
{
	return setSourceModelShowAct->isChecked();
}

bool Viewer::showTargetModel()
{
	return setTargetModelShowAct->isChecked();
}

bool Viewer::renderTransparent() {
	return setTransparencyAct->isChecked();
}

void Viewer::createContextMenu() {
	// Create menu bar.
	contextMenu = new QMenu(tr("&File"), this);

	openAct = new QAction(tr("&Open Sequence..."), this);
	contextMenu->addAction(openAct);
	connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

	addAct = new QAction(tr("&Add Additional Frame..."), this);
	contextMenu->addAction(addAct);
	connect(addAct, SIGNAL(triggered()), this, SLOT(add()));

	simplifyCurrentShapeAct = new QAction(tr("&Simplify Current Frame..."), this);
	contextMenu->addAction(simplifyCurrentShapeAct);
	connect(simplifyCurrentShapeAct, SIGNAL(triggered()), this, SLOT(simplifyCurrentShape()));

	batchSimplifyAct = new QAction(tr("&Batch Simplify..."), this);
	contextMenu->addAction(batchSimplifyAct);
	connect(batchSimplifyAct, SIGNAL(triggered()), this, SLOT(batchSimplify()));

	shapeSliceAct = new QAction(tr("&Slice Shape..."), this);
	contextMenu->addAction(shapeSliceAct);
	connect(shapeSliceAct, SIGNAL(triggered()), this, SLOT(CreateShapeSlicer()));

	assignTransformsAct = new QAction(tr("&Assign Transforms..."), this);
	contextMenu->addAction(assignTransformsAct);
	connect(assignTransformsAct, SIGNAL(triggered()), this, SLOT(CreateTransformViewer()));

	contextMenu->addSeparator();

	setWireframeModeAct = new QAction(tr("&Wireframe Mode"), this);
	setWireframeModeAct->setCheckable(true);
	contextMenu->addAction(setWireframeModeAct);

	setTransparencyAct = new QAction(tr("&Transparency Mode"), this);
	setTransparencyAct->setCheckable(true);
	contextMenu->addAction(setTransparencyAct);

	fitViewToShapeAct = new QAction(tr("&Fit View to Current Frame..."), this);
	contextMenu->addAction(fitViewToShapeAct);
	connect(fitViewToShapeAct, SIGNAL(triggered()), this, SLOT(fitViewToShape()));

	contextMenu->addSeparator();

	setSourceModelShowAct = new QAction(tr("&Show Source Model"), this);
	setSourceModelShowAct->setCheckable(true);
	contextMenu->addAction(setSourceModelShowAct);

	setTargetModelShowAct = new QAction(tr("&Show Target Model"), this);
	setTargetModelShowAct->setCheckable(true);
	contextMenu->addAction(setTargetModelShowAct);

	showAllFramesAct = new QAction(tr("&Show All Frames"), this);
	showAllFramesAct->setCheckable(true);
	contextMenu->addAction(showAllFramesAct);

	contextMenu->addSeparator();

	showSamplesAct = new QAction(tr("&Show Samples"), this);
	showSamplesAct->setCheckable(true);
	contextMenu->addAction(showSamplesAct);

	visualizeClusterAct = new QAction(tr("&Visualize Cluster Result"), this);
	visualizeClusterAct->setCheckable(true);
	contextMenu->addAction(visualizeClusterAct);

	showVerticesAct = new QAction(tr("&Show Vertices"), this);
	showVerticesAct->setCheckable(true);
	contextMenu->addAction(showVerticesAct);

	showVertexFramesAct = new QAction(tr("&Show Vertex Frames"), this);
	showVertexFramesAct->setCheckable(true);
	contextMenu->addAction(showVertexFramesAct);

	contextMenu->addSeparator();
	modelDisplayGroup = new QActionGroup(this);

	showGeometryAct = new QAction(tr("&Show Geometry"), this);
	showGeometryAct->setCheckable(true);
	contextMenu->addAction(showGeometryAct);
	modelDisplayGroup->addAction(showGeometryAct);
	showGeometryAct->setChecked(true);

	showStatusAct = new QAction(tr("&Show Vertex Status"), this);
	showStatusAct->setCheckable(true);
	contextMenu->addAction(showStatusAct);
	modelDisplayGroup->addAction(showStatusAct);

	showFeaturesAct = new QAction(tr("&Show Features"), this);
	showFeaturesAct->setCheckable(true);
	contextMenu->addAction(showFeaturesAct);
	modelDisplayGroup->addAction(showFeaturesAct);

	showCurvaturesAct = new QAction(tr("&Show Curvatures"), this);
	showCurvaturesAct->setCheckable(true);
	contextMenu->addAction(showCurvaturesAct);
	modelDisplayGroup->addAction(showCurvaturesAct);

	showFramesAct = new QAction(tr("&Show Frames"), this);
	showFramesAct->setCheckable(true);
	contextMenu->addAction(showFramesAct);
	modelDisplayGroup->addAction(showFramesAct);

	setBboxModeAct = new QAction(tr("&Show Bounding Box"), this);
	setBboxModeAct->setCheckable(true);
	contextMenu->addAction(setBboxModeAct);

	contextMenu->addSeparator();

	showCorrespViewerAct = new QAction(tr("&Show Correspondences..."), this);
	contextMenu->addAction(showCorrespViewerAct);
	connect(showCorrespViewerAct, SIGNAL(triggered()), this, SLOT(CreateCorrespViewer()));

	showCorrespEditorAct = new QAction(tr("&Change Correspondence Parameters..."), this);
	contextMenu->addAction(showCorrespEditorAct);
	connect(showCorrespEditorAct, SIGNAL(triggered()), this, SLOT(CreateCorrespEditor()));

	computeAlignDistanceAct = new QAction(tr("&Compute Alignment Distance..."), this);
	contextMenu->addAction(computeAlignDistanceAct);
	connect(computeAlignDistanceAct, SIGNAL(triggered()), this, SLOT(ComputeAlignDistance()));


	contextMenu->addSeparator();

	saveFramesAct = new QAction(tr("&Save Frames..."), this);
	contextMenu->addAction(saveFramesAct);
	connect(saveFramesAct, SIGNAL(triggered()), this, SLOT(saveFrames()));


	contextMenu->addSeparator();

	captureVideoAct = new QAction(tr("&Capture Animation Sequence"), this);
	contextMenu->addAction(captureVideoAct);
	connect(captureVideoAct, SIGNAL(triggered()), this, SLOT(captureVideo()));

	changeBackgroundColorAct = new QAction(tr("&Change Background Color..."), this);
	contextMenu->addAction(changeBackgroundColorAct);
	connect(changeBackgroundColorAct, SIGNAL(triggered()), this, SLOT(changeBackgroundColor()));

	changeForegroundColorAct = new QAction(tr("&Change Foreground Color..."), this);
	contextMenu->addAction(changeForegroundColorAct);
	connect(changeForegroundColorAct, SIGNAL(triggered()), this, SLOT(changeForegroundColor()));


	contextMenu->addSeparator();

	saveColormapAct = new QAction(tr("&Save Colormap..."), this);
	contextMenu->addAction(saveColormapAct);
	connect(saveColormapAct, SIGNAL(triggered()), this, SLOT(saveColormap()));

	loadColormapAct = new QAction(tr("&Load Colormap..."), this);
	contextMenu->addAction(loadColormapAct);
	connect(loadColormapAct, SIGNAL(triggered()), this, SLOT(loadColormap()));

	setColormapLimitsAct = new QAction(tr("&Set Colormap Range..."), this);
	contextMenu->addAction(setColormapLimitsAct);
	connect(setColormapLimitsAct, SIGNAL(triggered()), this, SLOT(setColormapLimits()));

	setFeatureThresholdAct = new QAction(tr("&Set Feature L2 norm Threshold..."), this);
	contextMenu->addAction(setFeatureThresholdAct);
	connect(setFeatureThresholdAct, SIGNAL(triggered()), this, SLOT(setFeatureThreshold()));
}

/**
 * Rotate the camera frame about an axis with angle degrees, about the revolve-around point.
 */
void Viewer::rotateAbout(qglviewer::Vec up, float angle) {
	// rotate camera about the current up vector
	qglviewer::Vec o, p, v, cup;
	qglviewer::Quaternion delta;

	up.normalize();
	o = camera()->revolveAroundPoint();
	p = camera()->position();
	v = p + camera()->viewDirection();
	cup = p + camera()->upVector();

	delta.setAxisAngle(up, angle/180.0*M_PI);

	p = delta.rotate(p-o) + o;
	v = delta.rotate(v-o) + o - p;
	cup = delta.rotate(cup-o) + o - p;
	
	camera()->setPosition(p);
	camera()->setViewDirection(v);
	camera()->setUpVector(cup);
}


void Viewer::keyPressEvent(QKeyEvent *e)
{
	// Get event modifiers key
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	bool handled = false;
	if ((e->key()==Qt::Key_Space) && (modifiers==Qt::NoButton)) {
		// Compute the center location of this window
		QRect win = this->geometry();
		QPoint center;
		center.setX(win.x() + win.width() / 7);
		center.setY(win.y() + win.height() / 7);
		//cout << "(x,y,w,h) = (" << win.x() << "," << win.y() << "," << win.width() << "," << win.height() << ")" << endl;

		contextMenu->exec(center);
		handled = true;
		updateGL();
	}
	else if ((e->key()==Qt::Key_Right) && (modifiers==Qt::CTRL)) {
		// increment the frame number
		if (++displayFrame >= scene->getAnimation()->numFrames()) {
			displayFrame = 0;
		}
		displayMessage(updateFrame(displayFrame));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Left) && (modifiers==Qt::CTRL)) {
		// decrement the frame number
		if (--displayFrame < 0) {
			displayFrame = scene->getAnimation()->numFrames() - 1;
		}
		displayMessage(updateFrame(displayFrame));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Return) && (modifiers==Qt::NoButton)) {
		// Solve for the correspondences and display the result
		scene->solveCorrespondences();
		CreateCorrespViewer();
		correspViewer->DisplayCorrespondenceResult(true);

		//scene->setupDeform(whichDeform());
		//// Display the deformed mesh
		//setDeformedModelShowAct->setChecked(true);
		//displayMessage(tr("Generated deformation graph with %1 nodes!").arg(scene->graph_n));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Left) && (modifiers==Qt::NoButton)) {
		rotateAbout(qglviewer::Vec(0, 1, 0), 10.0);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_Right) && (modifiers==Qt::NoButton)) {
		rotateAbout(qglviewer::Vec(0, 1, 0), -10.0);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_Up) && (modifiers==Qt::NoButton)) {
		rotateAbout(qglviewer::Vec(1, 0, 0), 10.0);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_Down) && (modifiers==Qt::NoButton)) {
		rotateAbout(qglviewer::Vec(1, 0, 0), -10.0);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_F) && (modifiers==Qt::NoButton)) {
		QString message;

		// Flip between displaying the source and the target meshes
		if ((showSourceModel() && showTargetModel()) || (!showSourceModel() && !showTargetModel()) || (!showSourceModel() && showTargetModel())) {
			setSourceModelShowAct->setChecked(true);
			setTargetModelShowAct->setChecked(false);

			// Find the frame number
			displayFrame = scene->getAnimation()->frameNumber(scene->source);
			message = tr("Showing source model (frame %1)").arg(displayFrame);
		}
		else {
			setSourceModelShowAct->setChecked(false);
			setTargetModelShowAct->setChecked(true);

			// Find the frame number
			displayFrame = scene->getAnimation()->frameNumber(scene->target);
			message = tr("Showing target model (frame %1)").arg(displayFrame);
		}

		displayMessage(message);
		handled = true;
	}
	else if ((e->key()==Qt::Key_PageDown) && (modifiers==Qt::NoButton)) {
		displayScale++;
		displayMessage(tr("Displaying feature component %1").arg(displayScale));
	}
	else if ((e->key()==Qt::Key_PageUp) && (modifiers==Qt::NoButton)) {
		displayScale = max(0, displayScale-1);
		displayMessage(tr("Displaying feature component %1").arg(displayScale));
	}
	else if ((e->key()==Qt::Key_PageDown) && (modifiers==Qt::CTRL)) {
		curvDisplayMode++;
		if (curvDisplayMode > 4) {
			curvDisplayMode = 1;
		}
	}
	else if ((e->key()==Qt::Key_PageUp) && (modifiers==Qt::CTRL)) {
		curvDisplayMode--;
		if (curvDisplayMode < 1) {
			curvDisplayMode = 4;
		}
	}
	else if ((e->key()==Qt::Key_T) && (modifiers==Qt::CTRL)) {
		// Generate a random point
		real3_type meshMin = scene->target->box_min;
		real3_type meshMax = scene->target->box_max;

		for (int d = 0; d < 3; d++) {
			real_type a = genrand_real1();
			testPoint[d] = (1.0 - a)*meshMin[d] + a*meshMax[d];
		}

		real3_type diffMesh;
		testDMesh = scene->target->pointSqDistance(testPoint, testCMesh);
		diffMesh = testPoint - testCMesh;
		cout << "Squared distance to mesh: " << testDMesh << " v.s. " << diffMesh*diffMesh << endl;

		//// Set the box and triangle
		//testBoxMin[0] = 0.3; testBoxMax[0] = 0.5;
		//testBoxMin[1] = 0.4; testBoxMax[1] = 0.6;
		//testBoxMin[2] = 0.5; testBoxMax[2] = 0.7;
		//testT0[0] = 0.1; testT1[0] = 0.4; testT2[0] = 0.0;
		//testT0[1] = 0.1; testT1[1] = 0.4; testT2[1] = 0.7;
		//testT0[2] = 0.1; testT1[2] = 0.4; testT2[2] = 0.8;

		//real3_type diffBox, diffTri;
		//real_type lbDTri = lowerBoundTriangleSqDistance(testPoint, testT0, testT1, testT2);

		//// Compute distances
		//testDBox = pointBoxSqDistance(testPoint, testBoxMin, testBoxMax, testCBox);
		//testDTri = pointTriangleSqDistance(testPoint, testT0, testT1, testT2, testCTri);

		//diffBox = testPoint - testCBox;
		//diffTri = testPoint - testCTri;

		//cout << "Squared distance to box: " << testDBox << " v.s. " << diffBox*diffBox << endl;
		//cout << "Squared distance to triangle: " << testDTri << " v.s. " << diffTri*diffTri << endl;
		//cout << "Lower-bound distance to triangle: " << lbDTri << endl;

		handled = true;
	}
	else if ((e->key()==Qt::Key_S) && (modifiers==Qt::CTRL+Qt::ALT)) {
		//if (scene->matchCandidates) {
		//	perfTimers["Spin image stack"].Start();
		//	scene->matchCandidates->sourceModel.computeSpinImages(scene->binSizeMultiple_SI, scene->imageWidth_SI, scene->maxSupportAngle_SI);
		//	perfTimers["Spin image stack"].Stop();
		//	printTime("Spin image stack");
		//}
		// Compute the spin image for the current selected vertex
		if (sourceQueryIndex >= 0) {
			cout << "Computing spin image for vertex " << sourceQueryIndex << endl;

			// Save the spin image to disk for viewing
			//scene->matchCandidates->sourceModel.computeSpinImage(sourceQueryIndex);
			scene->matchCandidates->sourceModel.saveSingleSpinImage("siTemp.data", sourceQueryIndex);
		}
	}

	if (!handled) {
		QGLViewer::keyPressEvent(e);
	}

	emit updateChildDisplay();
}

real3_type Viewer::closestPointOnLine( QPoint pos, real3_type x )
{
	// Find the line in 3D corresponding to the cursor
	qglviewer::Vec o, d;
	real3_type orig, dir;
	camera()->convertClickToLine(pos, o, d);
	orig = VecToVec3f(o);
	dir = VecToVec3f(d);

	// Find closest point on line
	real_type t = - ((orig-x) * dir) / (dir*dir);
	return orig + t*dir;
}

void Viewer::mousePressEvent( QMouseEvent *e )
{
	if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
		updateGL();

		// Move the selected vertex along the shape
		bool found;
		clickStartPos = VecToVec3f(camera()->pointUnderPixel(e->pos(), found));
		clickEndPos = clickStartPos;

		if (found) {
			//cout << "Setting start position " << clickStartPos << endl;
			interactionMode_ = DEFINE_SPHERE;
		}
		else {
			interactionMode_ = NONE;
		}
	}
	else {
		QGLViewer::mousePressEvent(e);
	}
}

void Viewer::mouseMoveEvent( QMouseEvent *e )
{
	if (interactionMode_ == DEFINE_SPHERE) {
		clickEndPos = closestPointOnLine(e->pos(), clickStartPos);

		//cout << "Move: Setting end position to " << clickEndPos << endl;
		updateGL();
	}
	else {
		QGLViewer::mouseMoveEvent(e);
	}
}

void Viewer::mouseReleaseEvent( QMouseEvent *e )
{
	if (interactionMode_ == DEFINE_SPHERE) {
		if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
			clickEndPos = closestPointOnLine(e->pos(), clickStartPos);

			//cout << "Release: Setting end position to " << clickEndPos << endl;

			if (scene->source) {
				// Perform sphere fit
				//sphereFit();

				// Find feature space points close in proximity
				featureProximity();
			}

			updateGL();
		}

		interactionMode_ = NONE;
	}
	else {
		QGLViewer::mouseReleaseEvent(e);
	}
}

void Viewer::Update( Subject* theChangedSubject )
{
	// Implement updating of opengl here by setting a variable and looping in an animation loop
	sceneMutex.lock();
	sceneChanged = true;
	sceneMutex.unlock();

	//cerr << "Updated by scene at: " << theChangedSubject << endl;
	//updateGL();
}

QString Viewer::updateFrame( int frame )
{
	QString message;

	// determine the mode
	if (showSourceModel() && !showTargetModel()) {
		scene->source = (*scene->getAnimation())[displayFrame];
		message = tr("Showing source frame %1").arg(displayFrame);
	}
	else if (!showSourceModel() && showTargetModel()) {
		scene->target = (*scene->getAnimation())[displayFrame];
		message = tr("Showing target frame %1").arg(displayFrame);
	}
	else {
		message = tr("View only the source/target frame to update.");
	}

	return message;
}

void Viewer::customSaveSnapshot()
{
	cout << "Saving snapshot to file: " << qPrintable(customSaveSnapshotFile) << endl << flush;
	saveSnapshot(customSaveSnapshotFile);
}


void Viewer::captureVideo()
{
	// Iterate through each frame and take a screenshot
	// Show only the source model
	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(true);

	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the animation", scene->lastDir, tr("*.%1").arg(snapshotFormat().toLower()));
	if (savefileName.length() == 0) {
		cout << "Video capture cancelled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	const int numFrames = scene->getAnimation()->numFrames();

	FileSequence seq(savefileName);
	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal

	for (unsigned int i = 0; i < numFrames; i++) {
		scene->target = (*scene->getAnimation())[i];

		// Update the display and take a screenshot
		customSaveSnapshotFile = seq[i];
		updateGL();
	}
	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal

}

void Viewer::CreateCorrespViewer()
{
	if (!correspViewer) {
		correspViewer = new CorrespViewer(scene, this);
		correspViewer->show();

		if (featureViewer) {
			featureViewer->correspViewer = correspViewer;
		}

		connect(this, SIGNAL(updateChildDisplay()), correspViewer, SLOT(updateDisplay()));
		connect(correspViewer, SIGNAL(destroyed()), this, SLOT(CorrespViewerDestroyed()));
	}
	else {
		// Focus on the current render controller
		// Raise the current path editor
		correspViewer->hide();
		correspViewer->show();
	}
}

void Viewer::CorrespViewerDestroyed() {
	//cerr << "Correspondence Viewer is: " << correspViewer << endl;
	correspViewer = NULL;
}


void Viewer::CreateCorrespEditor()
{
	if (!correspEdit) {
		correspEdit = new CorrespEdit();
		correspEdit->initRange(COLORMAP_RANGE_MIN, scene->colormapMin*2, 0, scene->colormapMin);
		correspEdit->initRange(COLORMAP_RANGE_MAX, 0, scene->colormapMax*2, scene->colormapMax);
		correspEdit->initRange(NUMBER_SAMPLES_SOURCE, 0, scene->featureSamplesSource*2, scene->featureSamplesSource);
		correspEdit->initRange(NUMBER_SAMPLES_TARGET, 0, scene->featureSamplesTarget*2, scene->featureSamplesTarget);
		correspEdit->initRange(FEATURE_SPACE_RANGE, 0, maxFeatureSpaceRange, scene->radius_FM);
		correspEdit->initRange(FEATURE_SPACE_THRESHOLD, 0, scene->threshold_FM*2, scene->threshold_FM);
		correspEdit->initRange(FEATURE_SPACE_MAX_MATCHES, 0, scene->maxCandidates_FM*2, scene->maxCandidates_FM);
		correspEdit->initRange(MEAN_SHIFT_BANDWIDTH, 0, scene->radius_MS*2, scene->radius_MS);
		correspEdit->initRange(MIN_CLUSTER_SIZE, 0, scene->minVerifyClusterSize*2, scene->minVerifyClusterSize);
		correspEdit->initRange(MIN_VERIFY_REGION_SIZE, 0, scene->minMatchSize_VR*2, scene->minMatchSize_VR);
		correspEdit->initRange(MAX_NUM_CLUSTERS, 0, scene->maxNumVerifyClusters_VR*2, scene->maxNumVerifyClusters_VR);
		correspEdit->initRange(MAX_VERIFIED_REGIONS, 0, scene->maxNumLabels_GC*2, scene->maxNumLabels_GC);
		correspEdit->initRange(CLUSTER_SELECT, 0, scene->showClusterMode, scene->showClusterMode);
		correspEdit->initRange(MAX_PAIR_DISTANCE, 0, scene->maxDist_FM*2, scene->maxDist_FM);
		
		correspEdit->show();

		connect(correspEdit, SIGNAL(valueChanged(int,double)), this, SLOT(CorrespEditorChanged(int,double)));
		connect(correspEdit, SIGNAL(destroyed()), this, SLOT(CorrespEditorDestroyed()));
	}
	else {
		correspEdit->hide();
		correspEdit->show();
	}
}

void Viewer::CorrespEditorDestroyed()
{
	//cerr << "Correspondence Editor is: " << correspEdit<< endl;
	correspEdit = NULL;
}

void Viewer::CorrespEditorChanged( int which, double value )
{
	switch(which) {
	case COLORMAP_RANGE_MIN:
		scene->colormapMin = value;
		break;
	case COLORMAP_RANGE_MAX:
		scene->colormapMax = value;
		break;
	case NUMBER_SAMPLES_SOURCE:
		scene->featureSamplesSource = value;
		testFeatureSelect();
		break;
	case NUMBER_SAMPLES_TARGET:
		scene->featureSamplesTarget = value;
		testFeatureSelect();
		break;
	case FEATURE_SPACE_RANGE:
		scene->radius_FM = value;
		//cout << "Changing feature space range query radius to: " << scene->radius_FM << endl;
		featureProximity();
		break;
	case FEATURE_SPACE_THRESHOLD:
		scene->threshold_FM = value;
		break;
	case FEATURE_SPACE_MAX_MATCHES:
		scene->maxCandidates_FM = value;
		featureProximity();
		break;
	case MEAN_SHIFT_BANDWIDTH:
		scene->radius_MS = value;
		break;
	case MIN_CLUSTER_SIZE:
		scene->minVerifyClusterSize = value;
		break;
	case CLUSTER_SELECT:
		scene->showClusterMode = value;
		break;
	case MAX_PAIR_DISTANCE:
		scene->maxDist_FM = value;
		featureProximity();
		break;
	case MIN_VERIFY_REGION_SIZE:
		scene->minMatchSize_VR = value;
		if (scene->clusterResultExists()) {
			scene->matchCandidates->sortVerifyResult(scene->minMatchSize_VR, scene->maxNumLabels_GC);
		}
		break;
	case MAX_NUM_CLUSTERS:
		scene->maxNumVerifyClusters_VR = value;
		break;
	case MAX_VERIFIED_REGIONS:
		scene->maxNumLabels_GC = value;
		if (scene->clusterResultExists()) {
			scene->matchCandidates->sortVerifyResult(scene->minMatchSize_VR, scene->maxNumLabels_GC);
		}
		break;
	case COMPUTE_FEATURES:
		// Test feature computation
		testFeatureSelect();
		break;
	case REFINE_CLUSTER:
		// Refine the selected cluster using ICP
		testRefineCluster();
		break;
	case COMPUTE_TRANSFORMS:
		// Test feature matching
		testFeatureMatch();
		break;
	case TRANSFORMS_SAVE:
		// Save transformations
		saveTransformSpacePoints();
		break;
	case COMPUTE_MEAN_SHIFT:
		testFeatureCluster();
		break;
	case TRANSFORMS_VERIFY:
		verifyTransformationSpace();
		break;
	case GRAPH_CUT_RELABEL:
		testGraphCutRelabel();
		break;
	case SAMPLE_CORRESP:
		testSampleCorrespondences();
		break;
	case FEATURE_VIEWER:
		updateFeatureViewer();
		break;
	default:
		break;
	}

	updateGL();
	emit updateChildDisplay();
}


void Viewer::CreateFeatureViewer()
{
	if (!featureViewer) {
		featureViewer = new FeatureViewer(scene);
		featureViewer->show();

		if (correspViewer) {
			featureViewer->correspViewer = correspViewer;
		}

		connect(this, SIGNAL(updateChildDisplay()), featureViewer, SLOT(updateGL()));
		connect(featureViewer, SIGNAL(destroyed()), this, SLOT(FeatureViewerDestroyed()));
	}
	else {
		featureViewer->hide();
		featureViewer->show();
	}
}

void Viewer::updateFeatureViewer()
{
	if (!featureViewer) {
		// Create a new one
		CreateFeatureViewer();
	}
	featureViewer->updateGL();
}

void Viewer::FeatureViewerDestroyed()
{
	//cerr << "Feature Viewer is: " << featureViewer << endl;
	featureViewer = NULL;
}



void Viewer::CreateShapeSlicer()
{
	ShapeModel *shape = selectCurrentModel();

	if (shape && !shapeSlicer) {
		shapeSlicer = new ShapeSlicer(scene, shape);
		shapeSlicer->show();

		connect(this, SIGNAL(updateChildDisplay()), shapeSlicer, SLOT(updateGL()));
		connect(shapeSlicer, SIGNAL(destroyed()), this, SLOT(ShapeSlicerDestroyed()));
	}
	else {
		// Focus on the current render controller
		// Raise the current path editor
		shapeSlicer->hide();
		shapeSlicer->show();
	}
}

void Viewer::ShapeSlicerDestroyed() {
	//cerr << "Shape Slicer is: " << shapeSlicer << endl;
	shapeSlicer = NULL;
}


void Viewer::CreateTransformViewer()
{
	ShapeModel *shape = selectCurrentModel();

	if (shape && !transformViewer) {
		transformViewer = new TransformViewer(scene, scene->source, scene->target);
		transformViewer->show();

		if (scene->matchCandidates && scene->matchCandidates->graphCutLabels_source.size()) {
			const ShapeModel &shape = scene->matchCandidates->sourceModel;

			// Assign the initial transformations
			for (index_type i = 0; i < shape.numVertices(); i++) {
				index_type label = scene->matchCandidates->graphCutLabels_source[i];
				transformViewer->setAssignedTf(i, scene->matchCandidates->sortedTFList[label]);
			}

			cout << "Assigned graph cut result as initial transformations!" << endl;
		}

		connect(this, SIGNAL(updateChildDisplay()), transformViewer, SLOT(updateGL()));
		connect(transformViewer, SIGNAL(destroyed()), this, SLOT(TransformViewerDestroyed()));
	}
	else {
		// Focus on the current render controller
		// Raise the current path editor
		transformViewer->hide();
		transformViewer->show();
	}
}

void Viewer::TransformViewerDestroyed() {
	//cerr << "TransformViewer is: " << transformViewer << endl;
	transformViewer = NULL;
}

void Viewer::saveColormap()
{
	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the colormap...", scene->lastDir, QString("*.txt"));
	if (savefileName.length() == 0) {
		cout << "Viewer::loadColormap(): Save canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	scene->colormap.save(qPrintable(savefileName));
}

void Viewer::loadColormap()
{
	QString loadfileName;
	loadfileName = QFileDialog::getOpenFileName(this, "Choose a file name to load the colormap...", scene->lastDir, tr("*.txt"));
	if (loadfileName.length() == 0) {
		cout << "Viewer::loadColormap(): Loading canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(loadfileName);
	}

	scene->colormap.load(qPrintable(loadfileName));
}

void Viewer::setColormapLimits()
{
	// Set the range of the colormap
	real_type newMin, newMax;

	// Set the resolution of the screen here.
	bool ok;

	newMin = QInputDialog::getDouble(this, tr("Colormap minimum range"),
				tr("Input Minimum:"), scene->colormapMin, -DBL_MAX, DBL_MAX, 5, &ok);
	if (!ok) return;

	newMax = QInputDialog::getDouble(this, tr("Colormap maximum range"),
				tr("Input Maximum:"), scene->colormapMax, -DBL_MAX, DBL_MAX, 5, &ok);
	if (!ok) return;

	// Update the window
	scene->colormapMin = newMin;
	scene->colormapMax = newMax;
	updateGL();
}

void Viewer::setFeatureThreshold()
{
	// Set the range of the colormap
	real_type newVal;

	// Set the resolution of the screen here.
	bool ok;

	newVal = QInputDialog::getDouble(this, tr("Feature threshold display"),
		tr("Input Threshold:"), scene->threshold_FM, -DBL_MAX, DBL_MAX, 5, &ok);
	if (!ok) return;

	// Update the window
	scene->threshold_FM = newVal;
	updateGL();
}

void Viewer::saveTransformSpacePoints()
{
	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the feature points...", scene->lastDir, QString("*.data"));
	if (savefileName.length() == 0) {
		cout << "Viewer::saveFeatureSpacePoints(): Save canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	scene->matchCandidates->writeData(qPrintable(savefileName));
}


void Viewer::sphereFit()
{
	PointList rangePts, rangeNormals;
	real3_type sample = clickStartPos;
	real_type radSq = (clickEndPos - clickStartPos).square_norm();

	cout << "Fitting sphere at: " << sample << " with radius^2 " << radSq << endl;

	scene->source->vertex.queryRange(sample, radSq, inRange, distRange, true);

	// Gather the point locations and the normals
	rangePts.resize(inRange.size());
	rangeNormals.resize(inRange.size());
	for (index_type j = 0; j < inRange.size(); j++) {
		// Throw out the point if the normal is zero
		if (scene->source->normal[inRange[j]].norm() >= REAL_TYPE_EPSILON) {
			rangePts[j] = scene->source->vertex[inRange[j]];
			rangeNormals[j] = scene->source->normal[inRange[j]];
		}
	}

	// Fit a circle at this position
	if (FitSphere(rangePts, rangeNormals, sample, radSq)) {
		// Store the mean curvature (inverse of the radius) as the feature
		//features.val[s] = 1.0 / sqrt(radSq);
		cout << "Fit sphere: center " << sample << " radius^2 " << radSq << endl;
		fitCenter = sample;
		fitRadius = sqrt(radSq);
	}
	else {
		// If the fit is degenerate, then the curvature is 0.
		//features.val[s] = 0;
		cout << "Fit plane: dist from origin " << radSq << endl;
	}
}



void Viewer::featureProximity()
{
	idxList inRange;

	if (!scene->matchCandidates) {
		return;
	}

	// Find closest vertex in source model
	ShapeModel &sourceShape = scene->matchCandidates->sourceModel;
	ClosestPointInfo cpi = sourceShape.findClosestPoint(clickStartPos);
	sourceQueryIndex = cpi.index;

	// Write the query spin image to disk
	sourceShape.saveSingleSpinImage("J:\\singleSpinImage.data", cpi.index);

	// Find closest point in target model
	//sourceQuery.resize(1);
	//sourceQuery[0] = sourceShape.vertex[sourceQueryIndex];

	//real_type temp;
	//real3_type closest;
	//scene->matchCandidates->targetModel.hybridClosestPoint(sourceQuery[0], closest, temp);
	//targetMatch.clear();
	//targetMatch.add(closest);

	//return;

	//// Match spin image features
	//if (scene->featureType == FEATURE_SPIN_IMAGES) {
	//	ShapeModel &targetShape = scene->matchCandidates->targetModel;
	//	sourceQuery.resize(1);
	//	sourceQuery[0] = sourceShape.vertex[sourceQueryIndex];

	//	const SpinImage &query = sourceShape.spinImageStack[sourceQueryIndex];
	//	targetShape.spinImageStack.querySimilar(query, inRange);

	//	targetMatch.clear();
	//	for (index_type i = 0, i_end = min(scene->maxCandidates_FM, inRange.size()); i < i_end; i++) {
	//		targetMatch.add(targetShape.vertex[inRange[i]]);
	//	}

	//	return;
	//}

	// Match spin image features
	if (scene->featureType == FEATURE_SPIN_IMAGES) {
		if (!scene->matchCandidates->sourceSample || !scene->matchCandidates->targetSample) {
			return;
		}

		ShapeModel &sourceSample = *scene->matchCandidates->sourceSample;
		ShapeModel &targetSample = *scene->matchCandidates->targetSample;

		distList distances;
		sourceSample.queryKNN(clickStartPos, 1, inRange, distances, true);
		sourceQuery.resize(1);
		sourceQuery[0] = sourceSample.vertex[inRange[0]];

		const SpinImage &query = sourceSample.spinImageStack[inRange[0]];
		targetSample.spinImageStack.querySimilar(query, inRange);
		//targetSample.spinImageStack.querySimilar(query, inRange, "J:\\scores.txt");

		targetMatch.clear();
		for (index_type i = 0, i_end = min(scene->maxCandidates_FM, (int)inRange.size()); i < i_end; i++) {
			targetMatch.add(targetSample.vertex[inRange[i]]);
		}

		return;
	}

	//////////////////////////////////////////////////////////////////////////////
	// Match other features
	distList distances;
	const real_type radSq = scene->radius_FM * scene->radius_FM;	/**< Range query radius in feature space */
	const int maxCandidates = scene->maxCandidates_FM;				/**< Maximum number of candidates extracted */

	if (!scene->matchCandidates || !scene->matchCandidates->sourceSample || !scene->matchCandidates->targetSample) {
		return;
	}
	ShapeModel *src = scene->matchCandidates->sourceSample;
	ShapeModel *tgt = scene->matchCandidates->targetSample;

	// Set ANN Epsilon for feature space query
	tgt->features->setANNEpsilon(scene->ANNEpsilon_FM);

	// Find closest feature point in source sample
	src->queryKNN(clickStartPos, 1, inRange, distances, true);
	sourceQuery.resize(1);
	sourceQuery[0] = src->vertex[inRange[0]];
	data_type queryFeature = (*src->features)[inRange[0]];

	// Query feature space
	tgt->features->queryRange(queryFeature, radSq, inRange, distances, true);
	//tgt->features->queryKNN(queryFeature, 7, inRange, distances, true);

	// Populate result
	int n = inRange.size();
	if (maxCandidates > 0) {
		n = min(n, maxCandidates);
	}
	targetMatch.clear();
	targetMatch.reserve(n);

	const real_type maxPairDist = scene->makeSceneRadius(scene->maxDist_FM);

	for (index_type i = 0; i < n; i++) {
		// Threshold the Euclidean distance
		real3_type targetPoint = tgt->vertex[inRange[i]];
		if ((sourceQuery[0] - targetPoint).norm() < maxPairDist) {
			targetMatch.add(tgt->vertex[inRange[i]]);
		}
	}
}


void Viewer::testFeatureSelect()
{
	// Just sample the source and target meshes
	scene->initMatchCandidates(*scene->source, *scene->target);
	scene->matchCandidates->subSampleGeometry(scene->featureSamplesSource, scene->featureSamplesTarget, scene->pruneMaxRatio_FM);

	// Test the sampling of smoothness edges on the boundary
	scene->matchCandidates->shapeBoundaryKNNEdges(scene->matchCandidates->sourceModel, scene->bdySmoothnessKNN_GC, scene->matchCandidates->sourceExtraEdges);
	scene->matchCandidates->shapeBoundaryKNNEdges(scene->matchCandidates->targetModel, scene->bdySmoothnessKNN_GC, scene->matchCandidates->targetExtraEdges);

	//// Recompute principal curvatures and update display for the currently displayed mesh
	//ShapeModel *shape = selectCurrentModel();

	//shape->computeFrames(scene->makeSceneSqRadii(scene->scaleFactors_PC), true);
}

void Viewer::testFeatureMatch()
{
	scene->initMatchCandidates(*scene->source, *scene->target);
	scene->matchCandidates->subSampleGeometry(scene->featureSamplesSource, scene->featureSamplesTarget, scene->pruneMaxRatio_FM);

	if (scene->featureType == FEATURE_SPIN_IMAGES) {
		scene->matchCandidates->spinImageCandidates(scene->maxCandidates_FM);
	}
	else {
		scene->matchCandidates->findCandidateCorrespondence(scene->radius_FM*scene->radius_FM,
															scene->maxCandidates_FM,
															scene->ANNEpsilon_FM);
	}


	//scene->matchCandidates->RANSAC_Candidates(scene->RANSAC_max_itr,
	//										  scene->RANSAC_sourceModel_ANNEpsilon,
	//										  scene->RANSAC_targetModel_ANNEpsilon,
	//										  scene->makeSceneRadiusSq(scene->RANSAC_inlier_queryRadius),
	//										  scene->makeSceneRadiusSq(scene->RANSAC_inlier_epsilonBand),
	//										  scene->RANSAC_minimumInliersPercentage / 100.0);
}

/**
 * Assumes that point correspondences are already defined between the source and target meshes.
 */
void Viewer::verifyTransformationSpace()
{
	scene->initMatchCandidates(*scene->source, *scene->target);
	scene->matchCandidates->groundTruthCorrespondence();
}

void Viewer::testFeatureCluster()
{
	// Test the mean-shift clustering
	scene->matchCandidates->meanShiftBasic(
		scene->radius_MS*scene->radius_MS,
		scene->rate_MS,
		scene->maxIter_MS,
		scene->ANNEpsilon_MS
		);

	//scene->matchCandidates->pruneClusterGeometricConsistency(0.25);
}

void Viewer::testRefineCluster()
{
	//const int n = 54;
	//real_type leftx[] = {1.1083, 1.29, 0.9052, 0.6289, 0.9923, 1.2578, 1.1075, 0.5549, 1.0859, 0.2728, 0.5033, 0.9631, 0.6627, 0.2577, 0.0999, -0.4068, -0.6887, -0.8531, -1.425, -1.2356, -1.3816, -1.0528, -0.7782, -0.6249, -1.0701, 0.1044, -0.3676, 0.9159, 1.2793, 1.0869, 0.5343, 0.716, 1.2469, 0.9465, 0.6702, 0.9357, 0.1059, 0.2211, 0.6901, 0.5399, 0.0983, -0.2399, -0.2339, -0.7978, -1.1265, -1.4125, -1.0712, -1.3632, -1.1988, -0.6496, -0.343, -0.7955, -0.2082, -0.2143};
	//real_type lefty[] = {1.3408, 1.0535, 1.2876, 1.3628, 0.7881, 0.3829, 0.5093, 0.6597, -0.1507, 0.651, 0.0934, -0.3888, -0.136, -0.3828, 0.9476, 0.6691, 0.6928, 0.905, 1.1257, 0.5715, 0.453, 0.0285, -0.0977, 0.1233, -0.3348, -0.6039, -0.3662, 1.4579, 0.8832, 1.0002, 1.1507, 0.8633, 0.0529, 0.3057, 0.3809, -0.0243, 0.3635, 0.0847, -0.5005, -0.3741, -0.1534, 0.9566, 0.3725, 0.4199, 0.8444, 0.9547, 0.3593, 0.1222, -0.09, -0.3425, 0.0997, -0.461, -0.5956, -0.1451};
	//real_type leftz[] = {-1.3836, -1.3241, -0.9513, -0.7652, -0.6462, -0.8966, -0.6831, -0.3109, -0.8118, -0.2333, -0.307, -0.8971, -0.47, -0.4777, -0.2677, -0.2207, -0.2879, -0.4311, -1.2739, -0.6398, -0.8478, -0.5615, -0.4434, -0.2862, -0.8595, -0.6349, -0.4661, -1.1972, -1.0782, -0.8918, -0.5196, -0.4601, -0.961, -0.5339, -0.3478, -0.5983, -0.1925, -0.2294, -0.7689, -0.5553, -0.3147, -0.2615, -0.1862, -0.3207, -0.607, -1.0284, -0.4966, -0.9127, -0.7695, -0.5333, -0.2189, -0.7414, -0.6291, -0.3089};
	//real_type rightx[] = {1.2175, 1.2861, 1.0737, 0.886, 1.0232, 1.1259, 1.0497, 0.6743, 0.8796, 0.447, 0.5, 0.7139, 0.5615, 0.1684, 0.3804, -0.0878, -0.319, -0.4024, -0.8895, -0.8149, -0.9846, -0.8178, -0.6234, -0.4169, -0.9628, -0.038, -0.3675, 1.1113, 1.2485, 1.1423, 0.7669, 0.8355, 1.0409, 0.8885, 0.7008, 0.8034, 0.2462, 0.2727, 0.4719, 0.3957, 0.1069, 0.113, -0.0212, -0.4835, -0.6503, -0.8939, -0.7315, -1.071, -0.9876, -0.5986, -0.1858, -0.7684, -0.3059, -0.1611};
	//real_type righty[] = {0.3913, 0.1783, 0.6062, 0.8012, 0.3753, -0.0696, 0.1516, 0.5416, -0.3483, 0.6347, 0.1872, -0.5291, -0.0867, -0.1744, 0.8477, 0.7994, 0.8486, 0.9523, 0.8502, 0.7284, 0.5827, 0.3754, 0.2823, 0.4719, -0.009, -0.364, -0.0057, 0.6052, 0.1793, 0.3932, 0.7833, 0.5703, -0.3196, 0.1228, 0.3178, -0.1271, 0.5041, 0.2804, -0.4887, -0.2675, 0.0996, 0.93, 0.5865, 0.6848, 0.8922, 0.8411, 0.6247, 0.3334, 0.2297, 0.0435, 0.4227, -0.1021, -0.2796, 0.1839};
	//real_type rightz[] = {-1.8853, -1.7139, -1.57, -1.489, -1.1461, -1.1272, -1.0377, -0.8757, -0.8379, -0.8364, -0.6196, -0.7893, -0.6103, -0.5223, -1.0073, -0.8828, -0.9634, -1.1748, -1.9117, -1.2056, -1.3168, -0.894, -0.7139, -0.7022, -0.9364, -0.534, -0.5644, -1.8134, -1.4705, -1.3985, -1.2365, -1.0651, -1.0273, -0.8483, -0.7673, -0.7484, -0.6886, -0.5803, -0.6511, -0.5616, -0.5316, -1.0305, -0.7118, -0.8731, -1.2959, -1.6643, -0.9942, -1.2166, -1.0053, -0.645, -0.6216, -0.7562, -0.555, -0.5526};
	//PointList left, right;
	//left.resize(n);
	//right.resize(n);
	//for (index_type i = 0; i < n; i++) {
	//	left[i][0] = leftx[i];
	//	left[i][1] = lefty[i];
	//	left[i][2] = leftz[i];
	//	right[i][0] = rightx[i];
	//	right[i][1] = righty[i];
	//	right[i][2] = rightz[i];
	//}
	//registerQuat(left, right);

	//// Refine the currently selected cluster using ICP
	//if (scene->clusterResultExists()) {
	//	scene->matchCandidates->refineCluster(scene->showClusterMode);
	//}

	if (scene->skipVerification_VR) {
		scene->matchCandidates->skipVerify(scene->minVerifyClusterSize, scene->maxNumVerifyClusters_VR);
	}
	else {
		scene->matchCandidates->verifyRegion(scene->minVerifyClusterSize, scene->maxNumVerifyClusters_VR, scene->makeSceneRadius(scene->growRegionDistThreshold_VR), scene->ANNEpsilon_VR);
	}
	scene->matchCandidates->sortVerifyResult(scene->minMatchSize_VR, scene->maxNumLabels_GC);
}

void Viewer::testGraphCutRelabel()
{
	if (scene->clusterResultExists()) {
		if (scene->graphCutInstance_GC == GRAPH_CUT_INSTANCE_SYMMETRIC) {
			scene->matchCandidates->GraphCutRelabel(
				scene->graphCutSolutionType_GC,
				scene->dataTermType_GC,
				scene->useDummyLabel_GC,
				scene->bdySmoothnessKNN_GC,
				scene->makeSceneRadius(scene->closestPtDistThreshold_GC),
				scene->makeSceneRadius(scene->mismatchError_GC),
				scene->makeSceneRadius(scene->dummyLabelSmoothError_GC),
				scene->makeSceneRadius(scene->dummyLabelDataError_GC),
				scene->smoothWeight_GC,
				scene->dataWeight_GC,
				scene->makeSceneRadius(scene->smoothTruncateValue_GC),
				scene->makeSceneRadius(scene->dataTruncateValue_GC));
		}
		else if (scene->graphCutInstance_GC == GRAPH_CUT_INSTANCE_FORWARD) {
			scene->matchCandidates->GraphCutRelabel_Forward(
				scene->graphCutSolutionType_GC,
				scene->dataTermType_GC,
				scene->smoothWeight_GC,
				scene->dataWeight_GC,
				scene->makeSceneRadius(scene->smoothTruncateValue_GC),
				scene->makeSceneRadius(scene->dataTruncateValue_GC));
		}
		else {
			cerr << "Unknown graph cut instance type " << scene->graphCutInstance_GC << "!" << endl;
		}
	}
}

void Viewer::testSampleCorrespondences()
{
	scene->solveCorrespondences();

	// Extract the correspondences
	//scene->packageCorrespondences();
}

ShapeModel* Viewer::selectCurrentModel()
{
	if (showAllFramesAct->isChecked()) {
		return (*scene->getAnimation())[0];
	}
	else {
		if (showTargetModel()) {
			return scene->target;
		}
		if (showSourceModel()) {
			return scene->source;
		}
		if (!showSourceModel() && !showTargetModel()) {
			return scene->source;
			//cout << "Please display a mesh for computing curvatures." << endl;
		}
	}

	return NULL;
}

void Viewer::saveFrames()
{
	//// Save the per-vertex frames of the source and target meshes.
	//const CurvInfoList *pCurvDataS = &scene->source->frames;
	//const CurvInfoList *pCurvDataT = &scene->target->frames;
	//if (!pCurvDataS || !pCurvDataT) {
	//	return;
	//}

	////////////////////////////////////////////////////////////////////////////
	//QString sourceName = scene->source->geometryFile();
	//QString savefileName;
	//savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save frames...",
	//	QString("%1/frames-%2%3.data").arg(scene->lastDir).arg(sourceName.left(sourceName.size()-4)).arg(useRobustCurvaturesAct->isChecked()?"-robust":""),
	//	QString("*.data"));

	//if (savefileName.length() == 0) {
	//	cout << "Viewer::saveCurvatures(): Save canceled." << endl;
	//	return;
	//}
	//else {
	//	scene->lastDir = grabPath(savefileName);
	//}
	//saveMeshFrames(qPrintable(savefileName), scene->source, *pCurvDataS);

	////////////////////////////////////////////////////////////////////////////
	//sourceName = scene->target->geometryFile();
	//savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save frames...",
	//	QString("%1/frames-%2%3.data").arg(scene->lastDir).arg(sourceName.left(sourceName.size()-4)).arg(useRobustCurvaturesAct->isChecked()?"-robust":""),
	//	QString("*.data"));

	//if (savefileName.length() == 0) {
	//	cout << "Viewer::saveCurvatures(): Save canceled." << endl;
	//	return;
	//}
	//else {
	//	scene->lastDir = grabPath(savefileName);
	//}
	//saveMeshFrames(qPrintable(savefileName), scene->target, *pCurvDataT);
}

void Viewer::testDistanceCode()
{
	glDisable(GL_LIGHTING);

	// Draw point
	glPointSize(5.0);
	glBegin(GL_POINTS);
	glColor3f(1.0, 0.0, 0.0);
	glVertex3f(testPoint[0],testPoint[1],testPoint[2]);
	glColor3f(0.0, 0.0, 1.0);
	glVertex3f(testCMesh[0],testCMesh[1],testCMesh[2]);
	//glColor3f(0.0, 1.0, 0.0);
	//glVertex3f(testCBox[0],testCBox[1],testCBox[2]);
	//glColor3f(0.0, 0.0, 1.0);
	//glVertex3f(testCTri[0],testCTri[1],testCTri[2]);
	glEnd();

	//// Draw box
	//glColor3f(1.0, 1.0, 1.0);
	//drawBox(testBoxMin, testBoxMax);

	//// Draw triangle
	//glBegin(GL_TRIANGLES);
	//glVertex3f(testT0[0],testT0[1],testT0[2]);
	//glVertex3f(testT1[0],testT1[1],testT1[2]);
	//glVertex3f(testT2[0],testT2[1],testT2[2]);
	//glEnd();
}

void Viewer::visualizeClusterMode()
{
	if (!scene->clusterResultExists()) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Visualize the cluster modes by applying the mode transformation to the source and comparing with the target.
	float color[4];
	int numClusters = scene->matchCandidates->clusterModes.size();
	scene->showClusterMode = max(0, min(numClusters-1, scene->showClusterMode));
	//SE3	ctf = scene->matchCandidates->getClusterTransform(scene->showClusterMode);
	SE3 ctf = scene->matchCandidates->tfList[scene->showClusterMode];

	// Draw source mesh w/ transformation
	glPushMatrix();
	ctf.applyGL();
	color[0] = 0.7f;  color[1] = 0.7f;  color[2] = 1.0f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	scene->source->draw();
	glPopMatrix();

	// Draw target mesh
	color[0] = 1.0f;  color[1] = 0.7f;  color[2] = 0.7f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	scene->target->draw();

	// Draw correspondence pairs for this cluster
	glDisable(GL_LIGHTING);
	glBegin(GL_POINTS);
	glPointSize(10.f);
	for (index_type i = 0, i_end = scene->matchCandidates->size(); i < i_end; i++ ) {
		// Draw all the pairs with this label
		if (scene->matchCandidates->clusterLabels[i] == scene->showClusterMode) {
			//real3_type sp = scene->matchCandidates->sourceSample->vertex[scene->matchCandidates->sourceIndex[i]];
			real3_type tp = scene->matchCandidates->targetSample->vertex[scene->matchCandidates->targetIndex[i]];

			//glColor3f(0.5, 0.5, 1.0);
			//glVertex3f(sp[0],sp[1],sp[2]);
			glColor3f(1.0, 0.2, 0.2);
			glVertex3f(tp[0],tp[1],tp[2]);
		}
	}
	glEnd();

	displayMessage(QString("Showing cluster %1 / %2 (%3 pairs)").arg(scene->showClusterMode).arg(numClusters).arg(scene->getClusterSize(scene->showClusterMode)));

	////////////////////////////////////////////////////////////////////////////
	//// Draw each verified region
	//size_type numRegions = scene->matchCandidates->matchList.size();
	//scene->showClusterMode = max(0, min(numRegions-1, scene->showClusterMode));
	//if (!scene->matchCandidates->validCluster(scene->showClusterMode)) {
	//	return;
	//}

	//CorrespondenceList &list = scene->matchCandidates->matchList[scene->showClusterMode];
	//glDisable(GL_LIGHTING);
	//glPointSize(3.0);
	//glBegin(GL_POINTS);
	//for (index_type i = 0; i < list.size(); i++) {
	//	real3_type temp;

	//	// Draw source vertex
	//	glColor3f(0.7,0.7,1.0);
	//	temp = list.sourcePos(i);
	//	glVertex3f(temp[0],temp[1],temp[2]);

	//	// Draw corresponding target vertex
	//	glColor3f(1.0,0.7,0.7);
	//	temp = list.targetPos(i);
	//	glVertex3f(temp[0],temp[1],temp[2]);
	//}
	//glEnd();

	//displayMessage(QString("Showing region %1 with %2 vertices").arg(scene->showClusterMode).arg(list.size()));

	////////////////////////////////////////////////////////////////////////////
	//// Draw each verified region  -->  Moved to CorrespViewer
	//size_type numRegions = scene->matchCandidates->matchList.size();

	//if (scene->clusterColors.size() < numRegions) {
	//	// If not enough colors, generate more
	//	scene->generateClusterColors(numRegions);
	//}

	//glDisable(GL_LIGHTING);
	//glPointSize(3.0);
	//glBegin(GL_POINTS);
	//for (index_type ci = 0; ci < numRegions; ci++) {
	//	if (!scene->matchCandidates->validCluster(ci)) {
	//		continue;
	//	}

	//	setOpenGLColor(scene->getClusterColor(ci));

	//	CorrespondenceList &list = scene->matchCandidates->matchList[ci];
	//	for (index_type i = 0; i < list.size(); i++) {
	//		real3_type temp;

	//		// Draw corresponding target vertex
	//		temp = list.targetPos(i);
	//		glVertex3f(temp[0],temp[1],temp[2]);
	//	}
	//}
	//glEnd();
}


void Viewer::fitViewToShape( const ShapeModel *shape /*= NULL */ )
{
	if (!shape) {
		shape = selectCurrentModel();
	}

	if (shape) {
		// Fit the view to the current shape
		setSceneBoundingBox(Vec3fToVec(shape->box_min), Vec3fToVec(shape->box_max));

		camera()->setZClippingCoefficient(10);
		camera()->setZNearCoefficient(1e-2);
		showEntireScene();

		//cout << "Camera near clipping plane: " << camera()->zNear() << endl;
		//cout << "Camera far clipping plane: " << camera()->zFar() << endl;
	}
}

void Viewer::simplifyCurrentShape()
{
	// Prompt for the simplification factor
	bool ok;
	double fac = QInputDialog::getDouble(this, tr("Simplification Factor"),
		tr("Factor Reduction:"), 0.5, 0.0, 1.0, 4, &ok);

	if (ok) {
		ShapeModel *newShape = ((Mesh3_real*)selectCurrentModel())->simplifyGeometry(fac);
		scene->getAnimation()->add(newShape);
	}
}

void Viewer::batchSimplify()
{
	// Get a list of files
	QStringList files = QFileDialog::getOpenFileNames(this, tr("Select meshes to simplify..."), scene->lastDir, tr("Meshes (*.obj)"));
	if (files.size() == 0) {
		return;
	}

	// Prompt for a simplification factor
	bool ok;
	double fac = QInputDialog::getDouble(this, tr("Simplification Factor"),
		tr("Factor Simplification:"), 0.0, 0.0, 1.0, 4, &ok);
	if (!ok) {
		return;
	}

	// Prompt for a scale factor
	double scale = QInputDialog::getDouble(this, tr("Scale Factor"),
		tr("Factor Scale:"), 1.0, 0.0, 1e10, 4, &ok);
	if (!ok) {
		return;
	}

	// Simplify each file
	QStringList list = files;
	QStringList::Iterator it = list.begin();
	while(it != list.end()) {
		QString filename = *it;

		// Load and simplify
		Mesh3_real loaded;
		loaded.setFile(QString(scene->rootDir.c_str()), filename);
		Mesh3_real *simplified = loaded.simplifyGeometry(fac, scale);
		delete simplified;

		++it;
	}

}

void Viewer::changeBackgroundColor()
{
	QColor color = QColorDialog::getColor(this->backgroundColor(), this);
	if (color.isValid()) {
		this->setBackgroundColor(color);
	}
}

void Viewer::changeForegroundColor()
{
	QColor color = QColorDialog::getColor(this->foregroundColor(), this);
	if (color.isValid()) {
		this->setForegroundColor(color);
	}
}

real_type Viewer::ComputeAlignDistance()
{
	real_type error = REAL_TYPE_MAX;

	if (scene->matchCandidates) {
		error = scene->matchCandidates->maxSymmetricHausdorff();

		cout << "Maximum Symmetric Hausdorff Error: " << error << " ("
			 << error / scene->totalScene_boundingBoxDiagonal * 100.0 << "% of bbox diagonal)" << endl;
	}

	return error;
}
