#include <QGLViewer/qglviewer.h>
#include <QtGui>
#include <QMutex>
#include "observer.h"
#include "subject.h"
#include "PointList.h"
#include "PointListN.h"
#include "FindCorrespondence.h"

class Scene;
class CorrespViewer;
class DeformEdit;
class CorrespEdit;
class MatchCandidates;
class FeatureViewer;
class ShapeSlicer;
class TransformViewer;

/**
 * Viewer class.
 *
 * Responsibilities:
 *  - Displaying meshes on screen
 *  - Implementing menu & menu functionality
 */
class Viewer : public QGLViewer, public Observer
{
	Q_OBJECT

public:
	Viewer(Scene* s);										/**< Constructor: initialize members here. */
	virtual ~Viewer();										/**< Destructor: destroy members here. */

	virtual void Update(Subject* theChangedSubject);		/**< Update the display according to the subject */

private slots:
	void open(const char *file = NULL, int maxFrames = 0, int skipFrames = 1);	/**< Open at most "maxFrames" frames of a mesh animation sequence. */
	void add();

	void customSaveSnapshot();								/**< Take a snapshot and save it in a specific file. */
	void captureVideo();									/**< Take a video of the mesh animation sequence. */

	void CreateCorrespViewer();								/**< Display the correspondence viewer */
	void CorrespViewerDestroyed();							/**< Cleanup after the correspondence viewer has been closed */

	void CreateCorrespEditor();								/**< Display the dialog for editing correspondence parameters */
	void CorrespEditorDestroyed();							/**< Cleanup after the editor has been closed */
	void CorrespEditorChanged(int which, double value);		/**< Handle changes in parameter values */

	void CreateFeatureViewer();								/**< Display the dialog for viewing feature/transformation spaces */
	void updateFeatureViewer();								/**< Update the feature viewer display */
	void FeatureViewerDestroyed();							/**< Cleanup after the viewer has been closed */

	void saveColormap();									/**< Save the current colormap */
	void loadColormap();									/**< Load the current colormap */
	void setColormapLimits();								/**< Set colormap limits */

	void setFeatureThreshold();								/**< Set threshold on feature for display */

	void autoLaunch();										/**< Code to execute after the first draw */
	void autoProcess();										/**< Automatic processing for batch jobs */

	void saveFrames();

	void simplifyCurrentShape();							/**< Simplify the current displayed shape. */
	void batchSimplify();									/**< Simplify a batch of shapes. */
	void fitViewToShape( const ShapeModel *shape = NULL );	/**< Fit the view to the current shape. */

	void changeBackgroundColor();							/**< Change background color */
	void changeForegroundColor();							/**< Change foreground color */

	void CreateShapeSlicer();								/**< Display the shape slicer */
	void ShapeSlicerDestroyed();							/**< Cleanup after the shape slicer has been closed */

	void CreateTransformViewer();							/**< Display the transform viewer */
	void TransformViewerDestroyed();						/**< Cleanup after the transform viewer has been closed */

	real_type ComputeAlignDistance();						/**< Compute the alignment distance. */

protected :
	virtual void draw();
	virtual void init();
	virtual void animate();
	virtual QString helpString() const;
	virtual	void keyPressEvent(QKeyEvent *e);
	virtual void mousePressEvent(QMouseEvent *e);
	virtual void mouseMoveEvent(QMouseEvent *e);
	virtual void mouseReleaseEvent(QMouseEvent *e);

private:
	//////////////////////////////////////////////////////////////////////////
	// Context Menu Related Methods and Members
	//////////////////////////////////////////////////////////////////////////
	QMenu *contextMenu;
	QAction *openAct;
	QAction *addAct;
	QAction *setWireframeModeAct;
	QAction *setBboxModeAct;
	QAction *setSourceModelShowAct;
	QAction *setTargetModelShowAct;
	QAction *setTransparencyAct;
	QAction *clearResultAct;
	QAction *captureVideoAct;
	QAction *showAllFramesAct;
	QAction *showCorrespViewerAct;
	QAction *saveColormapAct;
	QAction *loadColormapAct;
	QAction *setColormapLimitsAct;
	QAction *setFeatureThresholdAct;
	QAction *showCorrespEditorAct;
	QAction *saveFramesAct;
	QAction *visualizeClusterAct;
	QAction *simplifyCurrentShapeAct;
	QAction *batchSimplifyAct;
	QAction *fitViewToShapeAct;
	QAction *changeBackgroundColorAct;
	QAction *changeForegroundColorAct;
	QAction *shapeSliceAct;
	QAction *assignTransformsAct;
	QAction *showVerticesAct;
	QAction *showVertexFramesAct;
	QAction *computeAlignDistanceAct;

	QActionGroup *modelDisplayGroup;
	QAction *showFeaturesAct;
	QAction *showCurvaturesAct;
	QAction *showFramesAct;
	QAction *showGeometryAct;
	QAction	*showSamplesAct;
	QAction *showStatusAct;

	void createContextMenu();

	//////////////////////////////////////////////////////////////////////////
	// Scene data
	//////////////////////////////////////////////////////////////////////////
	Scene *scene;

	//////////////////////////////////////////////////////////////////////////
	// State methods and parameters
	//////////////////////////////////////////////////////////////////////////
	bool showWireframe();
	bool showBbox();
	bool showSourceModel();
	bool showTargetModel();
	bool renderTransparent();
	QString updateFrame(int frame);
	void drawSphereTest();
	void drawCurvatureScales();
	void colormapTest();
	void pointListTest();
	void saveTransformSpacePoints();
	ShapeModel *selectCurrentModel();
	void drawFrame( const ShapeModel* frame );


	void rotateAbout(qglviewer::Vec up, float angle);
	int displayFrame;

	// Scene change notify variable
	QMutex sceneMutex;
	bool sceneChanged;

	// Snapshot file
	QString customSaveSnapshotFile;

	//////////////////////////////////////////////////////////////////////////
	// Set deform parameters
	DeformEdit* deformEdit;
	float maxGraphNodes, maxGraphNeighbors, maxOverlapFactor;

	CorrespViewer *correspViewer;

	//////////////////////////////////////////////////////////////////////////
	// Test capabilities of PointList
	PointList *tempList;

	//////////////////////////////////////////////////////////////////////////
	// Test sphere-fitting
	enum InteractionMode { NONE = 0, DEFINE_SPHERE };
	InteractionMode interactionMode_;
	real3_type clickStartPos, clickEndPos;
	real3_type fitCenter;
	real_type fitRadius;

	real3_type closestPointOnLine( QPoint pos, real3_type x );
	void sphereFit();
	void testFeatureSelect();
	void featureProximity();
	void testFeatureMatch();
	void testFeatureCluster();
	void testRefineCluster();
	void verifyTransformationSpace();
	void visualizeClusterMode();
	void testGraphCutRelabel();
	void testSampleCorrespondences();

	// Range query for mesh
	idxList inRange;
	distList distRange;

	// Range query in feature space
	int sourceQueryIndex;
	PointList sourceQuery, targetMatch;

	// Curvature values
	int displayScale;
	PointListN *sourceFeatures, *targetFeatures;

	// Hold a list of correspondence candidates
	FeatureViewer *featureViewer;
	CorrespEdit *correspEdit;

	// Maximum values for various correspondence parameters
	float maxFeatureSpaceRange;

	// Principal curvatures testing variables
	int curvDisplayMode;

	bool autoLaunchDone;

	ShapeSlicer *shapeSlicer;
	TransformViewer *transformViewer;


	// Debugging: test point-box and point-triangle distance code.
	real3_type testBoxMin, testBoxMax;
	real3_type testT0, testT1, testT2;
	real3_type testPoint, testCBox, testCTri, testCMesh;
	real_type testDBox, testDTri, testDMesh;
	void testDistanceCode();


signals:
	void updateChildDisplay();
};
